MODULE dynhpg_tam
#ifdef key_tam
   !!======================================================================
   !!                       ***  MODULE  dynhpg_tam  ***
   !! Ocean dynamics:  hydrostatic pressure gradient trend
   !!                  Tangent and Adjoint module
   !!======================================================================
   !! History of the direct module:
   !!            1.0  !  87-09  (P. Andrich, M.-A. Foujols)  hpg_zco: Original code
   !!            5.0  !  91-11  (G. Madec)
   !!            7.0  !  96-01  (G. Madec)  hpg_sco: Original code for s-coordinates
   !!            8.0  !  97-05  (G. Madec)  split dynber into dynkeg and dynhpg
   !!            8.5  !  02-07  (G. Madec)  F90: Free form and module
   !!            8.5  !  02-08  (A. Bozec)  hpg_zps: Original code
   !!            9.0  !  05-10  (A. Beckmann, B.W. An)  various s-coordinate options
   !!                           Original code for hpg_ctl, hpg_hel hpg_wdj, hpg_djc, hpg_rot 
   !!            9.0  !  05-11  (G. Madec) style & small optimisation
   !! History of the TAM module:
   !!            9.0  !  08-06  (A. Vidard) Skeleton
   !!                 !  08-11  (A. Vidard) Nemo v3
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   dyn_hpg      : update the momentum trend with the now horizontal
   !!                  gradient of the hydrostatic pressure
   !!       hpg_ctl  : initialisation and control of options
   !!       hpg_zco  : z-coordinate scheme
   !!       hpg_zps  : z-coordinate plus partial steps (interpolation)
   !!       hpg_sco  : s-coordinate (standard jacobian formulation)
   !!       hpg_hel  : s-coordinate (helsinki modification)
   !!       hpg_wdj  : s-coordinate (weighted density jacobian)
   !!       hpg_djc  : s-coordinate (Density Jacobian with Cubic polynomial)
   !!       hpg_rot  : s-coordinate (ROTated axes scheme)
   !!----------------------------------------------------------------------
   USE par_kind,       ONLY: & ! Precision variables
      & wp
   USE par_oce,        ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1,               &
      & jpiglo
   USE oce_tam       , ONLY: & ! ocean dynamics and tracers
      & ua_tl,               &
      & va_tl,               &
      & rhd_tl,              &
      & ua_ad,               &
      & va_ad,               &
      & rhd_ad,              &
      & gru_tl,              &
      & grv_tl,              &
      & gru_ad,              &
      & grv_ad,              &    
      & tn_tl, sn_tl,        &
      & gtu_tl, gtv_tl,      &
      & gsu_tl, gsv_tl,      &
      & rhop_tl
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & lk_vvl,              &
      & e1u,                 &
      & e2u,                 &
      & e1v,                 &
      & e2v,                 &
#if defined key_zco
      & e3t_0,               &
      & e3w_0,               &
#else
      & e3u,                 &
      & e3v,                 &
      & e3w,                 &
#endif
      & mbathy,              &
      & umask,               &
      & vmask,               &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE phycst        , ONLY: & ! physical constants
      & grav
   USE in_out_manager, ONLY: & ! I/O manager
      & ctl_stop,                 &
      & lwp,                 &
      & numout,              & 
      & numnam,              & 
      & nit000,              & 
      & nitend
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
      & prntst_tlm,          & !
      & stdu,                & ! stdev for u-velocity
      & stdv,                & !           v-velocity
      & stdr,                & ! 
      & stdt,                & !
      & stds

   IMPLICIT NONE
   PRIVATE

   PUBLIC   dyn_hpg_tan    ! routine called by step_tam module
   PUBLIC   dyn_hpg_adj    ! routine called by step_tam module
   PUBLIC   dyn_hpg_adj_tst! routine called by test module
#if defined key_tst_tlm
   PUBLIC   dyn_hpg_tlm_tst! routine called by test module
#endif

   !!* Namelist nam_dynhpg : Choice of horizontal pressure gradient computation
   LOGICAL  ::   ln_hpg_zco = .TRUE.    ! z-coordinate - full steps
   LOGICAL  ::   ln_hpg_zps = .FALSE.   ! z-coordinate - partial steps (interpolation)
   LOGICAL  ::   ln_hpg_sco = .FALSE.   ! s-coordinate (standard jacobian formulation)
   LOGICAL  ::   ln_hpg_hel = .FALSE.   ! s-coordinate (helsinki modification)
   LOGICAL  ::   ln_hpg_wdj = .FALSE.   ! s-coordinate (weighted density jacobian)
   LOGICAL  ::   ln_hpg_djc = .FALSE.   ! s-coordinate (Density Jacobian with Cubic polynomial)
   LOGICAL  ::   ln_hpg_rot = .FALSE.   ! s-coordinate (ROTated axes scheme)
   REAL(wp) ::   gamm       = 0.e0      ! weighting coefficient

   INTEGER  ::   nhpg  =  0             ! = 0 to 6, type of pressure gradient scheme used
      !                                 ! (deduced from ln_hpg_... flags)

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE dyn_hpg_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_hpg_tan  ***
      !!
      !! ** Method of the direct routine:
      !!              Call the hydrostatic pressure gradient routine 
      !!              using the scheme defined in the namelist
      !!   
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!             - Save the trend (l_trddyn=T)
      !!----------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt   ! ocean time-step index
      !!
      !!----------------------------------------------------------------------
      IF( kt == nit000 )   CALL hpg_ctl_tam       ! initialisation & control of options

      SELECT CASE ( nhpg )      ! Hydrastatic pressure gradient computation
      CASE (  0 )   ;   CALL hpg_zco_tan    ( kt )      ! z-coordinate
      CASE (  1 )   ;   CALL hpg_zps_tan    ( kt )      ! z-coordinate plus partial steps (interpolation)
      CASE (  2 )   ;   CALL hpg_sco_tan    ( kt )      ! s-coordinate (standard jacobian formulation)
      CASE (  3 )   ;   CALL hpg_hel_tan    ( kt )      ! s-coordinate (helsinki modification)
      CASE (  4 )   ;   CALL hpg_wdj_tan    ( kt )      ! s-coordinate (weighted density jacobian)
      CASE (  5 )   ;   CALL hpg_djc_tan    ( kt )      ! s-coordinate (Density Jacobian with Cubic polynomial)
      CASE (  6 )   ;   CALL hpg_rot_tan    ( kt )      ! s-coordinate (ROTated axes scheme)
      END SELECT
   END SUBROUTINE dyn_hpg_tan
   SUBROUTINE dyn_hpg_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_hpg_adj  ***
      !!
      !! ** Method of the direct routine:
      !!              call the hydrostatic pressure gradient routine 
      !!              using the scheme defined in the namelist
      !!   
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!             - Save the trend (l_trddyn=T)
      !!----------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt   ! ocean time-step index
      !!
      !!----------------------------------------------------------------------
      IF( kt == nitend )   CALL hpg_ctl_tam      ! initialisation & control of options
      
      SELECT CASE ( nhpg )      ! Hydrastatic pressure gradient computation
      CASE (  0 )   ;   CALL hpg_zco_adj    ( kt )      ! z-coordinate
      CASE (  1 )   ;   CALL hpg_zps_adj    ( kt )      ! z-coordinate plus partial steps (interpolation)
      CASE (  2 )   ;   CALL hpg_sco_adj    ( kt )      ! s-coordinate (standard jacobian formulation)
      CASE (  3 )   ;   CALL hpg_hel_adj    ( kt )      ! s-coordinate (helsinki modification)
      CASE (  4 )   ;   CALL hpg_wdj_adj    ( kt )      ! s-coordinate (weighted density jacobian)
      CASE (  5 )   ;   CALL hpg_djc_adj    ( kt )      ! s-coordinate (Density Jacobian with Cubic polynomial)
      CASE (  6 )   ;   CALL hpg_rot_adj    ( kt )      ! s-coordinate (ROTated axes scheme)
      END SELECT
   END SUBROUTINE dyn_hpg_adj

   SUBROUTINE hpg_ctl_tam
      !!----------------------------------------------------------------------
      !!                 ***  ROUTINE hpg_ctl_tam  ***
      !!
      !! ** Purpose :   initializations for the hydrostatic pressure gradient
      !!              computation and consistency control
      !!
      !! ** Action  :   Read the namelist namdynhpg and check the consistency
      !!      with the type of vertical coordinate used (zco, zps, sco)
      !!----------------------------------------------------------------------
      INTEGER ::   ioptio = 0      ! temporary integer

      NAMELIST/nam_dynhpg/ ln_hpg_zco, ln_hpg_zps, ln_hpg_sco, ln_hpg_hel,   &
         &                 ln_hpg_wdj, ln_hpg_djc, ln_hpg_rot, gamm
      !!----------------------------------------------------------------------

      REWIND ( numnam )               ! Read Namelist nam_dynhpg : pressure gradient calculation options
      READ   ( numnam, nam_dynhpg )

      IF(lwp) THEN                    ! Control print
         WRITE(numout,*)
         WRITE(numout,*) 'dyn:hpg_ctl_tam : hydrostatic pressure gradient control'
         WRITE(numout,*) '~~~~~~~~~~~~~~~'
         WRITE(numout,*) '       Namelist nam_dynhpg : choice of hpg scheme'
         WRITE(numout,*) '          z-coord. - full steps                          ln_hpg_zco = ', ln_hpg_zco
         WRITE(numout,*) '          z-coord. - partial steps (interpolation)       ln_hpg_zps = ', ln_hpg_zps
         WRITE(numout,*) '          s-coord. (standard jacobian formulation)       ln_hpg_sco = ', ln_hpg_sco
         WRITE(numout,*) '          s-coord. (helsinki modification)               ln_hpg_hel = ', ln_hpg_hel
         WRITE(numout,*) '          s-coord. (weighted density jacobian)           ln_hpg_wdj = ', ln_hpg_wdj
         WRITE(numout,*) '          s-coord. (Density Jacobian: Cubic polynomial)  ln_hpg_djc = ', ln_hpg_djc
         WRITE(numout,*) '          s-coord. (ROTated axes scheme)                 ln_hpg_rot = ', ln_hpg_rot
         WRITE(numout,*) '          weighting coeff. (wdj scheme)                     gamm       = ', gamm
      ENDIF

      IF( lk_vvl .AND. .NOT. ln_hpg_sco )   THEN
         CALL ctl_stop( 'hpg_ctl_tam : variable volume key_vvl compatible only with the standard jacobian formulation hpg_sco')
      ENDIF

      !                               ! Set nhpg from ln_hpg_... flags
      IF( ln_hpg_zco )   nhpg = 0
      IF( ln_hpg_zps )   nhpg = 1
      IF( ln_hpg_sco )   nhpg = 2
      IF( ln_hpg_hel )   nhpg = 3
      IF( ln_hpg_wdj )   nhpg = 4
      IF( ln_hpg_djc )   nhpg = 5
      IF( ln_hpg_rot )   nhpg = 6

      !                               ! Consitency check
      ioptio = 0 
      IF( ln_hpg_zco )   ioptio = ioptio + 1
      IF( ln_hpg_zps )   ioptio = ioptio + 1
      IF( ln_hpg_sco )   ioptio = ioptio + 1
      IF( ln_hpg_hel )   ioptio = ioptio + 1
      IF( ln_hpg_wdj )   ioptio = ioptio + 1
      IF( ln_hpg_djc )   ioptio = ioptio + 1
      IF( ln_hpg_rot )   ioptio = ioptio + 1
      IF ( ioptio /= 1 )   CALL ctl_stop( ' NO or several hydrostatic pressure gradient options used' )

      !
   END SUBROUTINE hpg_ctl_tam
   SUBROUTINE hpg_zco_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_zco_tan  ***
      !!
      !! ** Method of the direct routine:
      !!      z-coordinate case, levels are horizontal surfaces.
      !!      The now hydrostatic pressure gradient at a given level, jk,
      !!      is computed by taking the vertical integral of the in-situ
      !!      density gradient along the model level from the suface to that
      !!      level:    zhpi = grav .....
      !!                zhpj = grav .....
      !!      add it to the general momentum trend (ua,va).
      !!            ua = ua - 1/e1u * zhpi
      !!            va = va - 1/e2v * zhpj
      !! 
      !! ** Action : - Update (ua_tl,va_tl) with the now hydrastatic pressure trend
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk       ! dummy loop indices
      REAL(wp) ::   zcoef0, zcoef1   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zhpitl, zhpjtl
      !!----------------------------------------------------------------------
      
      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_zco_tan : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~   z-coordinate case '
      ENDIF
      
      ! Local constant initialization 
      zcoef0 = - grav * 0.5_wp

      ! Surface value
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            zcoef1 = zcoef0 * fse3w(ji,jj,1)
            ! hydrostatic pressure gradient
            zhpitl(ji,jj,1) = zcoef1 * ( rhd_tl(ji+1,jj,1) - rhd_tl(ji,jj,1) ) / e1u(ji,jj)
            zhpjtl(ji,jj,1) = zcoef1 * ( rhd_tl(ji,jj+1,1) - rhd_tl(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua_tl(ji,jj,1) = ua_tl(ji,jj,1) + zhpitl(ji,jj,1)
            va_tl(ji,jj,1) = va_tl(ji,jj,1) + zhpjtl(ji,jj,1)
         END DO
      END DO
      !
      ! interior value (2=<jk=<jpkm1)
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zcoef1 = zcoef0 * fse3w(ji,jj,jk)
               ! hydrostatic pressure gradient
               zhpitl(ji,jj,jk) = zhpitl(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd_tl(ji+1,jj,jk)+rhd_tl(ji+1,jj,jk-1) )   &
                  &                       - ( rhd_tl(ji  ,jj,jk)+rhd_tl(ji  ,jj,jk-1) )  ) / e1u(ji,jj)

               zhpjtl(ji,jj,jk) = zhpjtl(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd_tl(ji,jj+1,jk)+rhd_tl(ji,jj+1,jk-1) )   &
                  &                       - ( rhd_tl(ji,jj,  jk)+rhd_tl(ji,jj  ,jk-1) )  ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua_tl(ji,jj,jk) = ua_tl(ji,jj,jk) + zhpitl(ji,jj,jk)
               va_tl(ji,jj,jk) = va_tl(ji,jj,jk) + zhpjtl(ji,jj,jk)
            END DO
         END DO
      END DO
      !
   END SUBROUTINE hpg_zco_tan
   SUBROUTINE hpg_zco_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_zco_tan  ***
      !!
      !! ** Method of the direct routine:
      !!      z-coordinate case, levels are horizontal surfaces.
      !!      The now hydrostatic pressure gradient at a given level, jk,
      !!      is computed by taking the vertical integral of the in-situ
      !!      density gradient along the model level from the suface to that
      !!      level:    zhpi = grav .....
      !!                zhpj = grav .....
      !!      add it to the general momentum trend (ua,va).
      !!            ua = ua - 1/e1u * zhpi
      !!            va = va - 1/e2v * zhpj
      !! 
      !! ** Action : - Update (ua_tl,va_tl) with the now hydrastatic pressure trend
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk       ! dummy loop indices
      REAL(wp) ::   zcoef0, zcoef1   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zhpiad, zhpjad
      !!----------------------------------------------------------------------
      
      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_zco_adj : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~   z-coordinate case '
      ENDIF
      ! adjoint variables initialization
      zhpiad = 0.0_wp
      zhpjad = 0.0_wp
      ! Local constant initialization 
      zcoef0 = - grav * 0.5

      ! interior value (2=<jk=<jpkm1)
      DO jk = jpkm1, 2, -1
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.
               zcoef1 = zcoef0 * fse3w(ji,jj,jk)
               ! add to the general momentum trend
               zhpiad(ji,jj,jk) = zhpiad(ji,jj,jk) + ua_ad(ji,jj,jk)
               zhpjad(ji,jj,jk) = zhpjad(ji,jj,jk) + va_ad(ji,jj,jk)
               ! hydrostatic pressure gradient
               rhd_ad(ji,jj+1,jk  ) = rhd_ad(ji,jj+1,jk  ) + zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               rhd_ad(ji,jj+1,jk-1) = rhd_ad(ji,jj+1,jk-1) + zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               rhd_ad(ji,jj  ,jk  ) = rhd_ad(ji,jj  ,jk  ) - zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               rhd_ad(ji,jj  ,jk-1) = rhd_ad(ji,jj  ,jk-1) - zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               zhpjad(ji,jj  ,jk-1) = zhpjad(ji,jj  ,jk-1) + zhpjad(ji,jj,jk) 
               zhpjad(ji,jj  ,jk  ) = 0.0_wp

               rhd_ad(ji+1,jj,jk  ) = rhd_ad(ji+1,jj,jk  ) + zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               rhd_ad(ji+1,jj,jk-1) = rhd_ad(ji+1,jj,jk-1) + zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               rhd_ad(ji  ,jj,jk  ) = rhd_ad(ji  ,jj,jk  ) - zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               rhd_ad(ji  ,jj,jk-1) = rhd_ad(ji  ,jj,jk-1) - zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               zhpiad(ji  ,jj,jk-1) = zhpiad(ji  ,jj,jk-1) + zhpiad(ji,jj,jk) 
               zhpiad(ji  ,jj,jk  ) = 0.0_wp

            END DO
         END DO
      END DO
      ! Surface value
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            zcoef1 = zcoef0 * fse3w(ji,jj,1)
            ! add to the general momentum trend
            zhpiad(ji,jj,1) = zhpiad(ji,jj,1) + ua_ad(ji,jj,1)
            zhpjad(ji,jj,1) = zhpjad(ji,jj,1) + va_ad(ji,jj,1)
            ! hydrostatic pressure gradient
            rhd_ad(ji,jj+1,1) = rhd_ad(ji,jj+1,1) + zhpjad(ji,jj,1) * zcoef1 / e2v(ji,jj)
            rhd_ad(ji,jj  ,1) = rhd_ad(ji,jj  ,1) - zhpjad(ji,jj,1) * zcoef1 / e2v(ji,jj)
            zhpjad(ji,jj,1) = 0.0_wp

            rhd_ad(ji+1,jj,1) = rhd_ad(ji+1,jj,1) + zhpiad(ji,jj,1) * zcoef1 / e1u(ji,jj)
            rhd_ad(ji  ,jj,1) = rhd_ad(ji  ,jj,1) - zhpiad(ji,jj,1) * zcoef1 / e1u(ji,jj)
            zhpiad(ji,jj,1) = 0.0_wp
         END DO
      END DO
      !
      !
   END SUBROUTINE hpg_zco_adj
   SUBROUTINE hpg_zps_tan( kt )
      !!---------------------------------------------------------------------
      !!                 ***  ROUTINE hpg_zps  ***
      !!                    
      !! ** Method of the direct routine:
      !!              z-coordinate plus partial steps case.  blahblah...
      !! 
      !! ** Action  : - Update (ua_tl,va_tl) with the now hydrastatic pressure trend
      !!---------------------------------------------------------------------- 
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk                       ! dummy loop indices
      INTEGER  ::   iku, ikv                         ! temporary integers
      REAL(wp) ::   zcoef0, zcoef1, zcoef2, zcoef3   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zhpitl, zhpjtl
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_zps_tan : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~   z-coordinate with partial steps - vector optimization'
      ENDIF

      ! Local constant initialization
      zcoef0 = - grav * 0.5

      !  Surface value
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            zcoef1 = zcoef0 * fse3w(ji,jj,1)
            ! hydrostatic pressure gradient
            zhpitl(ji,jj,1) = zcoef1 * ( rhd_tl(ji+1,jj  ,1) - rhd_tl(ji,jj,1) ) / e1u(ji,jj)
            zhpjtl(ji,jj,1) = zcoef1 * ( rhd_tl(ji  ,jj+1,1) - rhd_tl(ji,jj,1) ) / e2v(ji,jj)
            ! add to the general momentum trend
            ua_tl(ji,jj,1) = ua_tl(ji,jj,1) + zhpitl(ji,jj,1)
            va_tl(ji,jj,1) = va_tl(ji,jj,1) + zhpjtl(ji,jj,1)
         END DO
      END DO

      ! interior value (2=<jk=<jpkm1)
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zcoef1 = zcoef0 * fse3w(ji,jj,jk)
               ! hydrostatic pressure gradient
               zhpitl(ji,jj,jk) = zhpitl(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd_tl(ji+1,jj,jk) + rhd_tl(ji+1,jj,jk-1) )   &
                  &                       - ( rhd_tl(ji  ,jj,jk) + rhd_tl(ji  ,jj,jk-1) )  ) / e1u(ji,jj)

               zhpjtl(ji,jj,jk) = zhpjtl(ji,jj,jk-1)   &
                  &           + zcoef1 * (  ( rhd_tl(ji,jj+1,jk) + rhd_tl(ji,jj+1,jk-1) )   &
                  &                       - ( rhd_tl(ji,jj,  jk) + rhd_tl(ji,jj  ,jk-1) )  ) / e2v(ji,jj)
               ! add to the general momentum trend
               ua_tl(ji,jj,jk) = ua_tl(ji,jj,jk) + zhpitl(ji,jj,jk)
               va_tl(ji,jj,jk) = va_tl(ji,jj,jk) + zhpjtl(ji,jj,jk)
            END DO
         END DO
      END DO

      ! partial steps correction at the last level  (new gradient with  intgrd.F)
# if defined key_vectopt_loop
         jj = 1
         DO ji = jpi+2, jpij-jpi-1   ! vector opt. (forced unrolling)
# else
      DO jj = 2, jpjm1
         DO ji = 2, jpim1
# endif
            iku = MIN ( mbathy(ji,jj), mbathy(ji+1,jj) ) - 1
            ikv = MIN ( mbathy(ji,jj), mbathy(ji,jj+1) ) - 1
            zcoef2 = zcoef0 * MIN( fse3w(ji,jj,iku), fse3w(ji+1,jj  ,iku) )
            zcoef3 = zcoef0 * MIN( fse3w(ji,jj,ikv), fse3w(ji  ,jj+1,ikv) )
            ! on i-direction
            IF ( iku > 2 ) THEN
               ! subtract old value
               ua_tl(ji,jj,iku) = ua_tl(ji,jj,iku) - zhpitl(ji,jj,iku)
               ! compute the new one
               zhpitl (ji,jj,iku) = zhpitl(ji,jj,iku-1)   &
                  + zcoef2 * ( rhd_tl(ji+1,jj,iku-1) - rhd_tl(ji,jj,iku-1) + gru_tl(ji,jj) ) / e1u(ji,jj)
               ! add the new one to the general momentum trend
               ua_tl(ji,jj,iku) = ua_tl(ji,jj,iku) + zhpitl(ji,jj,iku)
            ENDIF
            ! on j-direction
            IF ( ikv > 2 ) THEN
               ! subtract old value
               va_tl(ji,jj,ikv) = va_tl(ji,jj,ikv) - zhpjtl(ji,jj,ikv)
               ! compute the new one
               zhpjtl (ji,jj,ikv) = zhpjtl(ji,jj,ikv-1)   &
                  + zcoef3 * ( rhd_tl(ji,jj+1,ikv-1) - rhd_tl(ji,jj,ikv-1) + grv_tl(ji,jj) ) / e2v(ji,jj)
               ! add the new one to the general momentum trend
               va_tl(ji,jj,ikv) = va_tl(ji,jj,ikv) + zhpjtl(ji,jj,ikv)
            ENDIF
# if ! defined key_vectopt_loop
         END DO
# endif
      END DO
      !
   END SUBROUTINE hpg_zps_tan
   SUBROUTINE hpg_zps_adj( kt )
      !!---------------------------------------------------------------------
      !!                 ***  ROUTINE hpg_zps  ***
      !!                    
      !! ** Method of the direct routine:
      !!              z-coordinate plus partial steps case.  blahblah...
      !! 
      !! ** Action  : - Update (ua_tl,va_tl) with the now hydrastatic pressure trend
      !!---------------------------------------------------------------------- 
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      !!
      INTEGER  ::   ji, jj, jk                       ! dummy loop indices
      INTEGER  ::   iku, ikv                         ! temporary integers
      REAL(wp) ::   zcoef0, zcoef1, zcoef2, zcoef3   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk):: zhpiad, zhpjad
      !!----------------------------------------------------------------------

      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn:hpg_zps_adj : hydrostatic pressure gradient trend'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~~   z-coordinate with partial steps - vector optimization'
      ENDIF
      zhpiad(:,:,:) = 0.0_wp
      zhpjad(:,:,:) = 0.0_wp
      ! Local constant initialization
      zcoef0 = - grav * 0.5

      ! partial steps correction at the last level  (new gradient with  intgrd.F)
# if defined key_vectopt_loop
         jj = 1
         DO ji = jpij-jpi-1, jpi+2, -1   ! vector opt. (forced unrolling)
# else
      DO jj = jpjm1, 2, -1
         DO ji = jpim1, 2, -1
# endif
            iku = MIN ( mbathy(ji,jj), mbathy(ji+1,jj) ) - 1
            ikv = MIN ( mbathy(ji,jj), mbathy(ji,jj+1) ) - 1
            zcoef2 = zcoef0 * MIN( fse3w(ji,jj,iku), fse3w(ji+1,jj  ,iku) )
            zcoef3 = zcoef0 * MIN( fse3w(ji,jj,ikv), fse3w(ji  ,jj+1,ikv) )
            ! on i-direction
            IF ( iku > 2 ) THEN
               ! add the new one to the general momentum trend
               zhpiad(ji,jj,iku) = zhpiad(ji,jj,iku) + ua_ad(ji,jj,iku)
               ! compute the new one
               rhd_ad(ji+1,jj,iku-1) = rhd_ad(ji+1,jj,iku-1) + zhpiad (ji,jj,iku) * zcoef2 / e1u(ji,jj)
               rhd_ad(ji,jj,iku-1) = rhd_ad(ji,jj,iku-1) - zhpiad (ji,jj,iku) * zcoef2 / e1u(ji,jj)
               gru_ad(ji,jj) = gru_ad(ji,jj) + zhpiad (ji,jj,iku) * zcoef2 / e1u(ji,jj)
               zhpiad(ji,jj,iku-1) = zhpiad(ji,jj,iku-1) + zhpiad (ji,jj,iku)
               zhpiad (ji,jj,iku) = 0.0_wp
               ! subtract old value
               zhpiad(ji,jj,iku) = zhpiad(ji,jj,iku) - ua_ad(ji,jj,iku)
            ENDIF
            ! on j-direction
            IF ( ikv > 2 ) THEN
               ! add the new one to the general momentum trend
               zhpjad(ji,jj,ikv) = zhpjad(ji,jj,ikv) + va_ad(ji,jj,ikv)
               ! compute the new one
               rhd_ad(ji,jj+1,ikv-1) = rhd_ad(ji,jj+1,ikv-1) + zhpjad (ji,jj,ikv) * zcoef3 / e2v(ji,jj)
               rhd_ad(ji,jj,ikv-1) = rhd_ad(ji,jj,ikv-1) -zhpjad (ji,jj,ikv) * zcoef3 / e2v(ji,jj)
               grv_ad(ji,jj) = grv_ad(ji,jj) +zhpjad (ji,jj,ikv) * zcoef3 / e2v(ji,jj)
               zhpjad(ji,jj,ikv-1) = zhpjad(ji,jj,ikv-1) + zhpjad(ji,jj,ikv)
               zhpjad (ji,jj,ikv) = 0.0_wp
               ! subtract old value
               zhpjad(ji,jj,ikv) = zhpjad(ji,jj,ikv) - va_ad(ji,jj,ikv)
            ENDIF
# if ! defined key_vectopt_loop
         END DO
# endif
      END DO
      !


      ! interior value (2=<jk=<jpkm1)
      DO jk = jpkm1, 2, -1
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.
               zcoef1 = zcoef0 * fse3w(ji,jj,jk)
               ! add to the general momentum trend
               zhpiad(ji,jj,jk) = zhpiad(ji,jj,jk) + ua_ad(ji,jj,jk)
               zhpjad(ji,jj,jk) = zhpjad(ji,jj,jk) + va_ad(ji,jj,jk)
               ! hydrostatic pressure gradient
               rhd_ad(ji,jj+1,jk  ) = rhd_ad(ji,jj+1,jk  ) + zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               rhd_ad(ji,jj+1,jk-1) = rhd_ad(ji,jj+1,jk-1) + zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               rhd_ad(ji,jj  ,jk  ) = rhd_ad(ji,jj  ,jk  ) - zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               rhd_ad(ji,jj  ,jk-1) = rhd_ad(ji,jj  ,jk-1) - zhpjad(ji,jj,jk) * zcoef1 / e2v(ji,jj)
               zhpjad(ji,jj  ,jk-1) = zhpjad(ji,jj  ,jk-1) + zhpjad(ji,jj,jk)
               zhpjad(ji,jj  ,jk  ) = 0.0_wp

               rhd_ad(ji+1,jj,jk  ) = rhd_ad(ji+1,jj,jk  ) + zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               rhd_ad(ji+1,jj,jk-1) = rhd_ad(ji+1,jj,jk-1) + zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               rhd_ad(ji  ,jj,jk  ) = rhd_ad(ji  ,jj,jk  ) - zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               rhd_ad(ji  ,jj,jk-1) = rhd_ad(ji  ,jj,jk-1) - zhpiad(ji,jj,jk) * zcoef1 / e1u(ji,jj)
               zhpiad(ji  ,jj,jk-1) = zhpiad(ji  ,jj,jk-1) + zhpiad(ji,jj,jk)
               zhpiad(ji  ,jj,jk  ) = 0.0_wp
            END DO
         END DO
      END DO
      !  Surface value
      DO jj = jpjm1, 2, -1
         DO ji = fs_jpim1, fs_2, -1   ! vector opt.
            zcoef1 = zcoef0 * fse3w(ji,jj,1)
            ! add to the general momentum trend
            zhpiad(ji,jj,1) = zhpiad(ji,jj,1) + ua_ad(ji,jj,1)
            zhpjad(ji,jj,1) = zhpjad(ji,jj,1) + va_ad(ji,jj,1)
            ! hydrostatic pressure gradient
            rhd_ad(ji+1,jj  ,1) = rhd_ad(ji+1,jj  ,1) + zhpiad(ji,jj,1) * zcoef1 / e1u(ji,jj)
            rhd_ad(ji  ,jj  ,1) = rhd_ad(ji  ,jj  ,1) - zhpiad(ji,jj,1) * zcoef1 / e1u(ji,jj)
            rhd_ad(ji  ,jj+1,1) = rhd_ad(ji  ,jj+1,1) + zhpjad(ji,jj,1) * zcoef1 / e2v(ji,jj)
            rhd_ad(ji  ,jj  ,1) = rhd_ad(ji  ,jj  ,1) - zhpjad(ji,jj,1) * zcoef1 / e2v(ji,jj)
            zhpiad(ji  ,jj  ,1) = 0.0_wp
            zhpjad(ji  ,jj  ,1) = 0.0_wp
         END DO
      END DO

   END SUBROUTINE hpg_zps_adj
   SUBROUTINE hpg_sco_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_sco_tan  ***
      !!
      !! ** Method of the direct routine:   s-coordinate case. Jacobian scheme.
      !!      The now hydrostatic pressure gradient at a given level, jk,
      !!      is computed by taking the vertical integral of the in-situ
      !!      density gradient along the model level from the suface to that
      !!      level. s-coordinates (ln_sco): a corrective term is added
      !!      to the horizontal pressure gradient :
      !!         zhpi = grav .....  + 1/e1u mi(rhd) di[ grav dep3w ]
      !!         zhpj = grav .....  + 1/e2v mj(rhd) dj[ grav dep3w ]
      !!      add it to the general momentum trend (ua,va).
      !!         ua = ua - 1/e1u * zhpi
      !!         va = va - 1/e2v * zhpj
      !!
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      CALL ctl_stop( 'hpg_sco_tan not available yet')
   END SUBROUTINE hpg_sco_tan
   SUBROUTINE hpg_sco_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_sco_adj  ***
      !!
      !! ** Method of the direct routine:   s-coordinate case. Jacobian scheme.
      !!      The now hydrostatic pressure gradient at a given level, jk,
      !!      is computed by taking the vertical integral of the in-situ
      !!      density gradient along the model level from the suface to that
      !!      level. s-coordinates (ln_sco): a corrective term is added
      !!      to the horizontal pressure gradient :
      !!         zhpi = grav .....  + 1/e1u mi(rhd) di[ grav dep3w ]
      !!         zhpj = grav .....  + 1/e2v mj(rhd) dj[ grav dep3w ]
      !!      add it to the general momentum trend (ua,va).
      !!         ua = ua - 1/e1u * zhpi
      !!         va = va - 1/e2v * zhpj
      !!
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      CALL ctl_stop( 'hpg_sco_adj not available yet')
   END SUBROUTINE hpg_sco_adj
   SUBROUTINE hpg_hel_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_hel_tan  ***
      !!
      !! ** Method of the direct routine:   s-coordinate case.
      !!      The now hydrostatic pressure gradient at a given level
      !!      jk is computed by taking the vertical integral of the in-situ 
      !!      density gradient along the model level from the suface to that 
      !!      level. s-coordinates (ln_sco): a corrective term is added
      !!      to the horizontal pressure gradient :
      !!         zhpi = grav .....  + 1/e1u mi(rhd) di[ grav dep3w ]
      !!         zhpj = grav .....  + 1/e2v mj(rhd) dj[ grav dep3w ]
      !!      add it to the general momentum trend (ua,va).
      !!         ua = ua - 1/e1u * zhpi
      !!         va = va - 1/e2v * zhpj
      !!
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!             - Save the trend (l_trddyn=T)
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      CALL ctl_stop( 'hpg_hel_tan not available yet')
   END SUBROUTINE hpg_hel_tan
   SUBROUTINE hpg_hel_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_hel_adj  ***
      !!
      !! ** Method of the direct routine:   s-coordinate case.
      !!      The now hydrostatic pressure gradient at a given level
      !!      jk is computed by taking the vertical integral of the in-situ 
      !!      density gradient along the model level from the suface to that 
      !!      level. s-coordinates (ln_sco): a corrective term is added
      !!      to the horizontal pressure gradient :
      !!         zhpi = grav .....  + 1/e1u mi(rhd) di[ grav dep3w ]
      !!         zhpj = grav .....  + 1/e2v mj(rhd) dj[ grav dep3w ]
      !!      add it to the general momentum trend (ua,va).
      !!         ua = ua - 1/e1u * zhpi
      !!         va = va - 1/e2v * zhpj
      !!
      !! ** Action : - Update (ua,va) with the now hydrastatic pressure trend
      !!             - Save the trend (l_trddyn=T)
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      CALL ctl_stop( 'hpg_hel_adj not available yet')
   END SUBROUTINE hpg_hel_adj
   SUBROUTINE hpg_wdj_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_wdj_tan  ***
      !!
      !! ** Method of the direct roiutine:
      !!      Weighted Density Jacobian (wdj) scheme (song 1998)
      !!      The weighting coefficients from the namelist parameter gamm
      !!      (alpha=0.5-gamm ; beta=1-alpha=0.5+gamm)
      !!
      !! Reference : Song, Mon. Wea. Rev., 126, 3213-3230, 1998.
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      CALL ctl_stop( 'hpg_wdj_tan not available yet')
   END SUBROUTINE hpg_wdj_tan
   SUBROUTINE hpg_wdj_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_wdj_adj  ***
      !!
      !! ** Method of the direct roiutine:
      !!      Weighted Density Jacobian (wdj) scheme (song 1998)
      !!      The weighting coefficients from the namelist parameter gamm
      !!      (alpha=0.5-gamm ; beta=1-alpha=0.5+gamm)
      !!
      !! Reference : Song, Mon. Wea. Rev., 126, 3213-3230, 1998.
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      CALL ctl_stop( 'hpg_wdj_adj not available yet')
   END SUBROUTINE hpg_wdj_adj
   SUBROUTINE hpg_djc_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_djc_tan  ***
      !!
      !! ** Method  :   Density Jacobian with Cubic polynomial scheme
      !! 
      !! Reference: Shchepetkin and McWilliams, J. Geophys. Res., 108(C3), 3090, 2003
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      !!
      CALL ctl_stop( 'hpg_djc_tan not available yet')
   END SUBROUTINE hpg_djc_tan
   SUBROUTINE hpg_djc_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_djc_adj  ***
      !!
      !! ** Method  :   Density Jacobian with Cubic polynomial scheme
      !! 
      !! Reference: Shchepetkin and McWilliams, J. Geophys. Res., 108(C3), 3090, 2003
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
      !!
      CALL ctl_stop( 'hpg_djc_adj not available yet')
   END SUBROUTINE hpg_djc_adj
   SUBROUTINE hpg_rot_tan( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_rot_tan  ***
      !!
      !! ** Method  :   rotated axes scheme (Thiem and Berntsen 2005)
      !!
      !! Reference: Thiem & Berntsen, Ocean Modelling, In press, 2005.
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
            !!
      CALL ctl_stop( 'hpg_rot_tan not available yet')
   END SUBROUTINE hpg_rot_tan
   SUBROUTINE hpg_rot_adj( kt )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE hpg_rot_adj  ***
      !!
      !! ** Method  :   rotated axes scheme (Thiem and Berntsen 2005)
      !!
      !! Reference: Thiem & Berntsen, Ocean Modelling, In press, 2005.
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt    ! ocean time-step index
            !!
      CALL ctl_stop( 'hpg_rot_adj not available yet')
   END SUBROUTINE hpg_rot_adj

   SUBROUTINE dyn_hpg_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dynhpg_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !! ** Action  : Separate tests are applied for the following dx and dy:
      !!                
      !!              1) dx = ( SSH ) and dy = ( SSH )
      !!                   
      !! History :
      !!        ! 08-07 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit

      REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::   &
         & zrhd_tlin,              &  ! in situ density anomalie
         & zua_tlin,               &  ! after u- velocity
         & zva_tlin,               &  ! after v- velocity
         & zua_tlout,              &  ! after u- velocity
         & zva_tlout                  ! after v- velocity
      REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::   &
         & zrhd_adout,             &  ! in situ density anomalie
         & zua_adout,              &  ! after u- velocity
         & zva_adout,              &  ! after v- velocity
         & zua_adin,               &  ! after u- velocity
         & zva_adin                   ! after v- velocity
      REAL(wp), DIMENSION(:,:), ALLOCATABLE ::   &
         & zgru_tlin,              &
         & zgrv_tlin,              &
         & zgru_adout,             &
         & zgrv_adout
         
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zrh,          & ! 3D random field for rhd
         & zau,          & ! 3D random field for u
         & zav             ! 3D random field for v
      REAL(wp), DIMENSION(:,:), ALLOCATABLE ::   &
         & zgru,         & ! 2D random field for gru 
         & zgrv            ! 2D random field for grv
      REAL(KIND=wp) ::   &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_1,       & !   scalar product components
         & zsp1_2,       & 
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       & !   scalar product components
         & zsp2_2,       & 
         & zsp2_3,       & 
         & zsp2_4,       & 
         & zsp2_5
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d           ! 2D seed for the random number generator
      INTEGER :: &
         & iseed, &
         & ji, &
         & jj, &
         & jk
      CHARACTER(LEN=14) :: cl_name

      ! Allocate memory
      ALLOCATE( & 
         & zrhd_tlin(jpi,jpj,jpk),  &
         & zua_tlin(jpi,jpj,jpk),   &
         & zva_tlin(jpi,jpj,jpk),   &
         & zgru_tlin(jpi,jpj),      & 
         & zgrv_tlin(jpi,jpj),      & 
         & zua_tlout(jpi,jpj,jpk),  &
         & zva_tlout(jpi,jpj,jpk),  &
         & zrhd_adout(jpi,jpj,jpk), &
         & zua_adout(jpi,jpj,jpk),  &
         & zva_adout(jpi,jpj,jpk),  &
         & zgru_adout(jpi,jpj),     & 
         & zgrv_adout(jpi,jpj),     & 
         & zua_adin(jpi,jpj,jpk),   &
         & zva_adin(jpi,jpj,jpk),   &
         & zrh(jpi,jpj,jpk),        & 
         & zau(jpi,jpj,jpk),        & 
         & zav(jpi,jpj,jpk),        & 
         & zgru(jpi,jpj),           &
         & zgrv(jpi,jpj)            &
         &     )
 

      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zrhd_tlin(:,:,:)  = 0.0_wp  
      zua_tlin(:,:,:)   = 0.0_wp   
      zva_tlin(:,:,:)   = 0.0_wp   
      zgru_tlin(:,:)    = 0.0_wp
      zgrv_tlin(:,:)    = 0.0_wp
      zua_tlout(:,:,:)  = 0.0_wp  
      zva_tlout(:,:,:)  = 0.0_wp  
      zgru_adout(:,:)   = 0.0_wp
      zgrv_adout(:,:)   = 0.0_wp
      zrhd_adout(:,:,:) = 0.0_wp 
      zua_adout(:,:,:)  = 0.0_wp  
      zva_adout(:,:,:)  = 0.0_wp  
      zua_adin(:,:,:)   = 0.0_wp   
      zva_adin(:,:,:)   = 0.0_wp   
      zrh(:,:,:)        = 0.0_wp         
      zau(:,:,:)        = 0.0_wp         
      zav(:,:,:)        = 0.0_wp 
      zgru(:,:)         = 0.0_wp
      zgrv(:,:)         = 0.0_wp


      gru_tl(:,:)   = 0.0_wp
      grv_tl(:,:)   = 0.0_wp
      gru_ad(:,:)   = 0.0_wp
      grv_ad(:,:)   = 0.0_wp
      ua_tl(:,:,:)  = 0.0_wp
      va_tl(:,:,:)  = 0.0_wp
      rhd_tl(:,:,:) = 0.0_wp
      ua_ad(:,:,:)  = 0.0_wp
      va_ad(:,:,:)  = 0.0_wp
      rhd_ad(:,:,:) = 0.0_wp

      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zau, 'U', 0.0_wp, stdu )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 523432 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zav, 'V', 0.0_wp, stdv )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 456953 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zrh, 'W', 0.0_wp, stdr )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 432545 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zgru, 'U', 0.0_wp, stdu )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 287503 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zgrv, 'V', 0.0_wp, stdv )

      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zrhd_tlin(ji,jj,jk) = zrh(ji,jj,jk)
               zua_tlin(ji,jj,jk)  = zau(ji,jj,jk)
               zva_tlin(ji,jj,jk)  = zav(ji,jj,jk)
            END DO
         END DO
      END DO
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zgru_tlin(ji,jj)   = zgru(ji,jj)
            zgrv_tlin(ji,jj)   = zgrv(ji,jj)
         END DO
      END DO
      ua_tl(:,:,:)  = zua_tlin(:,:,:)
      va_tl(:,:,:)  = zva_tlin(:,:,:)
      rhd_tl(:,:,:) = zrhd_tlin(:,:,:)
      gru_tl(:,:)   = zgru_tlin(:,:)
      grv_tl(:,:)   = zgrv_tlin(:,:) 

      CALL dyn_hpg_tan ( nit000 )

      zua_tlout(:,:,:) = ua_tl(:,:,:)
      zva_tlout(:,:,:) = va_tl(:,:,:)
      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zua_adin(ji,jj,jk) = zua_tlout(ji,jj,jk) &
                 &               * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
                 &               * umask(ji,jj,jk)
              zva_adin(ji,jj,jk) = zva_tlout(ji,jj,jk) &
                 &               * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
                 &               * vmask(ji,jj,jk)
            END DO
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1_1 = DOT_PRODUCT( zua_tlout, zua_adin )
      zsp1_2 = DOT_PRODUCT( zva_tlout, zva_adin )
      zsp1   = zsp1_1 + zsp1_2

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      ua_ad(:,:,:) = zua_adin(:,:,:)
      va_ad(:,:,:) = zva_adin(:,:,:)

      CALL dyn_hpg_adj ( nit000 )

      zgru_adout(:,:)   = gru_ad(:,:)
      zgrv_adout(:,:)   = grv_ad(:,:)
      zrhd_adout(:,:,:) = rhd_ad(:,:,:)
      zua_adout(:,:,:)  = ua_ad(:,:,:)
      zva_adout(:,:,:)  = va_ad(:,:,:)

      zsp2_1 = DOT_PRODUCT( zgru_tlin, zgru_adout )
      zsp2_2 = DOT_PRODUCT( zgrv_tlin, zgrv_adout )
      zsp2_3 = DOT_PRODUCT( zrhd_tlin, zrhd_adout )
      zsp2_4 = DOT_PRODUCT( zua_tlin, zua_adout )
      zsp2_5 = DOT_PRODUCT( zva_tlin, zva_adout )
      zsp2   = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4 + zsp2_5
      ! Compare the scalar products

      cl_name = 'dyn_hpg_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE( &
         & zrhd_tlin,  &
         & zua_tlin,   &
         & zva_tlin,   &
         & zgru_tlin,  & 
         & zgrv_tlin,  & 
         & zua_tlout,  &
         & zva_tlout,  &
         & zrhd_adout, &
         & zua_adout,  &
         & zva_adout,  &
         & zgru_adout, & 
         & zgrv_adout, & 
         & zua_adin,   &
         & zva_adin,   &
         & zrh,        & 
         & zau,        & 
         & zav,        & 
         & zgru,       &
         & zgrv        &
         &              )
   END SUBROUTINE dyn_hpg_adj_tst
#if defined key_tst_tlm
   SUBROUTINE dyn_hpg_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_hpg_tlm_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!                   
      !! History :
      !!        ! 09-08 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE dynhpg
      USE zpshde
      USE eosbn2,     ONLY: & ! horizontal & vertical advective trend
        & eos
      USE zpshde_tam
      USE eosbn2_tam,     ONLY: & ! horizontal & vertical advective trend
        & eos_tan
      USE tamtrj              ! writing out state trajectory
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE divcur             ! horizontal divergence and relative vorticity
      USE wzvmod             !  vertical velocity
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE oce       , ONLY: & ! ocean dynamics and tracers variables
        & ua, va, rhd,      &
        & gru, grv,         &
        & un, vn,           &      
        & tn, sn, gtu, gtv, &
        & gsu, gsv, rhop
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
      !! * Local declarations
      REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::   &
         & zrd_tlin,               &  ! in situ density anomalie
         & zua_tlin,               &  ! after u- velocity
         & zva_tlin,               &  ! after v- velocity
         & ztn_tlin,               &  ! after u- velocity
         & zsn_tlin,               &  ! after v- velocity
         & zua_wop,                &  ! after u- velocity
         & zva_wop,                &  ! after v- velocity
         & zua_out,                &  ! after u- velocity
         & zva_out                    ! after v- velocity

      REAL(wp), DIMENSION(:,:), ALLOCATABLE ::   &
         & zgru_tlin,              &
         & zgrv_tlin

      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zrh,          & ! 3D random field for rhd
         & zau,          & ! 3D random field for u
         & zav             ! 3D random field for v
      REAL(wp), DIMENSION(:,:), ALLOCATABLE ::   &
         & zgru,         & ! 2D random field for gru 
         & zgrv            ! 2D random field for grv

      REAL(KIND=wp) ::   &
         & zsp1,         & ! scalar product 
         & zsp1_1,       & ! scalar product
         & zsp1_2,       & 
         & zsp2,         & ! scalar product 
         & zsp2_1,       & ! scalar product
         & zsp2_2,       &
         & zsp3,         & ! scalar product
         & zsp3_1,       & 
         & zsp3_2,       & 
         & zzsp,         & ! scalar product
         & zzsp_1,       & 
         & zzsp_2,       &
         & gamma,            &
         & zgsp1,            &
         & zgsp2,            &
         & zgsp3,            &
         & zgsp4,            &
         & zgsp5,            &
         & zgsp6,            &
         & zgsp7 
      INTEGER :: &
         & ji, &
         & jj, &
         & jk
      CHARACTER(LEN=14)   :: cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_xdx
      CHARACTER (LEN=90)  ::  FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zscua, zscva,    &
         & zscerrua,        &
         & zscerrva
      INTEGER, DIMENSION(100):: &
         & iiposua, iiposva,    &
         & ijposua, ijposva,    & 
         & ikposua, ikposva
      INTEGER::             &
         & ii,              &
         & isamp=40,        &
         & jsamp=40,        &
         & ksamp=10,        &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj,jpk) :: &
         & zerrua, zerrva
      ! Allocate memory
      ALLOCATE( & 
         & zrd_tlin(jpi,jpj,jpk),   &
         & zua_tlin(jpi,jpj,jpk),   &
         & zva_tlin(jpi,jpj,jpk),   &
         & ztn_tlin(jpi,jpj,jpk),   &
         & zsn_tlin(jpi,jpj,jpk),   &
         & zgru_tlin(jpi,jpj),      & 
         & zgrv_tlin(jpi,jpj),      & 
         & zua_wop  (jpi,jpj,jpk),  &
         & zva_wop  (jpi,jpj,jpk),  &
         & zua_out  (jpi,jpj,jpk),  &
         & zva_out  (jpi,jpj,jpk),  &
         & zrh(      jpi,jpj,jpk),  & 
         & zau(      jpi,jpj,jpk),  & 
         & zav(      jpi,jpj,jpk),  & 
         & zgru(     jpi,jpj),      &
         & zgrv(     jpi,jpj)       &
         &     )

      !--------------------------------------------------------------------
      ! Reset variables
      !--------------------------------------------------------------------
      zrd_tlin( :,:,:) = 0.0_wp
      zgru_tlin(  :,:) = 0.0_wp
      zgrv_tlin(  :,:) = 0.0_wp
      zua_tlin( :,:,:) = 0.0_wp
      zva_tlin( :,:,:) = 0.0_wp
      ztn_tlin( :,:,:) = 0.0_wp
      zsn_tlin( :,:,:) = 0.0_wp
      zua_out ( :,:,:) = 0.0_wp
      zva_out ( :,:,:) = 0.0_wp
      zua_wop ( :,:,:) = 0.0_wp
      zva_wop ( :,:,:) = 0.0_wp

      zrh(:,:,:)        = 0.0_wp         
      zau(:,:,:)        = 0.0_wp         
      zav(:,:,:)        = 0.0_wp 
      zgru(:,:)         = 0.0_wp
      zgrv(:,:)         = 0.0_wp

      zscua(:)         = 0.0_wp
      zscva(:)         = 0.0_wp
      zscerrua(:)      = 0.0_wp
      zscerrva(:)      = 0.0_wp
      zerrua(:,:,:)    = 0.0_wp
      zerrva(:,:,:)    = 0.0_wp
      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio  
      file_wop='trj_wop_dynhpg'
      file_xdx='trj_xdx_dynhpg'
       !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 596035, zau,  'U', 0.0_wp, stdu)
         CALL grid_rd_sd( 523432, zav,  'V', 0.0_wp, stdv)
         CALL grid_rd_sd( 456953, zrh,  'W', 0.0_wp, stdr)
         CALL grid_rd_sd( 432545, zgru, 'U', 0.0_wp, stdu)
         CALL grid_rd_sd( 287503, zgrv, 'V', 0.0_wp, stdv) 
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zrd_tlin(ji,jj,jk) = zrh(ji,jj,jk)
                  zua_tlin(ji,jj,jk)  = zau(ji,jj,jk)
                  zva_tlin(ji,jj,jk)  = zav(ji,jj,jk)
               END DO
            END DO
         END DO 

         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgru_tlin(ji,jj)   = zgru(ji,jj)
               zgrv_tlin(ji,jj)   = zgrv(ji,jj)
            END DO
         END DO 
      ENDIF
      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      IF ( tlm_bch /= 2 )      CALL istate_p  

      ! *** initialize the reference trajectory
      ! ------------
      CALL  trj_rea( nit000-1, 1 )
!      CALL  trj_rea( nit000, 1 )
      ua(:,:,:)       = un(:,:,:)
      va(:,:,:)       = vn(:,:,:) 

      !  Compute rhd, gru and grv
      CALL eos(tn, sn, rhd, rhop)
      CALL zps_hde(nit000, tn, sn, rhd, gtu, gsu, gru, gtv, gsv, grv)


      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN

         zrd_tlin(:,:,:) = gamma * zrd_tlin(:,:,:)
         rhd(:,:,:)       = rhd(:,:,:) + zrd_tlin(:,:,:)

         zua_tlin(:,:,:) = gamma * zua_tlin(:,:,:)
         ua(:,:,:)       = ua(:,:,:) + zua_tlin(:,:,:)

         zva_tlin(:,:,:) = gamma * zva_tlin(:,:,:)
         va(:,:,:)       = va(:,:,:) + zva_tlin(:,:,:)

         zgru_tlin(:,:) = gamma * zgru_tlin(:,:)
         gru(:,:)       = gru(:,:) + zgru_tlin(:,:)

         zgrv_tlin(:,:) = gamma * zgrv_tlin(:,:)
         grv(:,:)       = grv(:,:) + zgrv_tlin(:,:)
      ENDIF 
      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      IF ( tlm_bch /= 2 ) CALL dyn_hpg(nit000)
      IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN
         !--------------------------------------------------------------------
         ! Initialize the tangent variables 
         !--------------------------------------------------------------------
         CALL  trj_rea( nit000-1, 1 ) 
!         CALL  trj_rea( nit000, 1 ) 
         ua(:,:,:)       = un(:,:,:)
         va(:,:,:)       = vn(:,:,:)
         gru_tl  (  :,:) = zgru_tlin (:,:  )
         grv_tl  (  :,:) = zgrv_tlin (:,:  )
         rhd_tl  (:,:,:) = zrd_tlin  (:,:,:)
         ua_tl   (:,:,:) = zua_tlin  (:,:,:)
         va_tl   (:,:,:) = zva_tlin  (:,:,:) 
         CALL dyn_hpg_tan(nit000)
         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------
         zsp2_1    = DOT_PRODUCT( ua_tl, ua_tl  )
         zsp2_2    = DOT_PRODUCT( va_tl, va_tl  )
         zsp2      = zsp2_1 + zsp2_2
         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------
         CALL trj_rd_spl(file_wop) 
         zua_wop  (:,:,:) = ua  (:,:,:)
         zva_wop  (:,:,:) = va  (:,:,:)
         CALL trj_rd_spl(file_xdx) 
         zua_out  (:,:,:) = ua  (:,:,:)
         zva_out  (:,:,:) = va  (:,:,:)
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zua_out   (ji,jj,jk) = zua_out    (ji,jj,jk) - zua_wop  (ji,jj,jk)
                  zua_wop   (ji,jj,jk) = zua_out    (ji,jj,jk) - ua_tl    (ji,jj,jk)
                  IF (  ua_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrua(ji,jj,jk) = zua_out(ji,jj,jk)/ua_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iiposua(ii) = ji
                      ijposua(ii) = jj
                      ikposua(ii) = jk
                      IF ( INT(umask(ji,jj,jk)) .NE. 0)  THEN
                         zscua (ii) =  zua_wop(ji,jj,jk)
                         zscerrua (ii) =  ( zerrua(ji,jj,jk) - 1.0_wp ) /gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zva_out   (ji,jj,jk) = zva_out    (ji,jj,jk) - zva_wop  (ji,jj,jk)
                  zva_wop   (ji,jj,jk) = zva_out    (ji,jj,jk) - va_tl    (ji,jj,jk)
                  IF (  va_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrva(ji,jj,jk) = zva_out(ji,jj,jk)/va_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iiposva(ii) = ji
                      ijposva(ii) = jj
                      ikposva(ii) = jk
                      IF ( INT(vmask(ji,jj,jk)) .NE. 0)  THEN
                         zscva (ii) =  zua_wop(ji,jj,jk)
                         zscerrva (ii) =  ( zerrva(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         zsp1_1    = DOT_PRODUCT( zua_out, zua_out )
         zsp1_2    = DOT_PRODUCT( zva_out, zva_out  )
         zsp1      = zsp1_1 + zsp1_2

         zsp3_1    = DOT_PRODUCT( zua_wop, zua_wop )
         zsp3_2    = DOT_PRODUCT( zva_wop, zva_wop  )
         zsp3      = zsp3_1 + zsp3_2
         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name  = 'dynhpg_tam:En '
         zzsp     = dsqrt(zsp3)
         zzsp_1   = dsqrt(zsp3_1)
         zzsp_2   = dsqrt(zsp3_2)

         zgsp5 = zzsp
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp     = SQRT(zsp2)
         zzsp_1   = SQRT(zsp2_1)
         zzsp_2   = SQRT(zsp2_2)
         zgsp4    = zzsp
         cl_name  = 'dynhpg_tam:Ln2'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         zzsp     = SQRT(zsp1)
         zzsp_1   = SQRT(zsp1_1)
         zzsp_2   = SQRT(zsp1_2)
         cl_name  = 'dynhpg:Mhdx-Mx'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         zgsp3    = SQRT( zsp3/zsp2 )
         zgsp7    = zgsp3/gamma
         zgsp1    = zzsp
         zgsp2    = zgsp1 / zgsp4
         zgsp6    = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'dynhpg  ', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7
         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'dynhpg  '
         IF (lwp) THEN
            DO ii=1, 100, 1
               IF ( zscua(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscua     ', &
                    & cur_loop, h_ratio, ii, iiposua(ii), ijposua(ii), zscua(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscva(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscva     ', &
                    & cur_loop, h_ratio, ii, iiposva(ii), ijposva(ii), zscva(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrua(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrua  ', &
                 & cur_loop, h_ratio, ii, iiposua(ii), ijposua(ii), zscerrua(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrva(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrva  ', &
                    & cur_loop, h_ratio, ii, iiposva(ii), ijposva(ii), zscerrva(ii)
            ENDDO

           ! write separator
           WRITE(numtan_sc,"(A4)") '===='
         ENDIF

      ENDIF
      DEALLOCATE(                           &
         & zrd_tlin, zgru_tlin, zgrv_tlin,  & 
         & zua_tlin, zva_tlin,              & 
         & ztn_tlin, zsn_tlin,              & 
         & zua_out, zva_out,                & 
         & zua_wop, zva_wop,                &
         & zrh,                             & 
         & zau,                             & 
         & zav,                             & 
         & zgru,                            &
         & zgrv                             &
         & )
   END SUBROUTINE dyn_hpg_tlm_tst
   !!======================================================================
#endif
#endif
END MODULE dynhpg_tam
