MODULE dynzdf_exp_tam
#ifdef key_tam
   !!==============================================================================
   !!                     ***  MODULE  dynzdf_exp_tam  ***
   !! Ocean dynamics:  vertical component(s) of the momentum mixing trend
   !!                 Tangent and Adjoint Module
   !!==============================================================================
   !! History of the direct module:
   !!                !  90-10  (B. Blanke)  Original code
   !!                !  97-05  (G. Madec)  vertical component of isopycnal
   !!           8.5  !  02-08  (G. Madec)  F90: Free form and module
   !! History of the TAM module:
   !!           9.0  !  08-0!  (A. Vidard) Skeleton
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   dyn_zdf_exp  : update the momentum trend with the vertical diffu-
   !!                  sion using an explicit time-stepping scheme.
   !!----------------------------------------------------------------------
   !! * Modules used

   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 & 
      & jpim1,               &
      & jpjm1,               & 
      & jpkm1
   USE oce_tam       , ONLY: & ! ocean dynamics and tracers
      & ub_tl,               &
      & vb_tl,               &
      & ua_tl,               &
      & va_tl,               &
      & ub_ad,               &
      & vb_ad,               &
      & ua_ad,               &
      & va_ad
   USE zdf_oce       , ONLY: & ! ocean vertical physics
      & avmu,                &
      & avmv,                &
      & n_zdfexp
   USE dom_oce       , ONLY: & ! ocean space and time domain
#if defined key_zco
      & e3t_0,               &
      & e3w_0,               &
#else
      & e3u,                 &
      & e3v,                 &
      & e3uw,                &
      & e3vw,                &
#endif
      & umask,               &
      & vmask
   USE phycst        , ONLY: & ! physical constants
      & rau0
   USE in_out_manager, ONLY: & ! I/O manager
      & nit000,              &
      & nitend,              &
      & numout,              &
      & lwp,                 &
      & ctl_stop
   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC dyn_zdf_exp_tan     ! called by dynzdf_tam.F90
   PUBLIC dyn_zdf_exp_adj     ! called by dynzdf_tam.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_zdf_exp_tan( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_zdf_exp_tan  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the trend due to the vert. momentum diffusion
      !!
      !! ** Method of the direct routine:
      !!      Explicit forward time stepping with a time splitting
      !!      technique. The vertical diffusion of momentum is given by:
      !!         diffu = dz( avmu dz(u) ) = 1/e3u dk+1( avmu/e3uw dk(ub) )
      !!      Surface boundary conditions: wind stress input
      !!      Bottom boundary conditions : bottom stress (cf zdfbfr.F90)
      !!      Add this trend to the general trend ua :
      !!         ua = ua + dz( avmu dz(u) )
      !!
      !! ** Action : - Update (ua,va) with the vertical diffusive trend
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER , INTENT( in ) ::   kt                           ! ocean time-step index
      REAL(wp), INTENT( in ) ::   p2dt                         ! time-step 

      !! * Local declarations
      INTEGER ::   ji, jj, jk, jl                              ! dummy loop indices
      REAL(wp) ::   zrau0r, zlavmr, zuatl, zvatl                   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpk) ::   zwxtl, zwytl, zwztl, zwwtl     ! temporary workspace arrays
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_zdf_exp_tan : vertical momentum diffusion explicit operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~ '
      ENDIF
      ! Local constant initialization
      ! -----------------------------
      zrau0r = 1. / rau0                                   ! inverse of the reference density
      zlavmr = 1. / float( n_zdfexp )                      ! inverse of the number of sub time step

      !                                                ! ===============
      DO jj = 2, jpjm1                                 !  Vertical slab
         !                                             ! ===============

         ! Surface boundary condition
         DO ji = 2, jpim1
            zwytl(ji,1) = 0.0_wp
            zwwtl(ji,1) = 0.0_wp
         END DO  

         ! Initialization of x, z and contingently trends array
         DO jk = 1, jpk
            DO ji = 2, jpim1
               zwxtl(ji,jk) = ub_tl(ji,jj,jk)
               zwztl(ji,jk) = vb_tl(ji,jj,jk)
            END DO  
         END DO  

         ! Time splitting loop
         DO jl = 1, n_zdfexp

            ! First vertical derivative
            DO jk = 2, jpk
               DO ji = 2, jpim1
                  zwytl(ji,jk) = avmu(ji,jj,jk) * ( zwxtl(ji,jk-1) - zwxtl(ji,jk) ) / fse3uw(ji,jj,jk) 
                  zwwtl(ji,jk) = avmv(ji,jj,jk) * ( zwztl(ji,jk-1) - zwztl(ji,jk) ) / fse3vw(ji,jj,jk)
               END DO  
            END DO  

            ! Second vertical derivative and trend estimation at kt+l*rdt/n_zdfexp
            DO jk = 1, jpkm1
               DO ji = 2, jpim1
                  zuatl = zlavmr*( zwytl(ji,jk) - zwytl(ji,jk+1) ) / fse3u(ji,jj,jk)
                  zvatl = zlavmr*( zwwtl(ji,jk) - zwwtl(ji,jk+1) ) / fse3v(ji,jj,jk)
                  ua_tl(ji,jj,jk) = ua_tl(ji,jj,jk) + zuatl
                  va_tl(ji,jj,jk) = va_tl(ji,jj,jk) + zvatl

                  zwxtl(ji,jk) = zwxtl(ji,jk) + p2dt*zuatl*umask(ji,jj,jk)
                  zwztl(ji,jk) = zwztl(ji,jk) + p2dt*zvatl*vmask(ji,jj,jk)
               END DO  
            END DO  

         END DO  

         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============


   END SUBROUTINE dyn_zdf_exp_tan
   SUBROUTINE dyn_zdf_exp_adj( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_zdf_exp_adj  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the trend due to the vert. momentum diffusion
      !!
      !! ** Method of the direct routine:
      !!      Explicit forward time stepping with a time splitting
      !!      technique. The vertical diffusion of momentum is given by:
      !!         diffu = dz( avmu dz(u) ) = 1/e3u dk+1( avmu/e3uw dk(ub) )
      !!      Surface boundary conditions: wind stress input
      !!      Bottom boundary conditions : bottom stress (cf zdfbfr.F90)
      !!      Add this trend to the general trend ua :
      !!         ua = ua + dz( avmu dz(u) )
      !!
      !! ** Action : - Update (ua,va) with the vertical diffusive trend
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER , INTENT( in ) ::   kt                           ! ocean time-step index
      REAL(wp), INTENT( in ) ::   p2dt                         ! time-step 

      !! * Local declarations
      INTEGER ::   ji, jj, jk, jl                              ! dummy loop indices
      REAL(wp) ::   zrau0r, zlavmr, zua, zva                   ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpk) ::   zwx, zwy, zwz, zww     ! temporary workspace arrays
      !!----------------------------------------------------------------------

      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_zdf_exp_adj : vertical momentum diffusion explicit operator'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~ '
      ENDIF

      CALL ctl_stop ('dyn_zdf_exp_adj not available yet')
   END SUBROUTINE dyn_zdf_exp_adj
#endif
   !!==============================================================================
END MODULE dynzdf_exp_tam
