MODULE solsor_tam
   !!======================================================================
   !!                       ***  MODULE solsor_tam ***
   !! NEMOVAR : Tangent linear ....
   !!======================================================================

   !!----------------------------------------------------------------------
   !!   sol_sor_tan : Red-Black Successive Over-Relaxation solver TL
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp        ! Variable kinds
   USE par_oce       , ONLY: &
      & jpi,                 &
      & jpj,                 &
      & jpk,                 &
      & jpr2di,              &
      & jpr2dj,              &
      & jpiglo
   USE in_out_manager, ONLY: & ! I/O manager 
      & nit000, lwp
   USE sol_oce       , ONLY: & ! solver variables
      & gcdmat,              &
      & rnorme,              &
      & eps,                 &
      & epsr,                &
      & nmax,                &
      & nmin,                &
      & ncut,                &
      & nmod,                &
      & res,                 &
      & resmax,              &
      & gcp,                 &
      & sor,                 &
      & nsol_arp,            &
      & c_solver_pt,         &
      & niter
   USE lib_mpp       , ONLY: & ! distributed memory computing
      & lk_mpp,              &
      &  mpp_max
   USE lbclnk       , ONLY: & ! ocean lateral boundary conditions (or mpp link)
      & lbc_lnk_e
   USE gridsum        , ONLY: & ! Global sums
      & global_sum
   USE lbclnk_tam              ! ocean lateral boundary conditions (or mpp link)
   USE sol_oce_tam   , ONLY: & ! solver variables (assimilation)
      & nitsor,              &
      & gcx_tl,              & 
      & gcb_tl,              & 
      & gcr_tl,              & 
      & gcx_ad,              & 
      & gcb_ad            
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & mig,                 &
      & mjg,                 &
      & nimpp,               &
      & njmpp,               &
      & e2t,                 &
      & e1t,                 &
      & tmask,               &
      & nlci,                &
      & nlcj,                &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE gridrandom    , ONLY: &     ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: &     ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          &
      & stdgc,               &
      & prntst_tlm
   

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC sol_sor_adj          ! 
   PUBLIC sol_sor_tan          ! 
   PUBLIC sol_sor_adj_tst      ! called by tamtst.F90
#if defined key_tst_tlm
   PUBLIC sol_sor_tlm_tst      ! called by tamtst.F90
#endif

CONTAINS

   SUBROUTINE sol_sor_tan( kt, kindic )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE sol_sor_tan : TL of sol_sor  ***
      !!                 
      !! ** Purpose :   Solve the ellipic equation for the barotropic stream 
      !!      function system (lk_dynspg_rl=T) or the transport divergence 
      !!      system (lk_dynspg_flt=T) using a red-black successive-over-
      !!      relaxation method.
      !!       In the former case, the barotropic stream function trend has a
      !!     zero boundary condition along all coastlines (i.e. continent
      !!     as well as islands) while in the latter the boundary condition
      !!     specification is not required.
      !!       This routine provides a MPI optimization to the existing solsor
      !!     by reducing the number of call to lbc.
      !!
      !! ** Method  :   Successive-over-relaxation method using the red-black 
      !!      technique. The former technique used was not compatible with 
      !!      the north-fold boundary condition used in orca configurations.
      !!      Compared to the classical sol_sor, this routine provides a 
      !!      mpp optimization by reducing the number of calls to lnc_lnk
      !!      The solution is computed on a larger area and the boudary
      !!      conditions only when the inside domain is reached.
      !!
      !! References :
      !!      Madec et al. 1988, Ocean Modelling, issue 78, 1-6.
      !!      Beare and Stevens 1997 Ann. Geophysicae 15, 1369-1377
      !!
      !! History of the direct routine:
      !!        !  90-10  (G. Madec)  Original code
      !!        !  91-11  (G. Madec)
      !!   7.1  !  93-04  (G. Madec)  time filter
      !!        !  96-05  (G. Madec)  merge sor and pcg formulations
      !!   9.0  !  03-04  (C. Deltel, G. Madec)  Red-Black SOR in free form
      !!   9.0  !  05-09  (R. Benshila, G. Madec)  MPI optimization
      !! History of the  T&A routine:
      !!        !  96-11  (A. Weaver)  correction to preconditioning
      !!   8.2  !  03-02  (C. Deltel) OPAVAR tangent-linear version
      !!   9.0  !  07-09  (K. Mogensen) tangent of the 03-04 version 
      !!   9.0  !  09-02  (A. Vidard)   tangent of the 05-09 version 
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in    ) ::   kt       ! Current timestep.
      INTEGER, INTENT( inout ) ::   kindic   ! solver indicator, < 0 if the conver-
      !                                      ! gence is not reached: the model is
      !                                      ! stopped in step
      !                                      ! set to zero before the call of solsor
      !! * Local declarations
      INTEGER  ::   ji, jj, jn               ! dummy loop indices
      INTEGER  ::   ishift, icount, istp
      REAL(wp) ::   ztmp, zres, zres2

      INTEGER  ::   ijmppodd, ijmppeven
      INTEGER  ::   ijpr2d
      !!----------------------------------------------------------------------
      
      ijmppeven = MOD(nimpp+njmpp+jpr2di+jpr2dj  ,2)
      ijmppodd  = MOD(nimpp+njmpp+jpr2di+jpr2dj+1,2)
      ijpr2d = MAX(jpr2di,jpr2dj)
      icount = 0
      !                                                       ! ==============
      DO jn = 1, nmax                                         ! Iterative loop 
         !                                                    ! ==============

         ! applied the lateral boundary conditions
         IF( MOD(icount,ijpr2d+1) == 0 ) CALL lbc_lnk_e( gcx_tl, c_solver_pt, 1.0_wp ) 
         
         ! Residuals
         ! ---------

         ! Guess black update
         DO jj = 2-jpr2dj, nlcj - 1 + jpr2dj
            ishift = MOD( jj-ijmppodd-jpr2dj, 2 )
            DO ji = 2-jpr2di+ishift, nlci - 1 + jpr2di, 2
               ztmp =                  gcb_tl(ji  ,jj  )   &
                  &   - gcp(ji,jj,1) * gcx_tl(ji  ,jj-1)   &
                  &   - gcp(ji,jj,2) * gcx_tl(ji-1,jj  )   &
                  &   - gcp(ji,jj,3) * gcx_tl(ji+1,jj  )   &
                  &   - gcp(ji,jj,4) * gcx_tl(ji  ,jj+1)
               ! Estimate of the residual
               zres = ztmp - gcx_tl(ji,jj)
               gcr_tl(ji,jj) = zres * gcdmat(ji,jj) * zres
               ! Guess update
               gcx_tl(ji,jj) = sor * ztmp + ( 1.0_wp - sor ) * gcx_tl(ji,jj)
            END DO
         END DO
         icount = icount + 1 
 
         ! applied the lateral boundary conditions
         IF( MOD(icount,ijpr2d+1) == 0 ) CALL lbc_lnk_e( gcx_tl, c_solver_pt, 1.0_wp ) 

         ! Guess red update
        DO jj = 2-jpr2dj, nlcj-1+jpr2dj
            ishift = MOD( jj-ijmppeven-jpr2dj, 2 )
            DO ji = 2-jpr2di+ishift, nlci-1+jpr2di, 2
               ztmp =                  gcb_tl(ji  ,jj  )   &
                  &   - gcp(ji,jj,1) * gcx_tl(ji  ,jj-1)   &
                  &   - gcp(ji,jj,2) * gcx_tl(ji-1,jj  )   &
                  &   - gcp(ji,jj,3) * gcx_tl(ji+1,jj  )   &
                  &   - gcp(ji,jj,4) * gcx_tl(ji  ,jj+1) 
               ! Estimate of the residual
               zres = ztmp - gcx_tl(ji,jj)
               gcr_tl(ji,jj) = zres * gcdmat(ji,jj) * zres
               ! Guess update
               gcx_tl(ji,jj) = sor * ztmp + ( 1.0_wp - sor ) * gcx_tl(ji,jj)
            END DO
         END DO
         icount = icount + 1
      
         ! test of convergence
         IF ( (jn > nmin .AND. MOD( jn-nmin, nmod ) == 0) .OR. jn==nmax) THEN

            SELECT CASE ( nsol_arp )
            CASE ( 0 )
               ! absolute precision (maximum value of the residual)
               zres2 = MAXVAL( gcr_tl(2:nlci - 1,2:nlcj - 1) )
               IF( lk_mpp )  CALL mpp_max( zres2 ) ! max over the global domain
               ! test of convergence
               IF( zres2 < resmax .OR. jn == nmax ) THEN
                  res = SQRT( zres2 )
                  niter = jn
                  ncut = 999
                  ! Store number of iterations for adjoint computation
                  istp = kt - nit000 + 1
                  nitsor(istp) = niter
               ENDIF
            CASE ( 1 )                 ! relative precision
!               rnorme = SUM( gcr_tl(2:nlci - 1,2:nlcj - 1) )
!               IF( lk_mpp ) CALL mpp_sum( rnorme ) ! sum over global domain
               rnorme = global_sum(gcr_tl)
               ! test of convergence
               IF( rnorme < epsr .OR. jn == nmax ) THEN
                  res = SQRT( rnorme )
                  niter = jn
                  ncut = 999
                  ! Store number of iterations for adjoint computation
                  istp = kt - nit000 + 1
                  nitsor(istp) = niter
               ENDIF
            END SELECT

            !****
            ! IF(lwp)WRITE(numsol,9300) jn, res, sqrt( epsr ) / eps
!9300        FORMAT('          niter :',i6,' res :',e20.10,' b :',e20.10)
            !****
         
         ENDIF
         ! indicator of non-convergence or explosion
	     IF( jn == nmax ) nitsor(istp) = jn
         IF( jn == nmax .OR. SQRT(epsr)/eps > 1.e+20 ) kindic = -2
         IF( ncut == 999 ) GOTO 999
         
         !                                              ! =====================
      END DO                                            ! END of iterative loop
      !                                                 ! =====================

999   CONTINUE
      
      !  Output in gcx_tl
      !  ----------------

      CALL lbc_lnk_e( gcx_tl, c_solver_pt, 1.0_wp )    ! Lateral BCs

   END SUBROUTINE sol_sor_tan
   SUBROUTINE sol_sor_adj( kt, kindic )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE sol_sor_adj : adjoint of sol_sor  ***
      !!                 
      !! ** Purpose :   Solve the ellipic equation for the barotropic stream 
      !!      function system (lk_dynspg_rl=T) or the transport divergence 
      !!      system (lk_dynspg_flt=T) using a red-black successive-over-
      !!      relaxation method.
      !!       In the former case, the barotropic stream function trend has a
      !!     zero boundary condition along all coastlines (i.e. continent
      !!     as well as islands) while in the latter the boundary condition
      !!     specification is not required.
      !!       This routine provides a MPI optimization to the existing solsor
      !!     by reducing the number of call to lbc.
      !!
      !! ** Method  :   Successive-over-relaxation method using the red-black 
      !!      technique. The former technique used was not compatible with 
      !!      the north-fold boundary condition used in orca configurations.
      !!      Compared to the classical sol_sor, this routine provides a 
      !!      mpp optimization by reducing the number of calls to lbc_lnk
      !!      The solution is computed on a larger area and the boudary
      !!      conditions only when the inside domain is reached.
      !! ** Comments on adjoint routine :
      !!      When the step in a tangent-linear DO loop is an arbitrary
      !!      integer then care must be taken in computing the lower bound
      !!      of the adjoint DO loop; i.e.,
      !!
      !!      If the tangent-linear DO loop is:  low_tl, up_tl, step
      !!
      !!      then the adjoint DO loop is:  low_ad, up_ad, -step
      !!
      !!      where  low_ad = low_tl + step * INT( ( up_tl - low_tl ) / step )
      !!             up_ad  = low_tl
      !!
      !!      NB. If step = 1 then low_ad = up_tl
      !!
      !! References :
      !!      Madec et al. 1988, Ocean Modelling, issue 78, 1-6.
      !!      Beare and Stevens 1997 Ann. Geophysicae 15, 1369-1377
      !!
      !! History of the direct routine:
      !!        !  90-10  (G. Madec)  Original code
      !!        !  91-11  (G. Madec)
      !!   7.1  !  93-04  (G. Madec)  time filter
      !!        !  96-05  (G. Madec)  merge sor and pcg formulations
      !!   9.0  !  03-04  (C. Deltel, G. Madec)  Red-Black SOR in free form
      !!   9.0  !  05-09  (R. Benshila, G. Madec)  MPI optimization
      !! History of the  T&A routine:
      !!        !  96-11  (A. Weaver)  correction to preconditioning
      !!   8.2  !  03-02  (C. Deltel) OPAVAR tangent-linear version
      !!   9.0  !  07-09  (K. Mogensen, A. Weaver) adjoint of the 03-04 version 
      !!   9.0  !  09-02  (A. Vidard)  adjoint of the 05-09 version 
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in    ) ::   kt       ! Current timestep.
      INTEGER, INTENT( inout ) ::   kindic   ! solver indicator, < 0 if the conver-
      !                                      ! gence is not reached: the model is
      !                                      ! stopped in step
      !                                      ! set to zero before the call of solsor
      !! * Local declarations
      INTEGER  ::   ji, jj, jn               ! dummy loop indices
      INTEGER  ::   ishift, icount, istp, iter, ilower
      REAL(wp) ::   ztmp, zres, zres2

      INTEGER  ::   ijmppodd, ijmppeven
      INTEGER  ::   ijpr2d
     !!----------------------------------------------------------------------
      
      ijmppeven = MOD(nimpp+njmpp+jpr2di+jpr2dj,2)
      ijmppodd  = MOD(nimpp+njmpp+jpr2di+jpr2dj+1,2)
      ijpr2d = MAX(jpr2di,jpr2dj)

      ! Fixed number of iterations
      istp = kt - nit000 + 1
      iter = nitsor(istp)
      icount = iter * 2
      !  Output in gcx_ad
      !  ----------------

      CALL lbc_lnk_e_adj( gcx_ad, c_solver_pt, 1.0_wp )    ! Lateral BCs

      !                                                    ! ==============
      DO jn = iter, 1, -1                                  ! Iterative loop 
         !                                                 ! ==============
         ! Guess red update
         DO jj =  nlcj-1+jpr2dj, 2-jpr2dj, -1
            ishift = MOD( jj-ijmppeven-jpr2dj, 2 )
            ilower = 2-jpr2dj+ishift + 2 * INT( ( ( nlci-1+jpr2dj )-( 2-jpr2dj+ishift ) ) / 2 )
            DO ji = ilower, 2-jpr2dj+ishift, -2
               gcb_ad(ji  ,jj  ) = gcb_ad(ji  ,jj  ) + sor * gcx_ad(ji,jj)
               gcx_ad(ji  ,jj-1) = gcx_ad(ji  ,jj-1) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,1)
               gcx_ad(ji-1,jj  ) = gcx_ad(ji-1,jj  ) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,2)
               gcx_ad(ji+1,jj  ) = gcx_ad(ji+1,jj  ) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,3)
               gcx_ad(ji  ,jj+1) = gcx_ad(ji  ,jj+1) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,4)
               gcx_ad(ji  ,jj  ) = gcx_ad(ji  ,jj  ) * ( 1.0_wp - sor )
            END DO
         END DO
	     icount = icount - 1         
         ! applied the lateral boundary conditions
         IF( MOD(icount,ijpr2d+1) == 0 ) CALL lbc_lnk_e_adj( gcx_ad, c_solver_pt, 1.0_wp )   ! Lateral BCs
         
         ! Residus
         ! -------

         ! Guess black update
         DO jj = nlcj - 1+jpr2dj, 2-jpr2dj, -1
            ishift = MOD( jj-ijmppodd-jpr2dj, 2 )
            ilower = 2-jpr2dj+ishift + 2 * INT( ( ( nlci-1+jpr2dj )-( 2-jpr2dj+ishift ) ) / 2 )
            DO ji = ilower, 2-jpr2dj+ishift, -2
               gcb_ad(ji  ,jj  ) = gcb_ad(ji  ,jj  ) + sor * gcx_ad(ji,jj)
               gcx_ad(ji  ,jj-1) = gcx_ad(ji  ,jj-1) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,1)
               gcx_ad(ji-1,jj  ) = gcx_ad(ji-1,jj  ) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,2)
               gcx_ad(ji+1,jj  ) = gcx_ad(ji+1,jj  ) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,3)
               gcx_ad(ji  ,jj+1) = gcx_ad(ji  ,jj+1) - sor * gcx_ad(ji,jj) &
                  &                                      * gcp(ji,jj,4)
               gcx_ad(ji  ,jj  ) = gcx_ad(ji  ,jj  ) * ( 1.0_wp - sor )
            END DO
         END DO

	     icount = icount - 1 
         ! applied the lateral boundary conditions
         IF( MOD(icount,ijpr2d+1) == 0 ) CALL lbc_lnk_e_adj( gcx_ad, c_solver_pt, 1.0_wp )   ! Lateral BCs

         !                                              ! =====================
      END DO                                            ! END of iterative loop
      !                                                 ! =====================

   END SUBROUTINE sol_sor_adj
   SUBROUTINE sol_sor_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk,    &        
         & kindic,&        ! flags fo solver convergence
         & kmod,  &        ! frequency of test for the SOR solver
         & kt              ! number of iteration
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp2            ! scalar product involving the adjoint routine
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgcb_tlin ,     & ! Tangent input
         & zgcx_tlin ,     & ! Tangent input
         & zgcx_tlout,     & ! Tangent output
         & zgcx_adin ,     & ! Adjoint input
         & zgcb_adout,     & ! Adjoint output
         & zgcx_adout,     & ! Adjoint output
         & zr             ! 3D random field 
      CHARACTER(LEN=14) :: cl_name
      ! Allocate memory


      ALLOCATE( &
         & zgcb_tlin( jpi,jpj),     &
         & zgcx_tlin( jpi,jpj),     &
         & zgcx_tlout(jpi,jpj),     &
         & zgcx_adin( jpi,jpj),     &
         & zgcx_adout(jpi,jpj),     &
         & zgcb_adout(jpi,jpj),     &
         & zr(        jpi,jpj)      &
         & )
      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zgcb_tlin( :,:) = 0.0_wp
      zgcx_tlin( :,:) = 0.0_wp
      zgcx_tlout(:,:) = 0.0_wp
      zgcx_adin( :,:) = 0.0_wp
      zgcx_adout(:,:) = 0.0_wp
      zgcb_adout(:,:) = 0.0_wp
      zr(        :,:) = 0.0_wp
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      kt=nit000
      kindic=0
!      kmod = nmod  ! store frequency of test for the SOR solver
!      nmod = 1     ! force frequency to one (remove adj_tst dependancy to nn_nmin)
         
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, c_solver_pt, 0.0_wp, stdgc )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zgcb_tlin(ji,jj) = zr(ji,jj)
         END DO
      END DO
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 264792 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, c_solver_pt, 0.0_wp, stdgc )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zgcx_tlin(ji,jj) = zr(ji,jj)
         END DO
      END DO
      ncut = 1 ! reinitilize the solver convergence flag
      gcr_tl(:,:) = 0.0_wp
      gcb_tl(:,:) = zgcb_tlin(:,:)
      gcx_tl(:,:) = zgcx_tlin(:,:)
      CALL sol_sor_tan(kt, kindic) 
      zgcx_tlout(:,:) = gcx_tl(:,:)
     

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zgcx_adin(ji,jj) = zgcx_tlout(ji,jj)         &
                 &               * e1t(ji,jj) * e2t(ji,jj) &
                 &               * tmask(ji,jj,1)
            END DO
         END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1 = DOT_PRODUCT( zgcx_tlout, zgcx_adin ) 

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------
      gcb_ad(:,:) = 0.0_wp
      gcx_ad(:,:) = zgcx_adin(:,:)
      CALL sol_sor_adj(kt, kindic) 
      zgcx_adout(:,:) = gcx_ad(:,:)
      zgcb_adout(:,:) = gcb_ad(:,:)
     

      zsp2 = DOT_PRODUCT( zgcx_tlin, zgcx_adout ) &
         & + DOT_PRODUCT( zgcb_tlin, zgcb_adout )

      ! 14 char:'12345678901234'
      cl_name = 'sol_sor_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

!      nmod = kmod  ! restore initial frequency of test for the SOR solver

      DEALLOCATE(      &
         & zgcb_tlin,  &
         & zgcx_tlin,  &
         & zgcx_tlout, &
         & zgcx_adin,  &
         & zgcx_adout, &
         & zgcb_adout, &
         & zr          &
         & )



   END SUBROUTINE sol_sor_adj_tst
#if defined key_tst_tlm
   SUBROUTINE sol_sor_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_adj_tst ***
      !!
      !! ** Purpose : Test the tangent routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!   
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!                  
      !! History :
      !!        ! 10-02 (A. Vigilant)
      !!-----------------------------------------------------------------------
#if defined key_tam
      !! * Modules used
      USE solsor             ! Red-Black Successive Over-Relaxation solver
      USE tamtrj              ! writing out state trajectory
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE wzvmod             !  vertical velocity
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE sol_oce           , ONLY: & ! ocean dynamics and tracers variables
        & gcb, gcx, ncut
      USE oce       , ONLY: & ! 
        & ua, ub, un
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
   
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk,    &        
         & kindic,&        ! flags fo solver convergence
         & kt              ! number of iteration
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) ::   &
         & zsp1, zsp2, zsp3, & ! scalar product
         & zsp_gcb, zsp_gcx, &
         & zsp,         &
         & gamma,        &
         & zgsp1,        &
         & zgsp2,        &
         & zgsp3,        &
         & zgsp4,        &
         & zgsp5,        &
         & zgsp6,        &
         & zgsp7
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgcb_tlin ,     & ! Tangent input
         & zgcx_tlin ,     & ! Tangent input
         & zgcb_out  ,     & ! Direct output
         & zgcx_out  ,     & ! Direct output
         & zgcb_wop  ,     & ! Direct output without perturbation
         & zgcx_wop  ,     & ! Direct output without perturbation
         & zr             ! 3D random field 
      CHARACTER(LEN=14) :: cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_xdx
      CHARACTER (LEN=90)  :: FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zscgcb,zscgcx,             &
         & zscerrgcb, zscerrgcx
      INTEGER, DIMENSION(100)::       &
         & iiposgcb, ijposgcb,          &
         & iiposgcx, ijposgcx
      INTEGER::             &
         & ii,              &
         & isamp=40,        &
         & jsamp=40,        &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj) :: &
         & zerrgcb, zerrgcx

      ! Allocate memory

      ALLOCATE( &
         & zgcb_tlin( jpi,jpj),     &
         & zgcx_tlin( jpi,jpj),     &
         & zgcb_out ( jpi,jpj),     &
         & zgcx_out ( jpi,jpj),     &
         & zgcb_wop ( jpi,jpj),     &
         & zgcx_wop ( jpi,jpj),     &
         & zr(        jpi,jpj)      &
         & )
      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zgcb_tlin( :,:) = 0.0_wp
      zgcx_tlin( :,:) = 0.0_wp
      zgcb_out ( :,:) = 0.0_wp
      zgcx_out ( :,:) = 0.0_wp
      zgcb_wop ( :,:) = 0.0_wp
      zgcx_wop ( :,:) = 0.0_wp
      zr(        :,:) = 0.0_wp

      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio   
      file_wop='trj_wop_solsor'
      file_xdx='trj_xdx_solsor'
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 596035, zr,  c_solver_pt, 0.0_wp, stdgc)    
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgcb_tlin(ji,jj) = zr(ji,jj)
            END DO
         END DO     
         CALL grid_rd_sd( 264792, zr,  c_solver_pt, 0.0_wp, stdgc)    
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgcx_tlin(ji,jj) = zr(ji,jj)
            END DO
         END DO
      ENDIF

      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      CALL istate_p

      ! *** initialize the reference trajectory
      ! ------------

!      gcx  (:,:) = ( ua(:,:,1) + ub(:,:,1) ) / 10.0_wp
!      gcb  (:,:) = ( ua(:,:,3) + ub(:,:,3) ) / 10.0_wp
      CALL  trj_rea( nit000-1, 1 ) 
      CALL  trj_rea( nit000, 1 )
      gcx  (:,:) =  un(:,:,1) / 10.0_wp
      gcb  (:,:) =  un(:,:,3) / 10.0_wp

      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         zgcb_tlin(:,:) = gamma * zgcb_tlin(:,:)
         gcb(:,:)       = gcb(:,:) + zgcb_tlin(:,:)

         zgcx_tlin(:,:) = gamma * zgcx_tlin(:,:)
         gcx(:,:)       = gcx(:,:) + zgcx_tlin(:,:)
      ENDIF 

      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      kindic=0
      ncut=1
      IF ( tlm_bch /= 2 ) CALL sol_sor(kindic)
      IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN
         !--------------------------------------------------------------------
         ! Initialize the tangent variables: dy^* = W dy  
         !--------------------------------------------------------------------
         gcr_tl(:,:) = 0.0_wp
         gcb_tl  (:,:) = zgcb_tlin  (:,:)
         gcx_tl  (:,:) = zgcx_tlin  (:,:)

         !-----------------------------------------------------------------------
         !  Initialization of the dynamics and tracer fields for the tangent
         !-----------------------------------------------------------------------
         ncut=1 !reset indicator of solver convergence
         CALL sol_sor_tan(nit000, kindic)
         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------

         zsp_gcx    = DOT_PRODUCT( gcx_tl, gcx_tl  )
         zsp2       = zsp_gcx
         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------
         CALL trj_rd_spl(file_wop) 
         zgcx_wop  (:,:) = gcx  (:,:)
         CALL trj_rd_spl(file_xdx) 
         zgcx_out  (:,:) = gcx  (:,:)
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgcx_out   (ji,jj) = zgcx_out    (ji,jj) - zgcx_wop  (ji,jj)
               zgcx_wop   (ji,jj) = zgcx_out    (ji,jj) - gcx_tl    (ji,jj)
               IF (  gcx_tl(ji,jj) .NE. 0.0_wp ) zerrgcx(ji,jj) = zgcx_out(ji,jj)/gcx_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgcx(ii) = ji
                   ijposgcx(ii) = jj
                   IF ( INT(tmask(ji,jj,1)) .NE. 0)  THEN
                      zscgcx (ii)    =  zgcx_wop(ji,jj)
                      zscerrgcx (ii) =  ( zerrgcx(ji,jj) - 1.0_wp ) / gamma
                   ENDIF
               ENDIF
            END DO
         END DO

         zsp_gcx   = DOT_PRODUCT( zgcx_out, zgcx_out  )

         zsp1      = zsp_gcx

         zsp_gcx   = DOT_PRODUCT( zgcx_wop, zgcx_wop  )

         zsp3      = zsp_gcx
         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'sol_sor: En   '
         zsp    = SQRT(zsp3)
         zgsp5   = zsp
         CALL prntst_tlm( cl_name, kumadt, zsp, h_ratio )

         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zsp      = SQRT(zsp2)

         zgsp4    = zsp
         cl_name  = 'sol_sor: Ln2 '
         CALL prntst_tlm( cl_name, kumadt, zsp, h_ratio )

         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         zsp     = SQRT(zsp1)

         cl_name = 'solsor:Mhdx-Mx'
         CALL prntst_tlm( cl_name, kumadt, zsp, h_ratio )

         zgsp3    = SQRT( zsp3/zsp2 )
         zgsp7    = zgsp3/gamma
         zgsp1    = zsp
         zgsp2    = zgsp1 / zgsp4
         zgsp6    = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'solsor  ', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7
         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name ='sol_sor '
         IF(lwp) THEN
            DO ii=1, 100, 1
               IF ( zscgcx(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgcx    ', &
                    & cur_loop, h_ratio, ii, iiposgcx(ii), ijposgcx(ii), zscgcx(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrgcx(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrgcx ', &
                    & cur_loop, h_ratio, ii, iiposgcx(ii), ijposgcx(ii), zscerrgcx(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc,"(A4)") '===='
         ENDIF

      ENDIF

      DEALLOCATE(      &
         & zgcb_tlin,  &
         & zgcx_tlin,  &
         & zgcb_out ,  &
         & zgcx_out ,  &
         & zgcb_wop ,  &
         & zgcx_wop ,  &
         & zr          &
         & )
#else
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
      ! dummy routine
#endif
   END SUBROUTINE sol_sor_tlm_tst
#endif   
END MODULE solsor_tam
