MODULE traadv_cen2_tam
#if defined key_tam
   !!======================================================================
   !!                     ***  MODULE  traadv_cen2_tam  ***
   !! Ocean active tracers:  horizontal & vertical advective trend
   !!                         Tangent and Adjoint module
   !!======================================================================
   !! History of the direct module:   
   !!             8.2  !  01-08  (G. Madec, E. Durand)  trahad+trazad=traadv 
   !!             8.5  !  02-06  (G. Madec)  F90: Free form and module
   !!             9.0  !  04-08  (C. Talandier) New trends organization
   !!             " "  !  05-11  (V. Garnier) Surface pressure gradient organization
   !!             " "  !  06-04  (R. Benshila, G. Madec) Step reorganization
   !!             " "  !  06-07  (G. madec)  add ups_orca_set routine
   !! History of the T&A module
   !!             9.0  !  08-12  (A. Vidard) original version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_adv_cen2 : update the tracer trend with the horizontal and
   !!                  vertical advection trends using a seconder order
   !!   ups_orca_set : allow mixed upstream/centered scheme in specific
   !!                  area (set for orca 2 and 4 only)
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1,               &
      & jpiglo,              &
      & jp_cfg
   USE oce           , ONLY: & ! ocean dynamics and active tracers
      & un,                  &
      & vn,                  &
      & tb, sb, &
      & wn,                  &
      & tn,                  &
      & sn
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & un_tl,               &
      & vn_tl,               &
      & tb_tl, sb_tl, &
      & wn_tl,               &
      & tn_tl,               &
      & sn_tl,               &
      & ta_tl,               &
      & sa_tl,               &
      & tn_ad,               &
      & sn_ad,               &
      & ta_ad,               &
      & sa_ad 
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & tmask,               &
      & e1t,                 &
      & e2t,                 &
      & e2u,                 &
      & e1v,                 &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0,               &
#  else
      & e3t,                 &
#  endif
# endif
# if defined key_zco
      & e3t_0,               &
# else
      & e3u,                 &
      & e3v,                 &
# endif
      & lk_vvl,              &   
      & tmask,               &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product

!   USE sbc_oce         ! surface boundary condition: ocean
   USE dynspg_oce    , ONLY: & ! choice/control of key cpp for surface pressure gradient
      & lk_dynspg_rl
!   USE eosbn2          ! equation of state
!   USE trdmod          ! ocean active tracers trends 
!   USE closea          ! closed sea
!   USE trabbl          ! advective term in the BBL
!   USE sbcmod          ! surface Boundary Condition
!   USE sbcrnf          ! river runoffs
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,                 &
      & numout,              & 
      & nitend,              & 
      & nit000
!   USE lib_mpp
!   USE lbclnk          ! ocean lateral boundary condition (or mpp link)
!   USE diaptr          ! poleward transport diagnostics
!   USE prtctl          ! Print control
   USE zdf_oce       , ONLY: & ! ocean vertical physics
      & avmb,                &
      & avtb
   USE tstool_tam    , ONLY: &
      & prntst_adj,          &    !
      & prntst_tlm,          &    !
      & stdu,                &    ! stdev for u-velocity
      & stdv,                &    ! stdev for v-velocity
      & stdw,                &    ! stdev for w-velocity
      & stdt,                &    ! stdev for temperature
      & stds                      !           salinity
   USE paresp        , ONLY: &
      & wesp_t,              &
      & wesp_s

   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_adv_cen2_tan    ! routine called by traadv_tam.F90
   PUBLIC   tra_adv_cen2_adj    ! routine called by traadv_tam.F90
   PUBLIC   tra_adv_cen2_adj_tst! routine called by tst.F90
#if defined key_tst_tlm
   PUBLIC   tra_adv_cen2_tlm_tst! routine called by tamtst.F90
#endif

   REAL(wp), DIMENSION(jpi,jpj) ::  &
     & btr2   ! inverse of T-point surface [1/(e1t*e2t)]

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2006) 
   !! $Id: traadv_cen2.F90 1201 2008-09-24 13:24:21Z rblod $
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE tra_adv_cen2_tan( kt, pun, pvn, pwn, pun_tl, pvn_tl, pwn_tl )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_cen2_tan  ***
      !!                 
      !! ** Purpose of the direct routine:
      !!      Compute the now trend due to the advection of tracers
      !!      and add it to the general trend of passive tracer equations.
      !!
      !! ** Method  :   The advection is evaluated by a second order centered
      !!      scheme using now fields (leap-frog scheme). In specific areas
      !!      (vicinity of major river mouths, some straits, or where tn is
      !!      approaching the freezing point) it is mixed with an upstream
      !!      scheme for stability reasons.
      !!         Part 0 : compute the upstream / centered flag
      !!                  (3D array, zind, defined at T-point (0<zind<1))
      !!         Part I : horizontal advection
      !!       * centered flux:
      !!               zcenu = e2u*e3u  un  mi(tn)
      !!               zcenv = e1v*e3v  vn  mj(tn)
      !!       * upstream flux:
      !!               zupsu = e2u*e3u  un  (tb(i) or tb(i-1) ) [un>0 or <0]
      !!               zupsv = e1v*e3v  vn  (tb(j) or tb(j-1) ) [vn>0 or <0]
      !!       * mixed upstream / centered horizontal advection scheme
      !!               zcofi = max(zind(i+1), zind(i))
      !!               zcofj = max(zind(j+1), zind(j))
      !!               zwx = zcofi * zupsu + (1-zcofi) * zcenu
      !!               zwy = zcofj * zupsv + (1-zcofj) * zcenv
      !!       * horizontal advective trend (divergence of the fluxes)
      !!               zta = 1/(e1t*e2t*e3t) { di-1[zwx] + dj-1[zwy] }
      !!       * Add this trend now to the general trend of tracer (ta,sa):
      !!              (ta,sa) = (ta,sa) + ( zta , zsa )
      !!       * trend diagnostic ('key_trdtra' defined): the trend is
      !!      saved for diagnostics. The trends saved is expressed as
      !!      Uh.gradh(T), i.e.
      !!                     save trend = zta + tn divn
      !!         In addition, the advective trend in the two horizontal direc-
      !!      tion is also re-computed as Uh gradh(T). Indeed hadt+tn divn is
      !!      equal to (in s-coordinates, and similarly in z-coord.):
      !!         zta+tn*divn=1/(e1t*e2t*e3t) { mi-1( e2u*e3u  un  di[tn] )
      !!                                      +mj-1( e1v*e3v  vn  mj[tn] )  }
      !!         NB:in z-coordinate - full step (ln_zco=T) e3u=e3v=e3t, so
      !!      they vanish from the expression of the flux and divergence.
      !!
      !!         Part II : vertical advection
      !!      For temperature (idem for salinity) the advective trend is com-
      !!      puted as follows :
      !!            zta = 1/e3t dk+1[ zwz ]
      !!      where the vertical advective flux, zwz, is given by :
      !!            zwz = zcofk * zupst + (1-zcofk) * zcent
      !!      with
      !!        zupsv = upstream flux = wn * (tb(k) or tb(k-1) ) [wn>0 or <0]
      !!        zcenu = centered flux = wn * mk(tn)
      !!         The surface boundary condition is :
      !!      rigid-lid (lk_dynspg_frd = T) : zero advective flux
      !!      free-surf (lk_dynspg_fsc = T) : wn(:,:,1) * tn(:,:,1)
      !!         Add this trend now to the general trend of tracer (ta,sa):
      !!            (ta,sa) = (ta,sa) + ( zta , zsa )
      !!         Trend diagnostic ('key_trdtra' defined): the trend is
      !!      saved for diagnostics. The trends saved is expressed as :
      !!             save trend =  w.gradz(T) = zta - tn divn.
      !!
      !! ** Action :  - update (ta,sa) with the now advective tracer trends
      !!              - save trends in (ztrdt,ztrds) ('key_trdtra')
      !!----------------------------------------------------------------------
      USE oce_tam, ONLY :   zwxtl => ua_tl   ! use ua as workspace
      USE oce_tam, ONLY :   zwytl => va_tl   ! use va as workspace
      !!
      INTEGER , INTENT(in)                         ::   kt    ! ocean time-step index
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pun_tl   ! ocean velocity u-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pvn_tl   ! ocean velocity v-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pwn_tl   ! ocean velocity w-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pun      ! ocean velocity u-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pvn      ! ocean velocity v-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pwn      ! ocean velocity w-component
      !!
      INTEGER  ::   ji, jj, jk                           ! dummy loop indices
      REAL(wp) ::   ztatl, zsatl, zbtr, zhw, zhwtl,   &  ! temporary scalars
         &          ze3tr, zfui  , zfuitl  ,          &  !    "         "
         &          zfvj  , zfvjtl                       !    "         "
      REAL(wp) ::   zice                                 !    -         -
      REAL(wp), DIMENSION(jpi,jpj)     ::   ztfreez            ! 2D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zwwtl, zwztl       ! 3D workspace
      !!----------------------------------------------------------------------
      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_adv_cen2_tam : 2nd order centered advection scheme'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~   Vector optimization case'
         IF(lwp) WRITE(numout,*)
         !
         !   
         btr2(:,:) = 1. / ( e1t(:,:) * e2t(:,:) )        ! inverse of T-point surface
            IF ( jp_cfg == 2 ) THEN
            !   Increase the background in the surface layers
               avmb(1) = 10.  * avmb(1)      ;      avtb(1) = 10.  * avtb(1)
               avmb(2) = 10.  * avmb(2)      ;      avtb(2) = 10.  * avtb(2)
               avmb(3) =  5.  * avmb(3)      ;      avtb(3) =  5.  * avtb(3)
               avmb(4) =  2.5 * avmb(4)      ;      avtb(4) =  2.5 * avtb(4)
            ENDIF
      ENDIF

      ! I. Horizontal advective fluxes
      ! ------------------------------
      !  Second order centered tracer flux at u and v-points
      ! -----------------------------------------------------
      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               ! volume fluxes * 1/2
#if defined key_zco
               zfuitl = 0.5 * e2u(ji,jj) * pun_tl(ji,jj,jk)
               zfvjtl = 0.5 * e1v(ji,jj) * pvn_tl(ji,jj,jk)
               zfui = 0.5 * e2u(ji,jj) * pun(ji,jj,jk)
               zfvj = 0.5 * e1v(ji,jj) * pvn(ji,jj,jk)
#else
               zfuitl = 0.5 * e2u(ji,jj) * fse3u(ji,jj,jk) * pun_tl(ji,jj,jk)
               zfvjtl = 0.5 * e1v(ji,jj) * fse3v(ji,jj,jk) * pvn_tl(ji,jj,jk)
               zfui = 0.5 * e2u(ji,jj) * fse3u(ji,jj,jk) * pun(ji,jj,jk)
               zfvj = 0.5 * e1v(ji,jj) * fse3v(ji,jj,jk) * pvn(ji,jj,jk)
#endif
               ! centered scheme
               zwxtl(ji,jj,jk) = zfuitl * ( tn(ji,jj,jk) + tn(ji+1,jj,jk) ) &
                  &            + zfui * ( tn_tl(ji,jj,jk) + tn_tl(ji+1,jj,jk) )
               zwytl(ji,jj,jk) = zfvjtl * ( tn(ji,jj,jk) + tn(ji,jj+1,jk) ) &
                  &            + zfvj * ( tn_tl(ji,jj,jk) + tn_tl(ji,jj+1,jk) )
               zwwtl(ji,jj,jk) = zfuitl * ( sn(ji,jj,jk) + sn(ji+1,jj,jk) ) &
                  &            + zfui * ( sn_tl(ji,jj,jk) + sn_tl(ji+1,jj,jk) )
               zwztl(ji,jj,jk) = zfvjtl * ( sn(ji,jj,jk) + sn(ji,jj+1,jk) ) &
                  &            + zfvj * ( sn_tl(ji,jj,jk) + sn_tl(ji,jj+1,jk) )
            END DO
         END DO

         !  Tracer flux divergence at t-point added to the general trend
         ! --------------------------------------------------------------
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
#if defined key_zco
               zbtr = btr2(ji,jj)
#else
               zbtr = btr2(ji,jj) / fse3t(ji,jj,jk)
#endif
               ! horizontal advective trends
               ztatl = - zbtr * (  zwxtl(ji,jj,jk) - zwxtl(ji-1,jj  ,jk)   &
                  &            + zwytl(ji,jj,jk) - zwytl(ji  ,jj-1,jk)  )
               zsatl = - zbtr * (  zwwtl(ji,jj,jk) - zwwtl(ji-1,jj  ,jk)   &
                  &            + zwztl(ji,jj,jk) - zwztl(ji  ,jj-1,jk)  )
               ! add it to the general tracer trends
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) + zsatl
            END DO
         END DO
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============

      ! "zonal" mean advective heat and salt transport 
      ! ----------------------------------------------

!      IF( ln_diaptr .AND. ( MOD( kt, nf_ptr ) == 0 ) ) THEN
!         IF( lk_zco ) THEN
!            DO jk = 1, jpkm1
!               DO jj = 2, jpjm1
!                  DO ji = fs_2, fs_jpim1   ! vector opt.
!                    zwytl(ji,jj,jk) = zwytl(ji,jj,jk) * fse3v(ji,jj,jk)
!                    zwztl(ji,jj,jk) = zwztl(ji,jj,jk) * fse3v(ji,jj,jk)
!                  END DO
!               END DO
!            END DO
!         ENDIF
!         pht_adv(:) = ptr_vj( zwy(:,:,:) )
!         pst_adv(:) = ptr_vj( zwz(:,:,:) )
!      ENDIF

      ! II. Vertical advection
      ! ----------------------

      ! Bottom value : flux set to zero
      zwxtl(:,:,jpk) = 0.e0     ;    zwytl(:,:,jpk) = 0.e0

      ! Surface value
      IF( lk_dynspg_rl .OR. lk_vvl ) THEN
         ! rigid lid or variable volume: flux set to zero
         zwxtl(:,:, 1 ) = 0.e0    ;    zwytl(:,:, 1 ) = 0.e0
      ELSE
         ! free surface
         zwxtl(:,:, 1 ) = pwn_tl(:,:,1) * tn(:,:,1) + pwn(:,:,1) * tn_tl(:,:,1)
         zwytl(:,:, 1 ) = pwn_tl(:,:,1) * sn(:,:,1) + pwn(:,:,1) * sn_tl(:,:,1)
      ENDIF

      ! 1. Vertical advective fluxes 
      ! ----------------------------
      ! Second order centered tracer flux at w-point
      DO jk = 2, jpk
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ! velocity * 1/2
               zhwtl = 0.5 * pwn_tl(ji,jj,jk)
               zhw   = 0.5 * pwn(   ji,jj,jk)
               ! centered scheme
               zwxtl(ji,jj,jk) = zhwtl * ( tn( ji,jj,jk) + tn(   ji,jj,jk-1) ) &
                  &            + zhw * ( tn_tl(ji,jj,jk) + tn_tl(ji,jj,jk-1) )
               zwytl(ji,jj,jk) = zhwtl * ( sn( ji,jj,jk) + sn(   ji,jj,jk-1) ) &
                  &            + zhw * ( sn_tl(ji,jj,jk) + sn_tl(ji,jj,jk-1) )
            END DO
         END DO
      END DO

      ! 2. Tracer flux divergence at t-point added to the general trend
      ! -------------------------
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ze3tr = 1. / fse3t(ji,jj,jk)
               ! vertical advective trends
               ztatl = - ze3tr * ( zwxtl(ji,jj,jk) - zwxtl(ji,jj,jk+1) )
               zsatl = - ze3tr * ( zwytl(ji,jj,jk) - zwytl(ji,jj,jk+1) )
               ! add it to the general tracer trends
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) + zsatl
            END DO
         END DO
      END DO


      !
   END SUBROUTINE tra_adv_cen2_tan
   
   SUBROUTINE tra_adv_cen2_adj( kt, pun, pvn, pwn, pun_ad, pvn_ad, pwn_ad )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_cen2_adj  ***
      !!                 
      !! ** Purpose of the direct routine:
      !!      Compute the now trend due to the advection of tracers
      !!      and add it to the general trend of passive tracer equations.
      !!
      !! ** Method  :   The advection is evaluated by a second order centered
      !!      scheme using now fields (leap-frog scheme). In specific areas
      !!      (vicinity of major river mouths, some straits, or where tn is
      !!      approaching the freezing point) it is mixed with an upstream
      !!      scheme for stability reasons.
      !!         Part 0 : compute the upstream / centered flag
      !!                  (3D array, zind, defined at T-point (0<zind<1))
      !!         Part I : horizontal advection
      !!       * centered flux:
      !!               zcenu = e2u*e3u  un  mi(tn)
      !!               zcenv = e1v*e3v  vn  mj(tn)
      !!       * upstream flux:
      !!               zupsu = e2u*e3u  un  (tb(i) or tb(i-1) ) [un>0 or <0]
      !!               zupsv = e1v*e3v  vn  (tb(j) or tb(j-1) ) [vn>0 or <0]
      !!       * mixed upstream / centered horizontal advection scheme
      !!               zcofi = max(zind(i+1), zind(i))
      !!               zcofj = max(zind(j+1), zind(j))
      !!               zwx = zcofi * zupsu + (1-zcofi) * zcenu
      !!               zwy = zcofj * zupsv + (1-zcofj) * zcenv
      !!       * horizontal advective trend (divergence of the fluxes)
      !!               zta = 1/(e1t*e2t*e3t) { di-1[zwx] + dj-1[zwy] }
      !!       * Add this trend now to the general trend of tracer (ta,sa):
      !!              (ta,sa) = (ta,sa) + ( zta , zsa )
      !!       * trend diagnostic ('key_trdtra' defined): the trend is
      !!      saved for diagnostics. The trends saved is expressed as
      !!      Uh.gradh(T), i.e.
      !!                     save trend = zta + tn divn
      !!         In addition, the advective trend in the two horizontal direc-
      !!      tion is also re-computed as Uh gradh(T). Indeed hadt+tn divn is
      !!      equal to (in s-coordinates, and similarly in z-coord.):
      !!         zta+tn*divn=1/(e1t*e2t*e3t) { mi-1( e2u*e3u  un  di[tn] )
      !!                                      +mj-1( e1v*e3v  vn  mj[tn] )  }
      !!         NB:in z-coordinate - full step (ln_zco=T) e3u=e3v=e3t, so
      !!      they vanish from the expression of the flux and divergence.
      !!
      !!         Part II : vertical advection
      !!      For temperature (idem for salinity) the advective trend is com-
      !!      puted as follows :
      !!            zta = 1/e3t dk+1[ zwz ]
      !!      where the vertical advective flux, zwz, is given by :
      !!            zwz = zcofk * zupst + (1-zcofk) * zcent
      !!      with
      !!        zupsv = upstream flux = wn * (tb(k) or tb(k-1) ) [wn>0 or <0]
      !!        zcenu = centered flux = wn * mk(tn)
      !!         The surface boundary condition is :
      !!      rigid-lid (lk_dynspg_frd = T) : zero advective flux
      !!      free-surf (lk_dynspg_fsc = T) : wn(:,:,1) * tn(:,:,1)
      !!         Add this trend now to the general trend of tracer (ta,sa):
      !!            (ta,sa) = (ta,sa) + ( zta , zsa )
      !!         Trend diagnostic ('key_trdtra' defined): the trend is
      !!      saved for diagnostics. The trends saved is expressed as :
      !!             save trend =  w.gradz(T) = zta - tn divn.
      !!
      !! ** Action :  - update (ta,sa) with the now advective tracer trends
      !!              - save trends in (ztrdt,ztrds) ('key_trdtra')
      !!----------------------------------------------------------------------
      USE oce_tam, ONLY :   zwxad => ua_ad   ! use ua as workspace
      USE oce_tam, ONLY :   zwyad => va_ad   ! use va as workspace
      !!
      INTEGER , INTENT(in)                         ::   kt    ! ocean time-step index
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pun_ad   ! ocean velocity u-component
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pvn_ad   ! ocean velocity v-component
      REAL(wp), INTENT(inout), DIMENSION(jpi,jpj,jpk) ::   pwn_ad   ! ocean velocity w-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pun      ! ocean velocity u-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pvn      ! ocean velocity v-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pwn      ! ocean velocity w-component
      !!
      INTEGER  ::   ji, jj, jk                           ! dummy loop indices
      REAL(wp) ::   ztaad, zsaad, zbtr, zhw, zhwad,   &  ! temporary scalars
         &          ze3tr, zfui  , zfuiad  ,          &  !    "         "
         &          zfvj  , zfvjad                       !    "         "
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zwwad, zwzad       ! 3D workspace
      !!----------------------------------------------------------------------
      ztaad = 0.0_wp  ;  zsaad = 0.0_wp  ;  zhwad = 0.0_wp  ;  zfuiad = 0.0_wp  ;  zfvjad = 0.0_wp
      zwxad(:,:,:) = 0.0_wp  ;  zwyad(:,:,:) = 0.0_wp  ;  zwwad(:,:,:) = 0.0_wp  ;  zwzad(:,:,:) = 0.0_wp

      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_adv_cen2_tam : 2nd order centered advection scheme'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~   Vector optimization case'
         IF(lwp) WRITE(numout,*)
         !
         !   
         btr2(:,:) = 1. / ( e1t(:,:) * e2t(:,:) )        ! inverse of T-point surface
            IF ( jp_cfg == 2 ) THEN
            !   Increase the background in the surface layers
               avmb(1) = 10.  * avmb(1)      ;      avtb(1) = 10.  * avtb(1)
               avmb(2) = 10.  * avmb(2)      ;      avtb(2) = 10.  * avtb(2)
               avmb(3) =  5.  * avmb(3)      ;      avtb(3) =  5.  * avtb(3)
               avmb(4) =  2.5 * avmb(4)      ;      avtb(4) =  2.5 * avtb(4)
            ENDIF
      ENDIF
      ! II. Vertical advection
      ! ----------------------
      ! 2. Tracer flux divergence at t-point added to the general trend
      ! -------------------------
      DO jk = jpkm1, 1, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ze3tr = 1. / fse3t(ji,jj,jk)
               ! add it to the general tracer trends
               ztaad = ta_ad(ji,jj,jk) + ztaad
               zsaad = sa_ad(ji,jj,jk) + zsaad
              ! vertical advective trends
               zwxad(ji,jj,jk)   = zwxad(ji,jj,jk)   - ze3tr * ztaad
               zwxad(ji,jj,jk+1) = zwxad(ji,jj,jk+1) + ze3tr * ztaad
               zwyad(ji,jj,jk)   = zwyad(ji,jj,jk)   - ze3tr * zsaad
               zwyad(ji,jj,jk+1) = zwyad(ji,jj,jk+1) + ze3tr * zsaad
               ztaad = 0.0_wp
               zsaad = 0.0_wp
            END DO
         END DO
      END DO
      ! 1. Vertical advective fluxes 
      ! ----------------------------
      ! Second order centered tracer flux at w-point
      DO jk = jpk, 2, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ! velocity * 1/2
               zhw   = 0.5 * pwn(   ji,jj,jk)
               ! centered scheme
               zhwad = zhwad + zwxad(ji,jj,jk) * ( tn( ji,jj,jk) + tn(   ji,jj,jk-1) )
               tn_ad(ji,jj,jk)   = tn_ad(ji,jj,jk)   + zwxad(ji,jj,jk) * zhw
               tn_ad(ji,jj,jk-1) = tn_ad(ji,jj,jk-1) + zwxad(ji,jj,jk) * zhw
               zwxad(ji,jj,jk) = 0.0_wp
               zhwad = zhwad + zwyad(ji,jj,jk) * ( sn( ji,jj,jk) + sn(   ji,jj,jk-1) )
               sn_ad(ji,jj,jk)   = sn_ad(ji,jj,jk)   + zwyad(ji,jj,jk) * zhw
               sn_ad(ji,jj,jk-1) = sn_ad(ji,jj,jk-1) + zwyad(ji,jj,jk) * zhw
               zwyad(ji,jj,jk) = 0.0_wp
               ! velocity * 1/2
               pwn_ad(ji,jj,jk) = pwn_ad(ji,jj,jk) + 0.5 * zhwad
               zhwad = 0.0_wp
           END DO
         END DO
      END DO
      ! Surface value
      IF( lk_dynspg_rl .OR. lk_vvl ) THEN
         ! rigid lid or variable volume: flux set to zero
         zwxad(:,:, 1 ) = 0.0_wp    ;    zwyad(:,:, 1 ) = 0.0_wp
      ELSE
         ! free surface
         pwn_ad(:,:,1) = pwn_ad(:,:,1) + zwxad(:,:, 1 ) * tn(:,:,1) 
         tn_ad(:,:,1)  = tn_ad(:,:,1)  + zwxad(:,:, 1 ) * pwn(:,:,1)
         pwn_ad(:,:,1) = pwn_ad(:,:,1) + zwyad(:,:, 1 ) * sn(:,:,1) 
         sn_ad(:,:,1)  = sn_ad(:,:,1)  + zwyad(:,:, 1 ) * pwn(:,:,1)
         zwxad(:,:, 1 ) = 0.0_wp
         zwyad(:,:, 1 ) = 0.0_wp
      ENDIF

      ! Bottom value : flux set to zero
      zwxad(:,:,jpk) = 0.0_wp     ;    zwyad(:,:,jpk) = 0.0_wp
      ! "zonal" mean advective heat and salt transport 
      ! ----------------------------------------------

!      IF( ln_diaptr .AND. ( MOD( kt, nf_ptr ) == 0 ) ) THEN
!         IF( lk_zco ) THEN
!            DO jk = 1, jpkm1
!               DO jj = 2, jpjm1
!                  DO ji = fs_2, fs_jpim1   ! vector opt.
!                    zwyad(ji,jj,jk) = zwyad(ji,jj,jk) * fse3v(ji,jj,jk)
!                    zwzad(ji,jj,jk) = zwzad(ji,jj,jk) * fse3v(ji,jj,jk)
!                  END DO
!               END DO
!            END DO
!         ENDIF
!         pht_adv(:) = ptr_vj( zwy(:,:,:) )
!         pst_adv(:) = ptr_vj( zwz(:,:,:) )
!      ENDIF
!
      ! I. Horizontal advective fluxes
      ! ------------------------------
      !  Second order centered tracer flux at u and v-points
      ! -----------------------------------------------------
      !                                                ! ===============
      DO jk = 1, jpkm1                                 ! Horizontal slab
         !                                             ! ===============
        !  Tracer flux divergence at t-point added to the general trend
         ! --------------------------------------------------------------
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.
#if defined key_zco
               zbtr = btr2(ji,jj)
#else
               zbtr = btr2(ji,jj) / fse3t(ji,jj,jk)
#endif
               ! add it to the general tracer trends
               ztaad = ta_ad(ji,jj,jk) + ztaad
               zsaad = sa_ad(ji,jj,jk) + zsaad
               ! horizontal advective trends
               zwxad(ji,jj,jk) = zwxad(ji,jj,jk) - zbtr * ztaad
               zwxad(ji-1,jj,jk) = zwxad(ji-1,jj,jk) + zbtr * ztaad
               zwyad(ji,jj,jk) =  zwyad(ji,jj,jk) - zbtr * ztaad
               zwyad(ji  ,jj-1,jk) = zwyad(ji  ,jj-1,jk) + zbtr * ztaad
               ztaad = 0.0_wp
               zwwad(ji,jj,jk) =  zwwad(ji,jj,jk) - zsaad * zbtr
               zwwad(ji-1,jj  ,jk) = zwwad(ji-1,jj  ,jk) + zsaad * zbtr
               zwzad(ji,jj,jk) = zwzad(ji,jj,jk) - zsaad * zbtr
               zwzad(ji  ,jj-1,jk) = zwzad(ji  ,jj-1,jk) + zsaad * zbtr
               zsaad = 0.0_wp
            END DO
         END DO
         DO jj = jpjm1, 1, -1
            DO ji = fs_jpim1, 1, -1   ! vector opt.
               ! volume fluxes * 1/2
#if defined key_zco
               zfui = 0.5 * e2u(ji,jj) * pun(ji,jj,jk)
               zfvj = 0.5 * e1v(ji,jj) * pvn(ji,jj,jk)
#else
               zfui = 0.5 * e2u(ji,jj) * fse3u(ji,jj,jk) * pun(ji,jj,jk)
               zfvj = 0.5 * e1v(ji,jj) * fse3v(ji,jj,jk) * pvn(ji,jj,jk)
#endif
               ! centered scheme
               zfuiad = zfuiad + zwxad(ji,jj,jk) * ( tn(ji,jj,jk) + tn(ji+1,jj,jk) )
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zwxad(ji,jj,jk) * zfui
               tn_ad(ji+1,jj,jk) = tn_ad(ji+1,jj,jk) + zwxad(ji,jj,jk) * zfui
               zwxad(ji,jj,jk) = 0.0_wp
               zfvjad = zfvjad + zwyad(ji,jj,jk) * ( tn(ji,jj,jk) + tn(ji,jj+1,jk) )
               tn_ad(ji,jj,jk) = tn_ad(ji,jj,jk) + zwyad(ji,jj,jk) * zfvj
               tn_ad(ji,jj+1,jk) = tn_ad(ji,jj+1,jk) + zwyad(ji,jj,jk) * zfvj
               zwyad(ji,jj,jk) = 0.0_wp
               zfuiad = zfuiad + zwwad(ji,jj,jk) * ( sn(ji,jj,jk) + sn(ji+1,jj,jk) ) 
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zwwad(ji,jj,jk) * zfui
               sn_ad(ji+1,jj,jk) = sn_ad(ji+1,jj,jk) + zwwad(ji,jj,jk) * zfui
               zwwad(ji,jj,jk) = 0.0_wp
               zfvjad = zfvjad + zwzad(ji,jj,jk) * ( sn(ji,jj,jk) + sn(ji,jj+1,jk) )
               sn_ad(ji,jj,jk) = sn_ad(ji,jj,jk) + zwzad(ji,jj,jk) * zfvj
               sn_ad(ji,jj+1,jk) = sn_ad(ji,jj+1,jk) + zwzad(ji,jj,jk) * zfvj
               zwzad(ji,jj,jk) = 0.0_wp
#if defined key_zco
               pun_ad(ji,jj,jk) = pun_ad(ji,jj,jk) + 0.5 * e2u(ji,jj) * zfuiad
               pvn_ad(ji,jj,jk) = pvn_ad(ji,jj,jk) + 0.5 * e1v(ji,jj) * zfvjad
               zfuiad = 0.0_wp
               zfvjad = 0.0_wp
#else
               pun_ad(ji,jj,jk) = pun_ad(ji,jj,jk) + 0.5 * e2u(ji,jj) * fse3u(ji,jj,jk) * zfuiad
               pvn_ad(ji,jj,jk) = pvn_ad(ji,jj,jk) + 0.5 * e1v(ji,jj) * fse3v(ji,jj,jk) * zfvjad
               zfuiad = 0.0_wp
               zfvjad = 0.0_wp
#endif
             END DO
         END DO
         !                                             ! ===============
      END DO                                           !   End of slab
      !                                                ! ===============


      !
   END SUBROUTINE tra_adv_cen2_adj
SUBROUTINE tra_adv_cen2_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_adv_cen2_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp2            ! scalar product involving the adjoint routine
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zun_tlin ,     & ! Tangent input
         & zvn_tlin ,     & ! Tangent input
         & zwn_tlin ,     & ! Tangent input
         & ztn_tlin ,     & ! Tangent input
         & zsn_tlin ,     & ! Tangent input
         & zta_tlin ,     & ! Tangent input
         & zsa_tlin ,     & ! Tangent input
         & zta_tlout,     & ! Tangent output
         & zsa_tlout,     & ! Tangent output
         & zta_adin ,     & ! Adjoint input
         & zsa_adin ,     & ! Adjoint input
         & zun_adout,     & ! Adjoint output
         & zvn_adout,     & ! Adjoint output
         & zwn_adout,     & ! Adjoint output
         & ztn_adout,     & ! Adjoint output
         & zsn_adout,     & ! Adjoint output
         & zta_adout,     & ! Adjoint output
         & zsa_adout,     & ! Adjoint output
         & zr             ! 3D random field 
      CHARACTER(LEN=14) ::&
         & cl_name
      ! Allocate memory

      ALLOCATE( &
         & zun_tlin( jpi,jpj,jpk),     &
         & zvn_tlin( jpi,jpj,jpk),     &
         & zwn_tlin( jpi,jpj,jpk),     &
         & ztn_tlin( jpi,jpj,jpk),     &
         & zsn_tlin( jpi,jpj,jpk),     &
         & zta_tlin( jpi,jpj,jpk),     &
         & zsa_tlin( jpi,jpj,jpk),     &
         & zta_tlout(jpi,jpj,jpk),     &
         & zsa_tlout(jpi,jpj,jpk),     &
         & zta_adin( jpi,jpj,jpk),     &
         & zsa_adin( jpi,jpj,jpk),     &
         & zun_adout(jpi,jpj,jpk),     &
         & zvn_adout(jpi,jpj,jpk),     &
         & zwn_adout(jpi,jpj,jpk),     &
         & ztn_adout(jpi,jpj,jpk),     &
         & zsn_adout(jpi,jpj,jpk),     &
         & zta_adout(jpi,jpj,jpk),     &
         & zsa_adout(jpi,jpj,jpk),     &
         & zr(       jpi,jpj,jpk)      &
         & )
      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zun_tlin( :,:,:) = 0.0_wp
      zvn_tlin( :,:,:) = 0.0_wp
      zwn_tlin( :,:,:) = 0.0_wp
      ztn_tlin( :,:,:) = 0.0_wp
      zsn_tlin( :,:,:) = 0.0_wp
      zta_tlin( :,:,:) = 0.0_wp
      zsa_tlin( :,:,:) = 0.0_wp
      zta_tlout(:,:,:) = 0.0_wp
      zsa_tlout(:,:,:) = 0.0_wp
      zta_adin( :,:,:) = 0.0_wp
      zsa_adin( :,:,:) = 0.0_wp
      zun_adout(:,:,:) = 0.0_wp
      zvn_adout(:,:,:) = 0.0_wp
      zwn_adout(:,:,:) = 0.0_wp
      ztn_adout(:,:,:) = 0.0_wp
      zsn_adout(:,:,:) = 0.0_wp
      zta_adout(:,:,:) = 0.0_wp
      zsa_adout(:,:,:) = 0.0_wp
      zr(       :,:,:) = 0.0_wp

      tn_ad(:,:,:) = 0.0_wp
      sn_ad(:,:,:) = 0.0_wp 
      
       
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'U', 0.0_wp, stdu )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zun_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 371836 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'V', 0.0_wp, stdv ) 
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zvn_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 148379 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'W', 0.0_wp, stdw )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zwn_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 264940 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              ztn_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 618304 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds ) 
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsn_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 481903 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 263871 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds ) 
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsa_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

      tn_tl(:,:,:) = ztn_tlin(:,:,:)
      sn_tl(:,:,:) = zsn_tlin(:,:,:)
      ta_tl(:,:,:) = zta_tlin(:,:,:)
      sa_tl(:,:,:) = zsa_tlin(:,:,:)

      CALL tra_adv_cen2_tan(nit000, un, vn, wn, zun_tlin, zvn_tlin, zwn_tlin) 

      zta_tlout(:,:,:) = ta_tl(:,:,:) 
      zsa_tlout(:,:,:) = sa_tl(:,:,:) 

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_adin(ji,jj,jk) = zta_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_t(jk)
              zsa_adin(ji,jj,jk) = zsa_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_s(jk)
            END DO
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1 = DOT_PRODUCT( zta_tlout, zta_adin ) &
         & + DOT_PRODUCT( zsa_tlout, zsa_adin )

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------
      ta_ad(:,:,:) = zta_adin(:,:,:)
      sa_ad(:,:,:) = zsa_adin(:,:,:)

      CALL tra_adv_cen2_adj(nit000, un, vn, wn, zun_adout, zvn_adout, zwn_adout)

      ztn_adout(:,:,:) = tn_ad(:,:,:)
      zsn_adout(:,:,:) = sn_ad(:,:,:)
      zta_adout(:,:,:) = ta_ad(:,:,:)
      zsa_adout(:,:,:) = sa_ad(:,:,:)


      zsp2 = DOT_PRODUCT( zun_tlin, zun_adout ) &
         & + DOT_PRODUCT( zvn_tlin, zvn_adout ) &
         & + DOT_PRODUCT( zwn_tlin, zwn_adout ) &
         & + DOT_PRODUCT( ztn_tlin, ztn_adout ) &
         & + DOT_PRODUCT( zsn_tlin, zsn_adout ) &
         & + DOT_PRODUCT( zta_tlin, zta_adout ) &
         & + DOT_PRODUCT( zsa_tlin, zsa_adout )

      ! 14 char:'12345678901234'
      cl_name = 'tra_adv_cen2  '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE(         &
         & zun_tlin ,     & ! Tangent input
         & zvn_tlin ,     & ! Tangent input
         & zwn_tlin ,     & ! Tangent input
         & ztn_tlin ,     & ! Tangent input
         & zsn_tlin ,     & ! Tangent input
         & zta_tlin ,     & ! Tangent input
         & zsa_tlin ,     & ! Tangent input
         & zta_tlout,     & ! Tangent output
         & zsa_tlout,     & ! Tangent output
         & zta_adin ,     & ! Adjoint input
         & zsa_adin ,     & ! Adjoint input
         & zun_adout,     & ! Adjoint output
         & zvn_adout,     & ! Adjoint output
         & zwn_adout,     & ! Adjoint output
         & ztn_adout,     & ! Adjoint output
         & zsn_adout,     & ! Adjoint output
         & zta_adout,     & ! Adjoint output
         & zsa_adout,     & ! Adjoint output
         & zr             & ! 3D random field 
         & )



   END SUBROUTINE tra_adv_cen2_adj_tst
#if defined key_tst_tlm
SUBROUTINE tra_adv_cen2_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_adv_cen2_tlm_tst ***
      !!
      !! ** Purpose : Test the tangent routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!   
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!                  
      !! History :
      !!        ! 09-08 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE traadv_cen2         ! horizontal & vertical advective trend
      USE tamtrj              ! writing out state trajectory
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE wzvmod             !  vertical velocity
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
        & tb, sb, tn, sn, ta,  &
        & sa, gtu, gsu, gtv,   &
        & gsv
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_Ta,      &
         & zsp1_Sa,      &
         & zsp2,         & ! scalar product involving the tangent routine
         & zsp2_Ta,      &
         & zsp2_Sa,      &
         & zsp3,         & ! scalar product involving the tangent routine
         & zsp3_Ta,      &
         & zsp3_Sa,      &
         & zzsp,         & ! scalar product involving the tangent routine
         & zzsp_Ta,      &
         & zzsp_Sa,      & 
         & gamma,            &
         & zgsp1,            &
         & zgsp2,            &
         & zgsp3,            &
         & zgsp4,            &
         & zgsp5,            &
         & zgsp6,            &
         & zgsp7 
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zun_tlin,      & ! Tangent input
         & zvn_tlin,      & ! Tangent input
         & zwn_tlin,      & ! Tangent input
         & ztn_tlin,      & ! Tangent input
         & zsn_tlin,      & ! Tangent input
         & zta_tlin,      & ! Tangent input
         & zsa_tlin,      & ! Tangent input
         & zta_out ,      & ! Direct output
         & zsa_out ,      & ! Direct output
         & zta_wop ,      & ! Direct output w/o perturbation
         & zsa_wop ,      & ! Direct output w/o perturbation
         & z3r
      CHARACTER(LEN=14)   :: cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_xdx
      CHARACTER (LEN=90)  :: FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zscta,zscsa,               &
         & zscerrta, zscerrsa
      INTEGER, DIMENSION(100)::           &
         & iiposta, iipossa,              &
         & ijposta, ijpossa,              & 
         & ikposta, ikpossa
      INTEGER::             &
         & ii,              &
         & isamp=40,        &
         & jsamp=40,        &
         & ksamp=10,        &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj,jpk) :: &
         & zerrta, zerrsa   
      ! Allocate memory

      ALLOCATE( &
         & zun_tlin( jpi,jpj,jpk),     &
         & zvn_tlin( jpi,jpj,jpk),     &
         & zwn_tlin( jpi,jpj,jpk),     &
         & ztn_tlin( jpi,jpj,jpk),     &
         & zsn_tlin( jpi,jpj,jpk),     &
         & zta_tlin( jpi,jpj,jpk),     &
         & zsa_tlin( jpi,jpj,jpk),     &
         & zta_out ( jpi,jpj,jpk),     &
         & zsa_out ( jpi,jpj,jpk),     &
         & zta_wop ( jpi,jpj,jpk),     &
         & zsa_wop ( jpi,jpj,jpk),     &
         & z3r     ( jpi,jpj,jpk)      &
         & )

      !--------------------------------------------------------------------
      ! Reset variables
      !--------------------------------------------------------------------
      zun_tlin( :,:,:) = 0.0_wp
      zvn_tlin( :,:,:) = 0.0_wp
      zwn_tlin( :,:,:) = 0.0_wp
      ztn_tlin( :,:,:) = 0.0_wp
      zsn_tlin( :,:,:) = 0.0_wp
      zta_tlin( :,:,:) = 0.0_wp
      zsa_tlin( :,:,:) = 0.0_wp
      zta_out ( :,:,:) = 0.0_wp
      zsa_out ( :,:,:) = 0.0_wp
      zta_wop ( :,:,:) = 0.0_wp
      zsa_wop ( :,:,:) = 0.0_wp

      zscta(:)         = 0.0_wp
      zscsa(:)         = 0.0_wp
      zscerrta(:)      = 0.0_wp
      zscerrsa(:)      = 0.0_wp

      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio     
      file_wop='trj_wop_tradv_cen2'
      file_xdx='trj_xdx_tradv_cen2'   
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 596035, z3r,  'U', 0.0_wp, stdu)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zun_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 371836, z3r,  'V', 0.0_wp, stdv)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zvn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 148379, z3r,  'W', 0.0_wp, stdw)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zwn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 264940, z3r,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  ztn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 618304, z3r,  'T', 0.0_wp, stds) 
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 481903, z3r,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 263871, z3r,  'T', 0.0_wp, stds) 
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsa_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF       
      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      IF ( tlm_bch /= 2 ) CALL istate_p  

      ! *** initialize the reference trajectory
      ! ------------
      CALL  trj_rea( nit000-1, 1 )
      CALL  trj_rea( nit000, 1 )

      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         zun_tlin(:,:,:) = gamma * zun_tlin(:,:,:)
         un(:,:,:)       = un(:,:,:) + zun_tlin(:,:,:)

         zvn_tlin(:,:,:) = gamma * zvn_tlin(:,:,:)
         vn(:,:,:)       = vn(:,:,:) + zvn_tlin(:,:,:)

         zwn_tlin(:,:,:) = gamma * zwn_tlin(:,:,:)
         wn(:,:,:)       = wn(:,:,:) + zwn_tlin(:,:,:)

         ztn_tlin(:,:,:) = gamma * ztn_tlin(:,:,:)
         tn(:,:,:)       = tn(:,:,:) + ztn_tlin(:,:,:)

         zsn_tlin(:,:,:) = gamma * zsn_tlin(:,:,:)
         sn(:,:,:)       = sn(:,:,:) + zsn_tlin(:,:,:)

         zta_tlin(:,:,:) = gamma * zta_tlin(:,:,:)
         ta(:,:,:)       = ta(:,:,:) + zta_tlin(:,:,:)

         zsa_tlin(:,:,:) = gamma * zsa_tlin(:,:,:)
         sa(:,:,:)       = sa(:,:,:) + zsa_tlin(:,:,:)
      ENDIF 

      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      IF ( tlm_bch /= 2 ) CALL tra_adv_cen2(nit000, un, vn, wn)
      IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN       
         !--------------------------------------------------------------------
         ! Initialize the tangent variables: dy^* = W dy  
         !--------------------------------------------------------------------
         CALL  trj_rea( nit000-1, 1 ) 
         CALL  trj_rea( nit000, 1 ) 
         tn_tl  (:,:,:) = ztn_tlin  (:,:,:)
         sn_tl  (:,:,:) = zsn_tlin  (:,:,:)
         ta_tl  (:,:,:) = zta_tlin  (:,:,:)
         sa_tl  (:,:,:) = zsa_tlin  (:,:,:)

         !-----------------------------------------------------------------------
         !  Initialization of the dynamics and tracer fields for the tangent
         !----------------------------------------------------------------------- 
         CALL tra_adv_cen2_tan(nit000, un, vn, wn, zun_tlin, zvn_tlin, zwn_tlin)

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------
         zsp2_Ta    = DOT_PRODUCT( ta_tl, ta_tl  )
         zsp2_Sa    = DOT_PRODUCT( sa_tl, sa_tl  )

         zsp2      = zsp2_Ta + zsp2_Sa
         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------
         CALL trj_rd_spl(file_wop) 
         zta_wop  (:,:,:) = ta  (:,:,:)
         zsa_wop  (:,:,:) = sa  (:,:,:)
         CALL trj_rd_spl(file_xdx) 
         zta_out  (:,:,:) = ta  (:,:,:)
         zsa_out  (:,:,:) = sa  (:,:,:)
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zta_out   (ji,jj,jk) = zta_out    (ji,jj,jk) - zta_wop  (ji,jj,jk)
                  zta_wop   (ji,jj,jk) = zta_out    (ji,jj,jk) - ta_tl    (ji,jj,jk)
                  IF (  ta_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrta(ji,jj,jk) = zta_out(ji,jj,jk)/ta_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iiposta(ii) = ji
                      ijposta(ii) = jj
                      ikposta(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                         zscta (ii) =  zta_wop(ji,jj,jk)
                         zscerrta (ii) =  ( zerrta(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zsa_out   (ji,jj,jk) = zsa_out    (ji,jj,jk) - zsa_wop  (ji,jj,jk)
                  zsa_wop   (ji,jj,jk) = zsa_out    (ji,jj,jk) - sa_tl    (ji,jj,jk)
                  IF (  sa_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrsa(ji,jj,jk) = zsa_out(ji,jj,jk)/sa_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iipossa(ii) = ji
                      ijpossa(ii) = jj
                      ikpossa(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                         zscsa (ii) =  zsa_wop(ji,jj,jk)
                         zscerrsa (ii) =  ( zerrsa(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO

         zsp1_Ta    = DOT_PRODUCT( zta_out, zta_out  )
         zsp1_Sa    = DOT_PRODUCT( zsa_out, zsa_out  )

         zsp1      = zsp1_Ta + zsp1_Sa

         zsp3_Ta    = DOT_PRODUCT( zta_wop, zta_wop  )
         zsp3_Sa    = DOT_PRODUCT( zsa_wop, zsa_wop  )

         zsp3      = zsp3_Ta + zsp3_Sa

         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'traadv_tam:En '
         zzsp    = SQRT(zsp3)
         zzsp_Ta = SQRT(zsp3_Ta)
         zzsp_Sa = SQRT(zsp3_Sa)
         zgsp5   = zzsp
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp     = SQRT(zsp2)
         zzsp_Ta  = SQRT(zsp2_Ta)
         zzsp_Sa  = SQRT(zsp2_Sa)
         zgsp4    = zzsp
         cl_name  = 'traadv_tam:Ln2'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         zzsp     = sqrt(zsp1)
         zzsp_Ta  = sqrt(zsp1_Ta)
         zzsp_Sa  = sqrt(zsp1_Sa)

         cl_name = 'traadv:Mhdx-Mx'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         zgsp3    = SQRT( zsp3/zsp2 )
         zgsp7    = zgsp3/gamma
         zgsp1    = zzsp
         zgsp2    = zgsp1 / zgsp4
         zgsp6    = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'tadvcen2', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7
         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'tadvcen2'
         IF(lwp) THEN
            DO ii=1, 100, 1
               IF ( zscta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscta     ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscsa     ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscsa(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrta  ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscerrta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrsa  ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscerrsa(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc,"(A4)") '===='
         ENDIF

      ENDIF

      DEALLOCATE(                           &
         & zun_tlin, zvn_tlin,              &
         & zwn_tlin, ztn_tlin,              &
         & zsn_tlin,                        &
         & zta_tlin, zsa_tlin,              & 
         & zta_out, zsa_out,                & 
         & zta_wop, zsa_wop,                &
         & z3r                              &
         & )

  END SUBROUTINE  tra_adv_cen2_tlm_tst
#endif
#endif
   !!======================================================================
END MODULE traadv_cen2_tam
