MODULE traadv_tam
#if defined key_tam
   !!==============================================================================
   !!                       ***  MODULE  traadv_tam  ***
   !! Ocean active tracers:  advection trend - 
   !!                        Tangent and Adjoint Module
   !!==============================================================================
   !! History of the direct routine:
   !!            9.0  !  05-11  (G. Madec)  Original code
   !! History of the TAM:
   !!                 !  08-06  (A. Vidard) Skeleton
   !!                 !  09-03  (F. Vigilant) Add tra_adv_ctl_tam routine
   !!                                         Allow call to tra_eiv_tam/adj
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_adv      : compute ocean tracer advection trend
   !!   tra_adv_ctl  : control the different options of advection scheme
   !!----------------------------------------------------------------------
   USE par_kind         , ONLY: & ! Precision variables
      & wp
   USE par_oce          , ONLY: & ! Ocean space and time domain variables
      & jpi,                    &
      & jpj,                    & 
      & jpk   
   USE oce_tam          , ONLY: & ! ocean dynamics and active tracers
      & un_tl,                  &
      & vn_tl,                  &
      & wn_tl,                  &
      & un_ad,                  &
      & vn_ad,                  &
      & wn_ad
   USE oce              , ONLY: & ! ocean dynamics and active tracers
      & un,                     &
      & vn,                     &
      & wn
   USE ldftra_oce       , ONLY: &
      & lk_traldf_eiv
   USE trabbl           , ONLY: &
      & lk_trabbl_adv 
   USE dom_oce         ! ocean space and time domain
   USE traadv_cen2_tam, ONLY: & ! 2nd order centered scheme (tra_adv_cen2   routine)
      & tra_adv_cen2_tan,     &
      & tra_adv_cen2_adj,     &
#if defined key_tst_tlm
      & tra_adv_cen2_tlm_tst, &
#endif
      & tra_adv_cen2_adj_tst
   USE traadv_eiv_tam, ONLY: & ! advection trend - eddy induced velocity (tra_adv_eiv   routine)
      & tra_adv_eiv_tan,     &
      & tra_adv_eiv_adj
   USE in_out_manager  ! I/O manager
   USE prtctl          ! Print control
 
   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_adv_tan     ! routine called by steptan module
   PUBLIC   tra_adv_adj     ! routine called by stepadj module
   PUBLIC   tra_adv_ctl_tam ! routine called by stepadj module
   PUBLIC   tra_adv_adj_tst ! routine called by tst module
#if defined key_tst_tlm
   PUBLIC   tra_adv_tlm_tst ! routine called by tst module
#endif
   !!* Namelist nam_traadv
   LOGICAL, PUBLIC ::   ln_traadv_cen2   = .TRUE.       ! 2nd order centered scheme flag
   LOGICAL, PUBLIC ::   ln_traadv_tvd    = .FALSE.      ! TVD scheme flag
   LOGICAL, PUBLIC ::   ln_traadv_muscl  = .FALSE.      ! MUSCL scheme flag
   LOGICAL, PUBLIC ::   ln_traadv_muscl2 = .FALSE.      ! MUSCL2 scheme flag
   LOGICAL, PUBLIC ::   ln_traadv_ubs    = .FALSE.      ! UBS scheme flag
   LOGICAL, PUBLIC ::   ln_traadv_qck    = .FALSE.      ! QUICKEST scheme flag

   INTEGER ::   nadv   ! choice of the type of advection scheme

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE tra_adv_tan( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_tan  ***
      !!
      !! ** Purpose of the direct routine:   
      !!              compute the ocean tracer advection trend.
      !!
      !! ** Method  : - Update (ua,va) with the advection term following nadv
      !!----------------------------------------------------------------------
#if ( defined key_trabbl_adv || defined key_traldf_eiv )
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zuntl, zvntl, zwntl   ! effective velocity
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zun, zvn, zwn         ! effective velocity
#else
      USE oce_tam, ONLY :                       zuntl => un_tl       ! the effective velocity is the
      USE oce_tam, ONLY :                       zvntl => vn_tl       ! Eulerian velocity
      USE oce_tam, ONLY :                       zwntl => wn_tl       ! 
      USE oce, ONLY :                       zun => un       ! the effective velocity is the
      USE oce, ONLY :                       zvn => vn       ! Eulerian velocity
      USE oce, ONLY :                       zwn => wn       ! 
#endif
      !!
      INTEGER, INTENT( in ) ::   kt   ! ocean time-step index
      !!----------------------------------------------------------------------


      IF( kt == nit000 )   CALL tra_adv_ctl_tam          ! initialisation & control of options

#if defined key_trabbl_adv
      zuntl(:,:,:) = un_tl(:,:,:) - u_bbl_tl(:,:,:)          ! add the bbl velocity
      zvntl(:,:,:) = vn_tl(:,:,:) - v_bbl_tl(:,:,:)
      zwntl(:,:,:) = wn_tl(:,:,:) + w_bbl_tl(:,:,:)
      zun(:,:,:) = un(:,:,:) - u_bbl(:,:,:)          ! add the bbl velocity
      zvn(:,:,:) = vn(:,:,:) - v_bbl(:,:,:)
      zwn(:,:,:) = wn(:,:,:) + w_bbl(:,:,:)
#endif
      IF( lk_traldf_eiv ) THEN                       ! commpute and add the eiv velocity
         IF( .NOT. lk_trabbl_adv ) THEN 
            zuntl(:,:,:) = un_tl(:,:,:)
            zvntl(:,:,:) = vn_tl(:,:,:)
            zwntl(:,:,:) = wn_tl(:,:,:)
            zun(:,:,:) = un(:,:,:)
            zvn(:,:,:) = vn(:,:,:)
            zwn(:,:,:) = wn(:,:,:)
         ENDIF
         CALL tra_adv_eiv_tan( kt, zuntl, zvntl, zwntl ) 
      ENDIF
      IF(lwp) WRITE(numout,*) ' tra_adv_tam: 2nd order scheme is forced in TAM'
      nadv = 1 ! force tra_adv_cen2 for tangent
      
      SELECT CASE ( nadv )                           ! compute advection trend and add it to general trend
      CASE ( 1 ) ; 
         CALL tra_adv_cen2_tan ( kt, zun, zvn, zwn, zuntl, zvntl, zwntl )    ! 2nd order centered scheme
      END SELECT
      !
   END SUBROUTINE tra_adv_tan

   SUBROUTINE tra_adv_adj( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_adj  ***
      !!
      !! ** Purpose of the direct routine:   
      !!              compute the ocean tracer advection trend.
      !!
      !! ** Method  : - Update (ua,va) with the advection term following nadv
      !!----------------------------------------------------------------------
#if ( defined key_trabbl_adv || defined key_traldf_eiv )
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zunad, zvnad, zwnad   ! effective velocity
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   zun, zvn, zwn   ! effective velocity
#else
      USE oce_tam, ONLY :                   zunad => un_ad       ! the effective velocity is the
      USE oce_tam, ONLY :                   zvnad => vn_ad       ! Eulerian velocity
      USE oce_tam, ONLY :                   zwnad => wn_ad       ! 
      USE oce, ONLY :                       zun => un       ! the effective velocity is the
      USE oce, ONLY :                       zvn => vn       ! Eulerian velocity
      USE oce, ONLY :                       zwn => wn       ! 
#endif
      !!
      INTEGER, INTENT( in ) ::   kt   ! ocean time-step index
      !!----------------------------------------------------------------------

      IF( kt == nitend )   CALL tra_adv_ctl_tam

#if ( defined key_trabbl_adv || defined key_traldf_eiv )
      zunad(:,:,:) = 0.0_wp 
      zvnad(:,:,:) = 0.0_wp
      zwnad(:,:,:) = 0.0_wp
#endif
#if defined key_trabbl_adv
      zun(:,:,:) = un(:,:,:) - u_bbl(:,:,:)          ! add the bbl velocity
      zvn(:,:,:) = vn(:,:,:) - v_bbl(:,:,:)
      zwn(:,:,:) = wn(:,:,:) + w_bbl(:,:,:)
#endif
      IF( lk_traldf_eiv ) THEN                       ! commpute and add the eiv velocity
         IF( .NOT. lk_trabbl_adv ) THEN 
            zun(:,:,:) = un(:,:,:)
            zvn(:,:,:) = vn(:,:,:)
            zwn(:,:,:) = wn(:,:,:)
         ENDIF
      ENDIF

     IF(lwp) WRITE(numout,*) ' tra_adv_tam: 2nd order scheme is forced in TAM'
      nadv = 1 ! force tra_adv_cen2 for adjoint
      
      SELECT CASE ( nadv )                           ! compute advection trend and add it to general trend
      CASE ( 1 ) ; CALL tra_adv_cen2_adj ( kt, zun, zvn, zwn, zunad, zvnad, zwnad )    ! 2nd order centered scheme
      END SELECT
      !
      IF( lk_traldf_eiv ) THEN                       ! commpute and add the eiv velocity
         CALL tra_adv_eiv_adj( kt, zunad, zvnad, zwnad ) 
         IF( .NOT. lk_trabbl_adv ) THEN 
            un_ad(:,:,:) = un_ad(:,:,:) + zunad(:,:,:)
            zunad(:,:,:) = 0.0_wp
            vn_ad(:,:,:) = vn_ad(:,:,:) + zvnad(:,:,:)
            zvnad(:,:,:) = 0.0_wp
            wn_ad(:,:,:) = wn_ad(:,:,:) + zwnad(:,:,:)
            zwnad(:,:,:) = 0.0_wp
         ENDIF
      ENDIF
# if defined key_trabbl_adv
      un_ad(:,:,:)    = un_ad(:,:,:)    + zunad(:,:,:)
      u_bbl_ad(:,:,:) = u_bbl_ad(:,:,:) - zunad(:,:,:)
      zunad(:,:,:) = 0.0_wp
      vn_ad(:,:,:)    = vn_ad(:,:,:)    + zvnad(:,:,:)
      v_bbl_ad(:,:,:) = v_bbl_ad(:,:,:) - zvnad(:,:,:)
      zvnad(:,:,:) = 0.0_wp
      wn_ad(:,:,:)    = wn_ad(:,:,:)    + zwnad(:,:,:)
      w_bbl_ad(:,:,:) = w_bbl_ad(:,:,:) + zwnad(:,:,:)
      zwnad(:,:,:) = 0.0_wp
# endif
   END SUBROUTINE tra_adv_adj
SUBROUTINE tra_adv_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit

      CALL tra_adv_cen2_adj_tst( kumadt )

   END SUBROUTINE tra_adv_adj_tst

   SUBROUTINE tra_adv_ctl_tam
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_ctl_tam  ***
      !!                
      !! ** Purpose :   Control the consistency between namelist options for 
      !!              tracer advection schemes and set nadv
      !!----------------------------------------------------------------------
      INTEGER ::   ioptio

      NAMELIST/nam_traadv/ ln_traadv_cen2 , ln_traadv_tvd,    &
         &                 ln_traadv_muscl, ln_traadv_muscl2, &
         &                 ln_traadv_ubs  , ln_traadv_qck
      !!----------------------------------------------------------------------

!      REWIND ( numnam )               ! Read Namelist nam_traadv : tracer advection scheme
!      READ   ( numnam, nam_traadv )

      IF(lwp) THEN                    ! Namelist print
         WRITE(numout,*)
         WRITE(numout,*) 'tra_adv_ctl : choice/control of the tracer advection scheme'
         WRITE(numout,*) '~~~~~~~~~~~'
         WRITE(numout,*) '       Namelist nam_traadv : chose a advection scheme for tracers'
         WRITE(numout,*) '          2nd order advection scheme     ln_traadv_cen2   = ', ln_traadv_cen2
         WRITE(numout,*) '          TVD advection scheme           ln_traadv_tvd    = ', ln_traadv_tvd
         WRITE(numout,*) '          MUSCL  advection scheme        ln_traadv_muscl  = ', ln_traadv_muscl
         WRITE(numout,*) '          MUSCL2 advection scheme        ln_traadv_muscl2 = ', ln_traadv_muscl2
         WRITE(numout,*) '          UBS    advection scheme        ln_traadv_ubs    = ', ln_traadv_ubs
         WRITE(numout,*) '          QUICKEST advection scheme      ln_traadv_qck    = ', ln_traadv_qck
    ENDIF

      ioptio = 0                      ! Parameter control
      IF( ln_traadv_cen2   )   ioptio = ioptio + 1
      IF( ln_traadv_tvd    )   ioptio = ioptio + 1
      IF( ln_traadv_muscl  )   ioptio = ioptio + 1
      IF( ln_traadv_muscl2 )   ioptio = ioptio + 1
      IF( ln_traadv_ubs    )   ioptio = ioptio + 1
      IF( ln_traadv_qck    )   ioptio = ioptio + 1
      IF( lk_esopa         )   ioptio =          1

      IF( ioptio /= 1 )   CALL ctl_stop( 'Choose ONE advection scheme in namelist nam_traadv' )

      IF( n_cla == 1 .AND. .NOT. ln_traadv_cen2 )   &
         &                CALL ctl_stop( 'cross-land advection only with 2nd order advection scheme' )

      !                              ! Set nadv
      IF( ln_traadv_cen2   )   nadv =  1
      IF( ln_traadv_tvd    )   nadv =  2
      IF( ln_traadv_muscl  )   nadv =  3
      IF( ln_traadv_muscl2 )   nadv =  4
      IF( ln_traadv_ubs    )   nadv =  5
      IF( ln_traadv_qck    )   nadv =  6
      IF( lk_esopa         )   nadv = -1

      IF(lwp) THEN                   ! Print the choice
         WRITE(numout,*)
         IF( nadv ==  1 )   WRITE(numout,*) '         2nd order scheme is used'
         IF( nadv ==  2 )   WRITE(numout,*) '         TVD       Not Available in NEMO TAM'
         IF( nadv ==  3 )   WRITE(numout,*) '         MUSCL     Not Available in NEMO TAM'
         IF( nadv ==  4 )   WRITE(numout,*) '         MUSCL2    Not Available in NEMO TAM'
         IF( nadv ==  5 )   WRITE(numout,*) '         UBS       Not Available in NEMO TAM'
         IF( nadv ==  6 )   WRITE(numout,*) '         QUICKEST  Not Available in NEMO TAM'
         IF( nadv == -1 )   WRITE(numout,*) '         esopa test: Not Available in NEMO TAM'
      ENDIF
      !
   END SUBROUTINE tra_adv_ctl_tam
#if defined key_tst_tlm
SUBROUTINE tra_adv_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_adv_tlm_tst ***
      !!
      !! ** Purpose : Test the tangent routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!                   
      !! History :
      !!        ! 09-08 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit

      CALL tra_adv_cen2_tlm_tst( kumadt )

   END SUBROUTINE tra_adv_tlm_tst
#endif
#endif


  !!======================================================================
END MODULE traadv_tam
