MODULE traldf_lap_tam
#if defined key_tam
   !!==============================================================================
   !!                       ***  MODULE  traldf_lap  ***
   !! Ocean active tracers:  horizontal component of the lateral tracer mixing trend
   !!                        Tangent and adjoint module
   !!==============================================================================
   !! History :   9.0  !  ?????
   !! History of the T&A module:
   !!             9.0  !  09-03  (F. Vigilant) original version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_ldf_lap  : update the tracer trend with the horizontal diffusion
   !!                 using a iso-level harmonic (laplacien) operator.
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpiglo,              &
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & tb_tl,               &
      & sb_tl,               &
      & ta_tl,               &
      & sa_tl,               &
      & gtu_tl,              &
      & gsu_tl,              &
      & gtv_tl,              &
      & gsv_tl,              &
      & tb_ad,               &
      & sb_ad,               &
      & ta_ad,               &
      & sa_ad,               &
      & gtu_ad,              &
      & gsu_ad,              &
      & gtv_ad,              &
      & gsv_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & e1u,                 &
      & e2u,                 &
      & e1v,                 &
      & e2v,                 &
      & e1t,                 &
      & e2t,                 &
#if defined key_zco
      & e3t_0,               &
#else
      & e3u,                 &
      & e3v,                 &
      & e3t,                 &
#endif
      & tmask,               &
      & umask,               &
      & vmask,               &
      & mbathy,              &
      & ln_zps,              &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej,                &
      &  lk_vvl
   USE ldftra_oce    , ONLY: & ! ocean active tracers: lateral physics
      & ahtv,                &
      & ahtu,                &
      & ahtw,                &
      & ahtb0,               &
      & aht0
!   USE trdmod          ! ocean active tracers trends 
!   USE trdmod_oce      ! ocean variables trends
   USE in_out_manager, ONLY: & ! I/O manager
      & lwp,                 &
      & numout,              &
      & nitend,              &
      & nit000,              &
      & nstop
!   USE diaptr          ! poleward transport diagnostics
!   USE prtctl          ! Print control

   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
      & prntst_tlm,          & !
      & stdt,                & ! stdev for u-velocity
      & stds                   !           v-velocity
   USE paresp        , ONLY: & ! Weights for an energy-type scalar product
      & wesp_t,              &
      & wesp_s

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC tra_ldf_lap_tan      ! routine called by tradldf_tam.F90
   PUBLIC tra_ldf_lap_adj      ! routine called by tradldf_tam.F90
   PUBLIC tra_ldf_lap_adj_tst  ! routine called by tradldf_tam.F90
#if defined key_tst_tlm
   PUBLIC tra_ldf_lap_tlm_tst  
#endif

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "ldftra_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id: traldf_lap.F90 1152 2008-06-26 14:11:13Z rblod $ 
   !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
   !!----------------------------------------------------------------------
   
CONTAINS

   SUBROUTINE tra_ldf_lap_tan( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_ldf_lap_tan  ***
      !!                   
      !! ** Purpose :   Compute the before horizontal tracer (t & s) diffusive 
      !!      trend and add it to the general trend of tracer equation.
      !!
      !! ** Method  :   Second order diffusive operator evaluated using before
      !!      fields (forward time scheme). The horizontal diffusive trends of 
      !!      temperature (idem for salinity) is given by:
      !!          difft = 1/(e1t*e2t*e3t) {  di-1[ aht e2u*e3u/e1u di(tb) ]
      !!                                   + dj-1[ aht e1v*e3v/e2v dj(tb) ] }
      !!     Note: key_zco defined, the e3t=e3u=e3v, the trend becomes:  
      !!          difft = 1/(e1t*e2t) {  di-1[ aht e2u/e1u di(tb) ]
      !!                               + dj-1[ aht e1v/e2v dj(tb) ] }
      !!      Add this trend to the general tracer trend (ta,sa):
      !!          (ta,sa) = (ta,sa) + ( difft , diffs )
      !!
      !! ** Action  : - Update (ta,sa) arrays with the before iso-level 
      !!                harmonic mixing trend.
      !!
      !! History :
      !!   1.0  !  87-06  (P. Andrich, D. L Hostis)  Original code
      !!        !  91-11  (G. Madec)
      !!        !  95-11  (G. Madec)  suppress volumetric scale factors
      !!        !  96-01  (G. Madec)  statement function for e3
      !!   8.5  !  02-06  (G. Madec)  F90: Free form and module
      !!   9.0  !  04-08  (C. Talandier) New trends organization
      !!        !  05-11  (G. Madec)  add zps case
      !! History of the tangent routine
      !!   9.0  !  03-09 (F. Vigilant) tangent of 9.0
      !!----------------------------------------------------------------------

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt       ! ocean time-step index
      
      !! * Local save
      REAL(wp), DIMENSION(jpi,jpj), SAVE ::   &
      &   ze1ur, ze2vr, zbtr2              ! scale factor coefficients
      
      !! * Local declarations
      INTEGER  ::   ji, jj, jk             ! dummy loop indices
      INTEGER  ::   iku, ikv               ! temporary integers
      REAL(wp) ::  zabe1, zabe2, zbtr      ! temporary scalars
      REAL(wp) ::  ztatl, zsatl            ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::  ztutl, ztvtl     ! 3D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::  zsutl, zsvtl     ! 3D workspace
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_ldf_lap_tan : iso-level laplacian diffusion'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~ '
         ze1ur(:,:) = e2u(:,:) / e1u(:,:)
         ze2vr(:,:) = e1v(:,:) / e2v(:,:)
         zbtr2(:,:) = 1. / ( e1t(:,:) * e2t(:,:) )
      ENDIF
      
      !                                                  ! =============
      DO jk = 1, jpkm1                                   ! Vertical slab
         !                                               ! =============
         ! 1. First derivative (gradient)
         ! -------------------
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
#if defined key_zco
               zabe1 = fsahtu(ji,jj,jk) * umask(ji,jj,jk) * ze1ur(ji,jj)
               zabe2 = fsahtv(ji,jj,jk) * vmask(ji,jj,jk) * ze2vr(ji,jj)
#else
               zabe1 = fsahtu(ji,jj,jk) * umask(ji,jj,jk) * ze1ur(ji,jj) * fse3u(ji,jj,jk)
               zabe2 = fsahtv(ji,jj,jk) * vmask(ji,jj,jk) * ze2vr(ji,jj) * fse3v(ji,jj,jk)
#endif
               ztutl(ji,jj,jk) = zabe1 * ( tb_tl(ji+1,jj  ,jk) - tb_tl(ji,jj,jk) )
               zsutl(ji,jj,jk) = zabe1 * ( sb_tl(ji+1,jj  ,jk) - sb_tl(ji,jj,jk) )
               ztvtl(ji,jj,jk) = zabe2 * ( tb_tl(ji  ,jj+1,jk) - tb_tl(ji,jj,jk) )
               zsvtl(ji,jj,jk) = zabe2 * ( sb_tl(ji  ,jj+1,jk) - sb_tl(ji,jj,jk) )
            END DO  
         END DO  
         IF( ln_zps ) THEN      ! set gradient at partial step level
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  ! last level
                  iku = MIN ( mbathy(ji,jj), mbathy(ji+1,jj  ) ) - 1
                  ikv = MIN ( mbathy(ji,jj), mbathy(ji  ,jj+1) ) - 1
                  IF( iku == jk ) THEN
                     zabe1 = fsahtu(ji,jj,iku) * umask(ji,jj,iku) * ze1ur(ji,jj) * fse3u(ji,jj,iku)
                     ztutl(ji,jj,jk) = zabe1 * gtu_tl(ji,jj)
                     zsutl(ji,jj,jk) = zabe1 * gsu_tl(ji,jj)
                  ENDIF
                  IF( ikv == jk ) THEN
                     zabe2 = fsahtv(ji,jj,ikv) * vmask(ji,jj,ikv) * ze2vr(ji,jj) * fse3v(ji,jj,ikv)
                     ztvtl(ji,jj,jk) = zabe2 * gtv_tl(ji,jj)
                     zsvtl(ji,jj,jk) = zabe2 * gsv_tl(ji,jj)
                  ENDIF
               END DO
            END DO
         ENDIF
         
         
         ! 2. Second derivative (divergence)
         ! --------------------
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
#if defined key_zco
               zbtr = zbtr2(ji,jj)
#else
               zbtr = zbtr2(ji,jj) / fse3t(ji,jj,jk)
#endif
               ! horizontal diffusive trends
               ztatl = zbtr * (  ztutl(ji,jj,jk) - ztutl(ji-1,jj  ,jk)   &
                  &            + ztvtl(ji,jj,jk) - ztvtl(ji  ,jj-1,jk)  )
               zsatl = zbtr * (  zsutl(ji,jj,jk) - zsutl(ji-1,jj  ,jk)   &
                  &            + zsvtl(ji,jj,jk) - zsvtl(ji  ,jj-1,jk)  )
               ! add it to the general tracer trends
               ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl
               sa_tl(ji,jj,jk) = sa_tl(ji,jj,jk) + zsatl
            END DO  
         END DO  
         !                                               ! =============
      END DO                                             !  End of slab  
      !                                                  ! =============

      ! "zonal" mean lateral diffusive heat and salt transport 
      ! Warning: no update of pht_ldf and /pst_ldf/ when /ln_diaptr/ is activated

   END SUBROUTINE tra_ldf_lap_tan


   SUBROUTINE tra_ldf_lap_adj( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_ldf_lap_adj  ***
      !!                   
      !! ** Purpose :   Compute the before horizontal tracer (t & s) diffusive 
      !!      trend and add it to the general trend of tracer equation.
      !!
      !! ** Method  :   Second order diffusive operator evaluated using before
      !!      fields (forward time scheme). The horizontal diffusive trends of 
      !!      temperature (idem for salinity) is given by:
      !!          difft = 1/(e1t*e2t*e3t) {  di-1[ aht e2u*e3u/e1u di(tb) ]
      !!                                   + dj-1[ aht e1v*e3v/e2v dj(tb) ] }
      !!     Note: key_zco defined, the e3t=e3u=e3v, the trend becomes:  
      !!          difft = 1/(e1t*e2t) {  di-1[ aht e2u/e1u di(tb) ]
      !!                               + dj-1[ aht e1v/e2v dj(tb) ] }
      !!      Add this trend to the general tracer trend (ta,sa):
      !!          (ta,sa) = (ta,sa) + ( difft , diffs )
      !!
      !! ** Action  : - Update (ta,sa) arrays with the before iso-level 
      !!                harmonic mixing trend.
      !!
      !! History :
      !!   1.0  !  87-06  (P. Andrich, D. L Hostis)  Original code
      !!        !  91-11  (G. Madec)
      !!        !  95-11  (G. Madec)  suppress volumetric scale factors
      !!        !  96-01  (G. Madec)  statement function for e3
      !!   8.5  !  02-06  (G. Madec)  F90: Free form and module
      !!   9.0  !  04-08  (C. Talandier) New trends organization
      !!        !  05-11  (G. Madec)  add zps case
      !! History of the tangent routine
      !!   9.0  !  03-09 (F. Vigilant) adjoint of 9.0
      !!----------------------------------------------------------------------

      !! * Arguments
      INTEGER, INTENT( in ) ::   kt       ! ocean time-step index
      
      !! * Local save
      REAL(wp), DIMENSION(jpi,jpj), SAVE ::   &
      &   ze1ur, ze2vr, zbtr2              ! scale factor coefficients
      
      !! * Local declarations
      INTEGER ::   ji, jj, jk             ! dummy loop indices
      INTEGER ::   iku, ikv               ! temporary integers
      REAL(wp) ::  zabe1, zabe2, zbtr     ! temporary scalars
      REAL(wp) ::  ztaad, zsaad           ! temporary scalars    
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: ztuad, ztvad    ! 3D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: zsuad, zsvad    ! 3D workspace
      !!----------------------------------------------------------------------
      ztvad(:,:,:) = 0.0_wp 
      zsvad(:,:,:) = 0.0_wp
      ztuad(:,:,:) = 0.0_wp
      zsuad(:,:,:) = 0.0_wp
      ztaad        = 0.0_wp
      zsaad        = 0.0_wp

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_ldf_lap_adj : iso-level laplacian diffusion'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~ '
         ze1ur(:,:) = e2u(:,:) / e1u(:,:)
         ze2vr(:,:) = e1v(:,:) / e2v(:,:)
         zbtr2(:,:) = 1. / ( e1t(:,:) * e2t(:,:) )
      ENDIF      


         
      !                                                  ! =============
      DO jk = 1, jpkm1                                   ! Vertical slab
         !                                               ! =============

         ! 2. Second derivative (divergence)
         ! --------------------
         DO jj = jpjm1, 2, -1
            DO ji = fs_jpim1, fs_2, -1   ! vector opt.
#if defined key_zco
               zbtr = zbtr2(ji,jj)
#else
               zbtr = zbtr2(ji,jj) / fse3t(ji,jj,jk)
#endif
               ! horizontal diffusive trends
               ztaad             = zbtr * ta_ad(ji,jj,jk)
               zsaad             = zbtr * sa_ad(ji,jj,jk)

               ztuad(ji  ,jj  ,jk) = ztuad(ji  ,jj  ,jk) + ztaad
               ztuad(ji-1,jj  ,jk) = ztuad(ji-1,jj  ,jk) - ztaad
               ztvad(ji  ,jj  ,jk) = ztvad(ji  ,jj  ,jk) + ztaad
               ztvad(ji  ,jj-1,jk) = ztvad(ji  ,jj-1,jk) - ztaad

               zsuad(ji  ,jj  ,jk) = zsuad(ji  ,jj  ,jk) + zsaad
               zsuad(ji-1,jj  ,jk) = zsuad(ji-1,jj  ,jk) - zsaad
               zsvad(ji  ,jj  ,jk) = zsvad(ji  ,jj  ,jk) + zsaad
               zsvad(ji  ,jj-1,jk) = zsvad(ji  ,jj-1,jk) - zsaad

               !ztaad   = 0.0_wp
               !zsaad   = 0.0_wp

            END DO  
         END DO  

         ! 1. First derivative (gradient)
         ! -------------------

         IF( ln_zps ) THEN      ! set gradient at partial step level
            DO jj = jpjm1, 1, -1
               DO ji = fs_jpim1, 1, -1   ! vector opt.?
                  ! last level
                  iku = MIN ( mbathy(ji,jj), mbathy(ji+1,jj  ) ) - 1
                  ikv = MIN ( mbathy(ji,jj), mbathy(ji  ,jj+1) ) - 1
                  IF( iku == jk ) THEN
                     zabe1 = fsahtu(ji,jj,iku) * umask(ji,jj,iku) * ze1ur(ji,jj) * fse3u(ji,jj,iku)

                     gtu_ad(ji,jj) = zabe1 * ztuad(ji,jj,jk) + gtu_ad(ji,jj)
                     gsu_ad(ji,jj) = zabe1 * zsuad(ji,jj,jk) + gsu_ad(ji,jj) 
 
                     ztuad(ji,jj,jk)  = 0.0_wp
                     zsuad(ji,jj,jk)  = 0.0_wp
                  ENDIF
                  IF( ikv == jk ) THEN
                     zabe2 = fsahtv(ji,jj,ikv) * vmask(ji,jj,ikv) * ze2vr(ji,jj) * fse3v(ji,jj,ikv)

                     gtv_ad(ji,jj) = zabe2 * ztvad(ji,jj,jk) + gtv_ad(ji,jj)
                     gsv_ad(ji,jj) = zabe2 * zsvad(ji,jj,jk) + gsv_ad(ji,jj)

                     ztvad(ji,jj,jk)  = 0.0_wp
                     zsvad(ji,jj,jk)  = 0.0_wp
                  ENDIF
               END DO
            END DO
         ENDIF

         DO jj = jpjm1, 1, -1
            DO ji = fs_jpim1, 1, -1   ! vector opt. ?
#if defined key_zco
               zabe1 = fsahtu(ji,jj,jk) * umask(ji,jj,jk) * ze1ur(ji,jj)
               zabe2 = fsahtv(ji,jj,jk) * vmask(ji,jj,jk) * ze2vr(ji,jj)
#else
               zabe1 = fsahtu(ji,jj,jk) * umask(ji,jj,jk) * ze1ur(ji,jj) * fse3u(ji,jj,jk)
               zabe2 = fsahtv(ji,jj,jk) * vmask(ji,jj,jk) * ze2vr(ji,jj) * fse3v(ji,jj,jk)
#endif
               
               tb_ad(ji  ,jj  ,jk) = tb_ad(ji  ,jj  ,jk) - (zabe1 * ztuad(ji,jj,jk) + zabe2 * ztvad(ji,jj,jk))
               tb_ad(ji+1,jj  ,jk) = tb_ad(ji+1,jj  ,jk) +  zabe1 * ztuad(ji,jj,jk)
               tb_ad(ji  ,jj+1,jk) = tb_ad(ji  ,jj+1,jk) +  zabe2 * ztvad(ji,jj,jk)

               sb_ad(ji  ,jj  ,jk) = sb_ad(ji  ,jj  ,jk) - (zabe1 * zsuad(ji,jj,jk) + zabe2 * zsvad(ji,jj,jk))
               sb_ad(ji+1,jj  ,jk) = sb_ad(ji+1,jj  ,jk) +  zabe1 * zsuad(ji,jj,jk)
               sb_ad(ji  ,jj+1,jk) = sb_ad(ji  ,jj+1,jk) +  zabe2 * zsvad(ji,jj,jk)

               ztuad(ji,jj,jk)   = 0.0_wp
               zsuad(ji,jj,jk)   = 0.0_wp
               ztvad(ji,jj,jk)   = 0.0_wp
               zsvad(ji,jj,jk)   = 0.0_wp

            END DO  
         END DO  
         
          !                                               ! =============
      END DO                                              !  End of slab  
      !                                                   ! =============                    

   END SUBROUTINE tra_ldf_lap_adj


   SUBROUTINE tra_ldf_lap_adj_tst ( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE example_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_T,       &
         & zsp1_S,       &
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       &
         & zsp2_2,       &
         & zsp2_3,       &
         & zsp2_4,       &
         & zsp2_5,       &
         & zsp2_6,       &
         & zsp2_7,       &
         & zsp2_8,       &
         & zsp2_T,       &
         & zsp2_S
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztb_tlin ,      & ! Tangent input
         & zsb_tlin ,      & ! Tangent input
         & zta_tlin ,      & ! Tangent input
         & zsa_tlin ,      & ! Tangent input
         & zta_tlout,      & ! Tangent output
         & zsa_tlout,      & ! Tangent output
         & zta_adin,       & ! Adjoint input
         & zsa_adin,       & ! Adjoint input
         & ztb_adout ,     & ! Adjoint output
         & zsb_adout ,     & ! Adjoint output
         & zta_adout ,     & ! Adjoint output
         & zsa_adout ,     & ! Adjoint output
         & z3r               ! 3D random field 
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgtu_tlin ,     & ! Tangent input
         & zgsu_tlin ,     & ! Tangent input
         & zgtv_tlin ,     & ! Tangent input
         & zgsv_tlin ,     & ! Tangent input
         & zgtu_adout ,    & ! Adjoint output
         & zgsu_adout ,    & ! Adjoint output
         & zgtv_adout ,    & ! Adjoint output
         & zgsv_adout ,    & ! Adjoint output
         & z2r               ! 2D random field 
      CHARACTER(LEN=14) :: cl_name
      ! Allocate memory

      ALLOCATE( &
         & ztb_tlin(jpi,jpj,jpk),      & 
         & zsb_tlin(jpi,jpj,jpk),      & 
         & zta_tlin(jpi,jpj,jpk),      & 
         & zsa_tlin(jpi,jpj,jpk),      & 
         & zgtu_tlin(jpi,jpj),         & 
         & zgsu_tlin(jpi,jpj),         & 
         & zgtv_tlin(jpi,jpj),         & 
         & zgsv_tlin(jpi,jpj),         & 
         & zta_tlout(jpi,jpj,jpk),     & 
         & zsa_tlout(jpi,jpj,jpk),     & 
         & zta_adin(jpi,jpj,jpk),      & 
         & zsa_adin(jpi,jpj,jpk),      & 
         & ztb_adout(jpi,jpj,jpk),     & 
         & zsb_adout(jpi,jpj,jpk),     & 
         & zta_adout(jpi,jpj,jpk),     & 
         & zsa_adout(jpi,jpj,jpk),     & 
         & zgtu_adout(jpi,jpj),        & 
         & zgsu_adout(jpi,jpj),        & 
         & zgtv_adout(jpi,jpj),        & 
         & zgsv_adout(jpi,jpj),        & 
         & z3r(jpi,jpj,jpk),           &
         & z2r(jpi,jpj)                &
         & )


      !=======================================================================
      ! 1) dx = ( tb_tl, ta_tl, sb_tl, sa_tl, gtu_tl, gtv_tl, gsu_tl, gsv_tl )
      !    dy = ( ta_tl, sa_tl )
      !=======================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------

      ztb_tlin(:,:,:)  = 0.0_wp
      zsb_tlin(:,:,:)  = 0.0_wp
      zta_tlin(:,:,:)  = 0.0_wp
      zsa_tlin(:,:,:)  = 0.0_wp
      zgtu_tlin(:,:)   = 0.0_wp
      zgsu_tlin(:,:)   = 0.0_wp
      zgtv_tlin(:,:)   = 0.0_wp
      zgsv_tlin(:,:)   = 0.0_wp
      zta_tlout(:,:,:) = 0.0_wp
      zsa_tlout(:,:,:) = 0.0_wp
      zta_adin(:,:,:)  = 0.0_wp
      zsa_adin(:,:,:)  = 0.0_wp
      ztb_adout(:,:,:) = 0.0_wp
      zsb_adout(:,:,:) = 0.0_wp
      zta_adout(:,:,:) = 0.0_wp
      zsa_adout(:,:,:) = 0.0_wp
      zgtu_adout(:,:)  = 0.0_wp
      zgsu_adout(:,:)  = 0.0_wp
      zgtv_adout(:,:)  = 0.0_wp
      zgsv_adout(:,:)  = 0.0_wp

      tb_tl(:,:,:) = 0.0_wp
      sb_tl(:,:,:) = 0.0_wp
      ta_tl(:,:,:) = 0.0_wp
      sa_tl(:,:,:) = 0.0_wp
      gtu_tl(:,:)  = 0.0_wp
      gsu_tl(:,:)  = 0.0_wp
      gtv_tl(:,:)  = 0.0_wp
      gsv_tl(:,:)  = 0.0_wp
      tb_ad(:,:,:) = 0.0_wp
      sb_ad(:,:,:) = 0.0_wp
      ta_ad(:,:,:) = 0.0_wp
      sa_ad(:,:,:) = 0.0_wp
      gtu_ad(:,:)  = 0.0_wp
      gsu_ad(:,:)  = 0.0_wp
      gtv_ad(:,:)  = 0.0_wp
      gsv_ad(:,:)  = 0.0_wp

      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
      ztb_tlin(:,:,:) = z3r(:,:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 727391 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds )
      zsb_tlin(:,:,:) = z3r(:,:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 249741 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stdt )
      zta_tlin(:,:,:) = z3r(:,:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 182029 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z3r, 'T', 0.0_wp, stds )
      zsa_tlin(:,:,:) = z3r(:,:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'U', 0.0_wp, stds )
      zgtu_tlin(:,:) = z2r(:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 727391 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'U', 0.0_wp, stds )
      zgsu_tlin(:,:) = z2r(:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 249741 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'V', 0.0_wp, stds )
      zgtv_tlin(:,:) = z2r(:,:) 

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 182029 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, z2r, 'V', 0.0_wp, stds )
      zgsv_tlin(:,:) = z2r(:,:) 

      tb_tl(:,:,:) = ztb_tlin(:,:,:) 
      sb_tl(:,:,:) = zsb_tlin(:,:,:) 
      ta_tl(:,:,:) = zta_tlin(:,:,:) 
      sa_tl(:,:,:) = zsa_tlin(:,:,:) 
      gtu_tl(:,:)  = zgtu_tlin(:,:)
      gsu_tl(:,:)  = zgsu_tlin(:,:)
      gtv_tl(:,:)  = zgtv_tlin(:,:)
      gsv_tl(:,:)  = zgsv_tlin(:,:)

      CALL tra_ldf_lap_tan( nit000 ) 
      
      zta_tlout(:,:,:) = ta_tl(:,:,:)
      zsa_tlout(:,:,:) = sa_tl(:,:,:)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsa_adin(ji,jj,jk) = zsa_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_s(jk)
              zta_adin(ji,jj,jk) = zta_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_t(jk)
            END DO
         END DO
      END DO

      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !-------------------------------------------------------------------

      zsp1_T = DOT_PRODUCT( zta_tlout, zta_adin )
      zsp1_S = DOT_PRODUCT( zsa_tlout, zsa_adin )
      zsp1 = zsp1_T + zsp1_S

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      ta_ad(:,:,:) = zta_adin(:,:,:)
      sa_ad(:,:,:) = zsa_adin(:,:,:)

      CALL tra_ldf_lap_adj( nit000 ) 

      ztb_adout(:,:,:) = tb_ad(:,:,:)
      zsb_adout(:,:,:) = sb_ad(:,:,:)
      zta_adout(:,:,:) = ta_ad(:,:,:)
      zsa_adout(:,:,:) = sa_ad(:,:,:)
      zgtu_adout(:,:)  = gtu_ad(:,:)
      zgsu_adout(:,:)  = gsu_ad(:,:)
      zgtv_adout(:,:)  = gtv_ad(:,:)
      zgsv_adout(:,:)  = gsv_ad(:,:)
          
      zsp2_1 = DOT_PRODUCT( ztb_tlin , ztb_adout  )
      zsp2_2 = DOT_PRODUCT( zta_tlin , zta_adout  )
      zsp2_3 = DOT_PRODUCT( zgtu_tlin, zgtu_adout )
      zsp2_4 = DOT_PRODUCT( zgtv_tlin, zgtv_adout )
      zsp2_5 = DOT_PRODUCT( zsb_tlin , zsb_adout  )
      zsp2_6 = DOT_PRODUCT( zsa_tlin , zsa_adout  )
      zsp2_7 = DOT_PRODUCT( zgsu_tlin, zgsu_adout )
      zsp2_8 = DOT_PRODUCT( zgsv_tlin, zgsv_adout ) 

      zsp2_T = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4
      zsp2_S = zsp2_5 + zsp2_6 + zsp2_7 + zsp2_8
      zsp2   = zsp2_T + zsp2_S

      cl_name = 'tra_ldf_lap_ad'
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      DEALLOCATE(         &
         & ztb_tlin,      & ! Tangent input
         & zsb_tlin,      & ! Tangent input
         & zta_tlin,      & ! Tangent input
         & zsa_tlin,      & ! Tangent input
         & zgtu_tlin,     & ! Tangent input
         & zgsu_tlin,     & ! Tangent input
         & zgtv_tlin,     & ! Tangent input
         & zgsv_tlin,     & ! Tangent input
         & zta_tlout,     & ! Tangent output
         & zsa_tlout,     & ! Tangent output
         & zta_adin,      & ! Adjoint input
         & zsa_adin,      & ! Adjoint input
         & ztb_adout,     & ! Adjoint output
         & zsb_adout,     & ! Adjoint output
         & zta_adout,     & ! Adjoint output
         & zsa_adout,     & ! Adjoint output
         & zgtu_adout,    & ! Adjoint output
         & zgsu_adout,    & ! Adjoint output
         & zgtv_adout,    & ! Adjoint output
         & zgsv_adout,    & ! Adjoint output
         & z3r,           & ! 3D random field 
         & z2r            &
         & )
      

   END SUBROUTINE tra_ldf_lap_adj_tst
#if defined key_tst_tlm
   SUBROUTINE tra_ldf_lap_tlm_tst ( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_ldf_lap_tlm_tst ***
      !!
      !! ** Purpose : Test the tangent routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+h*dx) = M(x) + L(h*dx) + O(h^2)
      !!
      !!              where  L   = Linear routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!    
      !!                   
      !! History :
      !!        ! 09-07 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE step_c1d            ! Time stepping loop for the 1D configuration
!!      USE asminc              ! assimilation increments       (asm_inc_init routine)
      USE tamtrj              ! writing out state trajectory  
      USE lib_mpp,    ONLY: & ! distributed memory computing
        & lk_mpp,           &
        & mpp_max 
      USE c1d,        ONLY: & ! 1D configuration
        & lk_c1d
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE zpshde
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
        & tb, sb, tn, sn, ta,  &
        & sa, gtu, gsu, gtv,   &
        & gsv, gru, grv, rhd
      USE traldf_lap          ! lateral mixing                   (tra_ldf routine)
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk,    &
         & istp      
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_Tb,      &
         & zsp1_Sb,      &
         & zsp1_Ta,      &
         & zsp1_Sa,      &
         & zsp1_Gtu,     &
         & zsp1_Gsu,     &
         & zsp1_Gtv,     &
         & zsp1_Gsv,     &
         & zsp2,         & ! scalar product involving the tangent routine
         & zsp2_Tb,      &
         & zsp2_Sb,      &
         & zsp2_Ta,      &
         & zsp2_Sa,      &
         & zsp2_Gtu,     &
         & zsp2_Gsu,     &
         & zsp2_Gtv,     &
         & zsp2_Gsv,     &
         & zsp3,         & ! scalar product involving the tangent routine
         & zsp3_Tb,      &
         & zsp3_Sb,      &
         & zsp3_Ta,      &
         & zsp3_Sa,      &
         & zsp3_Gtu,     &
         & zsp3_Gsu,     &
         & zsp3_Gtv,     &
         & zsp3_Gsv,     &
         & zzsp,         & ! scalar product involving the tangent routine
         & zzsp_Tb,      &
         & zzsp_Sb,      &
         & zzsp_Ta,      &
         & zzsp_Sa,      &
         & zzsp_Gtu,     &
         & zzsp_Gsu,     &
         & zzsp_Gtv,     &
         & zzsp_Gsv,     &
         & gamma,        &
         & zgsp1,        &
         & zgsp2,        &
         & zgsp3,        &
         & zgsp4,        &
         & zgsp5,        &
         & zgsp6,        &
         & zgsp7
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztb_tlin ,      & ! Tangent input
         & zsb_tlin ,      & ! Tangent input
         & zta_tlin ,      & ! Tangent input
         & zsa_tlin ,      & ! Tangent input
         & ztb_out ,       & ! Direct output
         & zsb_out ,       & ! Direct output
         & zta_out ,       & ! Direct output
         & zsa_out ,       & ! Direct output
         & ztb_wop ,       & ! 
         & zsb_wop ,       & ! 
         & zta_wop ,       & ! 
         & zsa_wop ,       & ! 
         & z3r               ! 3D random field 
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgtu_tlin ,     & ! Tangent input
         & zgsu_tlin ,     & ! Tangent input
         & zgtv_tlin ,     & ! Tangent input
         & zgsv_tlin ,     & ! Tangent input
         & zgtu_out ,      & ! Direct output
         & zgsu_out ,      & ! Direct output
         & zgtv_out ,      & ! Direct output
         & zgsv_out ,      & ! Direct output
         & zgtu_wop ,      & ! 
         & zgsu_wop ,      & ! 
         & zgtv_wop ,      & ! 
         & zgsv_wop ,      & ! 
         & z2r               ! 2D random field 
      CHARACTER(LEN=14) ::   cl_name
      CHARACTER (LEN=128) :: file_out_sc, file_wop, file_out, file_xdx
      CHARACTER (LEN=90) :: &
         & FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zsctb, zscta,    &
         & zscsb, zscsa,    &
         & zscgtu, zscgsu,  &
         & zscgtv, zscgsv,  &
         & zscerrtb,        &
         & zscerrta,        &
         & zscerrsb,        &
         & zscerrsa,        &
         & zscerrgtu,       &
         & zscerrgsu,       &
         & zscerrgtv,       &
         & zscerrgsv
      INTEGER, DIMENSION(100)::                    &
         & iipostb, iiposta, iipossb, iipossa,     &
         & iiposgtu, iiposgsu, iiposgtv, iiposgsv, &
         & ijpostb,  ijposta,  ijpossb,  ijpossa,  &
         & ijposgtu, ijposgsu, ijposgtv, ijposgsv, & 
         & ikpostb, ikposta, ikpossb, ikpossa
      INTEGER::             &
         & ii,              &
         & isamp=40,        &
         & jsamp=40,        &
         & ksamp=10,        &
         & numsctlm,        &
         & numtlm
     REAL(KIND=wp), DIMENSION(jpi,jpj,jpk) :: &
         & zerrtb, zerrta,  & 
         & zerrsb, zerrsa     

      REAL(KIND=wp), DIMENSION(jpi,jpj)    :: &
         & zerrgtu, zerrgsu, &  
         & zerrgtv, zerrgsv  

      ! Allocate memory
      ALLOCATE( &
         & ztb_tlin(jpi,jpj,jpk),      & 
         & zsb_tlin(jpi,jpj,jpk),      & 
         & zta_tlin(jpi,jpj,jpk),      & 
         & zsa_tlin(jpi,jpj,jpk),      & 
         & ztb_out(jpi,jpj,jpk),       & 
         & zsb_out(jpi,jpj,jpk),       & 
         & zta_out(jpi,jpj,jpk),       & 
         & zsa_out(jpi,jpj,jpk),       &
         & ztb_wop(jpi,jpj,jpk),       & 
         & zsb_wop(jpi,jpj,jpk),       & 
         & zta_wop(jpi,jpj,jpk),       & 
         & zsa_wop(jpi,jpj,jpk),       &
         & zgtu_tlin(jpi,jpj),         & 
         & zgsu_tlin(jpi,jpj),         & 
         & zgtv_tlin(jpi,jpj),         & 
         & zgsv_tlin(jpi,jpj),         &  
         & zgtu_out(jpi,jpj),         & 
         & zgsu_out(jpi,jpj),         & 
         & zgtv_out(jpi,jpj),         & 
         & zgsv_out(jpi,jpj),         &
         & zgtu_wop(jpi,jpj),         & 
         & zgsu_wop(jpi,jpj),         & 
         & zgtv_wop(jpi,jpj),         & 
         & zgsv_wop(jpi,jpj),         &
         & z3r(jpi,jpj,jpk),           &
         & z2r(jpi,jpj)                &
         & )

      !--------------------------------------------------------------------
      ! Reset variables
      !--------------------------------------------------------------------

      ztb_tlin(:,:,:)  = 0.0_wp
      zsb_tlin(:,:,:)  = 0.0_wp
      zta_tlin(:,:,:)  = 0.0_wp
      zsa_tlin(:,:,:)  = 0.0_wp
      ztb_out(:,:,:)   = 0.0_wp
      zsb_out(:,:,:)   = 0.0_wp
      zta_out(:,:,:)   = 0.0_wp
      zsa_out(:,:,:)   = 0.0_wp
      ztb_wop(:,:,:)   = 0.0_wp
      zsb_wop(:,:,:)   = 0.0_wp
      zta_wop(:,:,:)   = 0.0_wp
      zsa_wop(:,:,:)   = 0.0_wp
      zgtu_tlin(:,:)   = 0.0_wp
      zgsu_tlin(:,:)   = 0.0_wp
      zgtv_tlin(:,:)   = 0.0_wp
      zgsv_tlin(:,:)   = 0.0_wp
      zgtu_out(:,:)    = 0.0_wp
      zgsu_out(:,:)    = 0.0_wp
      zgtv_out(:,:)    = 0.0_wp
      zgsv_out(:,:)    = 0.0_wp
      zgtu_wop(:,:)    = 0.0_wp
      zgsu_wop(:,:)    = 0.0_wp
      zgtv_wop(:,:)    = 0.0_wp
      zgsv_wop(:,:)    = 0.0_wp
      IF ( tlm_bch == 2 ) THEN      
      tb_tl(:,:,:) = 0.0_wp
      sb_tl(:,:,:) = 0.0_wp
      ta_tl(:,:,:) = 0.0_wp
      sa_tl(:,:,:) = 0.0_wp
      gtu_tl(:,:)  = 0.0_wp
      gsu_tl(:,:)  = 0.0_wp
      gtv_tl(:,:)  = 0.0_wp
      gsv_tl(:,:)  = 0.0_wp
      ENDIF
      zsctb(:)     = 0.0_wp
      zscta(:)     = 0.0_wp
      zscsb(:)     = 0.0_wp
      zscsa(:)     = 0.0_wp
      zscgtu(:)     = 0.0_wp
      zscgsu(:)     = 0.0_wp
      zscgtv(:)     = 0.0_wp
      zscgsv(:)     = 0.0_wp
      zscerrtb(:)     = 0.0_wp
      zscerrta(:)     = 0.0_wp
      zscerrsb(:)     = 0.0_wp
      zscerrsa(:)     = 0.0_wp
      zscerrgtu(:)     = 0.0_wp
      zscerrgsu(:)     = 0.0_wp
      zscerrgtv(:)     = 0.0_wp
      zscerrgsv(:)     = 0.0_wp

      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio
      file_wop='trj_wop_tldf_lap'
      file_xdx='trj_xdx_tldf_lap'
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 596035, z3r,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  ztb_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 727391, z3r,  'S', 0.0_wp, stds)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsb_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 249741, z3r,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 182029, z3r,  'S', 0.0_wp, stds)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsa_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 596035, z2r, 'U', 0.0_wp, stds) 
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgtu_tlin(ji,jj) = z2r(ji,jj)
            END DO
         END DO
         CALL grid_rd_sd( 727391, z2r, 'U', 0.0_wp, stds)
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgsu_tlin(ji,jj) = z2r(ji,jj)
            END DO
         END DO
         CALL grid_rd_sd( 249741, z2r, 'V', 0.0_wp, stds)
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgtv_tlin(ji,jj) = z2r(ji,jj)
            END DO
         END DO
         CALL grid_rd_sd( 182029, z2r, 'V', 0.0_wp, stds)
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zgsv_tlin(ji,jj) = z2r(ji,jj)
            END DO
         END DO
      ENDIF
      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      IF ( tlm_bch /= 2 )      CALL istate_p  

      ! *** initialize the reference trajectory
      ! ------------
      CALL  trj_rea( nit000-1, 1 )
      CALL  trj_rea( nit000, 1 )

      ! Compute  gtu, gsu, gtv, gsv
      CALL zps_hde(nit000, tn, sn, rhd, gtu, gsu, gru, gtv, gsv, grv)

      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         ztb_tlin(:,:,:) = gamma * ztb_tlin(:,:,:)
         tb(:,:,:)       = tb(:,:,:) + ztb_tlin(:,:,:)

         zsb_tlin(:,:,:) = gamma * zsb_tlin(:,:,:)
         sb(:,:,:)       = sb(:,:,:) + zsb_tlin(:,:,:)

         zta_tlin(:,:,:) = gamma * zta_tlin(:,:,:)
         ta(:,:,:)       = ta(:,:,:) + zta_tlin(:,:,:)

         zsa_tlin(:,:,:) = gamma * zsa_tlin(:,:,:)
         sa(:,:,:)       = sa(:,:,:) + zsa_tlin(:,:,:)

         zgtu_tlin(:,:) = gamma * zgtu_tlin(:,:)
         gtu(:,:)       = gtu(:,:) + zgtu_tlin(:,:)

         zgsu_tlin(:,:) = gamma * zgsu_tlin(:,:)
         gsu(:,:)       = gsu(:,:) + zgsu_tlin(:,:)

         zgtv_tlin(:,:) = gamma * zgtv_tlin(:,:)
         gtv(:,:)       = gtv(:,:) + zgtv_tlin(:,:)

         zgsv_tlin(:,:) = gamma * zgsv_tlin(:,:)
         gsv(:,:)       = gsv(:,:) + zgsv_tlin(:,:)
      ENDIF 

      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      IF ( tlm_bch /= 2 )      CALL tra_ldf_lap( nit000 )
      IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN    
         !--------------------------------------------------------------------
         ! Initialize the tangent variables: dy^* = W dy  
         !--------------------------------------------------------------------
         CALL  trj_rea( nit000-1, 1 ) 
         CALL  trj_rea( nit000, 1 ) 
         tb_tl  (:,:,:) = ztb_tlin  (:,:,:)
         sb_tl  (:,:,:) = zsb_tlin  (:,:,:)
         ta_tl  (:,:,:) = zta_tlin  (:,:,:)
         sa_tl  (:,:,:) = zsa_tlin  (:,:,:)
         gtu_tl (  :,:) = zgtu_tlin (  :,:)
         gsu_tl (  :,:) = zgsu_tlin (  :,:)
         gtv_tl (  :,:) = zgtv_tlin (  :,:)
         gsv_tl (  :,:) = zgsv_tlin (  :,:)

         !-----------------------------------------------------------------------
         !  Initialization of the dynamics and tracer fields for the tangent
         !----------------------------------------------------------------------- 
         CALL tra_ldf_lap_tan(nit000)

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------

         zsp2_Tb    = DOT_PRODUCT( tb_tl  , tb_tl    )
         zsp2_Sb    = DOT_PRODUCT( sb_tl  , sb_tl    )
         zsp2_Ta    = DOT_PRODUCT( ta_tl  , ta_tl    )
         zsp2_Sa    = DOT_PRODUCT( sa_tl  , sa_tl    )
         zsp2_Gtu  = DOT_PRODUCT( gtu_tl, gtu_tl  )
         zsp2_Gsu  = DOT_PRODUCT( gsu_tl, gsu_tl  )
         zsp2_Gtv  = DOT_PRODUCT( gtv_tl, gtv_tl  )
         zsp2_Gsv  = DOT_PRODUCT( gsv_tl, gsv_tl  )

         zsp2      = zsp2_Tb + zsp2_Sb + zsp2_Ta + zsp2_Sa + &
                   & zsp2_Gtu + Zsp2_Gsu + zsp2_Gtv + zsp2_Gsv
  
         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------
         CALL trj_rd_spl(file_wop) 
         ztb_wop  (:,:,:) = tb  (:,:,:)
         zsb_wop  (:,:,:) = sb  (:,:,:)
         zta_wop  (:,:,:) = ta  (:,:,:)
         zsa_wop  (:,:,:) = sa  (:,:,:)
         zgtu_wop (:,:  ) = gtu (:,:  )
         zgsu_wop (:,:  ) = gsu (:,:  )
         zgtv_wop (:,:  ) = gtv (:,:  )
         zgsv_wop (:,:  ) = gsv (:,:  )
         CALL trj_rd_spl(file_xdx) 
         ztb_out  (:,:,:) = tb  (:,:,:)
         zsb_out  (:,:,:) = sb  (:,:,:)
         zta_out  (:,:,:) = ta  (:,:,:)
         zsa_out  (:,:,:) = sa  (:,:,:)
         zgtu_out (:,:  ) = gtu (:,:  )
         zgsu_out (:,:  ) = gsu (:,:  )
         zgtv_out (:,:  ) = gtv (:,:  )
         zgsv_out (:,:  ) = gsv (:,:  )
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  ztb_out   (ji,jj,jk) = ztb_out    (ji,jj,jk) - ztb_wop  (ji,jj,jk)
                  ztb_wop   (ji,jj,jk) = ztb_out    (ji,jj,jk) - tb_tl    (ji,jj,jk)
                  IF (  tb_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrtb(ji,jj,jk) = ztb_out(ji,jj,jk)/tb_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iipostb(ii) = ji
                      ijpostb(ii) = jj
                      ikpostb(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
!                         zsctb (ii) =  ztb_out(ji,jj,jk)
                         zsctb (ii) =  ztb_wop(ji,jj,jk)
                         zscerrtb (ii) =  ( zerrtb(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zsb_out   (ji,jj,jk) = zsb_out    (ji,jj,jk) - zsb_wop  (ji,jj,jk)
                  zsb_wop   (ji,jj,jk) = zsb_out    (ji,jj,jk) - sb_tl    (ji,jj,jk)
                  IF (  sb_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrsb(ji,jj,jk) = zsb_out(ji,jj,jk)/sb_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iipossb(ii) = ji
                      ijpossb(ii) = jj
                      ikpossb(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
!                         zscsb (ii) =  zsb_out(ji,jj,jk)
                         zscsb (ii) =  zsb_wop(ji,jj,jk)
                         zscerrsb (ii) =  ( zerrsb(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zta_out   (ji,jj,jk) = zta_out    (ji,jj,jk) - zta_wop  (ji,jj,jk)
                  zta_wop   (ji,jj,jk) = zta_out    (ji,jj,jk) - ta_tl    (ji,jj,jk)
                  IF (  ta_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrta(ji,jj,jk) = zta_out(ji,jj,jk)/ta_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iiposta(ii) = ji
                      ijposta(ii) = jj
                      ikposta(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
!                         zscta (ii) =  zta_out(ji,jj,jk)
                         zscta (ii) =  zta_wop(ji,jj,jk)
                         zscerrta (ii) =  ( zerrta(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zsa_out   (ji,jj,jk) = zsa_out    (ji,jj,jk) - zsa_wop  (ji,jj,jk)
                  zsa_wop   (ji,jj,jk) = zsa_out    (ji,jj,jk) - sa_tl    (ji,jj,jk)
                  IF (  sa_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrsa(ji,jj,jk) = zsa_out(ji,jj,jk)/sa_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iipossa(ii) = ji
                      ijpossa(ii) = jj
                      ikpossa(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
!                         zscsa (ii) =  zsa_out(ji,jj,jk)
                         zscsa (ii) =  zsa_wop(ji,jj,jk)
                         zscerrsa (ii) =  ( zerrsa(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         jk=1
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgtu_out   (ji,jj)= zgtu_out    (ji,jj) - zgtu_wop  (ji,jj)
               zgtu_wop   (ji,jj) = zgtu_out    (ji,jj) - gtu_tl    (ji,jj)
               IF (  gtu_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgtu(ji,jj) = zgtu_out(ji,jj)/gtu_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) .AND. &
               &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                   ii = ii+1
                   iiposgtu(ii) = ji
                   ijposgtu(ii) = jj
                   IF ( INT(umask(ji,jj,jk)) .NE. 0)  THEN
!                      zscgtu (ii) =  zgtu_out(ji,jj)
                      zscgtu (ii) =  zgtu_wop(ji,jj)
                      zscerrgtu (ii) =  ( zerrgtu(ji,jj) - 1.0_wp ) / gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         jk=1
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgtv_out   (ji,jj) = zgtv_out    (ji,jj) - zgtv_wop  (ji,jj)
               zgtv_wop   (ji,jj) = zgtv_out    (ji,jj) - gtv_tl    (ji,jj)
               IF (  gtv_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgtv(ji,jj) = zgtv_out(ji,jj)/gtv_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) .AND. &
               &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                   ii = ii+1
                   iiposgtv(ii) = ji
                   ijposgtv(ii) = jj
                   IF ( INT(vmask(ji,jj,jk)) .NE. 0)  THEN
!                      zscgtv (ii) =  zgtv_out(ji,jj)
                      zscgtv (ii) =  zgtv_wop(ji,jj)
                      zscerrgtv (ii) = ( zerrgtv(ji,jj) - 1.0_wp ) /gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         jk=1
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgsu_out   (ji,jj) = zgsu_out    (ji,jj) - zgsu_wop  (ji,jj)
               zgsu_wop   (ji,jj) = zgsu_out    (ji,jj) - gsu_tl    (ji,jj)
               IF (  gsu_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgsu(ji,jj) = zgsu_out(ji,jj)/gsu_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) .AND. &
               &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                   ii = ii+1
                   iiposgsu(ii) = ji
                   ijposgsu(ii) = jj
                   IF ( INT(umask(ji,jj,jk)) .NE. 0)  THEN
!                      zscgsu (ii) =  zgsu_out(ji,jj)
                      zscgsu (ii) =  zgsu_wop(ji,jj)
                      zscerrgsu (ii) =  ( zerrgsu(ji,jj) - 1.0_wp ) / gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         jk=1
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgsv_out   (ji,jj) = zgsv_out    (ji,jj) - zgsv_wop  (ji,jj)
               zgsv_wop   (ji,jj) = zgsv_out    (ji,jj) - gsv_tl    (ji,jj)
               IF (  gsv_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgsv(ji,jj) = zgsv_out(ji,jj)/gsv_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) .AND. &
               &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                   ii = ii+1
                   iiposgsv(ii) = ji
                   ijposgsv(ii) = jj
                   IF ( INT(vmask(ji,jj,jk)) .NE. 0)  THEN
!                      zscgsv (ii) =  zgsv_out(ji,jj)
                      zscgsv (ii) =  zgsv_wop(ji,jj)
                      zscerrgsv (ii) =  ( zerrgsv(ji,jj) - 1.0_wp ) /gamma
                   ENDIF
               ENDIF
            END DO
         END DO

         zsp1_Tb    = DOT_PRODUCT( ztb_out  , ztb_out    )
         zsp1_Sb    = DOT_PRODUCT( zsb_out  , zsb_out    )
         zsp1_Ta    = DOT_PRODUCT( zta_out  , zta_out    )
         zsp1_Sa    = DOT_PRODUCT( zsa_out  , zsa_out    )
         zsp1_Gtu   = DOT_PRODUCT( zgtu_out , zgtu_out   )
         zsp1_Gsu   = DOT_PRODUCT( zgsu_out , zgsu_out   )
         zsp1_Gtv   = DOT_PRODUCT( zgtv_out , zgtv_out   )
         zsp1_Gsv   = DOT_PRODUCT( zgsv_out , zgsv_out   )

         zsp1      = zsp1_Tb + zsp1_Sb + zsp1_Ta + zsp1_Sa + &
                   & zsp1_Gtu + Zsp1_Gsu + zsp1_Gtv + zsp1_Gsv

         zsp3_Tb    = DOT_PRODUCT( ztb_wop  , ztb_wop    )
         zsp3_Sb    = DOT_PRODUCT( zsb_wop  , zsb_wop    )
         zsp3_Ta    = DOT_PRODUCT( zta_wop  , zta_wop    )
         zsp3_Sa    = DOT_PRODUCT( zsa_wop  , zsa_wop    )
         zsp3_Gtu   = DOT_PRODUCT( zgtu_wop , zgtu_wop   )
         zsp3_Gsu   = DOT_PRODUCT( zgsu_wop , zgsu_wop   )
         zsp3_Gtv   = DOT_PRODUCT( zgtv_wop , zgtv_wop   )
         zsp3_Gsv   = DOT_PRODUCT( zgsv_wop , zgsv_wop   )

         zsp3      = zsp3_Tb + zsp3_Sb + zsp3_Ta + zsp3_Sa + &
                   & zsp3_Gtu + Zsp3_Gsu + zsp3_Gtv + zsp3_Gsv

         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name  = 'traldf_tam:En '
         zzsp     = SQRT(zsp3)
         zzsp_Tb  = SQRT(zsp3_Tb)
         zzsp_Sb  = SQRT(zsp3_Sb)
         zzsp_Ta  = SQRT(zsp3_Ta)
         zzsp_Sa  = SQRT(zsp3_Sa)
         zzsp_Gtu = SQRT(zsp3_Gtu)
         zzsp_Gsu = SQRT(zsp3_Gsu)
         zzsp_Gtv = SQRT(zsp3_Gtv)
         zzsp_Gsv = SQRT(zsp3_Gsv)

         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Compute the relative error Er of the linear model
         ! Er = norm(En) / norm(TL(gamma.dX0, t0,tn))
         !--------------------------------------------------------------------
         cl_name  = 'traldf:Er=En/L'
         zzsp     = SQRT( zsp3/zsp2 )
         zzsp_Tb  = SQRT( zsp3_Tb/zsp2_Tb)
         zzsp_Sb  = SQRT( zsp3_Sb/zsp2_Sb)
         zzsp_Ta  = SQRT(zsp3_Ta/zsp2_Ta)
         zzsp_Sa  = SQRT(zsp3_Sa/zsp2_Sa)
         zzsp_Gtu = SQRT(zsp3_Gtu/zsp2_Gtu)
         zzsp_Gsu = SQRT(zsp3_Gsu/zsp2_Gsu)
         zzsp_Gtv = SQRT(zsp3_Gtv/zsp2_Gtv)
         zzsp_Gsv = SQRT(zsp3_Gsv/zsp2_Gsv)

         zgsp3    = zzsp
         zgsp7    = zzsp/gamma
         zgsp4    = SQRT( zsp2 )
         zgsp5    = SQRT( zsp3 )
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp     =  SQRT(zsp2)
         zzsp_Tb  =  SQRT(zsp2_Tb)
         zzsp_Sb  =  SQRT(zsp2_Sb)
         zzsp_Ta  =  SQRT(zsp2_Ta)
         zzsp_Sa  =  SQRT(zsp2_Sa)
         zzsp_Gtu =  SQRT(zsp2_Gtu)
         zzsp_Gsu =  SQRT(zsp2_Gsu)
         zzsp_Gtv =  SQRT(zsp2_Gtv)
         zzsp_Gsv =  SQRT(zsp2_Gsv)

         cl_name = 'traldf:L_n2   '
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name  = 'traldf:Mhdx-Mx'
         zzsp     =  SQRT(zsp1)
         zzsp_Tb  =  SQRT(zsp1_Tb)
         zzsp_Sb  =  SQRT(zsp1_Sa)
         zzsp_Ta  =  SQRT(zsp1_Ta)
         zzsp_Sa  =  SQRT(zsp1_Sa)
         zzsp_Gtu =  SQRT(zsp1_Gtu)
         zzsp_Gsu =  SQRT(zsp1_Gsu)
         zzsp_Gtv =  SQRT(zsp1_Gtv)
         zzsp_Gsv =  SQRT(zsp1_Gsv)

         zgsp1    = zzsp
         zgsp2    = zgsp1 / SQRT(zsp2)
         zgsp6    = (zgsp2 - 1.0_wp)/gamma
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'tldf_lap', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7

         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'tldf_lap'
         IF(lwp) THEN
            DO ii=1, 100, 1
               IF ( zsctb(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zsctb     ', &
                    & cur_loop, h_ratio, ii, iipostb(ii), ijpostb(ii), zsctb(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscsb(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscsb     ', &
                    & cur_loop, h_ratio, ii, iipossb(ii), ijpossb(ii), zscsb(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscta     ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscsa     ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscsa(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrtb(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrtb     ', &
                    & cur_loop, h_ratio, ii, iipostb(ii), ijpostb(ii), zscerrtb(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrsb(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrsb     ', &
                    & cur_loop, h_ratio, ii, iipossb(ii), ijpossb(ii), zscerrsb(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrta     ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscerrta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrsa     ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscerrsa(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc,"(A4)") '===='
         ENDIF

      ENDIF

      DEALLOCATE(         &
         & ztb_tlin,      & ! Tangent input
         & zsb_tlin,      & ! Tangent input
         & zta_tlin,      & ! Tangent input
         & zsa_tlin,      & ! Tangent input
         & ztb_out,      & 
         & zsb_out,      & 
         & zta_out,      & 
         & zsa_out,      & 
         & ztb_wop,      & 
         & zsb_wop,      & 
         & zta_wop,      & 
         & zsa_wop,      &
         & zgtu_tlin,     & ! Tangent input
         & zgsu_tlin,     & ! Tangent input
         & zgtv_tlin,     & ! Tangent input
         & zgsv_tlin,     & ! Tangent input
         & zgtu_out,     & 
         & zgsu_out,     & 
         & zgtv_out,     &
         & zgsv_out,     & 
         & zgtu_wop,     & 
         & zgsu_wop,     & 
         & zgtv_wop,     &
         & zgsv_wop,     &
         & z3r,           & ! 3D random field 
         & z2r            &
         & )

  END SUBROUTINE tra_ldf_lap_tlm_tst



   SUBROUTINE asm_trj_wop_wri
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE asm_trj__wop_wri ***
      !!
      !! ** Purpose : Write to file the model state trajectory for use with
      !!              4D-Var.
      !!
      !! ** Method  : 
      !!
      !! ** Action  :
      !!
      !! History : 
      !!        ! 09-07 (F. Vigilant) 
      !!-----------------------------------------------------------------------
      !! *Module udes
      USE iom 
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
      & tb, sb, ta, sa, gtu, gsu, gtv, gsv 
      !! * Arguments
      !! * Local declarations
      INTEGER :: &
         & inum                  ! File unit number
      CHARACTER (LEN=50) :: &
         & filename
         ! Define the output file  
         filename='trj_wop'       
         WRITE(filename, FMT='(A,A)' ) TRIM( filename ), '.nc'
         filename = TRIM( filename )
         CALL iom_open( filename, inum, ldwrt = .TRUE., kiolib = jprstlib)

         ! Output trajectory fields
         CALL iom_rstput( 1, 1, inum, 'tb'   , tb  )
         CALL iom_rstput( 1, 1, inum, 'sb'   , sb  )
         CALL iom_rstput( 1, 1, inum, 'ta'   , ta  )
         CALL iom_rstput( 1, 1, inum, 'sa'   , sa  )
         CALL iom_rstput( 1, 1, inum, 'gtu'  , gtu )
         CALL iom_rstput( 1, 1, inum, 'gsu'  , gsu )
         CALL iom_rstput( 1, 1, inum, 'gtv'  , gtv )
         CALL iom_rstput( 1, 1, inum, 'gsv'  , gsv )
         CALL iom_close( inum )
   END SUBROUTINE asm_trj_wop_wri

   SUBROUTINE asm_trj_wop_rd
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE asm_trj__wop_rd ***
      !!
      !! ** Purpose : Read file the model state trajectory for use with
      !!              4D-Var.
      !!
      !! ** Method  : 
      !!
      !! ** Action  :
      !!
      !! History : 
      !!        ! 09-07 (F. Vigilant) 
      !!-----------------------------------------------------------------------
      !! *Module udes
      USE iom                 ! I/O module
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
      & tb, sb, ta, sa, gtu, gsu, gtv, gsv 
      !! * Arguments
      !! * Local declarations
      INTEGER :: &
         & inum                  ! File unit number
      CHARACTER (LEN=50) :: &
         & filename
      ! Define the output file  
      filename='trj_wop'       
      WRITE(filename, FMT='(A,A)' ) TRIM( filename ), '.nc'
      filename = TRIM( filename )
      CALL iom_open( filename, inum)

      ! Output trajectory fields
      CALL iom_get( inum, jpdom_data, 'tb'   , tb, 1  )
      CALL iom_get( inum, jpdom_data, 'sb'   , sb, 1  )
      CALL iom_get( inum, jpdom_data, 'ta'   , ta, 1  )
      CALL iom_get( inum, jpdom_data, 'sa'   , sa, 1  )
      CALL iom_get( inum, jpdom_data, 'gtu'  , gtu, 1 )
      CALL iom_get( inum, jpdom_data, 'gsu'  , gsu, 1 )
      CALL iom_get( inum, jpdom_data, 'gtv'  , gtv, 1 )
      CALL iom_get( inum, jpdom_data, 'gsv'  , gsv, 1 )
      CALL iom_close( inum )
   END SUBROUTINE asm_trj_wop_rd
#endif
#endif
   !!==============================================================================
END MODULE traldf_lap_tam
