MODULE traqsr_tam
#ifdef key_tam
   !!======================================================================
   !!                       ***  MODULE  traqsr_tam  ***
   !! Ocean physics: solar radiation penetration in the top ocean levels
   !!                Tangent and Adjoint Module
   !!======================================================================
   !! History of the direct module:  
   !!            6.0  !  90-10  (B. Blanke)  Original code
   !!            7.0  !  91-11  (G. Madec)
   !!                 !  96-01  (G. Madec)  s-coordinates
   !!            8.5  !  02-06  (G. Madec)  F90: Free form and module
   !!            9.0  !  05-11  (G. Madec) zco, zps, sco coordinate
   !! History of the TAM:
   !!                 !  08-05  (A. Vidard) Skeleton
   !!                 !  08-09  (A. Vidard) TAM of the 05-11 version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_qsr      : trend due to the solar radiation penetration
   !!   tra_qsr_init : solar radiation penetration initialization
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: &
      & jpi,                 &
      & jpj,                 &
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1,               &
      & jpiglo
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & ta_tl,               &
      & ta_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & tmask,               &
      & ln_zco,              &
      & ln_sco,              &
      & ln_zps,              &
      & e1t,                 &
      & e2t,                 &
#if ! defined key_zco
      & e3t,                 &
#endif
      & e3t_0,               &
      & gdepw_0,             &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,                 &
      & numout,              & 
      & nit000,              & 
      & nitend  
   USE sbc_oce       , ONLY: & ! thermohaline fluxes
      & qsr
   USE sbc_oce_tam   , ONLY: & ! thermohaline fluxes
      & qsr_tl,              &
      & qsr_ad
   USE phycst        , ONLY: & ! physical constants
      & ro0cpr
   USE prtctl        , ONLY: & ! Print control
      & prt_ctl
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE traqsr        , ONLY: & ! Solar radiation penetration
      & ln_traqsr,           &
      & tra_qsr_init,        &
      & rabs,                &
      & xsi1,                &
      & xsi2,                &
      & ln_qsr_sms,          &
      & nksr, gdsr
   USE trc_oce       , ONLY: & ! share SMS/Ocean variables
      & etot3,               &
      & lk_qsr_sms
   USE trc_oce_tam   , ONLY: & ! share SMS/Ocean variables
      & trc_oce_tam_init,    &
      & etot3_tl,            &
      & etot3_ad
   USE tstool_tam    , ONLY: &
      & prntst_adj,          &
      & stdqsr,              &
      & stdt
   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_qsr_tan      ! routine called by step_tam.F90 (ln_traqsr=T)
   PUBLIC   tra_qsr_adj      ! routine called by step_tam.F90 (ln_traqsr=T)
   PUBLIC   tra_qsr_adj_tst  ! routine called by tst.F90 

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE tra_qsr_tan( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_qsr_tan  ***
      !!
      !! ** Purpose of the direct routine:   
      !!      Compute the temperature trend due to the solar radiation
      !!      penetration and add it to the general temperature trend.
      !!
      !! ** Method of the direct routine:
      !!      The profile of the solar radiation within the ocean is
      !!      defined through two penetration length scale (xsr1,xsr2) and a 
      !!      ratio (rabs) as :
      !!         I(k) = Qsr*( rabs*EXP(z(k)/xsr1) + (1.-rabs)*EXP(z(k)/xsr2) )
      !!         The temperature trend associated with the solar radiation
      !!      penetration is given by :
      !!            zta = 1/e3t dk[ I ] / (rau0*Cp)
      !!         At the bottom, boudary condition for the radiation is no flux :
      !!      all heat which has not been absorbed in the above levels is put
      !!      in the last ocean level.
      !!         In z-coordinate case, the computation is only done down to the
      !!      level where I(k) < 1.e-15 W/m2. In addition, the coefficients 
      !!      used for the computation are calculated one for once as they
      !!      depends on k only.
      !!
      !! ** Action  : - update ta with the penetrative solar radiation trend
      !!----------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt     ! ocean time-step
      !
      !!
      INTEGER  ::    ji, jj, jk       ! dummy loop indexes
      REAL(wp) ::   zc0 , zc0tl , ztatl       ! temporary scalars
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_qsr_tan : penetration of the surface solar radiation'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~'
         CALL tra_qsr_init
         CALL tra_qsr_init_tan
      ENDIF
      ! ---------------------------------------------- !
      !  Biological fluxes  : all vertical coordinate  !
      ! ---------------------------------------------- !
      IF( lk_qsr_sms .AND. ln_qsr_sms ) THEN      
         !                                                   ! ===============
         DO jk = 1, jpkm1                                    ! Horizontal slab
            !                                                ! ===============
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zc0 = ro0cpr  / fse3t(ji,jj,jk)         ! compute the qsr trend
                  ztatl = zc0 * ( etot3_tl(ji,jj,jk  ) * tmask(ji,jj,jk)     &
                     &        - etot3_tl(ji,jj,jk+1) * tmask(ji,jj,jk+1) )
                  ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl       ! add qsr trend to the temperature trend
               END DO
            END DO
            !                                                ! ===============
         END DO                                              !   End of slab
         !                                                   ! ===============

      ! ---------------------------------------------- !
      !  Ocean alone :
      ! ---------------------------------------------- !
      ELSE
         !                                                ! =================== !
         IF( ln_sco ) THEN                                !    s-coordinate     !
            !                                             ! =================== !
            DO jk = 1, jpkm1
               ta_tl(:,:,jk) = ta_tl(:,:,jk) + etot3_tl(:,:,jk) * qsr(:,:) + etot3(:,:,jk) * qsr_tl(:,:)
            END DO
         ENDIF
         !                                                ! =================== !
         IF( ln_zps ) THEN                                !    partial steps    !
            !                                             ! =================== !
            DO jk = 1, nksr
               DO jj = 2, jpjm1
                  DO ji = fs_2, fs_jpim1   ! vector opt.
                     ! qsr trend from gdsr
                     zc0tl = qsr_tl(ji,jj) / fse3t(ji,jj,jk)
                     ztatl = zc0tl * ( gdsr(jk) * tmask(ji,jj,jk) - gdsr(jk+1) * tmask(ji,jj,jk+1) )
                     ! add qsr trend to the temperature trend
                     ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl
                  END DO
               END DO
            END DO
         ENDIF
         !                                                ! =================== !
         IF( ln_zco ) THEN                                !     z-coordinate    !
            !                                             ! =================== !
            DO jk = 1, nksr
               zc0 = 1. / e3t_0(jk)
               DO jj = 2, jpjm1
                  DO ji = fs_2, fs_jpim1   ! vector opt.
                     ! qsr trend
                     ztatl = qsr_tl(ji,jj) * zc0 * ( gdsr(jk)*tmask(ji,jj,jk) - gdsr(jk+1)*tmask(ji,jj,jk+1) )
                     ! add qsr trend to the temperature trend
                     ta_tl(ji,jj,jk) = ta_tl(ji,jj,jk) + ztatl      
                  END DO
               END DO
            END DO
         ENDIF
         !
      ENDIF

      !
   END SUBROUTINE tra_qsr_tan
   SUBROUTINE tra_qsr_adj( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_qsr_adj  ***
      !!
      !! ** Purpose of the direct routine:   
      !!      Compute the temperature trend due to the solar radiation
      !!      penetration and add it to the general temperature trend.
      !!
      !! ** Method of the direct routine:
      !!      The profile of the solar radiation within the ocean is
      !!      defined through two penetration length scale (xsr1,xsr2) and a 
      !!      ratio (rabs) as :
      !!         I(k) = Qsr*( rabs*EXP(z(k)/xsr1) + (1.-rabs)*EXP(z(k)/xsr2) )
      !!         The temperature trend associated with the solar radiation
      !!      penetration is given by :
      !!            zta = 1/e3t dk[ I ] / (rau0*Cp)
      !!         At the bottom, boudary condition for the radiation is no flux :
      !!      all heat which has not been absorbed in the above levels is put
      !!      in the last ocean level.
      !!         In z-coordinate case, the computation is only done down to the
      !!      level where I(k) < 1.e-15 W/m2. In addition, the coefficients 
      !!      used for the computation are calculated one for once as they
      !!      depends on k only.
      !!
      !! ** Action  : - update ta with the penetrative solar radiation trend
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt     ! ocean time-step
      !
      !!
      INTEGER  ::    ji, jj, jk       ! dummy loop indexes
      REAL(wp) ::   zc0 , zc0ad       ! temporary scalars
      !!----------------------------------------------------------------------
      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_qsr_adj : penetration of the surface solar radiation'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~'
         CALL tra_qsr_init
      ENDIF

      ! ---------------------------------------------- !
      !  Biological fluxes  : all vertical coordinate  !
      ! ---------------------------------------------- !
      IF( lk_qsr_sms .AND. ln_qsr_sms ) THEN  
         !                                                   ! ===============
         DO jk = jpkm1, 1, -1                                ! Horizontal slab
            !                                                ! ===============
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zc0 = ro0cpr  / fse3t(ji,jj,jk)         ! compute the qsr trend
                  etot3_ad(ji,jj,jk  ) = etot3_ad(ji,jj,jk  ) + ta_ad(ji,jj,jk) * zc0 * tmask(ji,jj,jk)
                  etot3_ad(ji,jj,jk+1) = etot3_ad(ji,jj,jk+1) - ta_ad(ji,jj,jk) * zc0 * tmask(ji,jj,jk+1)
               END DO
            END DO
            !                                                ! ===============
         END DO                                              !   End of slab
         !                                                   ! ===============

      ! ---------------------------------------------- !
      !  Ocean alone :
      ! ---------------------------------------------- !
      ELSE
         !                                                ! =================== !
         IF( ln_zco ) THEN                                !     z-coordinate    !
            !                                             ! =================== !
            DO jk = nksr, 1, -1
               zc0 = 1. / e3t_0(jk)
               DO jj = 2, jpjm1
                  DO ji = fs_2, fs_jpim1   ! vector opt.
                     ! qsr trend
                     qsr_ad(ji,jj) = qsr_ad(ji,jj) + ta_ad(ji,jj,jk) * zc0 * ( gdsr(jk)*tmask(ji,jj,jk) - gdsr(jk+1)*tmask(ji,jj,jk+1) )
                  END DO
               END DO
            END DO
         ENDIF
         !                                                ! =================== !
         IF( ln_zps ) THEN                                !    partial steps    !
            !                                             ! =================== !
            DO jk = 1, nksr
               DO jj = 2, jpjm1
                  DO ji = fs_2, fs_jpim1   ! vector opt.
                     ! qsr trend from gdsr
                     zc0ad = ta_ad(ji,jj,jk) * ( gdsr(jk) * tmask(ji,jj,jk) - gdsr(jk+1) * tmask(ji,jj,jk+1) )
                     qsr_ad(ji,jj) = qsr_ad(ji,jj) + zc0ad / fse3t(ji,jj,jk)
                  END DO
               END DO
            END DO
         ENDIF
         !                                                ! =================== !
         IF( ln_sco ) THEN                                !    s-coordinate     !
            !                                             ! =================== !
            DO jk = jpkm1, 1, -1
               etot3_ad(:,:,jk) = etot3_ad(:,:,jk) + ta_ad(:,:,jk) * qsr(:,:)
               qsr_ad(:,:)      = qsr_ad(:,:)      + ta_ad(:,:,jk) * etot3(:,:,jk)
            END DO
         ENDIF
         !
      ENDIF
      IF( kt == nit000 ) THEN
         CALL tra_qsr_init_adj
      ENDIF

   END SUBROUTINE tra_qsr_adj
   SUBROUTINE tra_qsr_adj_tst ( kumadt ) 
      !!-----------------------------------------------------------------------
      !!
      !!          ***  ROUTINE tra_sbc_adj_tst : TEST OF tra_sbc_adj  ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER ::  &
         & jstp,  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zta_tlin,     &! Tangent input : after temperature
         & zta_tlout,    &! Tangent output: after temperature 
         & zta_adout,    &! Adjoint output: after temperature 
         & zta_adin,     &! Adjoint input : after temperature
         & zta            ! temporary after temperature
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zqsr_tlin,    &! Tangent input : solar radiation (w/m2)
         & zqsr_adout,   &! Adjoint output: solar radiation (w/m2)
         & zqsr           ! temporary solar radiation (w/m2)
      REAL(KIND=wp) ::       &
         & zsp1,             & ! scalar product involving the tangent routine
         & zsp2,             & ! scalar product involving the adjoint routine
         & zsp2_1,           & ! scalar product involving the adjoint routine
         & zsp2_2              ! scalar product involving the adjoint routine
      CHARACTER(LEN=14) :: &
         & cl_name

      ALLOCATE( & 
         & zta_tlin(jpi,jpj,jpk),     &
         & zta_tlout(jpi,jpj,jpk),    &
         & zta_adout(jpi,jpj,jpk),    &
         & zta_adin(jpi,jpj,jpk),     &
         & zta(jpi,jpj,jpk),          &
         & zqsr_tlin(jpi,jpj),        &
         & zqsr_adout(jpi,jpj),       &
         & zqsr(jpi,jpj)              &
         & )
      ! Initialize the reference state
      qsr(:,:) = 1.0_wp ! ???
      ! Initialize random field standard deviations
      !=============================================================
      ! 1) dx = ( T ) and dy = ( T )
      !=============================================================

      CALL trc_oce_tam_init( 0 )

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zta_tlin(:,:,:)  = 0.0_wp     
      zta_tlout(:,:,:) = 0.0_wp    
      zta_adout(:,:,:) = 0.0_wp    
      zta_adin(:,:,:)  = 0.0_wp     
      zqsr_adout(:,:)  = 0.0_wp       
      zqsr_tlin(:,:)   = 0.0_wp 
      
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 358606 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zqsr, 'T', 0.0_wp, stdqsr )
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 232567 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zta, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zta_tlin(ji,jj,jk) = zta(ji,jj,jk)
            END DO
         END DO
      END DO 
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zqsr_tlin(ji,jj)  = zqsr(ji,jj)
         END DO
      END DO
      ! Test for time steps nit000 and nit000 + 1 (the matrix changes)
      DO jstp = nit000, nit000 + 1
         !--------------------------------------------------------------------
         ! Call the tangent routine: dy = L dx
         !--------------------------------------------------------------------
         
         ta_tl(:,:,:) = zta_tlin(:,:,:)
         qsr_tl(:,:)  = zqsr_tlin(:,:)
      
         CALL tra_qsr_tan( jstp )
         
         zta_tlout(:,:,:) = ta_tl(:,:,:)
         
         !--------------------------------------------------------------------
         ! Initialize the adjoint variables: dy^* = W dy
         !--------------------------------------------------------------------
      
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_adin(ji,jj,jk) = zta_tlout(ji,jj,jk) &
                     &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                     &               * tmask(ji,jj,jk)
               END DO
            END DO
         END DO

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L dx )^T W dy
         !--------------------------------------------------------------------

         zsp1 = DOT_PRODUCT( zta_tlout, zta_adin )

         !--------------------------------------------------------------------
         ! Call the adjoint routine: dx^* = L^T dy^*
         !--------------------------------------------------------------------

         qsr_ad(:,:)      = 0.0_wp       
         ta_ad(:,:,:) = zta_adin(:,:,:)
         
         CALL tra_qsr_adj( jstp )

         zta_adout(:,:,:) = ta_ad(:,:,:)
         zqsr_adout(:,:)  = qsr_ad(:,:)
      
         !--------------------------------------------------------------------
         ! Compute the scalar product: dx^T L^T W dy
         !--------------------------------------------------------------------

         zsp2_1 = DOT_PRODUCT( zta_tlin  , zta_adout   )
         zsp2_2 = DOT_PRODUCT( zqsr_tlin , zqsr_adout  )
      
         zsp2 = zsp2_1 + zsp2_2
      
         ! Compare the scalar products
      
         ! 14 char:   '12345678901234'
         IF (jstp == nit000) THEN
            cl_name = 'tra_qsr_adj  1'
         ELSE
            cl_name = 'tra_qsr_adj  2'
         END IF
         CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )
      END DO
      
      DEALLOCATE( & 
         & zta_tlin,     &
         & zta_tlout,    &
         & zta_adout,    &
         & zta_adin,     &
         & zta,          &
         & zqsr_adout,   &
         & zqsr_tlin,    &
         & zqsr          &
         & )

      !
   END SUBROUTINE tra_qsr_adj_tst
   SUBROUTINE tra_qsr_init_tan
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_qsr_init_tan  ***
      !!
      !! ** Purpose :   Initialization for the penetrative solar radiation
      !!
      !! ** Method  :   The profile of solar radiation within the ocean is set
      !!      from two length scale of penetration (xsr1,xsr2) and a ratio
      !!      (rabs). These parameters are read in the namqsr namelist. The
      !!      default values correspond to clear water (type I in Jerlov' 
      !!      (1968) classification.
      !!         called by tra_qsr at the first timestep (nit000)
      !!
      !! ** Action  : - initialize xsr1, xsr2 and rabs
      !!
      !! Reference : Jerlov, N. G., 1968 Optical Oceanography, Elsevier, 194pp.
      !!----------------------------------------------------------------------
      INTEGER  ::   ji, jj, jk   ! dummy loop index
      INTEGER  ::   indic        ! temporary integer
      REAL(wp) ::   zcst, zdp1    ! temporary scalars

      !                           ! Initialization of gdsr
      IF( ln_zco .OR. ln_zps ) THEN
         !
         ! Initialisation of Biological fluxes for light here because
         ! the optical biological model is call after the dynamical one
         IF( lk_qsr_sms .AND. ln_qsr_sms ) THEN
            DO jk = 1, jpkm1
               zcst = gdsr(jk) / ro0cpr
               etot3_tl(:,:,jk) = qsr_tl(:,:) * zcst * tmask(:,:,jk)
            END DO
         ENDIF
         !
      ENDIF

      ! Initialisation of etot3 (s-coordinate)
      ! -----------------------
      IF( ln_sco ) THEN
         etot3_tl(:,:,:) = 0.e0
      ENDIF
      !
   END SUBROUTINE tra_qsr_init_tan
   SUBROUTINE tra_qsr_init_adj
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_qsr_init_adj  ***
      !!
      !! ** Purpose :   Initialization for the penetrative solar radiation
      !!
      !! ** Method  :   The profile of solar radiation within the ocean is set
      !!      from two length scale of penetration (xsr1,xsr2) and a ratio
      !!      (rabs). These parameters are read in the namqsr namelist. The
      !!      default values correspond to clear water (type I in Jerlov' 
      !!      (1968) classification.
      !!         called by tra_qsr at the first timestep (nit000)
      !!
      !! ** Action  : - initialize xsr1, xsr2 and rabs
      !!
      !! Reference : Jerlov, N. G., 1968 Optical Oceanography, Elsevier, 194pp.
      !!----------------------------------------------------------------------
      INTEGER  ::   ji, jj, jk   ! dummy loop index
      INTEGER  ::   indic        ! temporary integer
      REAL(wp) ::   zcst, zdp1    ! temporary scalars

      ! Initialisation of etot3 (s-coordinate)
      ! -----------------------
      IF( ln_sco ) THEN
         etot3_ad(:,:,:) = 0.e0
      ENDIF
      !                           ! Initialization of gdsr
      IF( ln_zco .OR. ln_zps ) THEN
         !
         ! z-coordinate with or without partial step : same w-level everywhere inside the ocean
         ! Initialisation of Biological fluxes for light here because
         ! the optical biological model is call after the dynamical one
         IF( lk_qsr_sms .AND. ln_qsr_sms ) THEN
            DO jk = 1, jpkm1
               zcst = gdsr(jk) / ro0cpr
               qsr_ad(:,:) = qsr_ad(:,:) + etot3_ad(:,:,jk)  * zcst * tmask(:,:,jk)
               etot3_ad(:,:,jk) = 0.0_wp
            END DO
         ENDIF
         !
      ENDIF

      !
  END SUBROUTINE tra_qsr_init_adj

   !!======================================================================
#endif
END MODULE traqsr_tam
