MODULE trasbc_tam
#ifdef key_tam
   !!==============================================================================
   !!                       ***  MODULE  trasbc_tam  ***
   !! Ocean active tracers:  surface boundary condition
   !!                        Tangent and Adjoint Module
   !!==============================================================================
   !! History of the direct module:
   !!            8.2  !  98-10  (G. Madec, G. Roullet, M. Imbard)  Original code
   !!            8.2  !  01-02  (D. Ludicone)  sea ice and free surface
   !!            8.5  !  02-06  (G. Madec)  F90: Free form and module
   !! History of the TAM:
   !!                 !  08-05  (A. Vidard) Skeleton
   !!                 !  08-11  (A. Vidard) tam of the 02-06 version
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_sbc      : update the tracer trend at ocean surface
   !!----------------------------------------------------------------------
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 &
      & jpk,                 & 
      & jpim1,               &
      & jpjm1,               &
      & jpiglo
   USE oce           , ONLY: & ! ocean dynamics and active tracers
      & sn
   USE oce_tam       , ONLY: & ! tangent and adjoint ocean dynamics and active tracers
      & sn_tl,               &
      & sa_tl,               &
      & ta_tl,               &
      & sn_ad,               &
      & sa_ad,               &
      & ta_ad
   USE sbc_oce_tam   , ONLY: & ! surface thermohaline fluxes
      & qns_tl,              &
      & qsr_tl,              &
      & emps_tl,             &
      & qns_ad,              &
      & qsr_ad,              &
      & emps_ad
   USE sbc_oce       , ONLY: & ! surface thermohaline fluxes
      & emps
   USE dom_oce       , ONLY: & ! ocean space domain variables
      & rdt,                 &
      & e1t,                 &
      & e2t,                 &
#if defined key_zco
      & e3t_0,               &
#else
      & e3t,                 &
#endif
      & tmask,               &
      & lk_vvl,              &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE traqsr        , ONLY: & ! solar radiation penetration
      & ln_traqsr
   USE phycst        , ONLY: & ! physical constant
      & rauw,                &
      & ro0cpr
   USE in_out_manager, ONLY: & ! I/O manager 
      & ctl_stop,            &
      & lwp,                 &
      & numout,              & 
      & nit000,              &
      & nitend
   USE prtctl        , ONLY: & ! Print control
      & prt_ctl
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE paresp        , ONLY: & ! Weights for an energy-type scalar product
      & wesp_t,              &
      & wesp_s
   USE tstool_tam    , ONLY: &
      & prntst_adj,          &
      & prntst_tlm,          & !
      & stdqns,              &
      & stdt,                &
      & stds,                &
      & stdemp

   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_sbc_tan     ! routine called by step_tam.F90
   PUBLIC   tra_sbc_adj     ! routine called by step_tam.F90
   PUBLIC   tra_sbc_adj_tst ! routine called by tst.F90
#if defined key_tst_tlm
   PUBLIC   tra_sbc_tlm_tst ! routine calle  by tamtst.F90
#endif

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"


CONTAINS

   SUBROUTINE tra_sbc_tan ( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_sbc_tan  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the tracer surface boundary condition trend of
      !!      (flux through the interface, concentration/dilution effect)
      !!      and add it to the general trend of tracer equations.
      !!
      !! ** Method :
      !!      Following Roullet and Madec (2000), the air-sea flux can be divided 
      !!      into three effects: (1) Fext, external forcing; 
      !!      (2) Fwi, concentration/dilution effect due to water exchanged 
      !!         at the surface by evaporation, precipitations and runoff (E-P-R); 
      !!      (3) Fwe, tracer carried with the water that is exchanged. 
      !!
      !!      Fext, flux through the air-sea interface for temperature and salt: 
      !!            - temperature : heat flux q (w/m2). If penetrative solar
      !!         radiation q is only the non solar part of the heat flux, the
      !!         solar part is added in traqsr.F routine.
      !!            ta = ta + q /(rau0 rcp e3t)  for k=1
      !!            - salinity    : no salt flux
      !!
      !!      The formulation for Fwb and Fwi vary according to the free 
      !!      surface formulation (linear or variable volume). 
      !!      * Linear free surface
      !!            The surface freshwater flux modifies the ocean volume
      !!         and thus the concentration of a tracer and the temperature.
      !!         First order of the effect of surface freshwater exchange 
      !!         for salinity, it can be neglected on temperature (especially
      !!         as the temperature of precipitations and runoffs is usually
      !!         unknown).
      !!            - temperature : we assume that the temperature of both
      !!         precipitations and runoffs is equal to the SST, thus there
      !!         is no additional flux since in this case, the concentration
      !!         dilution effect is balanced by the net heat flux associated
      !!         to the freshwater exchange (Fwe+Fwi=0):
      !!            (Tp P - Te E) + SST (P-E) = 0 when Tp=Te=SST
      !!            - salinity    : evaporation, precipitation and runoff
      !!         water has a zero salinity (Fwe=0), thus only Fwi remains:
      !!            sa = sa + emp * sn / e3t   for k=1
      !!         where emp, the surface freshwater budget (evaporation minus
      !!         precipitation minus runoff) given in kg/m2/s is divided
      !!         by 1000 kg/m3 (density of plain water) to obtain m/s.    
      !!         Note: even though Fwe does not appear explicitly for 
      !!         temperature in this routine, the heat carried by the water
      !!         exchanged through the surface is part of the total heat flux
      !!         forcing and must be taken into account in the global heat
      !!         balance).
      !!      * nonlinear free surface (variable volume, lk_vvl)
      !!         contrary to the linear free surface case, Fwi is properly 
      !!         taken into account by using the true layer thicknesses to       
      !!         calculate tracer content and advection. There is no need to 
      !!         deal with it in this routine.
      !!           - temperature: Fwe=SST (P-E+R) is added to Fext.
      !!           - salinity:  Fwe = 0, there is no surface flux of salt.
      !!
      !! ** Action  : - Update the 1st level of (ta,sa) with the trend associated
      !!                with the tracer surface boundary condition 
      !!              - save the trend it in ttrd ('key_trdtra')
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt     ! ocean time-step index
      !!
      INTEGER  ::   ji, jj                   ! dummy loop indices
      REAL(wp) ::   ztatl, zsatl, zsrau, zse3t   ! temporary scalars

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_sbc_tan : TRAcer Surface Boundary Condition'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~ '
      ENDIF

      zsrau = 1.0_wp / rauw             ! initialization
#if defined key_zco
      zse3t = 1.0_wp / e3t_0(1)
#endif

      IF( .NOT.ln_traqsr )   qsr_tl(:,:) = 0.e0_wp   ! no solar radiation penetration

      ! Concentration dillution effect on (t,s)
      DO jj = 2, jpj
         DO ji = fs_2, fs_jpim1   ! vector opt.
#if ! defined key_zco
            zse3t = 1.0_wp / fse3t(ji,jj,1)
#endif
            IF( lk_vvl ) THEN
               CALL ctl_stop( 'key_vvl  not available in NEMOTAM' )
            ELSE
               ! temperature : heat flux
               ztatl = ro0cpr * zse3t * qns_tl(ji,jj)         

               ! salinity :  concent./dilut. effect
               zsatl = (   emps_tl(ji,jj) * sn   (ji,jj,1) &  
                  &      + emps   (ji,jj) * sn_tl(ji,jj,1) ) * zsrau * zse3t     
            ENDIF
            ta_tl(ji,jj,1) = ta_tl(ji,jj,1) + ztatl    ! add the trend to the general tracer trend
            sa_tl(ji,jj,1) = sa_tl(ji,jj,1) + zsatl
         END DO
      END DO

   END SUBROUTINE tra_sbc_tan

   SUBROUTINE tra_sbc_adj ( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_sbc_adj  ***
      !!                   
      !! ** Purpose of the direct routine:
      !!      Compute the tracer surface boundary condition trend of
      !!      (flux through the interface, concentration/dilution effect)
      !!      and add it to the general trend of tracer equations.
      !!
      !! ** Method :
      !!      Following Roullet and Madec (2000), the air-sea flux can be divided 
      !!      into three effects: (1) Fext, external forcing; 
      !!      (2) Fwi, concentration/dilution effect due to water exchanged 
      !!         at the surface by evaporation, precipitations and runoff (E-P-R); 
      !!      (3) Fwe, tracer carried with the water that is exchanged. 
      !!
      !!      Fext, flux through the air-sea interface for temperature and salt: 
      !!            - temperature : heat flux q (w/m2). If penetrative solar
      !!         radiation q is only the non solar part of the heat flux, the
      !!         solar part is added in traqsr.F routine.
      !!            ta = ta + q /(rau0 rcp e3t)  for k=1
      !!            - salinity    : no salt flux
      !!
      !!      The formulation for Fwb and Fwi vary according to the free 
      !!      surface formulation (linear or variable volume). 
      !!      * Linear free surface
      !!            The surface freshwater flux modifies the ocean volume
      !!         and thus the concentration of a tracer and the temperature.
      !!         First order of the effect of surface freshwater exchange 
      !!         for salinity, it can be neglected on temperature (especially
      !!         as the temperature of precipitations and runoffs is usually
      !!         unknown).
      !!            - temperature : we assume that the temperature of both
      !!         precipitations and runoffs is equal to the SST, thus there
      !!         is no additional flux since in this case, the concentration
      !!         dilution effect is balanced by the net heat flux associated
      !!         to the freshwater exchange (Fwe+Fwi=0):
      !!            (Tp P - Te E) + SST (P-E) = 0 when Tp=Te=SST
      !!            - salinity    : evaporation, precipitation and runoff
      !!         water has a zero salinity (Fwe=0), thus only Fwi remains:
      !!            sa = sa + emp * sn / e3t   for k=1
      !!         where emp, the surface freshwater budget (evaporation minus
      !!         precipitation minus runoff) given in kg/m2/s is divided
      !!         by 1000 kg/m3 (density of plain water) to obtain m/s.    
      !!         Note: even though Fwe does not appear explicitly for 
      !!         temperature in this routine, the heat carried by the water
      !!         exchanged through the surface is part of the total heat flux
      !!         forcing and must be taken into account in the global heat
      !!         balance).
      !!      * nonlinear free surface (variable volume, lk_vvl)
      !!         contrary to the linear free surface case, Fwi is properly 
      !!         taken into account by using the true layer thicknesses to       
      !!         calculate tracer content and advection. There is no need to 
      !!         deal with it in this routine.
      !!           - temperature: Fwe=SST (P-E+R) is added to Fext.
      !!           - salinity:  Fwe = 0, there is no surface flux of salt.
      !!
      !! ** Action  : - Update the 1st level of (ta,sa) with the trend associated
      !!                with the tracer surface boundary condition 
      !!              - save the trend it in ttrd ('key_trdtra')
      !!----------------------------------------------------------------------
      !!
      INTEGER, INTENT(in) ::   kt     ! ocean time-step index
      !!
      INTEGER  ::   ji, jj                   ! dummy loop indices
      REAL(wp) ::   ztaad, zsaad, zsrau, zse3t   ! temporary scalars

      IF( kt == nitend ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'tra_sbc_adj : TRAcer Surface Boundary Condition'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~ '
      ENDIF

      zsrau = 1.0_wp / rauw             ! initialization
#if defined key_zco
      zse3t = 1.0_wp / e3t_0(1)
#endif

      ! Concentration dillution effect on (t,s)
      DO jj = jpj, 2, -1
         DO ji = fs_jpim1, fs_2, -1   ! vector opt.
#if ! defined key_zco
            zse3t = 1.0_wp / fse3t(ji,jj,1)
#endif
            ztaad = ta_ad(ji,jj,1)          ! add the trend to the general tracer trend
            zsaad = sa_ad(ji,jj,1) 
            IF( lk_vvl) THEN
               CALL ctl_stop( 'key_vvl  not available in NEMOTAM' )
            ELSE
               ztaad = ztaad * ro0cpr * zse3t
               zsaad = zsaad * zsrau  * zse3t
               qns_ad(ji,jj)  = qns_ad(ji,jj)  + ztaad
               emps_ad(ji,jj) = emps_ad(ji,jj) + zsaad * sn(ji,jj,1)
               sn_ad(ji,jj,1) = sn_ad(ji,jj,1) + zsaad * emps(ji,jj)
            ENDIF
         END DO
      END DO

      IF( .NOT.ln_traqsr )   qsr_ad(:,:) = 0.e0_wp   ! no solar radiation penetration

   END SUBROUTINE tra_sbc_adj

   SUBROUTINE tra_sbc_adj_tst ( kumadt ) 
      !!-----------------------------------------------------------------------
      !!
      !!          ***  ROUTINE tra_sbc_adj_tst : TEST OF tra_sbc_adj  ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER :: &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zsn_tlin,     &! Tangent input : now salinity
         & zsa_tlin,     &! Tangent input : after salinity
         & zta_tlin,     &! Tangent input : after temperature
         & zqns_tlin,    &! Tangent input : solar radiation (w/m2)
         & zemps_tlin,   &! Tangent input : evaporation - precipitation (free surface)
         & zsa_tlout,    &! Tangent output: after salinity 
         & zta_tlout,    &! Tangent output: after temperature 
         & zsa_adin,     &! Adjoint input : after salinity
         & zta_adin,     &! Adjoint input : after temperature
         & zsn_adout,    &! Adjoint output: now salinity
         & zsa_adout,    &! Adjoint output: after salinity 
         & zta_adout,    &! Adjoint output: after temperature 
         & zqns_adout,   &! Adjoint output: solar radiation (w/m2)
         & zemps_adout,  &! Adjoint output: evaporation - precipitation (free surface)
         & zsn,          &! temporary now salinity
         & zsa,          &! temporary after salinity
         & zta,          &! temporary after temperature
         & zqns,         &! temporary solar radiation (w/m2)
         & zemps          ! temporary evaporation - precipitation (free surface)
      REAL(KIND=wp) ::       &
         & zsp1,             & ! scalar product involving the tangent routine
         & zsp1_1,           & ! scalar product involving the tangent routine
         & zsp1_2,           & ! scalar product involving the tangent routine
         & zsp2,             & ! scalar product involving the adjoint routine
         & zsp2_1,           & ! scalar product involving the adjoint routine
         & zsp2_2,           & ! scalar product involving the adjoint routine
         & zsp2_3,           & ! scalar product involving the adjoint routine
         & zsp2_4,           & ! scalar product involving the adjoint routine
         & zsp2_5,           & ! scalar product involving the adjoint routine
         & z2dt,             & ! temporary scalars
         & zraur
      CHARACTER (LEN=14) :: &
         & cl_name

      ALLOCATE( & 
         & zsn_tlin(jpi,jpj),     &
         & zsa_tlin(jpi,jpj),     &
         & zta_tlin(jpi,jpj),     &
         & zsa_tlout(jpi,jpj),    &
         & zta_tlout(jpi,jpj),    &
         & zsn_adout(jpi,jpj),    &
         & zsa_adout(jpi,jpj),    &
         & zta_adout(jpi,jpj),    &
         & zsa_adin(jpi,jpj),     &
         & zta_adin(jpi,jpj),     &
         & zsn(jpi,jpj),          &
         & zsa(jpi,jpj),          &
         & zta(jpi,jpj),          &
         & zqns_adout(jpi,jpj),   &
         & zqns_tlin(jpi,jpj),    &
         & zemps_tlin(jpi,jpj),   &
         & zemps_adout(jpi,jpj),  &
         & zqns(jpi,jpj),         &
         & zemps(jpi,jpj)         &
         & )


      ! Initialize constants
      z2dt  = 2.0_wp * rdt       ! time step: leap-frog
      zraur = 1.0_wp / rauw      ! inverse density of pure water (m3/kg)

      ! Initialize the reference state

      !===========================================================================
      ! 1) dx = ( qns_tl, sn_tl, emps_tl, ta_tl, sa_tl ) and dy = ( ta_tl, sa_tl )
      !===========================================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      zsn_tlin   (:,:) = 0.0_wp     
      zsa_tlin   (:,:) = 0.0_wp     
      zta_tlin   (:,:) = 0.0_wp     
      zsa_tlout  (:,:) = 0.0_wp    
      zta_tlout  (:,:) = 0.0_wp    
      zsn_adout  (:,:) = 0.0_wp    
      zsa_adout  (:,:) = 0.0_wp    
      zta_adout  (:,:) = 0.0_wp    
      zsa_adin   (:,:) = 0.0_wp     
      zta_adin   (:,:) = 0.0_wp     
      zqns_tlin  (:,:) = 0.0_wp        
      zqns_adout (:,:) = 0.0_wp       
      zemps_tlin (:,:) = 0.0_wp       
      zemps_adout(:,:) = 0.0_wp      
      
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 785483 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zemps, 'T', 0.0_wp, stdemp )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 358606 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zqns, 'T', 0.0_wp, stdqns )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 523432 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zsn, 'T', 0.0_wp, stds )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 297563 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zsa, 'T', 0.0_wp, stds )

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 232567 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zta, 'T', 0.0_wp, stdt )

      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zsn_tlin  (ji,jj) = zsn  (ji,jj)
            zsa_tlin  (ji,jj) = zsa  (ji,jj)
            zta_tlin  (ji,jj) = zta  (ji,jj)
            zemps_tlin(ji,jj) = zemps(ji,jj) / ( z2dt * zraur )
            zqns_tlin (ji,jj) = zqns (ji,jj)
         END DO
      END DO 

      !--------------------------------------------------------------------
      ! Call the tangent routine: dy = L dx
      !--------------------------------------------------------------------

      sn_tl  (:,:,1) = zsn_tlin  (:,:)
      sa_tl  (:,:,1) = zsa_tlin  (:,:)
      ta_tl  (:,:,1) = zta_tlin  (:,:)
      emps_tl(:,:)   = zemps_tlin(:,:)
      qns_tl (:,:)   = zqns_tlin (:,:)

      CALL tra_sbc_tan( nit000 )

      zsa_tlout(:,:) = sa_tl(:,:,1) 
      zta_tlout(:,:) = ta_tl(:,:,1)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jj = nldj, nlej
         DO ji = nldi, nlei
            zsa_adin(ji,jj) = zsa_tlout(ji,jj) &
               &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,1) &
               &               * tmask(ji,jj,1) * wesp_s(1)
            zta_adin(ji,jj) = zta_tlout(ji,jj) &
               &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,1) &
               &               * tmask(ji,jj,1) * wesp_t(1)
         END DO
      END DO

      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1_1 = DOT_PRODUCT( zsa_tlout, zsa_adin )
      zsp1_2 = DOT_PRODUCT( zta_tlout, zta_adin )
      zsp1   = zsp1_1 + zsp1_2

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      sa_ad(:,:,1) = zsa_adin(:,:)
      ta_ad(:,:,1) = zta_adin(:,:)
      sn_ad(:,:,1) = 0.0_wp
      emps_ad(:,:) = 0.0_wp
      qns_ad(:,:)  = 0.0_wp

      CALL tra_sbc_adj( nitend )

      zsn_adout  (:,:) = sn_ad(:,:,1)
      zsa_adout  (:,:) = sa_ad(:,:,1)
      zta_adout  (:,:) = ta_ad(:,:,1)
      zqns_adout (:,:) = qns_ad(:,: )
      zemps_adout(:,:) = emps_ad(:,:)     

      !--------------------------------------------------------------------
      ! Compute the scalar product: dx^T L^T W dy
      !--------------------------------------------------------------------

      zsp2_1 = DOT_PRODUCT( zsn_tlin  , zsn_adout   )
      zsp2_2 = DOT_PRODUCT( zsa_tlin  , zsa_adout   )
      zsp2_3 = DOT_PRODUCT( zta_tlin  , zta_adout   )
      zsp2_4 = DOT_PRODUCT( zqns_tlin , zqns_adout  )
      zsp2_5 = DOT_PRODUCT( zemps_tlin, zemps_adout )

      zsp2 = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4 + zsp2_5

      ! Compare the scalar products

      ! 14 char:'12345678901234'
      cl_name = 'tra_sbc_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )



      DEALLOCATE( & 
         & zsn_tlin,     &
         & zsa_tlin,     &
         & zta_tlin,     &
         & zsa_tlout,    &
         & zta_tlout,    &
         & zsn_adout,    &
         & zsa_adout,    &
         & zta_adout,    &
         & zsa_adin,     &
         & zta_adin,     &
         & zsn,          &
         & zsa,          &
         & zta,          &
         & zqns_adout,   &
         & zqns_tlin,    &
         & zemps_tlin,   &
         & zemps_adout,  &
         & zqns,         &
         & zemps         &
         & )
   END SUBROUTINE tra_sbc_adj_tst

#if defined key_tst_tlm
   SUBROUTINE tra_sbc_tlm_tst ( kumadt ) 
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_adv_tlm_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!                    
      !! History :
      !!        ! 09-08 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE trasbc
      USE tamtrj              ! writing out state trajectory
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE divcur             ! horizontal divergence and relative vorticity
      USE wzvmod             !  vertical velocity
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
        & sn, ta, sa
      USE sbc_oce       , ONLY: & ! ocean boundary condition
        & qns, emps
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER :: &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator

      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zsn_tlin,     &! Tangent input : now salinity
         & zsa_tlin,     &! Tangent input : after salinity
         & zta_tlin,     &! Tangent input : after temperature
         & zsa_out,      &! Tangent output: after salinity 
         & zta_out,      &! Tangent output: after temperature 
         & zsa_wop,      &! 
         & zta_wop,      &
         & z3r            ! 3D field   
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zqns_tlin,    &! Tangent input : solar radiation (w/m2)
         & zemps_tlin,   &! Tangent input : evaporation - precipitation (free surface)
         & z2r     
      REAL(KIND=wp) ::   &
         & zsp1,         & ! scalar product 
         & zsp1_1,       & ! scalar product
         & zsp1_2,       & ! scalar product
         & zsp2,         & ! scalar product
         & zsp2_1,       & ! scalar product
         & zsp2_2,       & ! scalar product 
         & zsp3,         & ! scalar product
         & zsp3_1,       & ! scalar product 
         & zsp3_2,       & ! scalar product
         & zzsp,         & ! scalar product
         & zzsp_1,       & ! scalar product 
         & zzsp_2,       & ! scalar product 
         & zraur,        &
         & gamma,        &
         & zgsp1,        &
         & zgsp2,        &
         & zgsp3,        &
         & zgsp4,        &
         & zgsp5,        &
         & zgsp6,        &
         & zgsp7 
      CHARACTER (LEN=14)  :: cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_xdx
      CHARACTER (LEN=90)  ::  FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zscta, zscsa, &
         & zscerrta,     &
         & zscerrsa
      INTEGER, DIMENSION(100):: &
         & iiposta, iipossa,    &
         & ijposta, ijpossa,    & 
         & ikposta, ikpossa
      INTEGER::          &
         & ii,           &
         & isamp=40,     &
         & jsamp=40,     &
         & ksamp=10,     &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj,jpk) :: &
         & zerrta, zerrsa 

      ALLOCATE( & 
         & zsn_tlin ( jpi,jpj,jpk),     &
         & zta_tlin ( jpi,jpj,jpk),     &
         & zsa_tlin ( jpi,jpj,jpk),     &
         & zta_out  ( jpi,jpj,jpk),     &
         & zsa_out  ( jpi,jpj,jpk),     &
         & zta_wop  ( jpi,jpj,jpk),     &
         & zsa_wop  ( jpi,jpj,jpk),     &
         & z3r      ( jpi,jpj,jpk),     &
         & zqns_tlin( jpi,jpj    ),     &
         & zemps_tlin(jpi,jpj    ),     &
         & z2r       (jpi,jpj    )      &
         & )

      !--------------------------------------------------------------------
      ! Reset variables
      !--------------------------------------------------------------------
      zsn_tlin  ( :,:,:) = 0.0_wp
      zsa_tlin  ( :,:,:) = 0.0_wp
      zta_tlin  ( :,:,:) = 0.0_wp
      zqns_tlin ( :,:  ) = 0.0_wp
      zemps_tlin( :,:  ) = 0.0_wp
      zta_out   ( :,:,:) = 0.0_wp
      zsa_out   ( :,:,:) = 0.0_wp
      zta_wop   ( :,:,:) = 0.0_wp
      zsa_wop   ( :,:,:) = 0.0_wp

      zscta(:)         = 0.0_wp
      zscsa(:)         = 0.0_wp
      zscerrta(:)      = 0.0_wp
      zscerrsa(:)      = 0.0_wp
      zerrta(:,:,:)    = 0.0_wp
      zerrsa(:,:,:)    = 0.0_wp
      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio
      file_wop='trj_wop_trasbc'
      file_xdx='trj_xdx_trasbc'
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 523432, z3r,    'T', 0.0_wp, stds  )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 232567, z3r,    'T', 0.0_wp, stdt  )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 297563, z3r,    'T', 0.0_wp, stds  )
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsa_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 358606, z2r,   'T', 0.0_wp, stdqns) 
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zqns_tlin(ji,jj) = z2r(ji,jj)
            END DO
         END DO
         CALL grid_rd_sd( 785483, z2r,  'T', 0.0_wp, stdemp)
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               zemps_tlin(ji,jj) = z2r(ji,jj)
            END DO
         END DO
      ENDIF    
      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      IF ( tlm_bch /= 2 )      CALL istate_p  

      ! *** initialize the reference trajectory
      ! ------------
      CALL  trj_rea( nit000-1, 1 )
      CALL  trj_rea( nit000, 1 )

      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         zsn_tlin(:,:,:) = gamma * zsn_tlin(:,:,:)
         sn(:,:,:)       = sn(:,:,:) + zsn_tlin(:,:,:)

         zta_tlin(:,:,:) = gamma * zta_tlin(:,:,:)
         ta(:,:,:)       = ta(:,:,:) + zta_tlin(:,:,:)

         zsa_tlin(:,:,:) = gamma * zsa_tlin(:,:,:)
         sa(:,:,:)       = sa(:,:,:) + zsa_tlin(:,:,:)

         zqns_tlin(:,:) = gamma * zqns_tlin(:,:)
         qns(:,:)         = qns(:,:) + zqns_tlin(:,:)

         zemps_tlin(:,:) = gamma * zemps_tlin(:,:)
         emps(:,:)       = emps(:,:) + zemps_tlin(:,:)
      ENDIF 
      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      IF ( tlm_bch /= 2 ) CALL tra_sbc(nit000)
      IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN         
         !--------------------------------------------------------------------
         ! Initialize the tangent variables: dy^* = W dy  
         !--------------------------------------------------------------------
         CALL  trj_rea( nit000-1, 1 ) 
         CALL  trj_rea( nit000, 1 ) 
         sn_tl   (:,:,:) = zsn_tlin  (:,:,:)
         ta_tl   (:,:,:) = zta_tlin  (:,:,:)
         sa_tl   (:,:,:) = zsa_tlin  (:,:,:)
         qns_tl  (:,:  ) = zqns_tlin (:,:  )
         emps_tl (:,:  ) = zemps_tlin(:,:  )
         !-----------------------------------------------------------------------
         !  Initialization of the dynamics and tracer fields for the tangent
         !----------------------------------------------------------------------- 
         CALL tra_sbc_tan(nit000)

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------
         zsp2_1    = DOT_PRODUCT( ta_tl, ta_tl  )
         zsp2_2    = DOT_PRODUCT( sa_tl, sa_tl  )

         zsp2      = zsp2_1 + zsp2_2
         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------
         CALL trj_rd_spl(file_wop) 
         zta_wop  (:,:,:) = ta  (:,:,:)
         zsa_wop  (:,:,:) = sa  (:,:,:)
         CALL trj_rd_spl(file_xdx) 
         zta_out  (:,:,:) = ta  (:,:,:)
         zsa_out  (:,:,:) = sa  (:,:,:)
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zta_out   (ji,jj,jk) = zta_out    (ji,jj,jk) - zta_wop  (ji,jj,jk)
                  zta_wop   (ji,jj,jk) = zta_out    (ji,jj,jk) - ta_tl    (ji,jj,jk)
                  IF (  ta_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrta(ji,jj,jk) = zta_out(ji,jj,jk)/ta_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iiposta(ii) = ji
                      ijposta(ii) = jj
                      ikposta(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                         zscta (ii) =  zta_wop(ji,jj,jk)
                         zscerrta (ii) =  ( zerrta(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zsa_out   (ji,jj,jk) = zsa_out    (ji,jj,jk) - zsa_wop  (ji,jj,jk)
                  zsa_wop   (ji,jj,jk) = zsa_out    (ji,jj,jk) - sa_tl    (ji,jj,jk)
                  IF (  sa_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrsa(ji,jj,jk) = zsa_out(ji,jj,jk)/sa_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iipossa(ii) = ji
                      ijpossa(ii) = jj
                      ikpossa(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                         zscsa (ii) =  zsa_wop(ji,jj,jk)
                         zscerrsa (ii) =  ( zerrsa(ji,jj,jk) - 1.0_wp ) /gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO

         zsp1_1    = DOT_PRODUCT( zta_out, zta_out )
         zsp1_2    = DOT_PRODUCT( zsa_out, zsa_out  )
         zsp1      = zsp1_1 + zsp1_2

         zsp3_1    = DOT_PRODUCT( zta_wop, zta_wop )
         zsp3_2    = DOT_PRODUCT( zsa_wop, zsa_wop  )
         zsp3      = zsp3_1 + zsp3_2
         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'trasbc_tam:En '
         zzsp    = SQRT(zsp3)
         zzsp_1  = SQRT(zsp3_1)
         zzsp_2  = SQRT(zsp3_2)
         zgsp5   = zzsp
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp    = SQRT(zsp2)
         zzsp_1  = SQRT(zsp2_1)
         zzsp_2  = SQRT(zsp2_2)
         zgsp4   = zzsp
         cl_name = 'trasbc_tam:Ln2'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         zzsp    = SQRT(zsp1)
         zzsp_1  = SQRT(zsp1_1)
         zzsp_2  = SQRT(zsp1_2)
         cl_name = 'trasbc:Mhdx-Mx'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         zgsp3   = SQRT( zsp3/zsp2 )
         zgsp7   = zgsp3/gamma
         zgsp1   = zzsp
         zgsp2   = zgsp1 / zgsp4
         zgsp6   = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'trasbc  ', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7
         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'trasbc  '
         IF(lwp) THEN
            DO ii=1, 100, 1
               IF ( zscta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscta     ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscsa     ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscsa(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrta  ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscerrta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrsa  ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscerrsa(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc,"(A4)") '===='
         ENDIF
      ENDIF

      DEALLOCATE(                           &
         & zsn_tlin, zta_tlin, zsa_tlin,    & 
         & zqns_tlin, zemps_tlin,           & 
         & zta_out, zsa_out,                & 
         & zta_wop, zsa_wop,                &
         & z3r, z2r                         &
         & )
   END SUBROUTINE tra_sbc_tlm_tst
!!======================================================================
#endif
#endif
END MODULE trasbc_tam
