MODULE trazdf_imp_tam
#ifdef key_tam
   !!==============================================================================
   !!                 ***  MODULE  trazdf_imp_tam  ***
   !! Ocean active tracers:  vertical component of the tracer mixing trend
   !!                        Tangent and Adjoint Module
   !!==============================================================================
   !! History of the direct module:
   !!   6.0  !  90-10  (B. Blanke)  Original code
   !!   7.0  !  91-11 (G. Madec)
   !!        !  92-06 (M. Imbard) correction on tracer trend loops
   !!        !  96-01 (G. Madec) statement function for e3
   !!        !  97-05 (G. Madec) vertical component of isopycnal
   !!        !  97-07 (G. Madec) geopotential diffusion in s-coord
   !!        !  00-08 (G. Madec) double diffusive mixing
   !!   8.5  !  02-08 (G. Madec)  F90: Free form and module
   !!   9.0  !  06-11 (G. Madec)  New step reorganisation
   !! History of the T&A module:
   !!        !  09-01 (A. Vidard) tam of the 06-11 version
   !!----------------------------------------------------------------------
   !!   tra_zdf_imp_tan : Update the tracer trend with the diagonal vertical  
   !!                 part of the mixing tensor (tangent).
   !!   tra_zdf_imp_adj : Update the tracer trend with the diagonal vertical  
   !!                 part of the mixing tensor (adjoint).
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 &
      & jpim1,               &
      & jpjm1,               &
      & jpkm1,               &
      & jpiglo
   USE oce_tam       , ONLY: & ! ocean dynamics and active tracers
      & tb_tl,               &
      & sb_tl,               &
      & ta_tl,               &
      & sa_tl,               &
      & tb_ad,               &
      & sb_ad,               &
      & ta_ad,               &
      & sa_ad
   USE dom_oce       , ONLY: & ! ocean space and time domain
      & e1t,                 &
      & e2t,                 &
# if defined key_vvl
      & e3t_1,               &
# else
#  if defined key_zco
      & e3t_0,               &
      & e3w_0,               &
#  else
      & e3t,                 &
      & e3w,                 &
#  endif
# endif
      & tmask,               &
      & lk_vvl,              &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej,                &
      & rdttra
   USE oce           , ONLY: & ! ocean dynamics and tracers variables
      & l_traldf_rot
   USE zdf_oce       , ONLY: & ! ocean vertical physics
      & avt
   USE ldftra_oce    , ONLY: & ! ocean active tracers: lateral physics
      & ahtw,                &
      & aht0
#if defined key_ldfslp
   USE ldfslp        , ONLY: & ! lateral physics: slope of diffusion
      & wslpi,               & !: i_slope at W-points
      & wslpj                  !: j-slope at W-points
#endif 
#if defined key_zdfddm
   USE zdfddm        , ONLY: &
      & avs
#endif 
   USE traldf_tam
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,          &
      & numout,       & 
      & nitend,       & 
      & nit000
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld    , ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
      & prntst_tlm,          & !
      & stdt,                & ! stdev for u-velocity
      & stds                   !           v-velocity
   USE paresp        , ONLY: & ! Weights for an energy-type scalar product
      & wesp_t,              &
      & wesp_s

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC tra_zdf_imp_tan       !  routine called by tra_zdf_tan.F90
   PUBLIC tra_zdf_imp_adj       !  routine called by tra_zdf_adj.F90
   PUBLIC tra_zdf_imp_adj_tst   !  routine called by tst.F90
#if defined key_tst_tlm
   PUBLIC tra_zdf_imp_tlm_tst   !  routine called by tamtst.F90 
#endif

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "ldftra_substitute.h90"
#  include "zdfddm_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!  OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id: trazdf_imp.F90 1156 2008-06-26 16:06:45Z rblod $
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------
CONTAINS
   
   SUBROUTINE tra_zdf_imp_tan( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_zdf_imp_tan  ***
      !!
      !! ** Purpose of the direct routine:
      !!     Compute the trend due to the vertical tracer diffusion
      !!     including the vertical component of lateral mixing (only for 2nd
      !!     order operator, for fourth order it is already computed and add
      !!     to the general trend in traldf.F) and add it to the general trend
      !!     of the tracer equations.
      !!
      !! ** Method of the direct routine :
      !!      The vertical component of the lateral diffusive trends
      !!      is provided by a 2nd order operator rotated along neutral or geo-
      !!      potential surfaces to which an eddy induced advection can be 
      !!      added. It is computed using before fields (forward in time) and 
      !!      isopycnal or geopotential slopes computed in routine ldfslp.
      !!
      !!      Second part: vertical trend associated with the vertical physics
      !!      ===========  (including the vertical flux proportional to dk[t]
      !!                  associated with the lateral mixing, through the
      !!                  update of avt)
      !!      The vertical diffusion of tracers (t & s) is given by:
      !!             difft = dz( avt dz(t) ) = 1/e3t dk+1( avt/e3w dk(t) )
      !!      It is computed using a backward time scheme (t=ta).
      !!      Surface and bottom boundary conditions: no diffusive flux on
      !!      both tracers (bottom, applied through the masked field avt).
      !!      Add this trend to the general trend ta,sa :
      !!         ta = ta + dz( avt dz(t) )
      !!         (sa = sa + dz( avs dz(t) ) if lk_zdfddm=T )
      !!
      !!      Third part: recover avt resulting from the vertical physics
      !!      ==========  alone, for further diagnostics (for example to
      !!                  compute the turbocline depth in zdfmxl.F90).
      !!         avt = zavt
      !!         (avs = zavs if lk_zdfddm=T )
      !!
      !! ** Remarks on the tangent routine : - key_vvl is not available in tangent yet. 
      !!    Once it will be this routine wil need to be rewritten
      !!                                     - simplified version, slopes (wslp[ij])
      !!     assumed to be constant (read from the trajectory). same for av[ts]
      !!
      !!---------------------------------------------------------------------
      !! * Modules used
      USE oce    , ONLY :   zwd   => ua,  &  ! ua used as workspace
                            zws   => va      ! va   "          "
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt          ! ocean time-step index
      REAL(wp), DIMENSION(jpk), INTENT( in ) ::   &
         p2dt                                ! vertical profile of tracer time-step

      !! * Local declarations
      INTEGER  ::   ji, jj, jk               ! dummy loop indices
      REAL(wp) ::   zavi, zrhstl, znvvl,   & ! temporary scalars
         ze3tb, ze3tn, ze3ta, zvsfvvl        ! variable vertical scale factors
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &
         zwi, zwt, zavsi                     ! workspace arrays
      !!---------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp)WRITE(numout,*)
         IF(lwp)WRITE(numout,*) 'tra_zdf_imp : implicit vertical mixing'
         IF(lwp)WRITE(numout,*) '~~~~~~~~~~~ '
         zavi = 0._wp      ! avoid warning at compilation phase when lk_ldfslp=F
      ENDIF

      ! I. Local initialization
      ! -----------------------
      zwd  (1,:, : ) = 0._wp     ;     zwd  (jpi,:,:) = 0._wp
      zws  (1,:, : ) = 0._wp     ;     zws  (jpi,:,:) = 0._wp
      zwi  (1,:, : ) = 0._wp     ;     zwi  (jpi,:,:) = 0._wp
      zwt  (1,:, : ) = 0._wp     ;     zwt  (jpi,:,:) = 0._wp
      zavsi(1,:, : ) = 0._wp     ;     zavsi(jpi,:,:) = 0._wp
      zwt  (:,:,jpk) = 0._wp     ;     zwt  ( : ,:,1) = 0._wp
      zavsi(:,:,jpk) = 0._wp     ;     zavsi( : ,:,1) = 0._wp

      ! I.1 Variable volume : to take into account vertical variable vertical scale factors
      ! -------------------
      ! ... not available  in tangent yet
      ! II. Vertical trend associated with the vertical physics
      ! =======================================================
      !     (including the vertical flux proportional to dk[t] associated
      !      with the lateral mixing, through the avt update)
      !     dk[ avt dk[ (t,s) ] ] diffusive trends

      ! II.0 Matrix construction
      ! ------------------------
#if defined key_ldfslp
      ! update and save of avt (and avs if double diffusive mixing)
      IF( l_traldf_rot ) THEN
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zavi = fsahtw(ji,jj,jk)                       &   ! vertical mixing coef. due to lateral mixing
                     & * (  wslpi(ji,jj,jk) * wslpi(ji,jj,jk)   &
                     &    + wslpj(ji,jj,jk) * wslpj(ji,jj,jk)  )
                  zwt(ji,jj,jk) = avt(ji,jj,jk) + zavi              ! zwt=avt+zavi (total vertical mixing coef. on temperature)
# if defined key_zdfddm
                  zavsi(ji,jj,jk) = fsavs(ji,jj,jk) + zavi          ! dd mixing: zavsi = total vertical mixing coef. on salinity
# endif
               END DO
            END DO
         END DO
      ENDIF   
#else
      ! No isopycnal diffusion
      zwt(:,:,:) = avt(:,:,:)
# if defined key_zdfddm
      zavsi(:,:,:) = avs(:,:,:)
# endif

#endif

      ! Diagonal, inferior, superior  (including the bottom boundary condition via avt masked)
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ze3ta = 1._wp                                ! after scale factor at T-point
               ze3tn = fse3t(ji,jj,jk)                      ! now   scale factor at T-point
               zwi(ji,jj,jk) = - p2dt(jk) * zwt(ji,jj,jk  ) / ( ze3tn * fse3w(ji,jj,jk  ) )
               zws(ji,jj,jk) = - p2dt(jk) * zwt(ji,jj,jk+1) / ( ze3tn * fse3w(ji,jj,jk+1) )
               zwd(ji,jj,jk) = ze3ta - zwi(ji,jj,jk) - zws(ji,jj,jk)
            END DO
         END DO
      END DO

      ! Surface boudary conditions
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ze3ta = 1._wp         ! after scale factor at T-point
            zwi(ji,jj,1) = 0._wp
            zwd(ji,jj,1) = ze3ta - zws(ji,jj,1)
         END DO
      END DO

      ! II.1. Vertical diffusion on t
      ! ---------------------------

      !! Matrix inversion from the first level
      !!----------------------------------------------------------------------
      !   solve m.x = y  where m is a tri diagonal matrix ( jpk*jpk )
      !
      !        ( zwd1 zws1   0    0    0  )( zwx1 ) ( zwy1 )
      !        ( zwi2 zwd2 zws2   0    0  )( zwx2 ) ( zwy2 )
      !        (  0   zwi3 zwd3 zws3   0  )( zwx3 )=( zwy3 )
      !        (        ...               )( ...  ) ( ...  )
      !        (  0    0    0   zwik zwdk )( zwxk ) ( zwyk )
      !
      !   m is decomposed in the product of an upper and lower triangular matrix
      !   The 3 diagonal terms are in 2d arrays: zwd, zws, zwi
      !   The second member is in 2d array zwy
      !   The solution is in 2d array zwx
      !   The 3d arry zwt is a work space array
      !   zwy is used and then used as a work space array : its value is modified!

      ! first recurrence:   Tk = Dk - Ik Sk-1 / Tk-1   (increasing k)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            zwt(ji,jj,1) = zwd(ji,jj,1)
         END DO
      END DO
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               zwt(ji,jj,jk) = zwd(ji,jj,jk) - zwi(ji,jj,jk) * zws(ji,jj,jk-1)  /zwt(ji,jj,jk-1)
            END DO
         END DO
      END DO

      ! second recurrence:    Zk = Yk - Ik / Tk-1  Zk-1
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            ze3tb = 1._wp
            ze3tn = 1._wp
            ta_tl(ji,jj,1) = ze3tb * tb_tl(ji,jj,1) + p2dt(1) * ze3tn * ta_tl(ji,jj,1)
         END DO
      END DO
       
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ze3tb = 1._wp
               ze3tn = 1._wp
               zrhstl = ze3tb * tb_tl(ji,jj,jk) + p2dt(jk) * ze3tn * ta_tl(ji,jj,jk)   ! zrhs=right hand side 
               ta_tl(ji,jj,jk) = zrhstl - zwi(ji,jj,jk) / zwt(ji,jj,jk-1) *ta_tl(ji,jj,jk-1)
            END DO
         END DO
      END DO

      ! third recurrence: Xk = (Zk - Sk Xk+1 ) / Tk
      ! Save the masked temperature after in ta
      ! (c a u t i o n:  temperature not its trend, Leap-frog scheme done it will not be done in tranxt)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            ta_tl(ji,jj,jpkm1) = ta_tl(ji,jj,jpkm1) / zwt(ji,jj,jpkm1) * tmask(ji,jj,jpkm1)
         END DO
      END DO
      DO jk = jpk-2, 1, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ta_tl(ji,jj,jk) = ( ta_tl(ji,jj,jk) - zws(ji,jj,jk) * ta_tl(ji,jj,jk+1) ) / zwt(ji,jj,jk) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO

      ! II.2 Vertical diffusion on salinity
      ! -----------------------------------

#if defined key_zdfddm
      ! Rebuild the Matrix as avt /= avs

      ! Diagonal, inferior, superior  (including the bottom boundary condition via avs masked)
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ze3ta = 1._wp         ! after scale factor at T-point
               ze3tn = fse3t(ji,jj,jk)        ! now   scale factor at T-point
               zwi(ji,jj,jk) = - p2dt(jk) * zavsi(ji,jj,jk  ) / ( ze3tn * fse3w(ji,jj,jk  ) )
               zws(ji,jj,jk) = - p2dt(jk) * zavsi(ji,jj,jk+1) / ( ze3tn * fse3w(ji,jj,jk+1) )
               zwd(ji,jj,jk) = ze3ta - zwi(ji,jj,jk) - zws(ji,jj,jk)
            END DO
         END DO
      END DO

      ! Surface boudary conditions
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ze3ta = 1._wp         ! after scale factor at T-point
            zwi(ji,jj,1) = 0._wp
            zwd(ji,jj,1) = ze3ta - zws(ji,jj,1)
         END DO
      END DO
#endif


      !! Matrix inversion from the first level
      !!----------------------------------------------------------------------
      !   solve m.x = y  where m is a tri diagonal matrix ( jpk*jpk )
      !
      !        ( zwd1 zws1   0    0    0  )( zwx1 ) ( zwy1 )
      !        ( zwi2 zwd2 zws2   0    0  )( zwx2 ) ( zwy2 )
      !        (  0   zwi3 zwd3 zws3   0  )( zwx3 )=( zwy3 )
      !        (        ...               )( ...  ) ( ...  )
      !        (  0    0    0   zwik zwdk )( zwxk ) ( zwyk )
      !
      !   m is decomposed in the product of an upper and lower triangular
      !   matrix
      !   The 3 diagonal terms are in 2d arrays: zwd, zws, zwi
      !   The second member is in 2d array zwy
      !   The solution is in 2d array zwx
      !   The 3d arry zwt is a work space array
      !   zwy is used and then used as a work space array : its value is modified!

      ! first recurrence:   Tk = Dk - Ik Sk-1 / Tk-1   (increasing k)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            zwt(ji,jj,1) = zwd(ji,jj,1)
         END DO
      END DO
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               zwt(ji,jj,jk) = zwd(ji,jj,jk) - zwi(ji,jj,jk) * zws(ji,jj,jk-1)  /zwt(ji,jj,jk-1)
            END DO
         END DO
      END DO

      ! second recurrence:    Zk = Yk - Ik / Tk-1  Zk-1
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            ze3tb = 1.0_wp                   ! before scale factor at T-point
            ze3tn = 1.0_wp                   ! now    scale factor at T-point
            sa_tl(ji,jj,1) = ze3tb * sb_tl(ji,jj,1) + p2dt(1) * ze3tn * sa_tl(ji,jj,1)
         END DO
      END DO
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ze3tb = 1.0_wp                  ! before scale factor at T-point
               ze3tn = 1.0_wp                  ! now    scale factor at T-point
               zrhstl = ze3tb * sb_tl(ji,jj,jk) + p2dt(jk) * ze3tn * sa_tl(ji,jj,jk)     ! zrhs=right hand side
               sa_tl(ji,jj,jk) = zrhstl - zwi(ji,jj,jk) / zwt(ji,jj,jk-1) *sa_tl(ji,jj,jk-1)
            END DO
         END DO
      END DO

      ! third recurrence: Xk = (Zk - Sk Xk+1 ) / Tk
      ! Save the masked temperature after in ta
      ! (c a u t i o n:  temperature not its trend, Leap-frog scheme done it will not be done in tranxt)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            sa_tl(ji,jj,jpkm1) = sa_tl(ji,jj,jpkm1) / zwt(ji,jj,jpkm1) * tmask(ji,jj,jpkm1)
         END DO
      END DO
      DO jk = jpk-2, 1, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               sa_tl(ji,jj,jk) = ( sa_tl(ji,jj,jk) - zws(ji,jj,jk) * sa_tl(ji,jj,jk+1) ) / zwt(ji,jj,jk) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO

   END SUBROUTINE tra_zdf_imp_tan
   SUBROUTINE tra_zdf_imp_adj( kt, p2dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_zdf_imp_adj  ***
      !!
      !! ** Purpose of the direct routine:
      !!     Compute the trend due to the vertical tracer diffusion
      !!     including the vertical component of lateral mixing (only for 2nd
      !!     order operator, for fourth order it is already computed and add
      !!     to the general trend in traldf.F) and add it to the general trend
      !!     of the tracer equations.
      !!
      !! ** Method of the direct routine :
      !!      The vertical component of the lateral diffusive trends
      !!      is provided by a 2nd order operator rotated along neutral or geo-
      !!      potential surfaces to which an eddy induced advection can be 
      !!      added. It is computed using before fields (forward in time) and 
      !!      isopycnal or geopotential slopes computed in routine ldfslp.
      !!
      !!      Second part: vertical trend associated with the vertical physics
      !!      ===========  (including the vertical flux proportional to dk[t]
      !!                  associated with the lateral mixing, through the
      !!                  update of avt)
      !!      The vertical diffusion of tracers (t & s) is given by:
      !!             difft = dz( avt dz(t) ) = 1/e3t dk+1( avt/e3w dk(t) )
      !!      It is computed using a backward time scheme (t=ta).
      !!      Surface and bottom boundary conditions: no diffusive flux on
      !!      both tracers (bottom, applied through the masked field avt).
      !!      Add this trend to the general trend ta,sa :
      !!         ta = ta + dz( avt dz(t) )
      !!         (sa = sa + dz( avs dz(t) ) if lk_zdfddm=T )
      !!
      !!      Third part: recover avt resulting from the vertical physics
      !!      ==========  alone, for further diagnostics (for example to
      !!                  compute the turbocline depth in zdfmxl.F90).
      !!         avt = zavt
      !!         (avs = zavs if lk_zdfddm=T )
      !!
      !! ** Remarks on the adjoint routine : - key_vvl is not available in adjoint yet. 
      !!    Once it will be this routine wil need to be rewritten
      !!                                     - simplified version, slopes (wslp[ij])
      !!     assumed to be constant (read from the trajectory). same for av[ts]
      !!
      !!---------------------------------------------------------------------
      !! * Modules used
      USE oce    , ONLY :   zwd   => ua,  &  ! ua used as workspace
                            zws   => va      ! va   "          "
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt          ! ocean time-step index
      REAL(wp), DIMENSION(jpk), INTENT( in ) ::   &
         p2dt                                ! vertical profile of tracer time-step

      !! * Local declarations
      INTEGER  ::   ji, jj, jk               ! dummy loop indices
      REAL(wp) ::   zavi, zrhsad, znvvl,   & ! temporary scalars
         ze3tb, ze3tn, ze3ta, zvsfvvl        ! variable vertical scale factors
      REAL(wp), DIMENSION(jpi,jpj,jpk) ::   &
         zwi, zwt, zavsi                     ! workspace arrays
      !!---------------------------------------------------------------------

      IF( kt == nitend ) THEN
         IF(lwp)WRITE(numout,*)
         IF(lwp)WRITE(numout,*) 'tra_zdf_imp_adj : implicit vertical mixing'
         IF(lwp)WRITE(numout,*) '~~~~~~~~~~~~~~~ '
         CALL ldf_ctl_tam  ! init of l_traldf_rot
         zavi = 0._wp      ! avoid warning at compilation phase when lk_ldfslp=F
      ENDIF

      ! I. Local initialization
      ! -----------------------
      zrhsad         = 0.0_wp
      zwd  (1,:, : ) = 0._wp     ;     zwd  (jpi,:,:) = 0._wp
      zws  (1,:, : ) = 0._wp     ;     zws  (jpi,:,:) = 0._wp
      zwi  (1,:, : ) = 0._wp     ;     zwi  (jpi,:,:) = 0._wp
      zwt  (1,:, : ) = 0._wp     ;     zwt  (jpi,:,:) = 0._wp
      zavsi(1,:, : ) = 0._wp     ;     zavsi(jpi,:,:) = 0._wp
      zwt  (:,:,jpk) = 0._wp     ;     zwt  ( : ,:,1) = 0._wp
      zavsi(:,:,jpk) = 0._wp     ;     zavsi( : ,:,1) = 0._wp

      ! I.1 Variable volume : to take into account vertical variable vertical scale factors
      ! -------------------
      ! ... not available  in adjoint yet
      ! II. Vertical trend associated with the vertical physics
      ! =======================================================
      !     (including the vertical flux proportional to dk[t] associated
      !      with the lateral mixing, through the avt update)
      !     dk[ avt dk[ (t,s) ] ] diffusive trends


      ! II.0 Matrix construction
      ! ------------------------

#if defined key_ldfslp
      ! update and save of avt (and avs if double diffusive mixing)
      IF( l_traldf_rot ) THEN
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zavi = fsahtw(ji,jj,jk)                       &   ! vertical mixing coef. due to lateral mixing
                     & * (  wslpi(ji,jj,jk) * wslpi(ji,jj,jk)   &
                     &    + wslpj(ji,jj,jk) * wslpj(ji,jj,jk)  )
                  zwt(ji,jj,jk) = avt(ji,jj,jk) + zavi              ! zwt=avt+zavi (total vertical mixing coef. on temperature)
# if defined key_zdfddm
                  zavsi(ji,jj,jk) = fsavs(ji,jj,jk) + zavi          ! dd mixing: zavsi = total vertical mixing coef. on salinity
# endif
               END DO
            END DO
         END DO
      ENDIF   
#else
      ! No isopycnal diffusion
      zwt(:,:,:) = avt(:,:,:)
# if defined key_zdfddm
      zavsi(:,:,:) = avs(:,:,:)
# endif

#endif

      ! Diagonal, inferior, superior  (including the bottom boundary condition via avt masked)
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ze3ta = 1._wp                                ! after scale factor at T-point
               ze3tn = fse3t(ji,jj,jk)                      ! now   scale factor at T-point
               zwi(ji,jj,jk) = - p2dt(jk) * zwt(ji,jj,jk  ) / ( ze3tn * fse3w(ji,jj,jk  ) )
               zws(ji,jj,jk) = - p2dt(jk) * zwt(ji,jj,jk+1) / ( ze3tn * fse3w(ji,jj,jk+1) )
               zwd(ji,jj,jk) = ze3ta - zwi(ji,jj,jk) - zws(ji,jj,jk)
            END DO
         END DO
      END DO

      ! Surface boudary conditions
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ze3ta = 1._wp         ! after scale factor at T-point
            zwi(ji,jj,1) = 0._wp
            zwd(ji,jj,1) = ze3ta - zws(ji,jj,1)
         END DO
      END DO


      ! II.1. Vertical diffusion on t
      ! ---------------------------

      !! Matrix inversion from the first level
      !!----------------------------------------------------------------------
      !   solve m.x = y  where m is a tri diagonal matrix ( jpk*jpk )
      !
      !        ( zwd1 zws1   0    0    0  )( zwx1 ) ( zwy1 )
      !        ( zwi2 zwd2 zws2   0    0  )( zwx2 ) ( zwy2 )
      !        (  0   zwi3 zwd3 zws3   0  )( zwx3 )=( zwy3 )
      !        (        ...               )( ...  ) ( ...  )
      !        (  0    0    0   zwik zwdk )( zwxk ) ( zwyk )
      !
      !   m is decomposed in the product of an upper and lower triangular matrix
      !   The 3 diagonal terms are in 2d arrays: zwd, zws, zwi
      !   The second member is in 2d array zwy
      !   The solution is in 2d array zwx
      !   The 3d arry zwt is a work space array
      !   zwy is used and then used as a work space array : its value is modified!

      ! first recurrence:   Tk = Dk - Ik Sk-1 / Tk-1   (increasing k)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            zwt(ji,jj,1) = zwd(ji,jj,1)
         END DO
      END DO
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               zwt(ji,jj,jk) = zwd(ji,jj,jk) - zwi(ji,jj,jk) * zws(ji,jj,jk-1)  /zwt(ji,jj,jk-1)
            END DO
         END DO
      END DO
      ! third recurrence: Xk = (Zk - Sk Xk+1 ) / Tk
      ! Save the masked temperature after in ta
      ! (c a u t i o n:  temperature not its trend, Leap-frog scheme done it will not be done in tranxt)
      DO jk = 1, jpk-2
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ta_ad(ji,jj,jk+1) = ta_ad(ji,jj,jk+1) - zws(ji,jj,jk) * ta_ad(ji,jj,jk) / zwt(ji,jj,jk) * tmask(ji,jj,jk)
               ta_ad(ji,jj,jk)   = ta_ad(ji,jj,jk) / zwt(ji,jj,jk) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            ta_ad(ji,jj,jpkm1) = ta_ad(ji,jj,jpkm1) / zwt(ji,jj,jpkm1) * tmask(ji,jj,jpkm1)
         END DO
      END DO
      ! second recurrence:    Zk = Yk - Ik / Tk-1  Zk-1
      DO jk = jpkm1, 2, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ze3tb = 1._wp
               ze3tn = 1._wp
               zrhsad = zrhsad + ta_ad(ji,jj,jk)
               ta_ad(ji,jj,jk-1) = ta_ad(ji,jj,jk-1) - zwi(ji,jj,jk) / zwt(ji,jj,jk-1) * ta_ad(ji,jj,jk)
               ta_ad(ji,jj,jk) = 0.0_wp
               tb_ad(ji,jj,jk) = tb_ad(ji,jj,jk) + ze3tb * zrhsad
               ta_ad(ji,jj,jk) = ta_ad(ji,jj,jk) + p2dt(jk) * ze3tn * zrhsad
               zrhsad = 0.0_wp
            END DO
         END DO
      END DO
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            ze3tb = 1._wp
            ze3tn = 1._wp
            tb_ad(ji,jj,1) = tb_ad(ji,jj,1) + ze3tb * ta_ad(ji,jj,1)
            ta_ad(ji,jj,1) = ta_ad(ji,jj,1) * p2dt(1) * ze3tn 
         END DO
      END DO
     ! II.2 Vertical diffusion on salinity
      ! -----------------------------------

#if defined key_zdfddm
      ! Rebuild the Matrix as avt /= avs

      ! Diagonal, inferior, superior  (including the bottom boundary condition via avs masked)
      DO jk = jpkm1, 1, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               ze3ta = 1._wp         ! after scale factor at T-point
               ze3tn = fse3t(ji,jj,jk)         ! now   scale factor at T-point
               zwi(ji,jj,jk) = - p2dt(jk) * zavsi(ji,jj,jk  ) / ( ze3tn * fse3w(ji,jj,jk  ) )
               zws(ji,jj,jk) = - p2dt(jk) * zavsi(ji,jj,jk+1) / ( ze3tn * fse3w(ji,jj,jk+1) )
               zwd(ji,jj,jk) = ze3ta - zwi(ji,jj,jk) - zws(ji,jj,jk)
            END DO
         END DO
      END DO

      ! Surface boudary conditions
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            ze3ta = 1._wp         ! after scale factor at T-point
            zwi(ji,jj,1) = 0._wp
            zwd(ji,jj,1) = ze3ta - zws(ji,jj,1)
         END DO
      END DO
#endif


      !! Matrix inversion from the first level
      !!----------------------------------------------------------------------
      !   solve m.x = y  where m is a tri diagonal matrix ( jpk*jpk )
      !
      !        ( zwd1 zws1   0    0    0  )( zwx1 ) ( zwy1 )
      !        ( zwi2 zwd2 zws2   0    0  )( zwx2 ) ( zwy2 )
      !        (  0   zwi3 zwd3 zws3   0  )( zwx3 )=( zwy3 )
      !        (        ...               )( ...  ) ( ...  )
      !        (  0    0    0   zwik zwdk )( zwxk ) ( zwyk )
      !
      !   m is decomposed in the product of an upper and lower triangular
      !   matrix
      !   The 3 diagonal terms are in 2d arrays: zwd, zws, zwi
      !   The second member is in 2d array zwy
      !   The solution is in 2d array zwx
      !   The 3d arry zwt is a work space array
      !   zwy is used and then used as a work space array : its value is modified!

      ! first recurrence:   Tk = Dk - Ik Sk-1 / Tk-1   (increasing k)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            zwt(ji,jj,1) = zwd(ji,jj,1)
         END DO
      END DO
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               zwt(ji,jj,jk) = zwd(ji,jj,jk) - zwi(ji,jj,jk) * zws(ji,jj,jk-1)  /zwt(ji,jj,jk-1)
            END DO
         END DO
      END DO
      ! third recurrence: Xk = (Zk - Sk Xk+1 ) / Tk
      ! Save the masked temperature after in ta
      ! (c a u t i o n:  temperature not its trend, Leap-frog scheme done it will not be done in tranxt)
      DO jk = 1, jpk-2
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               sa_ad(ji,jj,jk+1) = sa_ad(ji,jj,jk+1) - zws(ji,jj,jk) * sa_ad(ji,jj,jk) / zwt(ji,jj,jk) * tmask(ji,jj,jk)
               sa_ad(ji,jj,jk  ) = sa_ad(ji,jj,jk  ) / zwt(ji,jj,jk) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            sa_ad(ji,jj,jpkm1) = sa_ad(ji,jj,jpkm1) / zwt(ji,jj,jpkm1) * tmask(ji,jj,jpkm1)
         END DO
      END DO

      ! second recurrence:    Zk = Yk - Ik / Tk-1  Zk-1
     DO jk = jpkm1, 2, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ze3tb = 1.0_wp                  ! before scale factor at T-point
               ze3tn = 1.0_wp                  ! now    scale factor at T-point
               zrhsad = zrhsad + sa_ad(ji,jj,jk)
               sa_ad(ji,jj,jk-1) = sa_ad(ji,jj,jk-1) - zwi(ji,jj,jk) / zwt(ji,jj,jk-1) * sa_ad(ji,jj,jk)
               sa_ad(ji,jj,jk) = 0.0_wp
               sb_ad(ji,jj,jk) = sb_ad(ji,jj,jk) + ze3tb * zrhsad
               sa_ad(ji,jj,jk) = sa_ad(ji,jj,jk) + p2dt(jk) * ze3tn * zrhsad
               zrhsad = 0.0_wp
            END DO
         END DO
      END DO
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            ze3tb = 1.0_wp                   ! before scale factor at T-point
            ze3tn = 1.0_wp                   ! now    scale factor at T-point
            sb_ad(ji,jj,1) = sb_ad(ji,jj,1) + ze3tb * sa_ad(ji,jj,1)
            sa_ad(ji,jj,1) = p2dt(1) * ze3tn * sa_ad(ji,jj,1)
         END DO
      END DO
   END SUBROUTINE tra_zdf_imp_adj
SUBROUTINE tra_zdf_imp_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_zdf_imp_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & istp,  &
         & jstp,  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
      REAL(KIND=wp) ::   &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp2            ! scalar product involving the adjoint routine
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & zta_tlin ,     & ! Tangent input
         & ztb_tlin ,     & ! Tangent input
         & zsa_tlin ,     & ! Tangent input
         & zsb_tlin ,     & ! Tangent input
         & zta_tlout,     & ! Tangent output
         & zsa_tlout,     & ! Tangent output
         & zta_adin ,     & ! Adjoint input
         & zsa_adin ,     & ! Adjoint input
         & zta_adout,     & ! Adjoint output
         & ztb_adout,     & ! Adjoint output
         & zsa_adout,     & ! Adjoint output
         & zsb_adout,     & ! Adjoint output
         & zr             ! 3D random field 
      CHARACTER(LEN=14) :: cl_name
      ! Allocate memory

      ALLOCATE( &
         & zta_tlin( jpi,jpj,jpk),     &
         & zsa_tlin( jpi,jpj,jpk),     &
         & ztb_tlin( jpi,jpj,jpk),     &
         & zsb_tlin( jpi,jpj,jpk),     &
         & zta_tlout(jpi,jpj,jpk),     &
         & zsa_tlout(jpi,jpj,jpk),     &
         & zta_adin( jpi,jpj,jpk),     &
         & zsa_adin( jpi,jpj,jpk),     &
         & zta_adout(jpi,jpj,jpk),     &
         & zsa_adout(jpi,jpj,jpk),     &
         & ztb_adout(jpi,jpj,jpk),     &
         & zsb_adout(jpi,jpj,jpk),     &
         & zr(       jpi,jpj,jpk)      &
         & )
      !==================================================================
      ! 1) dx = ( un_tl, vn_tl, hdivn_tl ) and 
      !    dy = ( hdivb_tl, hdivn_tl )
      !==================================================================

      ! initialization (normally done in traldf)
      l_traldf_rot = .TRUE. 

      ! Test for time steps nit000 and nit000 + 1 (the matrix changes)

      DO jstp = nit000, nit000 + 2
         istp = jstp
         IF ( jstp == nit000+2 ) istp = nitend

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
          zta_tlin( :,:,:) = 0.0_wp
          ztb_tlin( :,:,:) = 0.0_wp
          zsa_tlin( :,:,:) = 0.0_wp
          zsb_tlin( :,:,:) = 0.0_wp
          zta_tlout(:,:,:) = 0.0_wp
          zsa_tlout(:,:,:) = 0.0_wp
          zta_adin( :,:,:) = 0.0_wp
          zsa_adin( :,:,:) = 0.0_wp
          zta_adout(:,:,:) = 0.0_wp
          zsa_adout(:,:,:) = 0.0_wp
          ztb_adout(:,:,:) = 0.0_wp
          zsb_adout(:,:,:) = 0.0_wp
          zr(       :,:,:) = 0.0_wp
          tb_ad(:,:,:)     = 0.0_wp
          sb_ad(:,:,:)     = 0.0_wp
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------

      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 596035 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

     DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 352791 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stdt )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              ztb_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

     DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 142746 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsa_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO

     DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 214934 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zr, 'T', 0.0_wp, stds )
      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zsb_tlin(ji,jj,jk) = zr(ji,jj,jk) 
            END DO
         END DO
      END DO


      ta_tl(:,:,:) = zta_tlin(:,:,:)
      sa_tl(:,:,:) = zsa_tlin(:,:,:)
      tb_tl(:,:,:) = ztb_tlin(:,:,:)
      sb_tl(:,:,:) = zsb_tlin(:,:,:)
      CALL tra_zdf_imp_tan ( istp, rdttra )
      zta_tlout(:,:,:) = ta_tl(:,:,:)
      zsa_tlout(:,:,:) = sa_tl(:,:,:)

      !--------------------------------------------------------------------
      ! Initialize the adjoint variables: dy^* = W dy
      !--------------------------------------------------------------------

      DO jk = 1, jpk
        DO jj = nldj, nlej
           DO ji = nldi, nlei
              zta_adin(ji,jj,jk) = zta_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_t(jk)
              zsa_adin(ji,jj,jk) = zsa_tlout(ji,jj,jk) &
                 &               * e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) &
                 &               * tmask(ji,jj,jk) * wesp_s(jk)
            END DO
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1 = DOT_PRODUCT( zta_tlout, zta_adin ) &
         & + DOT_PRODUCT( zsa_tlout, zsa_adin )

      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------

      ta_ad(:,:,:) = zta_adin(:,:,:)
      sa_ad(:,:,:) = zsa_adin(:,:,:)
      
      CALL tra_zdf_imp_adj ( istp, rdttra )
      
      zta_adout(:,:,:) = ta_ad(:,:,:)
      zsa_adout(:,:,:) = sa_ad(:,:,:)
      ztb_adout(:,:,:) = tb_ad(:,:,:)
      zsb_adout(:,:,:) = sb_ad(:,:,:)
      zsp2 = DOT_PRODUCT( zta_tlin, zta_adout ) &
         & + DOT_PRODUCT( zsa_tlin, zsa_adout ) &
         & + DOT_PRODUCT( ztb_tlin, ztb_adout ) &
         & + DOT_PRODUCT( zsb_tlin, zsb_adout ) 

      ! 14 char:'12345678901234'
      IF ( istp == nit000 ) THEN
         cl_name = 'trazdfimpadjT1'
      ELSEIF ( istp == nit000 +1 ) THEN
         cl_name = 'trazdfimpadjT2'
      ELSEIF ( istp == nitend ) THEN
         cl_name = 'trazdfimpadjT3'
      END IF
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      END DO

      DEALLOCATE(   &
         & zta_tlin,  &
         & ztb_tlin,  &
         & zsa_tlin,  &
         & zsb_tlin,  &
         & zta_tlout, &
         & zsa_tlout, &
         & zta_adin,  &
         & zsa_adin,  &
         & zta_adout, &
         & ztb_adout, &
         & zsa_adout, &
         & zsb_adout, &
         & zr       &
         & )



   END SUBROUTINE tra_zdf_imp_adj_tst
#if defined key_tst_tlm
   SUBROUTINE tra_zdf_imp_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE tra_adv_zdf_imp_tlm_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!  
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!                   
      !! History :
      !!        ! 09-07 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE oce           , ONLY: & ! ocean dynamics and active tracers
      & tb, sb, ta, sa
      USE tamtrj              ! writing out state trajectory
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE wzvmod             !  vertical velocity
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
        & tb, sb, ta, sa
      USE trazdf_imp          ! vertical component of the tracer mixing trend
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER ::  &
         & istp,  &
         & jstp,  &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk     
      REAL(KIND=wp) :: &
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_Ta,      &
         & zsp1_Sa,      &
         & zsp2,         & ! scalar product involving the tangent routine
         & zsp2_Ta,      &
         & zsp2_Sa,      &
         & zsp3,         & ! scalar product involving the tangent routine
         & zsp3_Ta,      &
         & zsp3_Sa,      &
         & zzsp,         & ! scalar product involving the tangent routine
         & zzsp_Ta,      &
         & zzsp_Sa,      & 
         & gamma,            &
         & zgsp1,            &
         & zgsp2,            &
         & zgsp3,            &
         & zgsp4,            &
         & zgsp5,            &
         & zgsp6,            &
         & zgsp7  
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztb_tlin,      & ! Tangent input
         & zsb_tlin,      & ! Tangent input
         & zta_tlin,      & ! Tangent input
         & zsa_tlin,      & ! Tangent input
         & zta_out ,      & ! Direct output
         & zsa_out ,      & ! Direct output
         & zta_wop ,      & ! Direct output w/o perturbation
         & zsa_wop ,      & ! Direct output w/o perturbation
         & zr               ! 3D random field 
      CHARACTER(LEN=14) ::&
         & cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_xdx
      CHARACTER (LEN=90) :: &
         & FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zscta, zscsa,    &
         & zscerrta,        &
         & zscerrsa
      INTEGER, DIMENSION(100):: &
         & iipostb, iipossb,    &
         & iiposta, iipossa,    &
         & ijpostb, ijpossb,    &
         & ijposta, ijpossa,    & 
         & ikpostb, ikpossb,    &
         & ikposta, ikpossa
      INTEGER::             &
         & ii,              &
         & isamp=40,        &
         & jsamp=40,        &
         & ksamp=10,        &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj,jpk) :: &
         & zerrta, zerrsa   
      ! Allocate memory
      ALLOCATE( &
         & ztb_tlin( jpi,jpj,jpk),     &
         & zsb_tlin( jpi,jpj,jpk),     &
         & zta_tlin( jpi,jpj,jpk),     &
         & zsa_tlin( jpi,jpj,jpk),     &
         & zta_out ( jpi,jpj,jpk),     &
         & zsa_out ( jpi,jpj,jpk),     &
         & zta_wop ( jpi,jpj,jpk),     &
         & zsa_wop ( jpi,jpj,jpk),     &
         & zr(       jpi,jpj,jpk)      &
         & )
      !--------------------------------------------------------------------
      ! Reset variables
      !--------------------------------------------------------------------
      ztb_tlin( :,:,:) = 0.0_wp
      zsb_tlin( :,:,:) = 0.0_wp
      zta_tlin( :,:,:) = 0.0_wp
      zsa_tlin( :,:,:) = 0.0_wp
      zta_out ( :,:,:) = 0.0_wp
      zsa_out ( :,:,:) = 0.0_wp
      zta_wop ( :,:,:) = 0.0_wp
      zsa_wop ( :,:,:) = 0.0_wp
      zr(       :,:,:) = 0.0_wp
      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio
      file_wop='trj_wop_trazdf_imp'
      file_xdx='trj_xdx_trazdf_imp'
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 596035, zr,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  ztb_tlin(ji,jj,jk) = zr(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 352791, zr,  'T', 0.0_wp, stds) 
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsb_tlin(ji,jj,jk) = zr(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 142746, zr,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zta_tlin(ji,jj,jk) = zr(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 214934, zr,  'T', 0.0_wp, stds)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsa_tlin(ji,jj,jk) = zr(ji,jj,jk)
               END DO
            END DO
         END DO 
      ENDIF 
      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      IF ( tlm_bch /= 2 )      CALL istate_p  

      ! *** initialize the reference trajectory
      ! ------------
      CALL  trj_rea( nit000-1, 1 )
      CALL  trj_rea( nit000, 1 )

      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         ztb_tlin(:,:,:) = gamma * ztb_tlin(:,:,:)
         tb(:,:,:)       = tb(:,:,:) + ztb_tlin(:,:,:)

         zsb_tlin(:,:,:) = gamma * zsb_tlin(:,:,:)
         sb(:,:,:)       = sb(:,:,:) + zsb_tlin(:,:,:)

         zta_tlin(:,:,:) = gamma * zta_tlin(:,:,:)
         ta(:,:,:)       = ta(:,:,:) + zta_tlin(:,:,:)

         zsa_tlin(:,:,:) = gamma * zsa_tlin(:,:,:)
         sa(:,:,:)       = sa(:,:,:) + zsa_tlin(:,:,:)
      ENDIF 
      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      IF ( tlm_bch /= 2 )  CALL tra_zdf_imp(nit000, rdttra)
      IF ( tlm_bch == 0 )  CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 )  CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN
         !--------------------------------------------------------------------
         ! Initialize the tangent variables: dy^* = W dy  
         !--------------------------------------------------------------------
         CALL  trj_rea( nit000-1, 1 ) 
         CALL  trj_rea( nit000, 1 ) 
         tb_tl  (:,:,:) = ztb_tlin  (:,:,:)
         sb_tl  (:,:,:) = zsb_tlin  (:,:,:)
         ta_tl  (:,:,:) = zta_tlin  (:,:,:)
         sa_tl  (:,:,:) = zsa_tlin  (:,:,:)

         !-----------------------------------------------------------------------
         !  Initialization of the dynamics and tracer fields for the tangent
         !----------------------------------------------------------------------- 
         CALL tra_zdf_imp_tan(nit000, rdttra)

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------

         zsp2_Ta    = DOT_PRODUCT( ta_tl, ta_tl  )
         zsp2_Sa    = DOT_PRODUCT( sa_tl, sa_tl  )

         zsp2      = zsp2_Ta + zsp2_Sa

         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------
         CALL trj_rd_spl(file_wop) 
         zta_wop  (:,:,:) = ta  (:,:,:)
         zsa_wop  (:,:,:) = sa  (:,:,:)
         CALL trj_rd_spl(file_xdx) 
         zta_out  (:,:,:) = ta  (:,:,:)
         zsa_out  (:,:,:) = sa  (:,:,:)
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zta_out   (ji,jj,jk) = zta_out    (ji,jj,jk) - zta_wop  (ji,jj,jk)
                  zta_wop   (ji,jj,jk) = zta_out    (ji,jj,jk) - ta_tl    (ji,jj,jk)
                  IF (  ta_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrta(ji,jj,jk) = zta_out(ji,jj,jk)/ta_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iiposta(ii) = ji
                      ijposta(ii) = jj
                      ikposta(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                         zscta (ii) =  zta_wop(ji,jj,jk)
                         zscerrta (ii) =  ( zerrta(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         ii=0
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zsa_out   (ji,jj,jk) = zsa_out    (ji,jj,jk) - zsa_wop  (ji,jj,jk)
                  zsa_wop   (ji,jj,jk) = zsa_out    (ji,jj,jk) - sa_tl    (ji,jj,jk)
                  IF (  sa_tl(ji,jj,jk) .NE. 0.0_wp ) &
                     & zerrsa(ji,jj,jk) = zsa_out(ji,jj,jk)/sa_tl(ji,jj,jk)
                  IF( (MOD(ji, isamp) .EQ. 0) .AND. &
                  &   (MOD(jj, jsamp) .EQ. 0) .AND. &
                  &   (MOD(jk, ksamp) .EQ. 0)     ) THEN
                      ii = ii+1
                      iipossa(ii) = ji
                      ijpossa(ii) = jj
                      ikpossa(ii) = jk
                      IF ( INT(tmask(ji,jj,jk)) .NE. 0)  THEN
                         zscsa (ii) =  zsa_wop(ji,jj,jk)
                         zscerrsa (ii) =  ( zerrsa(ji,jj,jk) - 1.0_wp ) / gamma
                      ENDIF
                  ENDIF
               END DO
            END DO
         END DO

         zsp1_Ta    = DOT_PRODUCT( zta_out, zta_out  )
         zsp1_Sa    = DOT_PRODUCT( zsa_out, zsa_out  )

         zsp1      = zsp1_Ta + zsp1_Sa

         zsp3_Ta    = DOT_PRODUCT( zta_wop, zta_wop  )
         zsp3_Sa    = DOT_PRODUCT( zsa_wop, zsa_wop  )

         zsp3      = zsp3_Ta + zsp3_Sa
 
         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'trazdf_tam:En '
         zzsp    = SQRT(zsp3)
         zzsp_Ta = SQRT(zsp3_Ta)
         zzsp_Sa = SQRT(zsp3_Sa)
         zgsp5   = zzsp
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp      = SQRT(zsp2)
         zzsp_Ta   = SQRT(zsp2_Ta)
         zzsp_Sa   = SQRT(zsp2_Sa)
         zgsp4     = zzsp
         cl_name = 'trazdf_tam:Ln2'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         zzsp      = SQRT(zsp1)
         zzsp_Ta   = SQRT(zsp1_Ta)
         zzsp_Sa   = SQRT(zsp1_Sa)

         cl_name = 'trazdf:Mhdx-Mx'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         zgsp3 = SQRT( zsp3/zsp2 )
         zgsp7 = zgsp3/gamma
         zgsp1 = zzsp
         zgsp2 = zgsp1 / zgsp4
         zgsp6 = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'tzdfimp ', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7

         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'tzdfimp '
         IF(lwp) THEN
            DO ii=1, 100, 1
               IF ( zscta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscta     ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscsa     ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscsa(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrta(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrta ', &
                    & cur_loop, h_ratio, ii, iiposta(ii), ijposta(ii), zscerrta(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscerrsa(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscerrsa ', &
                    & cur_loop, h_ratio, ii, iipossa(ii), ijpossa(ii), zscerrsa(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc, "(A4)") '===='
         ENDIF

      ENDIF

      DEALLOCATE(                  &
         & ztb_tlin, zsb_tlin,     & 
         & zta_tlin, zsa_tlin,     &  
         & zta_out, zsa_out,       &  
         & zta_wop, zsa_wop,       & 
         & zr                      &
         & )
   END SUBROUTINE tra_zdf_imp_tlm_tst
#endif
#endif
END MODULE trazdf_imp_tam
