MODULE zpshde_tam
#ifdef key_tam
   !!==============================================================================
   !!                       ***  MODULE zpshde_tam   ***
   !! z-coordinate - partial step : Horizontal Derivative
   !!                               Tangent and Adjoint Module
   !!==============================================================================
   
   !!----------------------------------------------------------------------
   !!   zps_hde      :  Horizontal DErivative of T, S and rd at the last
   !!                   ocean level (Z-coord. with Partial Steps)
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE par_oce       , ONLY: & ! Ocean space and time domain variables
      & jpi,                 &
      & jpj,                 & 
      & jpk,                 & 
      & jpim1,               & 
      & jpjm1,               & 
      & jpiglo
   USE oce           , ONLY: & ! ocean variables
      & tn,                  &
      & sn
   USE dom_oce       , ONLY: & ! ocean space domain variables
      & umask,               &
      & vmask,               &
      & e2u,                 &
      & e1u,                 &
      & e2v,                 &
      & e1v,                 &
# if defined key_zco
      & gdept_0,             &
      & e3t_0,               &
      & e3w_0,               &
# else
      & gdept,               &
      & e3u,                 &
      & e3v,                 &
      & e3w,                 &
# endif
      & mbathy,              &
      & mig,                 &
      & mjg,                 &
      & nldi,                &
      & nldj,                &
      & nlei,                &
      & nlej
   USE in_out_manager, ONLY: & ! I/O manager
      & lwp,                 &
      & numout,              & 
      & nit000,              &
      & nitend
   USE eosbn2_tam    , ONLY: & ! ocean equation of state
      & eos_tan,             &
      & eos_adj
   USE lbclnk        , ONLY: & ! lateral boundary conditions (or mpp link)
      & lbc_lnk
   USE lbclnk_tam    , ONLY: & ! lateral boundary conditions (or mpp link)
      & lbc_lnk_adj
   USE gridrandom    , ONLY: & ! Random Gaussian noise on grids
      & grid_random
   USE dotprodfld,     ONLY: & ! Computes dot product for 3D and 2D fields
      & dot_product
   USE tstool_tam    , ONLY: &
      & prntst_adj,          & !
      & prntst_tlm,          & !
      & stdt,                & ! stdev for temperature
      & stds,                & !           salinity
      & stdr                   !           

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC zps_hde_tan      ! routine called by step_tam.F90
   PUBLIC zps_hde_adj      ! routine called by step_tam.F90
   PUBLIC zps_hde_adj_tst  ! routine called by tst.F90
#if defined key_tst_tlm
   PUBLIC  zps_hde_tlm_tst ! routine called by tamtst.F90
#endif

   !! * module variables
   INTEGER, DIMENSION(jpi,jpj) ::   &
   &   mbatu, mbatv      ! bottom ocean level index at U- and V-points

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE zps_hde_tan ( kt, ptem, psal,              &
      &                     ptem_tl, psal_tl, prd_tl ,   &
                            pgtu_tl, pgsu_tl, pgru_tl,   &
                            pgtv_tl, pgsv_tl, pgrv_tl     )
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE zps_hde_tan  ***
      !!                    
      !! ** Purpose of the direct routine:
      !!      Compute the horizontal derivative of T, S and rd
      !!      at u- and v-points with a linear interpolation for z-coordinate
      !!      with partial steps.
      !!
      !! ** Method of the direct routine:
      !!      In z-coord with partial steps, scale factors on last 
      !!      levels are different for each grid point, so that T, S and rd 
      !!      points are not at the same depth as in z-coord. To have horizontal
      !!      gradients again, we interpolate T and S at the good depth : 
      !!      Linear interpolation of T, S   
      !!         Computation of di(tb) and dj(tb) by vertical interpolation:
      !!          di(t) = t~ - t(i,j,k) or t(i+1,j,k) - t~
      !!          dj(t) = t~ - t(i,j,k) or t(i,j+1,k) - t~
      !!         This formulation computes the two cases:
      !!                 CASE 1                   CASE 2  
      !!         k-1  ___ ___________   k-1   ___ ___________
      !!                    Ti  T~                  T~  Ti+1
      !!                  _____                        _____
      !!         k        |   |Ti+1     k           Ti |   |
      !!                  |   |____                ____|   |
      !!              ___ |   |   |           ___  |   |   |
      !!                  
      !!      case 1->   e3w(i+1) >= e3w(i) ( and e3w(j+1) >= e3w(j) ) then
      !!          t~ = t(i+1,j  ,k) + (e3w(i+1) - e3w(i)) * dk(Ti+1)/e3w(i+1)
      !!        ( t~ = t(i  ,j+1,k) + (e3w(j+1) - e3w(j)) * dk(Tj+1)/e3w(j+1)  )
      !!          or
      !!      case 2->   e3w(i+1) <= e3w(i) ( and e3w(j+1) <= e3w(j) ) then
      !!          t~ = t(i,j,k) + (e3w(i) - e3w(i+1)) * dk(Ti)/e3w(i )
      !!        ( t~ = t(i,j,k) + (e3w(j) - e3w(j+1)) * dk(Tj)/e3w(j ) )
      !!          Idem for di(s) and dj(s)          
      !!
      !!      For rho, we call eos_insitu_2d which will compute rd~(t~,s~) at 
      !!      the good depth zh from interpolated T and S for the different
      !!      formulation of the equation of state (eos).
      !!      Gradient formulation for rho :
      !!          di(rho) = rd~ - rd(i,j,k) or rd (i+1,j,k) - rd~
      !!
      !! ** Action  : - pgtu, pgsu, pgru: horizontal gradient of T, S
      !!                and rd at U-points 
      !!              - pgtv, pgsv, pgrv: horizontal gradient of T, S
      !!                and rd at V-points 
      !!
      !! History of the direct routine:
      !!   8.5  !  02-04  (A. Bozec)  Original code
      !!   8.5  !  02-08  (G. Madec E. Durand)  Optimization and Free form
      !! History of the TAM routine:
      !!   9.0  !  08-06 (A. Vidard) Skeleton
      !!        !  08-06 (A. Vidard) tangent of the 02-08 version
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt          ! ocean time-step index
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( in ) ::   &
         ptem, psal                          ! 3D T, S and rd direct fields
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( in ) ::   &
         ptem_tl, psal_tl, prd_tl            ! 3D T, S and rd tangent fields
      REAL(wp), DIMENSION(jpi,jpj), INTENT( out ) ::   &
         pgtu_tl, pgsu_tl, pgru_tl,       &  ! horizontal grad. of T, S and rd at u- 
         pgtv_tl, pgsv_tl, pgrv_tl           ! and v-points of the partial step level


      !! * Local declarations
      INTEGER ::   ji, jj,       &  ! Dummy loop indices
                   iku,ikv          ! partial step level at u- and v-points
      REAL(wp), DIMENSION(jpi,jpj) ::   &
         ztmpi, ztmpj,           &
         zti, ztj, zsi, zsj,     &  ! interpolated value of T, S 
         zri, zrj,               &  ! and rd
         ztitl, ztjtl,           &  ! interpolated value of Ttl
         zsitl, zsjtl,           &  !, Stl 
         zritl, zrjtl,           &  ! and rdtl
         zhgi, zhgj                 ! depth of interpolation for eos2d
      REAL(wp) ::   &
         ze3wu, ze3wv,           &  ! temporary scalars
         zmaxu1, zmaxu2,         &  !    "         "
         zmaxv1, zmaxv2             !    "         "

      ! Initialization (first time-step only): compute mbatu and mbatv
      IF( kt == nit000 ) THEN
         mbatu(:,:) = 0
         mbatv(:,:) = 0
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               mbatu(ji,jj) = MAX( MIN( mbathy(ji,jj), mbathy(ji+1,jj  ) ) - 1, 2 )
               mbatv(ji,jj) = MAX( MIN( mbathy(ji,jj), mbathy(ji  ,jj+1) ) - 1, 2 )
            END DO
         END DO
         ztmpi(:,:) = FLOAT( mbatu(:,:) )
         ztmpj(:,:) = FLOAT( mbatv(:,:) )
         ! lateral boundary conditions: T-point, sign unchanged
         CALL lbc_lnk( ztmpi , 'U', 1. )
         CALL lbc_lnk( ztmpj , 'V', 1. )
         mbatu(:,:) = MAX( INT( ztmpi(:,:) ), 2 )
         mbatv(:,:) = MAX( INT( ztmpj(:,:) ), 2 )
      ENDIF
      

      ! Interpolation of T and S at the last ocean level
# if defined key_vectopt_loop
         jj = 1
         DO ji = 1, jpij-jpi   ! vector opt. (forced unrolled)
# else
      DO jj = 1, jpjm1
         DO ji = 1, jpim1
# endif
            ! last level
            iku = mbatu(ji,jj)
            ikv = mbatv(ji,jj)

            ze3wu  = fse3w(ji+1,jj  ,iku) - fse3w(ji,jj,iku)
            ze3wv  = fse3w(ji  ,jj+1,ikv) - fse3w(ji,jj,ikv)
            zmaxu1 =  ze3wu / fse3w(ji+1,jj  ,iku)
            zmaxu2 = -ze3wu / fse3w(ji  ,jj  ,iku)
            zmaxv1 =  ze3wv / fse3w(ji  ,jj+1,ikv)
            zmaxv2 = -ze3wv / fse3w(ji  ,jj  ,ikv)

            ! i- direction

            IF( ze3wu >= 0. ) THEN      ! case 1
               ! interpolated values of T and S
               zti(ji,jj) = ptem(ji+1,jj,iku)            &
                  &       + zmaxu1 * ( ptem(ji+1,jj,iku-1) - ptem(ji+1,jj,iku) )
               zsi(ji,jj) = psal(ji+1,jj,iku)            &
                  &       + zmaxu1 * ( psal(ji+1,jj,iku-1) - psal(ji+1,jj,iku) )
               ztitl(ji,jj) = ptem_tl(ji+1,jj,iku)       &
                  &         + zmaxu1 * ( ptem_tl(ji+1,jj,iku-1) - ptem_tl(ji+1,jj,iku) )
               zsitl(ji,jj) = psal_tl(ji+1,jj,iku)       &
                  &         + zmaxu1 * ( psal_tl(ji+1,jj,iku-1) - psal_tl(ji+1,jj,iku) )
               ! depth of the partial step level
               zhgi(ji,jj) = fsdept(ji,jj,iku)
               ! gradient of T and S
               pgtu_tl(ji,jj) = umask(ji,jj,1) * ( ztitl(ji,jj) - ptem_tl(ji,jj,iku) )
               pgsu_tl(ji,jj) = umask(ji,jj,1) * ( zsitl(ji,jj) - psal_tl(ji,jj,iku) )
            ELSE                        ! case 2
               ! interpolated values of T and S
               zti(ji,jj) = ptem(ji,jj,iku)              &
                  &       + zmaxu2 * ( ptem(ji,jj,iku-1) - ptem(ji,jj,iku) )
               zsi(ji,jj) = psal(ji,jj,iku)              &
                  &       + zmaxu2 * ( psal(ji,jj,iku-1) - psal(ji,jj,iku) )
               ! interpolated values of T and S
               ztitl(ji,jj) = ptem_tl(ji,jj,iku)         &
                  &         + zmaxu2 * ( ptem_tl(ji,jj,iku-1) - ptem_tl(ji,jj,iku) )
               zsitl(ji,jj) = psal_tl(ji,jj,iku)         &
                  &         + zmaxu2 * ( psal_tl(ji,jj,iku-1) - psal_tl(ji,jj,iku) )
               ! depth of the partial step level
               zhgi(ji,jj) = fsdept(ji+1,jj,iku)
               ! gradient of T and S 
               pgtu_tl(ji,jj) = umask(ji,jj,1) * ( ptem_tl(ji+1,jj,iku) - ztitl (ji,jj) )
               pgsu_tl(ji,jj) = umask(ji,jj,1) * ( psal_tl(ji+1,jj,iku) - zsitl (ji,jj) )
            ENDIF

            ! j- direction

            IF( ze3wv >= 0. ) THEN      ! case 1
               ! interpolated values of direct T and S
               ztj(ji,jj) = ptem(ji,jj+1,ikv)            &
                  &       + zmaxv1 * ( ptem(ji,jj+1,ikv-1) - ptem(ji,jj+1,ikv) )
               zsj(ji,jj) = psal(ji,jj+1,ikv)            &
                  &       + zmaxv1 * ( psal(ji,jj+1,ikv-1) - psal(ji,jj+1,ikv) )
               ! interpolated values of tangent T and S
               ztjtl(ji,jj) = ptem_tl(ji,jj+1,ikv)       &
                  &         + zmaxv1 * ( ptem_tl(ji,jj+1,ikv-1) - ptem_tl(ji,jj+1,ikv) )
               zsjtl(ji,jj) = psal_tl(ji,jj+1,ikv)       &
                  &         + zmaxv1 * ( psal_tl(ji,jj+1,ikv-1) - psal_tl(ji,jj+1,ikv) )
               ! depth of the partial step level
               zhgj(ji,jj) = fsdept(ji,jj,ikv) 
               ! gradient of T and S
               pgtv_tl(ji,jj) = vmask(ji,jj,1) * ( ztjtl(ji,jj) - ptem_tl(ji,jj,ikv) )
               pgsv_tl(ji,jj) = vmask(ji,jj,1) * ( zsjtl(ji,jj) - psal_tl(ji,jj,ikv) )

            ELSE                        ! case 2
               ! interpolated values of T and S
               ztj(ji,jj) = ptem(ji,jj,ikv)       &
                  &       + zmaxv2 * ( ptem(ji,jj,ikv-1) - ptem(ji,jj,ikv) )
               zsj(ji,jj) = psal(ji,jj,ikv)       &
                  &       + zmaxv2 * ( psal(ji,jj,ikv-1) - psal(ji,jj,ikv) ) 
               ! interpolated values of T and S
               ztjtl(ji,jj) = ptem_tl(ji,jj,ikv)  &
                  &         + zmaxv2 * ( ptem_tl(ji,jj,ikv-1) - ptem_tl(ji,jj,ikv) )
               zsjtl(ji,jj) = psal_tl(ji,jj,ikv)  &
                  &         + zmaxv2 * ( psal_tl(ji,jj,ikv-1) - psal_tl(ji,jj,ikv) ) 
               ! depth of the partial step level
               zhgj(ji,jj) = fsdept(ji,jj+1,ikv) 
               ! gradient of T and S
               pgtv_tl(ji,jj) = vmask(ji,jj,1) * ( ptem_tl(ji,jj+1,ikv) - ztjtl(ji,jj) )
               pgsv_tl(ji,jj) = vmask(ji,jj,1) * ( psal_tl(ji,jj+1,ikv) - zsjtl(ji,jj) )
            ENDIF
# if ! defined key_vectopt_loop
         END DO
# endif
      END DO

      ! Compute interpolated rd from zti, zsi, ztj, zsj for the 2 cases at the depth of the partial
      ! step and store it in  zri, zrj for each  case
      CALL eos_tan( zti, zsi, zhgi, ztitl, zsitl, zritl )
      CALL eos_tan( ztj, zsj, zhgj, ztjtl, zsjtl, zrjtl )


      ! Gradient of density at the last level 
# if defined key_vectopt_loop 
         jj = 1
         DO ji = 1, jpij-jpi   ! vector opt. (forced unrolled)
# else
      DO jj = 1, jpjm1
         DO ji = 1, jpim1
# endif
            iku = mbatu(ji,jj)
            ikv = mbatv(ji,jj)
            ze3wu  = fse3w(ji+1,jj  ,iku) - fse3w(ji,jj,iku)
            ze3wv  = fse3w(ji  ,jj+1,ikv) - fse3w(ji,jj,ikv)
            IF( ze3wu >= 0. ) THEN    ! i-direction: case 1
               pgru_tl(ji,jj) = umask(ji,jj,1) * ( zritl(ji,jj) - prd_tl(ji,jj,iku) )
            ELSE                      ! i-direction: case 2
               pgru_tl(ji,jj) = umask(ji,jj,1) * ( prd_tl(ji+1,jj,iku) - zritl(ji,jj) )
            ENDIF
            IF( ze3wv >= 0. ) THEN    ! j-direction: case 1
               pgrv_tl(ji,jj) = vmask(ji,jj,1) * ( zrjtl(ji,jj) - prd_tl(ji,jj,ikv) )  
            ELSE                      ! j-direction: case 2
               pgrv_tl(ji,jj) = vmask(ji,jj,1) * ( prd_tl(ji,jj+1,ikv) - zrjtl(ji,jj) )
            ENDIF
# if ! defined key_vectopt_loop 
         END DO
# endif
      END DO

      ! Lateral boundary conditions on each gradient
      CALL lbc_lnk( pgtu_tl , 'U', -1. )  ;  CALL lbc_lnk( pgtv_tl , 'V', -1. )
      CALL lbc_lnk( pgsu_tl , 'U', -1. )  ;  CALL lbc_lnk( pgsv_tl , 'V', -1. )
      CALL lbc_lnk( pgru_tl , 'U', -1. )  ;  CALL lbc_lnk( pgrv_tl , 'V', -1. )

   END SUBROUTINE zps_hde_tan

   SUBROUTINE zps_hde_adj ( kt, ptem, psal,              &
      &                     ptem_ad, psal_ad, prd_ad ,   &
                            pgtu_ad, pgsu_ad, pgru_ad,   &
                            pgtv_ad, pgsv_ad, pgrv_ad     )
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE zps_hde_adj  ***
      !!                    
      !! ** Purpose of the direct routine:
      !!      Compute the horizontal derivative of T, S and rd
      !!      at u- and v-points with a linear interpolation for z-coordinate
      !!      with partial steps.
      !!
      !! ** Method of the direct routine:
      !!      In z-coord with partial steps, scale factors on last 
      !!      levels are different for each grid point, so that T, S and rd 
      !!      points are not at the same depth as in z-coord. To have horizontal
      !!      gradients again, we interpolate T and S at the good depth : 
      !!      Linear interpolation of T, S   
      !!         Computation of di(tb) and dj(tb) by vertical interpolation:
      !!          di(t) = t~ - t(i,j,k) or t(i+1,j,k) - t~
      !!          dj(t) = t~ - t(i,j,k) or t(i,j+1,k) - t~
      !!         This formulation computes the two cases:
      !!                 CASE 1                   CASE 2  
      !!         k-1  ___ ___________   k-1   ___ ___________
      !!                    Ti  T~                  T~  Ti+1
      !!                  _____                        _____
      !!         k        |   |Ti+1     k           Ti |   |
      !!                  |   |____                ____|   |
      !!              ___ |   |   |           ___  |   |   |
      !!                  
      !!      case 1->   e3w(i+1) >= e3w(i) ( and e3w(j+1) >= e3w(j) ) then
      !!          t~ = t(i+1,j  ,k) + (e3w(i+1) - e3w(i)) * dk(Ti+1)/e3w(i+1)
      !!        ( t~ = t(i  ,j+1,k) + (e3w(j+1) - e3w(j)) * dk(Tj+1)/e3w(j+1)  )
      !!          or
      !!      case 2->   e3w(i+1) <= e3w(i) ( and e3w(j+1) <= e3w(j) ) then
      !!          t~ = t(i,j,k) + (e3w(i) - e3w(i+1)) * dk(Ti)/e3w(i )
      !!        ( t~ = t(i,j,k) + (e3w(j) - e3w(j+1)) * dk(Tj)/e3w(j ) )
      !!          Idem for di(s) and dj(s)          
      !!
      !!      For rho, we call eos_insitu_2d which will compute rd~(t~,s~) at 
      !!      the good depth zh from interpolated T and S for the different
      !!      formulation of the equation of state (eos).
      !!      Gradient formulation for rho :
      !!          di(rho) = rd~ - rd(i,j,k) or rd (i+1,j,k) - rd~
      !!
      !! ** Action  : - pgtu, pgsu, pgru: horizontal gradient of T, S
      !!                and rd at U-points 
      !!              - pgtv, pgsv, pgrv: horizontal gradient of T, S
      !!                and rd at V-points 
      !!
      !! History of the direct routine:
      !!   8.5  !  02-04  (A. Bozec)  Original code
      !!   8.5  !  02-08  (G. Madec E. Durand)  Optimization and Free form
      !! History of the TAM routine:
      !!   9.0  !  08-06 (A. Vidard) Skeleton
      !!        !  08-08 (A. Vidard) adjoint of the 02-08 version
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::   kt          ! ocean time-step index
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( in ) ::   &
         ptem, psal                          ! 3D T, S and rd direct fields
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( inout ) ::   &
         ptem_ad, psal_ad, prd_ad            ! 3D T, S and rd tangent fields
      REAL(wp), DIMENSION(jpi,jpj), INTENT( inout ) ::   &
         pgtu_ad, pgsu_ad, pgru_ad,       &  ! horizontal grad. of T, S and rd at u- 
         pgtv_ad, pgsv_ad, pgrv_ad           ! and v-points of the partial step level


      !! * Local declarations
      INTEGER ::   ji, jj,       &  ! Dummy loop indices
                   iku,ikv          ! partial step level at u- and v-points
      REAL(wp), DIMENSION(jpi,jpj) ::    &
         ztmpi, ztmpj,                   &
         zti, ztj, zsi, zsj,             &  ! interpolated value of T, S 
         zri, zrj,                       &  ! and rd
         ztiad, ztjad, zsiad, zsjad,     &  ! interpolated value of Tad, Sad 
         zriad, zrjad,                   &  ! and rdad
         zhgi, zhgj                         ! depth of interpolation for eos2d
      REAL(wp) ::   &
         ze3wu, ze3wv,           &  ! temporary scalars
         zmaxu1, zmaxu2,         &  !    "         "
         zmaxv1, zmaxv2             !    "         "

      ! Initialization (first time-step only): compute mbatu and mbatv
      IF( kt == nitend ) THEN
         mbatu(:,:) = 0
         mbatv(:,:) = 0
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               mbatu(ji,jj) = MAX( MIN( mbathy(ji,jj), mbathy(ji+1,jj  ) ) - 1, 2 )
               mbatv(ji,jj) = MAX( MIN( mbathy(ji,jj), mbathy(ji  ,jj+1) ) - 1, 2 )
            END DO
         END DO
         ztmpi(:,:) = FLOAT( mbatu(:,:) )
         ztmpj(:,:) = FLOAT( mbatv(:,:) )
         ! lateral boundary conditions: T-point, sign unchanged
         CALL lbc_lnk( ztmpi , 'U', 1. )
         CALL lbc_lnk( ztmpj , 'V', 1. )
         mbatu(:,:) = MAX( INT( ztmpi(:,:) ), 2 )
         mbatv(:,:) = MAX( INT( ztmpj(:,:) ), 2 )
      ENDIF
      ! 1. Direct model recomputation
      ! Interpolation of T and S at the last ocean level
# if defined key_vectopt_loop  
         jj = 1
         DO ji = 1, jpij-jpi   ! vector opt. (forced unrolled)
# else
      DO jj = 1, jpjm1
         DO ji = 1, jpim1
# endif
            ! last level
            iku = mbatu(ji,jj)
            ikv = mbatv(ji,jj)

            ze3wu  = fse3w(ji+1,jj  ,iku) - fse3w(ji,jj,iku)
            ze3wv  = fse3w(ji  ,jj+1,ikv) - fse3w(ji,jj,ikv)
            zmaxu1 =  ze3wu / fse3w(ji+1,jj  ,iku)
            zmaxu2 = -ze3wu / fse3w(ji  ,jj  ,iku)
            zmaxv1 =  ze3wv / fse3w(ji  ,jj+1,ikv)
            zmaxv2 = -ze3wv / fse3w(ji  ,jj  ,ikv)

            ! i- direction

            IF( ze3wu >= 0. ) THEN      ! case 1
               ! interpolated values of T and S
               zti(ji,jj) = ptem(ji+1,jj,iku)       &
                  &       + zmaxu1 * ( ptem(ji+1,jj,iku-1) - ptem(ji+1,jj,iku) )
               zsi(ji,jj) = psal(ji+1,jj,iku)       &
                  &       + zmaxu1 * ( psal(ji+1,jj,iku-1) - psal(ji+1,jj,iku) )
               ! depth of the partial step level
               zhgi(ji,jj) = fsdept(ji,jj,iku)
            ELSE                        ! case 2
               ! interpolated values of T and S
               zti(ji,jj) = ptem(ji,jj,iku)         &
                  &       + zmaxu2 * ( ptem(ji,jj,iku-1) - ptem(ji,jj,iku) )
               zsi(ji,jj) = psal(ji,jj,iku) + zmaxu2 * ( psal(ji,jj,iku-1) - psal(ji,jj,iku) )
               ! depth of the partial step level
               zhgi(ji,jj) = fsdept(ji+1,jj,iku)
            ENDIF

            ! j- direction

            IF( ze3wv >= 0. ) THEN      ! case 1
               ! interpolated values of T and S
               ztj(ji,jj) = ptem(ji,jj+1,ikv)       &
                  &       + zmaxv1 * ( ptem(ji,jj+1,ikv-1) - ptem(ji,jj+1,ikv) )
               zsj(ji,jj) = psal(ji,jj+1,ikv)       &
                  &       + zmaxv1 * ( psal(ji,jj+1,ikv-1) - psal(ji,jj+1,ikv) )
               ! depth of the partial step level
               zhgj(ji,jj) = fsdept(ji,jj,ikv) 
            ELSE                        ! case 2
               ! interpolated values of T and S
               ztj(ji,jj) = ptem(ji,jj,ikv)         &
                  &       + zmaxv2 * ( ptem(ji,jj,ikv-1) - ptem(ji,jj,ikv) )
               zsj(ji,jj) = psal(ji,jj,ikv)         &
                  &       + zmaxv2 * ( psal(ji,jj,ikv-1) - psal(ji,jj,ikv) ) 
               ! depth of the partial step level
               zhgj(ji,jj) = fsdept(ji,jj+1,ikv) 
            ENDIF
# if ! defined key_vectopt_loop  
         END DO
# endif
      END DO

      !2. Adjoint model counterpart
      ztiad = 0.0_wp ; ztjad = 0.0_wp ; zsiad = 0.0_wp
      zsjad = 0.0_wp ; zriad = 0.0_wp ; zrjad = 0.0_wp

      ! Lateral boundary conditions on each gradient
      CALL lbc_lnk_adj( pgtu_ad , 'U', -1. )  ;  CALL lbc_lnk_adj( pgtv_ad , 'V', -1. )
      CALL lbc_lnk_adj( pgsu_ad , 'U', -1. )  ;  CALL lbc_lnk_adj( pgsv_ad , 'V', -1. )
      CALL lbc_lnk_adj( pgru_ad , 'U', -1. )  ;  CALL lbc_lnk_adj( pgrv_ad , 'V', -1. )

      ! Gradient of density at the last level 
# if defined key_vectopt_loop  
         jj = 1
         DO ji = jpij-jpi, -1    ! vector opt. (forced unrolled)
# else
      DO jj = jpjm1, 1, -1
         DO ji = jpim1, 1, -1
# endif
            iku = mbatu(ji,jj)
            ikv = mbatv(ji,jj)
            ze3wu  = fse3w(ji+1,jj  ,iku) - fse3w(ji,jj,iku)
            ze3wv  = fse3w(ji  ,jj+1,ikv) - fse3w(ji,jj,ikv)
            IF( ze3wv >= 0. ) THEN    ! j-direction: case 1
               zrjad(ji,jj)        = zrjad(ji,jj)                     &
                  &                + pgrv_ad(ji,jj) * vmask(ji,jj,1)
               prd_ad(ji,jj,ikv)   = prd_ad(ji,jj,ikv)                &
                  &                - pgrv_ad(ji,jj) * vmask(ji,jj,1)
               pgrv_ad(ji,jj)      = 0.0_wp
            ELSE                      ! j-direction: case 2
               prd_ad(ji,jj+1,ikv) = prd_ad(ji,jj+1,ikv)            &
                  &                + pgrv_ad(ji,jj) * vmask(ji,jj,1)
               zrjad(ji,jj)        = zrjad(ji,jj)                     &
                  &                - pgrv_ad(ji,jj) * vmask(ji,jj,1)
               pgrv_ad(ji,jj)      = 0.0_wp
            ENDIF
            IF( ze3wu >= 0. ) THEN    ! i-direction: case 1
               zriad(ji,jj)        = zriad(ji,jj)                     &
                  &                + pgru_ad(ji,jj) * umask(ji,jj,1)
               prd_ad(ji,jj,iku)   = prd_ad(ji,jj,iku)                &
                  &                - pgru_ad(ji,jj) * umask(ji,jj,1)
               pgru_ad(ji,jj)      = 0.0_wp
            ELSE                      ! i-direction: case 2
               prd_ad(ji+1,jj,iku) = prd_ad(ji+1,jj,iku)            &
                  &                + pgru_ad(ji,jj) * umask(ji,jj,1)
               zriad(ji,jj)        = zriad(ji,jj)                   &
                  &                - pgru_ad(ji,jj) * umask(ji,jj,1)
               pgru_ad(ji,jj)      = 0.0_wp
            ENDIF

# if ! defined key_vectopt_loop   
         END DO
# endif
      END DO


      ! Compute interpolated rd from zti, zsi, ztj, zsj for the 2 cases at the depth of the partial
      ! step and store it in  zri, zrj for each  case
      CALL eos_adj( ztj, zsj, zhgj, ztjad, zsjad, zrjad )
      CALL eos_adj( zti, zsi, zhgi, ztiad, zsiad, zriad )


# if defined key_vectopt_loop  
         jj = 1
         DO ji = jpij-jpi, 1, -1   ! vector opt. (forced unrolled)
# else
      DO jj = jpjm1, 1, -1
         DO ji = jpim1, 1, -1
# endif
            ! last level
            iku = mbatu(ji,jj)
            ikv = mbatv(ji,jj)

            ze3wu  = fse3w(ji+1,jj  ,iku) - fse3w(ji,jj,iku)
            ze3wv  = fse3w(ji  ,jj+1,ikv) - fse3w(ji,jj,ikv)
            zmaxu1 =  ze3wu / fse3w(ji+1,jj  ,iku)
            zmaxu2 = -ze3wu / fse3w(ji  ,jj  ,iku)
            zmaxv1 =  ze3wv / fse3w(ji  ,jj+1,ikv)
            zmaxv2 = -ze3wv / fse3w(ji  ,jj  ,ikv)

            ! j- direction

            IF( ze3wv >= 0. ) THEN      ! case 1
               ! gradient of T and S
               ztjad(ji,jj)           = ztjad(ji,jj)                &
                  &                   + pgtv_ad(ji,jj) * vmask(ji,jj,1)
               ptem_ad(ji,jj,ikv)     = ptem_ad(ji,jj,ikv)          &
                  &                   - pgtv_ad(ji,jj) * vmask(ji,jj,1)
               pgtv_ad(ji,jj)         = 0.0_wp

               zsjad(ji,jj)           = zsjad(ji,jj)                &
                  &                   + pgsv_ad(ji,jj) * vmask(ji,jj,1)
               psal_ad(ji,jj,ikv)     = psal_ad(ji,jj,ikv)          &
                  &                   - pgsv_ad(ji,jj) * vmask(ji,jj,1)
               pgsv_ad(ji,jj)         = 0.0_wp
               ! interpolated values of T and S
               ptem_ad(ji,jj+1,ikv)   = ptem_ad(ji,jj+1,ikv)        &
                  &                   + ztjad(ji,jj) * (1 - zmaxv1)
               ptem_ad(ji,jj+1,ikv-1) = ptem_ad(ji,jj+1,ikv-1)      &
                  &                   + ztjad(ji,jj)* zmaxv1
               ztjad(ji,jj)           = 0.0_wp

               psal_ad(ji,jj+1,ikv)   = psal_ad(ji,jj+1,ikv)        &
                  &                   + zsjad(ji,jj) * (1 - zmaxv1)
               psal_ad(ji,jj+1,ikv-1) = psal_ad(ji,jj+1,ikv-1)      &
                  &                   + zsjad(ji,jj) * zmaxv1 
               zsjad(ji,jj)           = 0.0_wp
            ELSE                        ! case 2
                ! gradient of T and S
               ptem_ad(ji,jj+1,ikv)   = ptem_ad(ji,jj+1,ikv)        &
                  &                   + pgtv_ad(ji,jj) * vmask(ji,jj,1) 
               ztjad(ji,jj)           = ztjad(ji,jj)                &
                  &                   - pgtv_ad(ji,jj) * vmask(ji,jj,1) 
               pgtv_ad(ji,jj)         = 0.0_wp

               psal_ad(ji,jj+1,ikv)   = psal_ad(ji,jj+1,ikv)        &
                  &                   + pgsv_ad(ji,jj) * vmask(ji,jj,1) 
               zsjad(ji,jj)           = zsjad(ji,jj)                &
                  &                   - pgsv_ad(ji,jj) * vmask(ji,jj,1) 
               pgsv_ad(ji,jj)         = 0.0_wp
              ! interpolated values of T and S
               ptem_ad(ji,jj,ikv)     = ptem_ad(ji,jj,ikv)          &
                  &                   + ztjad(ji,jj) * (1 - zmaxv2)
               ptem_ad(ji,jj,ikv-1)   = ptem_ad(ji,jj,ikv-1)        &
                  &                   + ztjad(ji,jj) * zmaxv2
               ztjad(ji,jj)           = 0.0_wp

               psal_ad(ji,jj,ikv)     = psal_ad(ji,jj,ikv)          &
                  &                   + zsjad(ji,jj) * (1 - zmaxv2) 
               psal_ad(ji,jj,ikv-1)   = psal_ad(ji,jj,ikv-1)        &
                  &                   + zsjad(ji,jj) * zmaxv2
               zsjad(ji,jj)           = 0.0_wp
            ENDIF

            ! i- direction

            IF( ze3wu >= 0. ) THEN      ! case 1
               ! gradient of T and S
               ztiad(ji,jj)       = ztiad(ji,jj)                    &
                  &               + pgtu_ad(ji,jj) * umask(ji,jj,1)
               ptem_ad(ji,jj,iku) = ptem_ad(ji,jj,iku)              &
                  &               - pgtu_ad(ji,jj) * umask(ji,jj,1)
               pgtu_ad(ji,jj)     = 0.0_wp

               zsiad(ji,jj)       = zsiad(ji,jj)                    &
                  &               + pgsu_ad(ji,jj) * umask(ji,jj,1)
               psal_ad(ji,jj,iku) = psal_ad(ji,jj,iku)              &
                  &               - pgsu_ad(ji,jj) * umask(ji,jj,1)
               pgsu_ad(ji,jj)     = 0.0_wp
               ! interpolated values of T and S
               ptem_ad(ji+1,jj,iku)   = ptem_ad(ji+1,jj,iku)        &
                  &                   + ztiad(ji,jj) * (1 - zmaxu1)
               ptem_ad(ji+1,jj,iku-1) = ptem_ad(ji+1,jj,iku-1)      &
                  &                   + ztiad(ji,jj) * zmaxu1 
               ztiad(ji,jj)           = 0.0_wp

               psal_ad(ji+1,jj,iku)   = psal_ad(ji+1,jj,iku)        &
                  &                   + zsiad(ji,jj) * (1 - zmaxu1)
               psal_ad(ji+1,jj,iku-1) = psal_ad(ji+1,jj,iku-1)      &
                  &                   + zsiad(ji,jj) * zmaxu1
               zsiad(ji,jj)           = 0.0_wp

            ELSE                        ! case 2
              ! gradient of T and S 
               ptem_ad(ji+1,jj,iku)   = ptem_ad(ji+1,jj,iku)        &
                  &                   + pgtu_ad(ji,jj) * umask(ji,jj,1)
               ztiad (ji,jj)          = ztiad (ji,jj)               &
                  &                   - pgtu_ad(ji,jj) * umask(ji,jj,1)
               pgtu_ad(ji,jj)         = 0.0_wp

               psal_ad(ji+1,jj,iku)   = psal_ad(ji+1,jj,iku)        &
                  &                   + pgsu_ad(ji,jj) * umask(ji,jj,1)
               zsiad (ji,jj)          = zsiad (ji,jj)               &
                  &                   - pgsu_ad(ji,jj) * umask(ji,jj,1)
               pgsu_ad(ji,jj)         = 0.0_wp
              ! interpolated values of T and S
               ptem_ad(ji,jj,iku)     = ptem_ad(ji,jj,iku)          &
                  &                   + ztiad(ji,jj) * (1 - zmaxu2)
               ptem_ad(ji,jj,iku-1)   = ptem_ad(ji,jj,iku-1)        &
                  &                   + ztiad(ji,jj) * zmaxu2
               ztiad(ji,jj)           = 0.0_wp

               psal_ad(ji,jj,iku)     = psal_ad(ji,jj,iku)          &
                  &                   + zsiad(ji,jj) * (1 - zmaxu2)
               psal_ad(ji,jj,iku-1)   = psal_ad(ji,jj,iku-1)        &
                  &                   + zsiad(ji,jj) * zmaxu2
               zsiad(ji,jj)           = 0.0_wp
            ENDIF

# if ! defined key_vectopt_loop  
         END DO
# endif
      END DO

   END SUBROUTINE zps_hde_adj
   SUBROUTINE zps_hde_adj_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_adv_adj_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the scalar product 
      !!           
      !!                 ( L dx )^T W dy  =  dx^T L^T W dy
      !!
      !!              where  L   = tangent routine
      !!                     L^T = adjoint routine
      !!                     W   = diagonal matrix of scale factors
      !!                     dx  = input perturbation (random field)
      !!                     dy  = L dx 
      !!
      !!                   
      !! History :
      !!        ! 08-08 (A. Vidard)
      !!-----------------------------------------------------------------------
      !! * Modules used

      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      INTEGER :: &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk,    & 
         & kt
      INTEGER, DIMENSION(jpi,jpj) :: &
         & iseed_2d        ! 2D seed for the random number generator
 
      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztem,         & ! Direct field : temperature
         & zsal,         & ! Direct field : salinity
         & ztem_tlin,    & ! Tangent input: temperature
         & zsal_tlin,    & ! Tangent input: salinity
         & zrd_tlin,     & ! Tangent input:
         & ztem_adout,   & ! Adjoint output: temperature
         & zsal_adout,   & ! Adjoint output: salinity
         & zrd_adout,    & ! Adjoint output: 
         & zar,          & ! 3D random field for rd
         & zat,          & ! 3D random field for t
         & zas             ! 3D random field for s
         
      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgtu_tlout,   & ! Tangent output: horizontal gradient
         & zgtv_tlout,   & ! Tangent output: horizontal gradient
         & zgsu_tlout,   & ! Tangent output: horizontal gradient
         & zgsv_tlout,   & ! Tangent output: horizontal gradient
         & zgru_tlout,   & ! Tangent output: horizontal gradient
         & zgrv_tlout,   & ! Tangent output: horizontal gradient
         & zgtu_adin,    & ! Adjoint input : horizontal gradient
         & zgtv_adin,    & ! Adjoint input : horizontal gradient
         & zgsu_adin,    & ! Adjoint input : horizontal gradient
         & zgsv_adin,    & ! Adjoint input : horizontal gradient
         & zgru_adin,    & ! Adjoint input : horizontal gradient
         & zgrv_adin       ! Adjoint input : horizontal gradient

      REAL(KIND=wp) :: &
                           ! random field standard deviation for:
         & zsp1,         & ! scalar product involving the tangent routine
         & zsp1_1,       & !   scalar product components
         & zsp1_2,       & 
         & zsp1_3,       & !   scalar product components
         & zsp1_4,       & 
         & zsp1_5,       & !   scalar product components
         & zsp1_6,       & 
         & zsp2,         & ! scalar product involving the adjoint routine
         & zsp2_1,       & !   scalar product components
         & zsp2_2,       & 
         & zsp2_3
      CHARACTER (LEN=14) :: &
         & cl_name

      kt = nit000
      ! Allocate memory
      ALLOCATE( &
         & ztem(jpi,jpj,jpk),         & 
         & zsal(jpi,jpj,jpk),         & 
         & ztem_tlin(jpi,jpj,jpk),    & 
         & zsal_tlin(jpi,jpj,jpk),    & 
         & zrd_tlin(jpi,jpj,jpk),     & 
         & ztem_adout(jpi,jpj,jpk),   & 
         & zsal_adout(jpi,jpj,jpk),   & 
         & zrd_adout(jpi,jpj,jpk),    & 
         & zar(jpi,jpj,jpk),          & 
         & zat(jpi,jpj,jpk),          & 
         & zas(jpi,jpj,jpk),          &       
         & zgtu_tlout(jpi,jpj),       & 
         & zgtv_tlout(jpi,jpj),       & 
         & zgsu_tlout(jpi,jpj),       & 
         & zgsv_tlout(jpi,jpj),       & 
         & zgru_tlout(jpi,jpj),       & 
         & zgrv_tlout(jpi,jpj),       & 
         & zgtu_adin(jpi,jpj),        & 
         & zgtv_adin(jpi,jpj),        & 
         & zgsu_adin(jpi,jpj),        & 
         & zgsv_adin(jpi,jpj),        & 
         & zgru_adin(jpi,jpj),        & 
         & zgrv_adin(jpi,jpj)         &    
         & ) 
      ! Initialize random field standard deviationsthe reference state
      ztem = tn
      zsal = sn
 
      !=============================================================
      ! 1) dx = ( T ) and dy = ( T )
      !=============================================================

      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      ztem_tlin(:,:,:)  = 0.0_wp     
      zsal_tlin(:,:,:)  = 0.0_wp     
      zrd_tlin(:,:,:)   = 0.0_wp     
      ztem_adout(:,:,:) = 0.0_wp    
      zsal_adout(:,:,:) = 0.0_wp    
      zrd_adout(:,:,:)  = 0.0_wp    
      zgtu_tlout(:,:)   = 0.0_wp    
      zgtv_tlout(:,:)   = 0.0_wp    
      zgsu_tlout(:,:)   = 0.0_wp    
      zgsv_tlout(:,:)   = 0.0_wp    
      zgru_tlout(:,:)   = 0.0_wp    
      zgrv_tlout(:,:)   = 0.0_wp    
      zgtu_adin(:,:)    = 0.0_wp     
      zgtv_adin(:,:)    = 0.0_wp     
      zgsu_adin(:,:)    = 0.0_wp     
      zgsv_adin(:,:)    = 0.0_wp     
      zgru_adin(:,:)    = 0.0_wp     
      zgrv_adin(:,:)    = 0.0_wp         
      
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 456953 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zat, 'T', 0.0_wp, stdt )
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 526791 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zas, 'T', 0.0_wp, stds )
      
      DO jj = 1, jpj
         DO ji = 1, jpi
            iseed_2d(ji,jj) = - ( 395703 + &
               &                  mig(ji) + ( mjg(jj) - 1 ) * jpiglo )
         END DO
      END DO
      CALL grid_random( iseed_2d, zar, 'T', 0.0_wp, stdr )
 
      DO jk = 1, jpk
         DO jj = nldj, nlej
            DO ji = nldi, nlei
               ztem_tlin(ji,jj,jk) = zat(ji,jj,jk)
               zsal_tlin(ji,jj,jk) = zas(ji,jj,jk)
               zrd_tlin( ji,jj,jk) = zar(ji,jj,jk) 
            END DO
         END DO
      END DO

      CALL zps_hde_tan ( kt, ztem      , zsal      ,               &
         &                   ztem_tlin , zsal_tlin , zrd_tlin  ,   &
         &                   zgtu_tlout, zgsu_tlout, zgru_tlout,   &
         &                   zgtv_tlout, zgsv_tlout, zgrv_tlout     )
      DO jj = nldj, nlej
         DO ji = nldi, nlei
            jk = mbatu(ji,jj)
            zgtu_adin(ji,jj) = zgtu_tlout(ji,jj) &
               &             * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
               &             * umask(ji,jj,jk)
            zgsu_adin(ji,jj) = zgsu_tlout(ji,jj) &
               &             * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
               &             * umask(ji,jj,jk)
            zgru_adin(ji,jj) = zgru_tlout(ji,jj) &
               &             * e1u(ji,jj) * e2u(ji,jj) * fse3u(ji,jj,jk) &
               &             * umask(ji,jj,jk)
            jk = mbatv(ji,jj)
            zgtv_adin(ji,jj) = zgtv_tlout(ji,jj) &
               &             * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
               &             * vmask(ji,jj,jk)
            zgsv_adin(ji,jj) = zgsv_tlout(ji,jj) &
               &             * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
               &             * vmask(ji,jj,jk)
            zgrv_adin(ji,jj) = zgrv_tlout(ji,jj) &
               &             * e1v(ji,jj) * e2v(ji,jj) * fse3v(ji,jj,jk) &
               &             * vmask(ji,jj,jk)
         END DO
      END DO
      !--------------------------------------------------------------------
      ! Compute the scalar product: ( L dx )^T W dy
      !--------------------------------------------------------------------

      zsp1_1 = DOT_PRODUCT( zgtu_adin, zgtu_tlout )
      zsp1_2 = DOT_PRODUCT( zgsu_adin, zgsu_tlout )
      zsp1_3 = DOT_PRODUCT( zgru_adin, zgru_tlout )
      zsp1_4 = DOT_PRODUCT( zgtv_adin, zgtv_tlout )
      zsp1_5 = DOT_PRODUCT( zgsv_adin, zgsv_tlout )
      zsp1_6 = DOT_PRODUCT( zgrv_adin, zgrv_tlout )
      zsp1 = zsp1_1 + zsp1_2 + zsp1_3 + zsp1_4 + zsp1_5 + zsp1_6


      !--------------------------------------------------------------------
      ! Call the adjoint routine: dx^* = L^T dy^*
      !--------------------------------------------------------------------
      CALL zps_hde_adj ( kt, ztem      , zsal      ,               &
         &                   ztem_adout, zsal_adout, zrd_adout ,   &
         &                   zgtu_adin , zgsu_adin , zgru_adin ,   &
         &                   zgtv_adin , zgsv_adin , zgrv_adin      )
      zsp2_1 = DOT_PRODUCT( ztem_tlin, ztem_adout )
      zsp2_2 = DOT_PRODUCT( zsal_tlin, zsal_adout )
      zsp2_3 = DOT_PRODUCT( zrd_tlin , zrd_adout  )
      zsp2   = zsp2_1 + zsp2_2 + zsp2_3

      ! Compare the scalar products

      cl_name = 'zps_hde_adj   '
      CALL prntst_adj( cl_name, kumadt, zsp1, zsp2 )

      ! Deallocate memory
      DEALLOCATE( &
         & ztem,         & 
         & zsal,         & 
         & ztem_tlin,    & 
         & zsal_tlin,    & 
         & zrd_tlin,     & 
         & ztem_adout,   & 
         & zsal_adout,   & 
         & zrd_adout,    & 
         & zar,          & 
         & zat,          & 
         & zas,          &       
         & zgtu_tlout,   & 
         & zgtv_tlout,   & 
         & zgsu_tlout,   & 
         & zgsv_tlout,   & 
         & zgru_tlout,   & 
         & zgrv_tlout,   & 
         & zgtu_adin,    & 
         & zgtv_adin,    & 
         & zgsu_adin,    & 
         & zgsv_adin,    & 
         & zgru_adin,    & 
         & zgrv_adin     &    
         & ) 


      
   END SUBROUTINE zps_hde_adj_tst
#if defined key_tst_tlm
   SUBROUTINE zps_hde_tlm_tst( kumadt )
      !!-----------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE dyn_adv_tlm_tst ***
      !!
      !! ** Purpose : Test the adjoint routine.
      !!
      !! ** Method  : Verify the tangent with Taylor expansion 
      !!           
      !!                 M(x+hdx) = M(x) + L(hdx) + O(h^2)
      !!
      !!              where  L   = tangent routine
      !!                     M   = direct routine
      !!                     dx  = input perturbation (random field)
      !!                     h   = ration on perturbation 
      !!
      !!    In the tangent test we verify that:
      !!                M(x+h*dx) - M(x)
      !!        g(h) = ------------------ --->  1    as  h ---> 0
      !!                    L(h*dx)
      !!    and
      !!                g(h) - 1
      !!        f(h) = ----------         --->  k (costant) as  h ---> 0
      !!                    p
      !!                    
      !! History :
      !!        ! 09-08 (A. Vigilant)
      !!-----------------------------------------------------------------------
      !! * Modules used
      USE zpshde
      USE oce_tam,        ONLY: &
        & gtu_tl, gtv_tl,       &
        & gsu_tl, gsv_tl,       &
        & gru_tl, grv_tl,       &
        & tn_tl, sn_tl, rhd_tl
      USE tamtrj              ! writing out state trajectory
      USE par_tlm,    ONLY: &
        & tlm_bch,          &
        & cur_loop,         &
        & h_ratio
      USE istate_mod
      USE divcur             ! horizontal divergence and relative vorticity
      USE wzvmod             !  vertical velocity
      USE gridrandom, ONLY: &
        & grid_rd_sd
      USE trj_tam
      USE oce           , ONLY: & ! ocean dynamics and tracers variables
        & tn, sn, rhd,          &
        & gtu, gtv, gsu, gsv,   &
        & gru, grv
      USE opatam_tst_ini, ONLY: &
       & tlm_namrd
      USE tamctl,         ONLY: & ! Control parameters
       & numtan, numtan_sc
      !! * Arguments
      INTEGER, INTENT(IN) :: &
         & kumadt             ! Output unit
 
      !! * Local declarations
      INTEGER :: &
         & ji,    &        ! dummy loop indices
         & jj,    &        
         & jk   
      !! * Local declarations
      REAL(KIND=wp), DIMENSION(:,:,:), ALLOCATABLE :: &
         & ztem,         & ! Direct field : temperature
         & zsal,         & ! Direct field : salinity
         & ztn_tlin,     & ! Tangent input: temperature
         & zsn_tlin,     & ! Tangent input: salinity
         & zrd_tlin,     &
         & z3r             ! 3D field   

      REAL(KIND=wp), DIMENSION(:,:), ALLOCATABLE :: &
         & zgtu_out, zgtu_wop,    & ! Tangent output: horizontal gradient
         & zgtv_out, zgtv_wop,    & ! Tangent output: horizontal gradient
         & zgsu_out, zgsu_wop,    & ! Tangent output: horizontal gradient
         & zgsv_out, zgsv_wop,    & ! Tangent output: horizontal gradient
         & zgru_out, zgru_wop,    & ! Tangent output: horizontal gradient
         & zgrv_out, zgrv_wop

      REAL(KIND=wp) :: &
         & zsp1,                 & ! scalar product involving the tangent routine
         & zsp1_1, zsp1_2,       & 
         & zsp1_3, zsp1_4,       & 
         & zsp1_5, zsp1_6,       & 
         & zsp2,                 & 
         & zsp2_1, zsp2_2,       & 
         & zsp2_3, zsp2_4,       & 
         & zsp2_5, zsp2_6,       & 
         & zsp3,                 &
         & zsp3_1, zsp3_2,       & 
         & zsp3_3, zsp3_4,       & 
         & zsp3_5, zsp3_6,       & 
         & zzsp,                 & 
         & zzsp_1, zzsp_2,       & 
         & zzsp_3, zzsp_4,       & 
         & zzsp_5, zzsp_6,       &
         & gamma,                &
         & zgsp1, zgsp2,         &
         & zgsp3, zgsp4,         &
         & zgsp5, zgsp6,         &
         & zgsp7
      CHARACTER(LEN=14)   :: cl_name
      CHARACTER (LEN=128) :: file_out, file_wop, file_xdx
      CHARACTER (LEN=90)  ::  FMT
      REAL(KIND=wp), DIMENSION(100):: &
         & zscgtu, zscgtv, zscgsu,    &
         & zscgsv, zscgru, zscgrv,    &
         & zscerrgtu, zscerrgtv,      &
         & zscerrgsu, zscerrgsv,      &
         & zscerrgru, zscerrgrv
      INTEGER, DIMENSION(100):: &
         & iiposgtu, iiposgtv, iiposgsu,   &
         & iiposgsv, iiposgru, iiposgrv,   &
         & ijposgtu, ijposgtv, ijposgsu,   &
         & ijposgsv, ijposgru, ijposgrv  
      INTEGER::             &
         & ii,              &
         & isamp=40,        &
         & jsamp=40,        &
         & numsctlm
     REAL(KIND=wp), DIMENSION(jpi,jpj) :: &
         & zerrgtu, zerrgtv, zerrgsu,     &
         & zerrgsv, zerrgru, zerrgrv  
      ! Allocate memory

      ALLOCATE( &
         & ztem(jpi,jpj,jpk),       & 
         & zsal(jpi,jpj,jpk),       & 
         & ztn_tlin(jpi,jpj,jpk),   & 
         & zsn_tlin(jpi,jpj,jpk),   & 
         & zrd_tlin(jpi,jpj,jpk),   &   
         & z3r     (jpi,jpj,jpk),   &       
         & zgtu_out(jpi,jpj),       & 
         & zgtv_out(jpi,jpj),       & 
         & zgsu_out(jpi,jpj),       & 
         & zgsv_out(jpi,jpj),       & 
         & zgru_out(jpi,jpj),       & 
         & zgrv_out(jpi,jpj),       &
         & zgtu_wop(jpi,jpj),       & 
         & zgtv_wop(jpi,jpj),       & 
         & zgsu_wop(jpi,jpj),       & 
         & zgsv_wop(jpi,jpj),       & 
         & zgru_wop(jpi,jpj),       & 
         & zgrv_wop(jpi,jpj)        &
         & ) 
      !--------------------------------------------------------------------
      ! Reset the tangent and adjoint variables
      !--------------------------------------------------------------------
      ztn_tlin(:,:,:) = 0.0_wp     
      zsn_tlin(:,:,:) = 0.0_wp     
      zrd_tlin(:,:,:) = 0.0_wp         
      zgtu_out(:,:)   = 0.0_wp    
      zgtv_out(:,:)   = 0.0_wp    
      zgsu_out(:,:)   = 0.0_wp    
      zgsv_out(:,:)   = 0.0_wp    
      zgru_out(:,:)   = 0.0_wp    
      zgrv_out(:,:)   = 0.0_wp
      IF ( tlm_bch == 2 ) THEN
      gtu_tl(:,:)     = 0.0_wp
      gtv_tl(:,:)     = 0.0_wp    
      gsu_tl(:,:)     = 0.0_wp    
      gsv_tl(:,:)     = 0.0_wp    
      gru_tl(:,:)     = 0.0_wp    
      grv_tl(:,:)     = 0.0_wp 
      ENDIF
      zscgtu(:)         = 0.0_wp
      zscgtv(:)         = 0.0_wp
      zscgsu(:)         = 0.0_wp
      zscgsv(:)         = 0.0_wp
      zscgru(:)         = 0.0_wp
      zscgrv(:)         = 0.0_wp
      zscerrgtu(:)      = 0.0_wp
      zscerrgtv(:)      = 0.0_wp
      zscerrgsu(:)      = 0.0_wp
      zscerrgsv(:)      = 0.0_wp
      zscerrgru(:)      = 0.0_wp
      zscerrgrv(:)      = 0.0_wp
      zerrgtu(:,:)      = 0.0_wp
      zerrgtv(:,:)      = 0.0_wp
      zerrgsu(:,:)      = 0.0_wp
      zerrgsv(:,:)      = 0.0_wp
      zerrgru(:,:)      = 0.0_wp
      zerrgrv(:,:)      = 0.0_wp
      !--------------------------------------------------------------------
      ! Output filename Xn=F(X0)
      !--------------------------------------------------------------------
      CALL tlm_namrd
      gamma = h_ratio
      file_wop='trj_wop_zps'
      file_xdx='trj_xdx_zps'
      !--------------------------------------------------------------------
      ! Initialize the tangent input with random noise: dx
      !--------------------------------------------------------------------
      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         CALL grid_rd_sd( 596035, z3r,  'T', 0.0_wp, stdt)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  ztn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 523432, z3r,  'T', 0.0_wp, stds)
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zsn_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
         CALL grid_rd_sd( 432545, z3r,  'T', 0.0_wp, stdr) 
         DO jk = 1, jpk
            DO jj = nldj, nlej
               DO ji = nldi, nlei
                  zrd_tlin(ji,jj,jk) = z3r(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF 
      !--------------------------------------------------------------------
      ! Complete Init for Direct
      !-------------------------------------------------------------------
      IF ( tlm_bch /= 2 )  CALL istate_p  
      ! *** initialize the reference trajectory
      ! ------------
      CALL  trj_rea( nit000-1, 1 )
      CALL  trj_rea( nit000, 1 )

      IF (( cur_loop .NE. 0) .OR. ( gamma .NE. 0.0_wp) )THEN
         ztn_tlin(:,:,:) = gamma * ztn_tlin(:,:,:)
         tn(:,:,:)       = tn(:,:,:) + ztn_tlin(:,:,:)

         zsn_tlin(:,:,:) = gamma * zsn_tlin(:,:,:)
         sn(:,:,:)       = sn(:,:,:) + zsn_tlin(:,:,:)

         zrd_tlin(:,:,:) = gamma * zrd_tlin(:,:,:)
         rhd(:,:,:)       = rhd(:,:,:) + zrd_tlin(:,:,:)
      ENDIF     
      !--------------------------------------------------------------------
      !  Compute the direct model F(X0,t=n) = Xn
      !--------------------------------------------------------------------
      IF ( tlm_bch /= 2 ) CALL zps_hde(nit000, tn, sn, rhd, gtu, gsu, gru, gtv, gsv, grv)
      IF ( tlm_bch == 0 ) CALL trj_wri_spl(file_wop)
      IF ( tlm_bch == 1 ) CALL trj_wri_spl(file_xdx)
      !--------------------------------------------------------------------
      !  Compute the Tangent 
      !--------------------------------------------------------------------
      IF ( tlm_bch == 2 ) THEN
         !--------------------------------------------------------------------
         ! Initialize the tangent variables:
         !--------------------------------------------------------------------
         CALL  trj_rea( nit000-1, 1 ) 
         CALL  trj_rea( nit000, 1 ) 
         tn_tl  (:,:,:)  = ztn_tlin  (:,:,:)
         sn_tl  (:,:,:)  = zsn_tlin  (:,:,:)
         rhd_tl  (:,:,:) = zrd_tlin  (:,:,:)
         CALL zps_hde_tan(nit000, tn  , sn       ,            &
         &                   ztn_tlin , zsn_tlin , zrd_tlin,  &
         &                   gtu_tl, gsu_tl, gru_tl,          &
         &                   gtv_tl, gsv_tl, grv_tl           )

         !--------------------------------------------------------------------
         ! Compute the scalar product: ( L(t0,tn) gamma dx0 ) )
         !--------------------------------------------------------------------

         zsp2_1    = DOT_PRODUCT( gtu_tl, gtu_tl  )
         zsp2_2    = DOT_PRODUCT( gtv_tl, gtv_tl  )
         zsp2_3    = DOT_PRODUCT( gsu_tl, gsu_tl  )
         zsp2_4    = DOT_PRODUCT( gsv_tl, gsv_tl  )
         zsp2_5    = DOT_PRODUCT( gru_tl, gru_tl  )
         zsp2_6    = DOT_PRODUCT( grv_tl, grv_tl  )

         zsp2      = zsp2_1 + zsp2_2 + zsp2_3 + zsp2_4 + zsp2_5 + zsp2_6
         !--------------------------------------------------------------------
         !  Storing data
         !--------------------------------------------------------------------

         CALL trj_rd_spl(file_wop) 
         zgtu_wop  (:,:) = gtu  (:,:)
         zgtv_wop  (:,:) = gtv  (:,:)
         zgsu_wop  (:,:) = gsu  (:,:)
         zgsv_wop  (:,:) = gsv  (:,:)
         zgru_wop  (:,:) = gru  (:,:)
         zgrv_wop  (:,:) = grv  (:,:)
         CALL trj_rd_spl(file_xdx) 
         zgtu_out  (:,:) = gtu  (:,:)
         zgtv_out  (:,:) = gtv  (:,:)
         zgsu_out  (:,:) = gsu  (:,:)
         zgsv_out  (:,:) = gsv  (:,:)
         zgru_out  (:,:) = gru  (:,:)
         zgrv_out  (:,:) = grv  (:,:)
         !--------------------------------------------------------------------
         ! Compute the Linearization Error 
         ! Nn = M( X0+gamma.dX0, t0,tn) - M(X0, t0,tn)
         ! and 
         ! Compute the Linearization Error 
         ! En = Nn -TL(gamma.dX0, t0,tn)
         !--------------------------------------------------------------------
         ! Warning: Here we re-use local variables z()_out and z()_wop
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgtu_out   (ji,jj) = zgtu_out (ji,jj) - zgtu_wop  (ji,jj)
               zgtu_wop   (ji,jj) = zgtu_out (ji,jj) - gtu_tl    (ji,jj)
               IF (  gtu_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgtu(ji,jj) = zgtu_out(ji,jj)/gtu_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgtu(ii) = ji
                   ijposgtu(ii) = jj
                   IF ( INT(umask(ji,jj,1)) .NE. 0)  THEN
                      zscgtu (ii) =  zgtu_wop(ji,jj)
                      zscerrgtu (ii) =  ( zerrgtu(ji,jj) - 1.0_wp ) /gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgtv_out   (ji,jj) = zgtv_out (ji,jj) - zgtv_wop  (ji,jj)
               zgtv_wop   (ji,jj) = zgtv_out (ji,jj) - gtv_tl    (ji,jj)
               IF (  gtv_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgtv(ji,jj) = zgtv_out(ji,jj)/gtv_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgtv(ii) = ji
                   ijposgtv(ii) = jj
                   IF ( INT(vmask(ji,jj,1)) .NE. 0)  THEN
                      zscgtv (ii) =  zgtv_wop(ji,jj)
                      zscerrgtv (ii) =  ( zerrgtv(ji,jj) - 1.0_wp ) / gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgsu_out   (ji,jj) = zgsu_out (ji,jj) - zgsu_wop  (ji,jj)
               zgsu_wop   (ji,jj) = zgsu_out (ji,jj) - gsu_tl    (ji,jj)
               IF (  gsu_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgsu(ji,jj) = zgsu_out(ji,jj)/gsu_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgsu(ii) = ji
                   ijposgsu(ii) = jj
                   IF ( INT(umask(ji,jj,1)) .NE. 0)  THEN
                      zscgsu (ii) =  zgsu_wop(ji,jj)
                      zscerrgsu (ii) =  ( zerrgsu(ji,jj) - 1.0_wp ) /gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgsv_out   (ji,jj) = zgsv_out (ji,jj) - zgsv_wop  (ji,jj)
               zgsv_wop   (ji,jj) = zgsv_out (ji,jj) - gsv_tl    (ji,jj)
               IF (  gtv_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgsv(ji,jj) = zgsv_out(ji,jj)/gsv_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgsv(ii) = ji
                   ijposgsv(ii) = jj
                   IF ( INT(vmask(ji,jj,1)) .NE. 0)  THEN
                      zscgsv (ii) =  zgsv_wop(ji,jj)
                      zscerrgsv (ii) =  ( zerrgsv(ji,jj) - 1.0_wp ) /gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgru_out   (ji,jj) = zgru_out (ji,jj) - zgru_wop  (ji,jj)
               zgru_wop   (ji,jj) = zgru_out (ji,jj) - gru_tl    (ji,jj)
               IF (  gru_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgru(ji,jj) = zgru_out(ji,jj)/gru_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgru(ii) = ji
                   ijposgru(ii) = jj
                   IF ( INT(umask(ji,jj,1)) .NE. 0)  THEN
                      zscgru (ii) =  zgru_wop(ji,jj)
                      zscerrgru (ii) =  ( zerrgru(ji,jj) - 1.0_wp ) /gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         ii=0
         DO jj = 1, jpj
            DO ji = 1, jpi
               zgrv_out   (ji,jj) = zgrv_out (ji,jj) - zgrv_wop  (ji,jj)
               zgrv_wop   (ji,jj) = zgrv_out (ji,jj) - grv_tl    (ji,jj)
               IF (  grv_tl(ji,jj) .NE. 0.0_wp ) &
                  & zerrgrv(ji,jj) = zgrv_out(ji,jj)/grv_tl(ji,jj)
               IF( (MOD(ji, isamp) .EQ. 0) .AND. &
               &   (MOD(jj, jsamp) .EQ. 0) ) THEN
                   ii = ii+1
                   iiposgrv(ii) = ji
                   ijposgrv(ii) = jj
                   IF ( INT(vmask(ji,jj,1)) .NE. 0)  THEN
                      zscgrv (ii) =  zgrv_wop(ji,jj)
                      zscerrgrv (ii) =  ( zerrgrv(ji,jj) - 1.0_wp ) / gamma
                   ENDIF
               ENDIF
            END DO
         END DO
         zsp1_1    = DOT_PRODUCT( zgtu_out, zgtu_out  )
         zsp1_2    = DOT_PRODUCT( zgtv_out, zgtv_out  )
         zsp1_3    = DOT_PRODUCT( zgsu_out, zgsu_out  )
         zsp1_4    = DOT_PRODUCT( zgsv_out, zgsv_out  )
         zsp1_5    = DOT_PRODUCT( zgru_out, zgru_out  )
         zsp1_6    = DOT_PRODUCT( zgrv_out, zgrv_out  )
         zsp1      = zsp1_1 + zsp1_2 + zsp1_3 + zsp1_4 + zsp1_5 + zsp1_6
         zsp3_1    = DOT_PRODUCT( zgtu_wop, zgtu_wop  )
         zsp3_2    = DOT_PRODUCT( zgtv_wop, zgtv_wop  )
         zsp3_3    = DOT_PRODUCT( zgsu_wop, zgsu_wop  )
         zsp3_4    = DOT_PRODUCT( zgsv_wop, zgsv_wop  )
         zsp3_5    = DOT_PRODUCT( zgru_wop, zgru_wop  )
         zsp3_6    = DOT_PRODUCT( zgrv_wop, zgrv_wop  )
         zsp3      = zsp3_1 + zsp3_2 + zsp3_3 + zsp3_4 + zsp3_5 + zsp3_6
         !--------------------------------------------------------------------
         ! Print the linearization error En - norme 2
         !--------------------------------------------------------------------
         ! 14 char:'12345678901234'
         cl_name = 'zps_tam:En '
         zzsp    = SQRT(zsp3)
         zzsp_1  = SQRT(zsp3_1)
         zzsp_2  = SQRT(zsp3_2)
         zzsp_3  = SQRT(zsp3_3)
         zzsp_4  = SQRT(zsp3_4)
         zzsp_5  = SQRT(zsp3_5)
         zzsp_6  = SQRT(zsp3_6)
         zgsp5   = zzsp
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         !--------------------------------------------------------------------
         ! Compute TLM norm2
         !--------------------------------------------------------------------
         zzsp    = SQRT(zsp2)
         zzsp_1  = SQRT(zsp2_1)
         zzsp_2  = SQRT(zsp2_2)
         zzsp_3  = SQRT(zsp2_3)
         zzsp_4  = SQRT(zsp2_4)
         zzsp_5  = SQRT(zsp2_5)
         zzsp_6  = SQRT(zsp2_6)
         zgsp4   = zzsp
         cl_name = 'zps_tam:Ln2'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )
         !--------------------------------------------------------------------
         ! Print the linearization error Nn - norme 2
         !--------------------------------------------------------------------
         zzsp     = SQRT(zsp1)
         zzsp_1   = SQRT(zsp1_1)
         zzsp_2   = SQRT(zsp1_2)
         zzsp_3   = SQRT(zsp1_3)
         zzsp_4   = SQRT(zsp1_4)
         zzsp_5   = SQRT(zsp1_5)
         zzsp_6   = SQRT(zsp1_6)
         cl_name  = 'zpshde:Mhdx-Mx'
         CALL prntst_tlm( cl_name, kumadt, zzsp, h_ratio )

         zgsp3    = SQRT( zsp3/zsp2 )
         zgsp7    = zgsp3/gamma
         zgsp1    = zzsp
         zgsp2    = zgsp1 / zgsp4
         zgsp6    = (zgsp2 - 1.0_wp)/gamma

         FMT = "(A8,2X,I4.4,2X,E6.1,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13,2X,E20.13)"
         WRITE(numtan,FMT) 'zpshde  ', cur_loop, h_ratio, zgsp1, zgsp2, zgsp3, zgsp4, zgsp5, zgsp6, zgsp7
         !--------------------------------------------------------------------
         ! Unitary calculus
         !--------------------------------------------------------------------
         FMT = "(A8,2X,A8,2X,I4.4,2X,E6.1,2X,I4.4,2X,I4.4,2X,I4.4,2X,E20.13,1X)"
         cl_name = 'zpshde  '
         IF (lwp) THEN
            DO ii=1, 100, 1
               IF ( zscgtu(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgtu    ', &
                    & cur_loop, h_ratio, ii, iiposgtu(ii), ijposgtu(ii), zscgtu(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscgtv(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgtv    ', &
                    & cur_loop, h_ratio, ii, iiposgtv(ii), ijposgtv(ii), zscgtv(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscgsu(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgsu    ', &
                    & cur_loop, h_ratio, ii, iiposgsu(ii), ijposgsu(ii), zscgsu(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscgsv(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgsv    ', &
                    & cur_loop, h_ratio, ii, iiposgsv(ii), ijposgsv(ii), zscgsv(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscgru(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgru    ', &
                    & cur_loop, h_ratio, ii, iiposgru(ii), ijposgru(ii), zscgru(ii)
            ENDDO
            DO ii=1, 100, 1
               IF ( zscgrv(ii) .NE. 0.0_wp ) WRITE(numtan_sc,FMT) cl_name, 'zscgrv    ', &
                    & cur_loop, h_ratio, ii, iiposgrv(ii), ijposgrv(ii), zscgrv(ii)
            ENDDO
            ! write separator
            WRITE(numtan_sc,"(A4)") '===='
         ENDIF
      ENDIF
      DEALLOCATE(              &
         & ztem, zsal,         & 
         & ztn_tlin, zsn_tlin, & 
         & zrd_tlin, zgtu_out, & 
         & zgtv_out, zgsu_out, & 
         & zgsv_out, zgru_out, & 
         & zgrv_out,           &
         & z3r                 &
         & ) 
   END SUBROUTINE zps_hde_tlm_tst
   !!======================================================================
#endif
#endif
END MODULE zpshde_tam
