MODULE cla_dynspg_tam
   !!----------------------------------------------------------------------
   !!    This software is governed by the CeCILL licence (Version 2)
   !!----------------------------------------------------------------------
#if defined key_tam
   !!======================================================================
   !!  ***  MODULE  cla_dynspg_tam : TANGENT/ADJOINT OF MODULE cla_dynspg  ***
   !!
   !!  Update the momentum trend with the cross land advection from surface 
   !!  pressure gradient in the free surface constant volume case with 
   !!  vector optimization.
   !!                  
   !!======================================================================
   !! History of the direct module:
   !!        !         (A. Bozec)  Original code
   !!   8.5  !  02-11  (A. Bozec)  F90: Free form and module
   !! History of the TAM module:
   !!   9.0  !  09-03  (A. Weaver) Original version
   !!----------------------------------------------------------------------
   !!   dyn_spg_cla_tan  : update the momentum trend with the surface pressure
   !!                      gradient in the free surface constant volume case
   !!                      with vector optimization (tangent routine)
   !!   dyn_spg_cla_adj  : update the momentum trend with the surface pressure
   !!                      gradient in the free surface constant volume case
   !!                      with vector optimization (adjoint routine)
   !!----------------------------------------------------------------------
   !! * Modules used
   USE par_kind      , ONLY: & ! Precision variables
      & wp
   USE in_out_manager, ONLY: & ! I/O manager 
      & lwp,                 &
      & numout,              &  
      & nit000,              &
      & nitend
   USE dom_oce       , ONLY: & ! Ocean space and time domain
      & mi0,                 &
      & mi1,                 &
      & mj0,                 &
      & mj1,                 &
      & e1t,                 &
      & e2t,                 &
      & e2u,                 &
#if defined key_zco
      & e3t_0,               &
#else
      & e3t,                 &
#endif
      & tmask
   USE oce_tam       , ONLY: & ! Tangent-linear and adjoint model variables
      & ua_tl,               &
      & va_tl,               &
      & ua_ad,               &
      & va_ad
   USE sbc_oce_tam   , ONLY: & ! Surface BCs for tangent and adjoint model
      & emp_tl,              &
      & emp_ad
   USE lib_mpp       , ONLY: & ! Massively parallel processing library
      & mpp_sum,             &
      & lk_mpp
     
   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC dyn_spg_cla_tan,   & ! routine called by dyn_spg_tan
      &   dyn_spg_cla_adj      ! routine called by dyn_spg_adj

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"

CONTAINS

   SUBROUTINE dyn_spg_cla_tan( kt ) 
      !!----------------------------------------------------------------------
      !!              ***  routine dyn_spg_cla_tan  ***
      !!
      !! ** Purpose of direct routine :
      !!
      !! ** Method of direct routine : 
      !!
      !! ** Action :
      !!
      !! History
      !!        !  09-03  (A. Weaver) Original version
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::  &
         & kt              ! ocean time-step

      !! * Local declarations
      INTEGER  :: &
         & ji, &           ! dummy loop indices
         & jj, &
         & jk                
      INTEGER  :: &
         & ii0, &          ! temporary integers
         & ii1, &
         & ij0, &
         & ij1        
      REAL(wp) ::  &    
         & zempmed, &      ! EMP on Med Sea ans Red Sea
         & zempred, &                
         & zwei,    &                            !              
         & zisw_rs, &      ! imposed transport Red sea
         & zurw_rs, &
         & zbrw_rs, &    
         & zisw_ms, &      ! imposed transport Med Sea
         & zurw_ms, &      
         & zbrw_ms, &
         & zmrw_ms
      !!----------------------------------------------------------------------

      ! Different velocities for straits ( Gibraltar, Bab el Mandeb...)
         
      ! Control print
      ! -------------
      IF( kt == nit000 ) THEN 
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'cla_dynspg_tan : cross land advection on surface '
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~   pressure '
         IF(lwp) WRITE(numout,*) ' '
      ENDIF

      ! EMP on Mediterranean Sea and Red Sea 
      ! ------------------------------------
      ! compute the emp in Mediterranean Sea
      zempmed = 0.0_wp
      zwei = 0.0_wp
      ij0 =  96   ;   ij1 = 110
      ii0 = 141   ;   ii1 = 181
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempmed = zempmed + emp_tl(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempmed )      ! sum with other processors value

      ! minus 2 points in Red Sea and 3 in Atlantic 
      ij0 =  96   ;   ij1 =  96
      ii0 = 148   ;   ii1 = 148
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            zempmed = zempmed - emp_tl(ji  ,jj) * tmask(ji  ,jj,1)             &
               &                                * e1t(ji  ,jj) * e2t(ji  ,jj)  &
               &              - emp_tl(ji+1,jj) * tmask(ji+1,jj,1)             &
               &                                * e1t(ji+1,jj) * e2t(ji+1,jj)   
         END DO
      END DO
      ! we convert in m3
      zempmed = zempmed * 1.e-3_wp

      ! compute the emp in Red Sea   
      zempred = 0.0_wp
      zwei = 0.0_wp
      ij0 =  87   ;   ij1 =  96
      ii0 = 148   ;   ii1 = 160
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            zwei    = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            zempred = zempred + emp_tl(ji,jj) * zwei
         END DO
      END DO
      IF( lk_mpp )   CALL mpp_sum( zempred )      ! sum with other processors value

      ! we convert in m3
      zempred = zempred * 1.e-3

      ! New Transport at Bab el Mandeb and Gibraltar
      ! --------------------------------------------

      ! imposed transport at Bab el Mandeb
      zisw_rs = 0.4e6_wp        ! inflow surface water
      zurw_rs = 0.2e6_wp        ! upper recirculation water
      zbrw_rs = 0.5e6_wp        ! bottom recirculation water

      ! imposed transport at Gibraltar
      zisw_ms = 0.8e6_wp        ! atlantic-mediterranean  water
      zmrw_ms = 0.7e6_wp        ! middle recirculation water
      zurw_ms = 2.5e6_wp        ! upper  recirculation water 
      zbrw_ms = 3.5e6_wp        ! bottom recirculation water 

      ! Different velocities for straits ( Gibraltar, Bab el Mandeb )
      ! -------------------------------------------------------------

      ! Bab el Mandeb
      ! -------------
      ! 160,88 north point Bab el Mandeb
      ij0 =  88   ;   ij1 =  88
      ii0 = 160   ;   ii1 = 160
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji,jj  ,: ) = 0.0_wp  !  North East Bab el Mandeb 
         END DO
      END DO
      !                              ! surface
      DO jk = 1,  8                                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               ua_tl(ji, jj,jk) = - ( zempred / 8.0_wp ) &
                  &               / ( e2u(ji, jj) * fse3t(ji, jj,jk) )     
            END DO
         END DO
      END DO
      !                              ! deeper
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji, jj,21) = 0.0_wp
         END DO
      END DO

      ! 160,87 south point Bab el Mandeb
      ij0 =  87   ;   ij1 =  87
      ii0 = 160   ;   ii1 = 160
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji,jj  ,: ) = 0.0_wp  !  South East Bab el Mandeb 
         END DO
      END DO
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji, jj,21) = 0.0_wp
         END DO
      END DO

      ! Gibraltar
      ! ---------

      ! initialisation of velocity at concerned points 
      ! 139, 101 south point in Gibraltar 
      ij0 = 101   ;   ij1 = 101
      ii0 = 139   ;   ii1 = 139
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji,jj  ,: ) = 0.0_wp  !  South West Gibraltar
            ua_tl(ji,jj+1,: ) = 0.0_wp  !  North West Gibraltar
         END DO
      END DO
      !                            ! surface
      DO jk = 1, 14                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               ua_tl(ji,jj,jk) = ( zempmed / 14.0_wp ) &
                  &              / ( e2u(ji,jj) * fse3t(ji,jj,jk) ) 
            END DO
         END DO
      END DO
      !                            ! middle circulation
      DO jk = 15, 20                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               ua_tl(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
      !                            ! deeper 
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji,jj,21) = 0.0_wp
            ua_tl(ji,jj,22) = 0.0_wp
         END DO
      END DO

      ! 139,102 north point in Gibraltar
      ij0 = 102   ;   ij1 = 102
      ii0 = 139   ;   ii1 = 139
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji,jj  ,: ) = 0.0_wp !  North West Gibraltar
         END DO
      END DO
      DO jk = 15, 20                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               ua_tl(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
      !                            ! deeper
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_tl(ji,jj,22) = 0.0_wp
         END DO
      END DO

   END SUBROUTINE dyn_spg_cla_tan

   SUBROUTINE dyn_spg_cla_adj( kt ) 
      !!----------------------------------------------------------------------
      !!              ***  routine dyn_spg_cla_adj  ***
      !!
      !! ** Purpose of direct routine :
      !!
      !! ** Method of direct routine : 
      !!
      !! ** Action :
      !!
      !! History
      !!        !  09-03  (A. Weaver) Original version
      !!---------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in ) ::  &
         & kt              ! ocean time-step

      !! * Local declarations
      INTEGER  :: &
         & ji, &           ! dummy loop indices
         & jj, &
         & jk                
      INTEGER  :: &
         & ii0, &          ! temporary integers
         & ii1, &
         & ij0, &
         & ij1        
      REAL(wp) ::  &    
         & zempmed, &      ! EMP on Med Sea ans Red Sea
         & zempred, &                
         & zwei,    &                            !              
         & zisw_rs, &      ! imposed transport Red sea
         & zurw_rs, &
         & zbrw_rs, &    
         & zisw_ms, &      ! imposed transport Med Sea
         & zurw_ms, &      
         & zbrw_ms, &
         & zmrw_ms
      !!----------------------------------------------------------------------

      ! Different velocities for straits ( Gibraltar, Bab el Mandeb...)
         
      ! Control print
      ! -------------
      IF( kt == nitend ) THEN 
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'cla_dynspg_adj : cross land advection on surface '
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~   pressure '
         IF(lwp) WRITE(numout,*) ' '
      ENDIF

      ! Local initialization
      ! --------------------
      zempmed = 0.0_wp
      zempred = 0.0_wp

      ! New Transport at Bab el Mandeb and Gibraltar
      ! --------------------------------------------

      ! imposed transport at Bab el Mandeb
      zisw_rs = 0.4e6_wp        ! inflow surface water
      zurw_rs = 0.2e6_wp        ! upper recirculation water
      zbrw_rs = 0.5e6_wp        ! bottom recirculation water

      ! imposed transport at Gibraltar
      zisw_ms = 0.8e6_wp        ! atlantic-mediterranean  water
      zmrw_ms = 0.7e6_wp        ! middle recirculation water
      zurw_ms = 2.5e6_wp        ! upper  recirculation water 
      zbrw_ms = 3.5e6_wp        ! bottom recirculation water 

      ! Different velocities for straits ( Gibraltar, Bab el Mandeb )
      ! -------------------------------------------------------------

      ! Gibraltar
      ! ---------

      ! initialisation of velocity at concerned points

      ! 139,102 north point in Gibraltar
      ij0 = 102   ;   ij1 = 102
      ii0 = 139   ;   ii1 = 139
      !                            ! deeper
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji,jj,22) = 0.0_wp
         END DO
      END DO

      DO jk = 15, 20                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               ua_ad(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji,jj  ,: ) = 0.0_wp !  North West Gibraltar
         END DO
      END DO

      ! 139, 101 south point in Gibraltar 
      ij0 = 101   ;   ij1 = 101
      ii0 = 139   ;   ii1 = 139
      !                            ! deeper 
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji,jj,21) = 0.0_wp
            ua_ad(ji,jj,22) = 0.0_wp
         END DO
      END DO
      !                            ! middle circulation
      DO jk = 15, 20                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               ua_ad(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
      !                            ! surface
      DO jk = 1, 14                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               zempmed = zempmed + ua_ad(ji,jj,jk) / 14.0_wp &
                  &            / ( e2u(ji,jj) * fse3t(ji,jj,jk) ) 
               ua_ad(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji,jj+1,: ) = 0.0_wp  !  North West Gibraltar
            ua_ad(ji,jj  ,: ) = 0.0_wp  !  South West Gibraltar
         END DO
      END DO

      ! Bab el Mandeb
      ! -------------

      ! 160,87 south point Bab el Mandeb
      ij0 =  87   ;   ij1 =  87
      ii0 = 160   ;   ii1 = 160

      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji, jj,21) = 0.0_wp
         END DO
      END DO
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji,jj  ,: ) = 0.0_wp  !  South East Bab el Mandeb 
         END DO
      END DO

      ! 160,88 north point Bab el Mandeb
      ij0 =  88   ;   ij1 =  88
      ii0 = 160   ;   ii1 = 160
      !                              ! deeper
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji, jj,21) = 0.0_wp
         END DO
      END DO
      !                              ! surface
      DO jk = 1, 8                                      
         DO jj = mj0(ij0), mj1(ij1)
            DO ji = mi0(ii0), mi1(ii1)
               zempred = zempred - ua_ad(ji,jj,jk) / 8.0_wp &
                  &            / ( e2u(ji,jj) * fse3t(ji,jj,jk) ) 
               ua_ad(ji,jj,jk) = 0.0_wp
            END DO
         END DO
      END DO
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            ua_ad(ji,jj  ,: ) = 0.0_wp  !  North East Bab el Mandeb 
         END DO
      END DO

      ! EMP on Mediterranean Sea and Red Sea 
      ! ------------------------------------

      ! we convert in m3
      zempred = zempred * 1.e-3_wp

!!!! AW: Adjoint of this????
      IF( lk_mpp )   CALL mpp_sum( zempred )      ! sum with other processors value

      ! compute the emp in Red Sea   
      ij0 =  87   ;   ij1 =  96
      ii0 = 148   ;   ii1 = 160
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            zwei = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            emp_ad(ji,jj) = emp_ad(ji,jj) + zempred * zwei
         END DO
      END DO

      ! we convert in m3
      zempmed = zempmed * 1.e-3_wp

      ! minus 2 points in Red Sea and 3 in Atlantic 
      ij0 =  96   ;   ij1 =  96
      ii0 = 148   ;   ii1 = 148
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            emp_ad(ji  ,jj) = emp_ad(ji  ,jj)                        &
               &            - zempmed * tmask(ji  ,jj,1)             &
               &                      * e1t(ji  ,jj) * e2t(ji  ,jj)
            emp_ad(ji+1,jj) = emp_ad(ji+1,jj)                        &
               &            - zempmed * tmask(ji+1,jj,1)             &
               &                      * e1t(ji+1,jj) * e2t(ji+1,jj)   
         END DO
      END DO

      IF( lk_mpp )   CALL mpp_sum( zempmed )      ! sum with other processors value

      ! compute the emp in Mediterranean Sea
      ij0 =  96   ;   ij1 = 110
      ii0 = 141   ;   ii1 = 181
      DO jj = mj0(ij0), mj1(ij1)
         DO ji = mi0(ii0), mi1(ii1)
            zwei = tmask(ji,jj,1) * e1t(ji,jj) * e2t(ji,jj)
            emp_ad(ji,jj) = emp_ad(ji,jj) + zempmed * zwei
         END DO
      END DO

   END SUBROUTINE dyn_spg_cla_adj
   !!======================================================================

#endif
END MODULE cla_dynspg_tam
