MODULE hltinc
#if defined key_tam
   !!======================================================================
   !!                       ***  MODULE hltinc  ***
   !! Linear-Tangent test increment : Apply an increment to the control vect
   !!                                 Increment is either generated by  
   !!                                 difference between two states or either
   !!                                 directly read from an input file
   !!======================================================================

   !!----------------------------------------------------------------------
   !!   'key_asminc' : Switch on the assimilation increment interface
   !!----------------------------------------------------------------------
   !!   asm_inc_init : Initialize the increment arrays and IAU weights
   !!   calc_date    : Compute the calendar date YYYYMMDD on a given step
   !!   tra_asm_inc  : Apply the tracer (T and S) increments
   !!   dyn_asm_inc  : Apply the dynamic (u and v) increments
   !!   ssh_asm_inc  : Apply the SSH increment
   !!----------------------------------------------------------------------
   !! * Modules used   
   USE par_kind, ONLY : &       ! Precision variables
      & wp
   USE in_out_manager, ONLY : & ! I/O manager 
      & lwp,      &
      & numnam,   &
      & numout,   &
      & ctl_warn, &
      & ctl_stop, &
      & nit000,   &
      & nstop,    &
      & ln_rstart
   USE par_oce, ONLY : & ! Ocean space and time domain variables
      & jpi,  & 
      & jpj,  &
      & jpk,  &
      & jpkm1
   USE dom_oce, ONLY : & ! Ocean space and time domain
      & rdt,    &
      & n_cla,  &
      & neuler, &
      & ln_zps, &
      & tmask,  &
      & umask,  &
      & vmask,  &
      & nldi,   &
      & nldj,   &
      & nlei,   &
      & nlej
   USE c1d, ONLY : &    ! 1D initialization
      & lk_c1d
   USE oce, ONLY : &      ! Dynamics and active tracers defined in memory
      & ub, un, ua,    &
      & vb, vn, va,    &
      & tb, tn, ta,    &
      & sb, sn, sa,    &
      & sshb, sshn,    &
      & rhd, rhop,     &
      & rotb, rotn,    &
      & hdivb, hdivn,  &
      & gtu, gsu, gru, &
      & gtv, gsv, grv 
   USE oce_tam, ONLY : &      ! Dynamics and active tracers defined in memory
      & ub_tl, un_tl, ua_tl,    &
      & vb_tl, vn_tl, va_tl,    &
      & tb_tl, tn_tl, ta_tl,    &
      & sb_tl, sn_tl, sa_tl,    &
      & sshb_tl, sshn_tl,    &
      & rhd_tl, rhop_tl,     &
      & rotb_tl, rotn_tl,    &
      & hdivb_tl, hdivn_tl,  &
      & gtu_tl, gsu_tl, gru_tl, &
      & gtv_tl, gsv_tl, grv_tl
   USE divcur, ONLY : &   ! Horizontal divergence and relative vorticity
      & div_cur
   USE cla_div, ONLY : &  ! Specific update of the horizontal divergence
      & div_cla           ! (specific to ORCA_R2)
   USE wzvmod, ONLY : &   ! Vertical velocity
      & wzv
   USE eosbn2, ONLY : &   ! Equation of state - in situ and potential density
      & eos
   USE zpshde, ONLY : &   ! Partial step : Horizontal Derivative
      & zps_hde 
   USE divcur_tam, ONLY : &   ! Horizontal divergence and relative vorticity
      & div_cur_tan
   USE cla_div_tam, ONLY : &  ! Specific update of the horizontal divergence
      & div_cla_tan           ! (specific to ORCA_R2)
   USE wzvmod_tam, ONLY : &   ! Vertical velocity
      & wzv_tan
   USE eosbn2_tam, ONLY : &   ! Equation of state - in situ and potential density
      & eos_tan
   USE zpshde_tam, ONLY : &   ! Partial step : Horizontal Derivative
      & zps_hde_tan
   USE iom                ! Library to read input files
   USE par_tlm

   IMPLICIT NONE

   !! * Routine accessibility
   PRIVATE
   PUBLIC hlt_inc_bld      !: Initialize the increment arrays

   !! * Private Module variables
   REAL(wp), PRIVATE, DIMENSION(:,:,:), ALLOCATABLE :: &
      & t_hltinc1, &       !: Increment to the background temperature
      & s_hltinc1, &       !: Increment to the background salinity
      & u_hltinc1, &       !: Increment to the u-component velocity 
      & v_hltinc1, &       !: Increment to the v-component velocity 
      & t_hltinc2, &       !: Increment to the background temperature
      & s_hltinc2, &       !: Increment to the background salinity
      & u_hltinc2, &       !: Increment to the u-component velocity 
      & v_hltinc2          !: Increment to the v-component velocity

#if defined key_dynspg_flt         
   REAL(wp), PRIVATE, DIMENSION(:,:), ALLOCATABLE :: &
      & ssh_hltinc1, &     !: Increment to the background sea surface height
      & ssh_hltinc2        !: Increment to the background sea surface height
#endif
   CHARACTER (LEN=40), PRIVATE, PARAMETER :: &
      & c_hltincwri = 'increments_01'                 !: Filename for storing the 
                                                   !: hlt increment dX

CONTAINS

   SUBROUTINE hlt_inc_bld ( pstg )
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE hlt_inc_bld  ***
      !!          
      !! ** Purpose : Initialize the assimilation increment and IAU weights.
      !!
      !! ** Method  : Initialize the assimilation increment and IAU weights.
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE
      INTEGER, INTENT(IN) :: &
         & pstg               ! Current stage
      !! * Modules used
      !! * Local declarations
      INTEGER :: &
         & jk, jj, ji
      INTEGER :: &
         & inum
      REAL(wp), DIMENSION(5) :: &
         & zvalmax
      REAL(wp) :: &
         & zt, zs, zu, zv, zssh, &
         & znormt, znorms, znormu, znormv, znormssh
      REAL(wp) :: &
         & znorm
      LOGICAL :: &
         & zlnrm = .FALSE.

      !--------------------------------------------------------------------
      ! Initialize the Incremental Analysis Updating weighting function
      !--------------------------------------------------------------------

      IF ( .NOT. ln_incdx ) THEN      ! No increment file defined

         c_hltrst1 = 'restart.nc'
         c_hltrst2 = 'restart2.nc'

          IF(lwp) THEN
             WRITE(numout,*)
             WRITE(numout,*) 'hlt_inc_init : Open restart file 1 ', c_hltrst1
             WRITE(numout,*) '~~~~~~~~~~~~'
          ENDIF

          !--------------------------------------------------------------------
          ! Allocate and initialize the increment arrays 1
          !--------------------------------------------------------------------
          ALLOCATE( t_hltinc1(jpi,jpj,jpk) )
          ALLOCATE( s_hltinc1(jpi,jpj,jpk) )
          ALLOCATE( u_hltinc1(jpi,jpj,jpk) )
          ALLOCATE( v_hltinc1(jpi,jpj,jpk) )
#if defined key_dynspg_flt
          ALLOCATE( ssh_hltinc1(jpi,jpj  ) )
#endif
          t_hltinc1(:,:,:) = 0.0_wp
          s_hltinc1(:,:,:) = 0.0_wp
          u_hltinc1(:,:,:) = 0.0_wp
          v_hltinc1(:,:,:) = 0.0_wp
#if defined key_dynspg_flt
          ssh_hltinc1(:,:) = 0.0_wp
#endif
         CALL iom_open( c_hltrst1, inum )

         IF ( ln_hltt  )   CALL iom_get( inum, jpdom_autoglo, 'tn',   t_hltinc1 ) 
         IF ( ln_hlts  )   CALL iom_get( inum, jpdom_autoglo, 'sn',   s_hltinc1 )
         IF ( ln_hltuv )   CALL iom_get( inum, jpdom_autoglo, 'un',   u_hltinc1 )
         IF ( ln_hltuv )   CALL iom_get( inum, jpdom_autoglo, 'vn',   v_hltinc1 )
#if defined key_dynspg_flt
         IF ( ln_hltuv )   CALL iom_get( inum, jpdom_autoglo, 'sshn', ssh_hltinc1 )
#endif
         CALL iom_close( inum )

          IF(lwp) THEN
             WRITE(numout,*)
             WRITE(numout,*) 'hlt_inc_init : Open restart file 2 ', c_hltrst2
             WRITE(numout,*) '~~~~~~~~~~~~'
          ENDIF

          !--------------------------------------------------------------------
          ! Allocate and initialize the increment arrays 2
          !--------------------------------------------------------------------
          ALLOCATE( t_hltinc2(jpi,jpj,jpk) )
          ALLOCATE( s_hltinc2(jpi,jpj,jpk) )
          ALLOCATE( u_hltinc2(jpi,jpj,jpk) )
          ALLOCATE( v_hltinc2(jpi,jpj,jpk) )
#if defined key_dynspg_flt
          ALLOCATE( ssh_hltinc2(jpi,jpj  ) )
#endif
          t_hltinc2(:,:,:) = 0.0_wp
          s_hltinc2(:,:,:) = 0.0_wp
          u_hltinc2(:,:,:) = 0.0_wp
          v_hltinc2(:,:,:) = 0.0_wp
#if defined key_dynspg_flt
          ssh_hltinc2(:,:) = 0.0_wp
#endif
         CALL iom_open( c_hltrst2, inum )

         IF ( ln_hltt  )   CALL iom_get( inum, jpdom_autoglo, 'tn',   t_hltinc2 ) 
         IF ( ln_hlts  )   CALL iom_get( inum, jpdom_autoglo, 'sn',   s_hltinc2 )
         IF ( ln_hltuv )   CALL iom_get( inum, jpdom_autoglo, 'un',   u_hltinc2 )
         IF ( ln_hltuv )   CALL iom_get( inum, jpdom_autoglo, 'vn',   v_hltinc2 )
#if defined key_dynspg_flt
         IF ( ln_hltuv )   CALL iom_get( inum, jpdom_autoglo, 'sshn', ssh_hltinc2 )
#endif
         CALL iom_close( inum )

          !--------------------------------------------------------------------
          ! Apply difference restart1 - restart2
          !--------------------------------------------------------------------
          t_hltinc1(:,:,:) = t_hltinc1(:,:,:) - t_hltinc2(:,:,:)
          s_hltinc1(:,:,:) = s_hltinc1(:,:,:) - s_hltinc2(:,:,:)
          u_hltinc1(:,:,:) = u_hltinc1(:,:,:) - u_hltinc2(:,:,:)
          v_hltinc1(:,:,:) = v_hltinc1(:,:,:) - v_hltinc2(:,:,:)
#if defined key_dynspg_flt
          ssh_hltinc1(:,:) = ssh_hltinc1(:,:) - ssh_hltinc2(:,:)
#endif
          !--------------------------------------------------------------------
          ! Deallocate and initialize the increment arrays 2
          !--------------------------------------------------------------------
          DEALLOCATE( t_hltinc2 )
          DEALLOCATE( s_hltinc2 )
          DEALLOCATE( u_hltinc2 )
          DEALLOCATE( v_hltinc2 )
#if defined key_dynspg_flt
          DEALLOCATE( ssh_hltinc2 )
#endif
      ELSE    ! increment file defined

!          c_hltinc = c_hltincwri
          WRITE(c_hltinc, FMT='(A,A)' ) TRIM( c_hltincwri ), '.nc'
          IF(lwp) THEN
             WRITE(numout,*)
             WRITE(numout,*) 'hlt_inc_init : Open increment file ', c_hltinc
             WRITE(numout,*) '~~~~~~~~~~~~'
          ENDIF

          !--------------------------------------------------------------------
          ! Allocate and initialize the increment arrays 1
          !--------------------------------------------------------------------
          ALLOCATE( t_hltinc1(jpi,jpj,jpk) )
          ALLOCATE( s_hltinc1(jpi,jpj,jpk) )
          ALLOCATE( u_hltinc1(jpi,jpj,jpk) )
          ALLOCATE( v_hltinc1(jpi,jpj,jpk) )
#if defined key_dynspg_flt
          ALLOCATE( ssh_hltinc1(jpi,jpj)   )
#endif
          t_hltinc1(:,:,:) = 0.0_wp
          s_hltinc1(:,:,:) = 0.0_wp
          u_hltinc1(:,:,:) = 0.0_wp
          v_hltinc1(:,:,:) = 0.0_wp
          ssh_hltinc1(:,:) = 0.0_wp

         CALL iom_open( c_hltinc, inum )

         IF ( ln_hltt   ) CALL iom_get( inum, jpdom_autoglo, 'bckint',   t_hltinc1 ) 
         IF ( ln_hlts   ) CALL iom_get( inum, jpdom_autoglo, 'bckins',   s_hltinc1 )
         IF ( ln_hltuv  ) CALL iom_get( inum, jpdom_autoglo, 'bckinu',   u_hltinc1 )
         IF ( ln_hltuv  ) CALL iom_get( inum, jpdom_autoglo, 'bckinv',   v_hltinc1 )
#if defined key_dynspg_flt
         IF ( ln_hltssh ) CALL iom_get( inum, jpdom_autoglo, 'bckineta', ssh_hltinc1 )
#endif
         CALL iom_close( inum )
         
      ENDIF

      !--------------------------------------------------------------------
      ! Check X max value and normalize if needed
      !--------------------------------------------------------------------

      IF ( ln_hnorm ) THEN

         zvalmax(1) = maxval(ABS(t_hltinc1(:,:,:)))
         zvalmax(2) = maxval(ABS(s_hltinc1(:,:,:)))
         zvalmax(3) = maxval(ABS(u_hltinc1(:,:,:)))
         zvalmax(4) = maxval(ABS(v_hltinc1(:,:,:)))
#if defined key_dynspg_flt
         IF ( ln_incdx ) THEN  
            zvalmax(5) = maxval(ABS(ssh_hltinc1(:,:)))
         ENDIF
#endif
         IF(lwp) THEN
            WRITE(numout,*)
            WRITE(numout,*) '        Increments max value before normalization'
            WRITE(numout,*) '        Temperature increment max value         : ', zvalmax(1)
            WRITE(numout,*) '        Salinity increment max value            : ', zvalmax(2)
            WRITE(numout,*) '        Horizontal velocity increment max value : ', zvalmax(3)
            WRITE(numout,*) '        Zonal velocity increment max value      : ', zvalmax(4)
            IF ( ln_incdx ) THEN
               WRITE(numout,*) '        Sea Surface increment max value         : ', zvalmax(5)
            ENDIF
            WRITE(numout,*) 
            WRITE(numout,*) '        Normalization bounds:'
            WRITE(numout,*) '        Temperature rhstdt                      : ',rhstdt
            WRITE(numout,*) '        Salinity    rhstds                      : ',rhstds
            WRITE(numout,*) '        Velocity    rhstduv                     : ',rhstduv
#if defined key_dynspg_flt
            IF ( ln_incdx ) THEN
               WRITE(numout,*) '        Sea Surface Height    rhstdssh          : ',rhstdssh
            ENDIF
#endif
         ENDIF
         
         IF ( zvalmax(1) > rhstdt )   THEN 
            zlnrm = .TRUE.
         ELSE
            zvalmax(1) = rhstdt
         ENDIF
         IF ( zvalmax(2) > rhstds )   THEN 
            zlnrm = .TRUE.
         ELSE
            zvalmax(2) = rhstds
         ENDIF
         IF ( zvalmax(3) > rhstduv )   THEN 
            zlnrm = .TRUE.
         ELSE
            zvalmax(3) = rhstduv
         ENDIF
         IF ( zvalmax(4) > rhstduv )   THEN 
            zlnrm = .TRUE.
         ELSE
            zvalmax(4) = rhstduv
         ENDIF
#if defined key_dynspg_flt
         IF ( ln_incdx ) THEN
            IF ( zvalmax(5) > rhstdssh )   THEN 
               zlnrm = .TRUE.
            ELSE
               zvalmax(5) = rhstdssh
            ENDIF
         ENDIF
#endif
         
         IF ( zlnrm ) THEN
!!$         IF ( .NOT. ln_incdx ) THEN
!!$            znorm = MIN( stdt/zvalmax(1), stds/zvalmax(2), stdu/zvalmax(3), &
!!$                       & stdv/zvalmax(4), stdssh/zvalmax(5))
!!$         ELSE
!!$            znorm = MIN( stdt/zvalmax(1), stds/zvalmax(2), stdu/zvalmax(3), &
!!$                       & stdv/zvalmax(4))
!!$
!!$         ENDIF
            znormt = rhstdt/zvalmax(1)
            znorms = rhstds/zvalmax(2)
            znormu = rhstduv/zvalmax(3)
            znormv = rhstduv/zvalmax(4)
#if defined key_dynspg_flt
            IF ( ln_incdx ) znormssh = rhstdssh/zvalmax(5)
#endif            
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     t_hltinc1(ji,jj,jk) = t_hltinc1(ji,jj,jk) * znormt
                  END DO
               END DO
            END DO
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     s_hltinc1(ji,jj,jk) = s_hltinc1(ji,jj,jk) * znorms
                  END DO
               END DO
            END DO
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     u_hltinc1(ji,jj,jk) = u_hltinc1(ji,jj,jk) * znormu
                  END DO
               END DO
            END DO
            DO jk = 1, jpk
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     v_hltinc1(ji,jj,jk) = v_hltinc1(ji,jj,jk) * znormv
                  END DO
               END DO
            END DO
#if defined key_dynspg_flt
            IF ( ln_incdx ) THEN
               DO jj = nldj, nlej
                  DO ji = nldi, nlei
                     ssh_hltinc1(ji,jj) = ssh_hltinc1(ji,jj) * znormssh
                  END DO
               END DO
            ENDIF
#endif            
         ENDIF

      ENDIF

      zt = maxval(ABS(t_hltinc1(:,:,:)))
      zs = maxval(ABS(s_hltinc1(:,:,:)))
      zu = maxval(ABS(u_hltinc1(:,:,:)))
      zv = maxval(ABS(v_hltinc1(:,:,:)))
#if defined key_dynspg_flt
        zssh =  maxval(ABS(ssh_hltinc1(:,:)))
#endif
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) '        Increments max value after normalization'
         WRITE(numout,*) '        Temperature increment max value         : ', zt
         WRITE(numout,*) '        Salinity increment max value            : ', zs
         WRITE(numout,*) '        Horizontal velocity increment max value : ', zu
         WRITE(numout,*) '        Zonal velocity increment max value      : ', zv
#if defined key_dynspg_flt
            WRITE(numout,*) '        Sea Surface increment max value         : ', zssh
#endif
         WRITE(numout,*) 
      ENDIF

      SELECT CASE( nstg )
      CASE ( 1 )                                 ! X = X + dX
         IF ( .NOT. ln_incdx ) CALL hlt_inc_wri( ln_incdx, nstg)
                               CALL tra_hlt_inc            ! Tracers
                               CALL dyn_hlt_inc            ! Dynamics
#if defined key_dynspg_flt
         IF ( ln_incdx )       CALL ssh_hlt_inc            ! SSH
#endif
      CASE ( 2 )                                 ! X_tl = dX
         IF ( .NOT. ln_incdx ) CALL hlt_inc_wri( ln_incdx, nstg)
                               CALL tra_hlt_inc_tan        ! Tracers
                               CALL dyn_hlt_inc_tan        ! Dynamics
#if defined key_dynspg_flt
                               CALL ssh_hlt_inc_tan        ! SSH
#endif
      CASE ( 3 )            ! save dX to file if compute from restart
         IF ( .NOT. ln_incdx ) CALL hlt_inc_wri( ln_incdx, nstg)
      END SELECT

   END SUBROUTINE hlt_inc_bld

   SUBROUTINE tra_hlt_inc
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE tra_hlt_inc  ***
      !!          
      !! ** Purpose : Apply the tracer (T and S)  increments
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE

      !! * Arguments
      !! * Local declarations
            
      neuler = 0  ! Force Euler forward step

      ! Add increment
      tn(:,:,:) = tn(:,:,:) + t_hltinc1(:,:,:)   
      sn(:,:,:) = sn(:,:,:) + s_hltinc1(:,:,:)   

      tb(:,:,:) = tn(:,:,:)                        ! Update before fields
      sb(:,:,:) = sn(:,:,:)

      CALL eos( tb, sb, rhd, rhop )                ! Before potential and in situ densities

      IF( ln_zps .AND. .NOT. lk_c1d ) &
          &  CALL zps_hde( nit000, tb, sb, rhd,  &  ! Partial steps: before horizontal derivative
          &                gtu, gsu, gru,        &  ! of T, S, rd at the bottom ocean level
          &                gtv, gsv, grv )

      DEALLOCATE( t_hltinc1 )
      DEALLOCATE( s_hltinc1 )

   END SUBROUTINE tra_hlt_inc

   SUBROUTINE dyn_hlt_inc
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE dyn_hlt_inc  ***
      !!          
      !! ** Purpose : Apply the dynamics (u and v) increments.
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE

      !! * Arguments
      !! * Local declarations
      INTEGER :: &
         & kt

      kt = nit000
      neuler = 0                    ! Force Euler forward step

      ! Add increment
      un(:,:,:) = un(:,:,:) + u_hltinc1(:,:,:)
      vn(:,:,:) = vn(:,:,:) + v_hltinc1(:,:,:)  

      ub(:,:,:) = un(:,:,:)         ! Update before fields
      vb(:,:,:) = vn(:,:,:)
 
      CALL div_cur( kt )                  ! Compute divergence and curl for now fields
      IF( n_cla == 1 ) CALL div_cla( kt ) ! Cross Land Advection (Update Hor. divergence)

      rotb (:,:,:) = rotn (:,:,:)   ! Update before fields
      hdivb(:,:,:) = hdivn(:,:,:)

      CALL wzv( kt )                ! Vertical velocity

      DEALLOCATE( u_hltinc1 )
      DEALLOCATE( v_hltinc1 )

   END SUBROUTINE dyn_hlt_inc

   SUBROUTINE ssh_hlt_inc
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE ssh_hlt_inc  ***
      !!          
      !! ** Purpose : Apply the sea surface height increment.
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE

      !! * Arguments

      !! * Local declarations
      INTEGER :: &
         & it

      neuler = 0                    ! Force Euler forward step

      ! Add increment
      sshn(:,:) = sshn(:,:) + ssh_hltinc1(:,:)  

      sshb(:,:) = sshn(:,:)         ! Update before fields

      DEALLOCATE( ssh_hltinc1 )

   END SUBROUTINE ssh_hlt_inc

   SUBROUTINE tra_hlt_inc_tan
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE tra_hlt_inc_tan  ***
      !!          
      !! ** Purpose : Apply the tracer (T and S)  increments
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE

      !! * Arguments
      !! * Local declarations
            
      neuler = 0  ! Force Euler forward step

      ! Add increment
      tn_tl(:,:,:) = t_hltinc1(:,:,:)   
      sn_tl(:,:,:) = s_hltinc1(:,:,:)   

      CALL lbc_lnk( tn_tl, 'T',  1.0 )
      CALL lbc_lnk( sn_tl, 'T',  1.0 )

      tb_tl(:,:,:) = tn_tl(:,:,:)                        ! Update before fields
      sb_tl(:,:,:) = sn_tl(:,:,:)

      CALL eos_tan( tb, sb, tb_tl, sb_tl, rhd_tl, rhop_tl )                ! Before potential and in situ densities
         
      IF( ln_zps .AND. .NOT. lk_c1d ) &
          &  CALL zps_hde_tan( nit000, tb, sb, tb_tl, sb_tl, rhd_tl,  &  ! Partial steps: before horizontal derivative
          &                gtu_tl, gsu_tl, gru_tl,        &  ! of T, S, rd at the bottom ocean level
          &                gtv_tl, gsv_tl, grv_tl )

      DEALLOCATE( t_hltinc1 )
      DEALLOCATE( s_hltinc1 )

   END SUBROUTINE tra_hlt_inc_tan

   SUBROUTINE dyn_hlt_inc_tan
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE dyn_hlt_inc_tan  ***
      !!          
      !! ** Purpose : Apply the dynamics (u and v) increments.
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE

      !! * Arguments
      !! * Local declarations
      INTEGER :: &
         & kt

      kt = nit000
      neuler = 0                    ! Force Euler forward step

      ! Add increment
      un_tl(:,:,:) = u_hltinc1(:,:,:)
      vn_tl(:,:,:) = v_hltinc1(:,:,:)  

      CALL lbc_lnk( un_tl, 'U', -1.0 )
      CALL lbc_lnk( vn_tl, 'V', -1.0 )
 
      ub_tl(:,:,:) = un_tl(:,:,:)         ! Update before fields
      vb_tl(:,:,:) = vn_tl(:,:,:)
 
      CALL div_cur_tan( kt )                  ! Compute divergence and curl for now fields
      IF( n_cla == 1 ) CALL div_cla_tan( kt ) ! Cross Land Advection (Update Hor. divergence)

      rotb_tl (:,:,:) = rotn_tl (:,:,:)   ! Update before fields
      hdivb_tl(:,:,:) = hdivn_tl(:,:,:)

      CALL wzv_tan( kt )                ! Vertical velocity

      DEALLOCATE( u_hltinc1 )
      DEALLOCATE( v_hltinc1 )

   END SUBROUTINE dyn_hlt_inc_tan

   SUBROUTINE ssh_hlt_inc_tan
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE ssh_hlt_inc_tan  ***
      !!          
      !! ** Purpose : Apply the sea surface height increment.
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      IMPLICIT NONE

      !! * Arguments

      !! * Local declarations
      INTEGER :: &
         & it

      neuler = 0                    ! Force Euler forward step

      ! Add increment
      sshn_tl(:,:) = ssh_hltinc1(:,:)  

      sshb_tl(:,:) = sshn_tl(:,:)         ! Update before fields

      DEALLOCATE( ssh_hltinc1 )

   END SUBROUTINE ssh_hlt_inc_tan

   SUBROUTINE hlt_inc_wri ( plrst, pstg ) 
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE hlt_inc_wri  ***
      !!          
      !! ** Purpose : Write the  increment.
      !!
      !! ** Method  : 
      !!
      !! ** Action  : 
      !!
      !! History :
      !!        !  10-07  (F. Vigilant) Original code
      !!----------------------------------------------------------------------

      !! *Module udes
      USE iom  
      !! * Arguments
      LOGICAL, INTENT(in) :: &
         & plrst            ! logical to write or not ssh
      !! * Local declarations
      INTEGER :: &
         & inum, &                  ! File unit number
         & fd                       ! field number
      INTEGER :: &
         & it, pstg
      CHARACTER (LEN=100) :: &
         & cl_hltinc          

      cl_hltinc = c_hltincwri
      fd = 1

      WRITE(cl_hltinc, FMT='(A,A)' ) TRIM( cl_hltinc ), '_output.nc'
!      WRITE(cl_hltinc, FMT='(A)' ) TRIM( cl_hltinc )
      cl_hltinc = TRIM( cl_hltinc )
      CALL iom_open( cl_hltinc, inum, ldwrt = .TRUE., kiolib = jprstlib)

      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'Writing increment in file : ', c_hltinc
         WRITE(numout,*)
      ENDIF

      ! Output increment fields
      CALL iom_rstput( fd, fd, inum, 'bckinu'   , u_hltinc1   )
      CALL iom_rstput( fd, fd, inum, 'bkcinv'   , v_hltinc1   )
      CALL iom_rstput( fd, fd, inum, 'bckint'   , t_hltinc1   )
      CALL iom_rstput( fd, fd, inum, 'bckins'   , s_hltinc1   )
#if defined key_dynspg_flt
      IF ( plrst ) THEN
         CALL iom_rstput( fd, fd, inum, 'bckineta' , ssh_hltinc1 )
      ENDIF
#endif
      CALL iom_close( inum )

      IF ( pstg == 3 ) THEN
         DEALLOCATE( t_hltinc1 )
         DEALLOCATE( s_hltinc1 )
         DEALLOCATE( u_hltinc1 )
         DEALLOCATE( v_hltinc1 )
#if defined key_dynspg_flt
         IF (  plrst ) THEN
            DEALLOCATE( ssh_hltinc1 )
         ENDIF
#endif
      ENDIF

   END SUBROUTINE hlt_inc_wri
#endif
END MODULE hltinc
