MODULE opahlt
#if defined key_tam
   !!==============================================================================
   !!                       ***  MODULE opa_htl   ***
   !! Ocean system   : OPA ocean dynamics (including on-line tracers and sea-ice)
   !!==============================================================================

   !!----------------------------------------------------------------------
   !!   opa_hlt            : solve ocean dynamics, tracer and/or sea-ice
   !!   opa_htl_init       : initialization of the opa model 
   !!   opa_hlt_closefile  : close remaining files
   !!----------------------------------------------------------------------
   !! History :
   !!                  ()  Original code from opa
   !!        !  10-07  (F. Vigilant) Modification for tangent linear hyp
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! dynamics and tracers variables
   USE dom_oce         ! ocean space domain variables
   USE sbc_oce         ! surface boundary condition: ocean
   USE trdmod_oce      ! ocean variables trends
   USE daymod          ! calendar
   USE in_out_manager  ! I/O manager
   USE lib_mpp         ! distributed memory computing

   USE domcfg          ! domain configuration               (dom_cfg routine)
   USE mppini          ! shared/distributed memory setting (mpp_init routine)
   USE domain          ! domain initialization             (dom_init routine)
   USE obc_par         ! open boundary cond. parameters
   USE obcini          ! open boundary cond. initialization (obc_ini routine)
   USE bdy_par         ! unstructured open boundary cond. parameters
   USE bdyini          ! unstructured open boundary cond. initialization (bdy_init routine)
   USE istate          ! initial state setting          (istate_init routine)
   USE eosbn2          ! equation of state            (eos bn2 routine)
   USE zpshde          ! partial step: hor. derivative (zps_hde routine)

   ! ocean physics
   USE ldfdyn          ! lateral viscosity setting      (ldfdyn_init routine)
   USE ldftra          ! lateral diffusivity setting    (ldftra_init routine)
   USE zdfini

   USE phycst          ! physical constant                  (par_cst routine)
   USE trdmod          ! momentum/tracers trends       (trd_mod_init routine)

!   USE diaobs          ! Observation diagnostics       (dia_obs_init routine)

   USE asminc          ! assimilation increments       (asm_inc_init routine)
!   USE asmbkg          ! writing out background state    
   USE tamtrj          ! writing out state trajectory 
!   USE bias            ! bias (T,S,SSH)                   (bias_init routine)
!   USE biaspar         ! bias  parameters

   USE step            ! OPA time-stepping                  (stp     routine)
   USE dynspg_oce      ! Control choice of surface pressure gradient schemes
   USE prtctl          ! Print control                 (prt_ctl_init routine)
   USE c1d             ! 1D configuration
   USE dyncor_c1d      ! Coriolis factor at T-point
!   USE step_c1d        ! Time stepping loop for the 1D configuration

   USE trcini          ! Initialization of the passive tracers
   USE par_tlm

   USE hltinc, ONLY : &
     & hlt_inc_bld

   IMPLICIT NONE
   PRIVATE

   !! * Module variables
   CHARACTER (len=64) ::        &
      cform_aaa="( /, 'AAAAAAAA', / ) "     ! flag for output listing

   !! * Routine accessibility
   PUBLIC opa_hlt        ! called by hlttst.F90

CONTAINS

   SUBROUTINE opa_hlt
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_hlt  ***
      !!
      !! ** Purpose :   opa_hlt solves the primitive equations on an orthogonal 
      !!      curvilinear mesh on the sphere.
      !!
      !! ** Method  : - model general initialization
      !!              - launch the time-stepping (stp routine)
      !!
      !! References :
      !!      Madec, Delecluse,Imbard, and Levy, 1997: reference manual.
      !!              internal report, IPSL.
      !!----------------------------------------------------------------------
      INTEGER ::   istp       ! time step index
      !!----------------------------------------------------------------------

#if defined key_agrif
      CALL ctl_stop( '       Agrif not available')
#endif
      
      CALL opa_hlt_init  ! Initializations

      ! check that all process are still there... If some process have an error,
      ! they will never enter in step and other processes will wait until the end of the cpu time!
      IF( lk_mpp )   CALL mpp_max( nstop )

      istp = nit000
      IF( lk_c1d ) THEN                 ! 1D configuration (no AGRIF zoom)
         CALL ctl_stop( '       lk_c1d not available')
      ELSE

         istp = nit000 - 1                
         IF( (nstg == 1) .OR. (nstg == 3) ) CALL hlt_inc_bld( nstg )      ! increment builder


IF( nstg .NE. 3 ) THEN
         istp = nit000
PRINT*,'STARTING LOOP STP'
         DO WHILE ( istp <= nitend .AND. nstop == 0 )
#if defined key_agrif
            CALL ctl_stop( '       Agrif not available')
#else
            CALL stp( istp )
#endif
            istp = istp + 1
            IF( lk_mpp )   CALL mpp_max( nstop )
         END DO
ENDIF
      ENDIF

      !                                     ! ========= !
      !                                     !  Job end  !
      !                                     ! ========= !

      IF(lwp) WRITE(numout,cform_aaa)       ! Flag AAAAAAA

      IF( nstop /= 0 .AND. lwp ) THEN                 ! error print
         WRITE(numout,cform_err)
         WRITE(numout,*) nstop, ' error have been found' 
      ENDIF

      CALL opa_hlt_closefile

      IF( lk_mpp )   CALL mppstop                          ! Close all files (mpp)
      !
   END SUBROUTINE opa_hlt

   SUBROUTINE opa_hlt_init
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_init  ***
      !!
      !! ** Purpose :   initialization of the opa model
      !!
      !!----------------------------------------------------------------------
#if defined key_coupled
      INTEGER ::   itro, istp0        ! ???
#endif
#if defined key_oasis3 || defined key_oasis4
      INTEGER :: localComm
#endif
!      CHARACTER (len=20) ::   namelistname
      CHARACTER (len=28) ::   file_out
      NAMELIST/namctl/ ln_ctl, nprint, nictls, nictle,   &
         &             isplt , jsplt , njctls, njctle, nbench, nbit_cmp
      !!----------------------------------------------------------------------

      ! Initializations
      ! ===============

      file_out = 'nemohlt.output'

      ! Namelist namctl : Control prints & Benchmark
      REWIND( numnam )
      READ  ( numnam, namctl )

#if defined key_oasis3 || defined key_oasis4
      CALL ctl_stop( '       Key_oasis3 and key_oasis4 not available')
#else
      ! Nodes selection
      nproc = mynode()
#endif
      narea = nproc + 1    ! mynode return the rank of proc (0 --> jpnij -1 )
      lwp   = narea == 1

      ! open additionnal listing
      IF( ln_ctl )   THEN
         IF( narea-1 > 0 )   THEN
            WRITE(file_out,FMT="('nemohlt.output_',I4.4)") narea-1
            CALL ctlopn( numout, file_out, 'UNKNOWN', 'FORMATTED',   &
               &         'SEQUENTIAL', 1, numout, .FALSE., 1 )
            lwp = .TRUE.
            !
            WRITE(numout,*)
            WRITE(numout,*) '                 L O D Y C - I P S L'
            WRITE(numout,*) '                     O P A model'
            WRITE(numout,*) '            Ocean General Circulation Model'
            WRITE(numout,*) '               version OPA 9.0  (2005) '
            WRITE(numout,*) '                   MPI Ocean output '
            WRITE(numout,*)
            WRITE(numout,*)
         ENDIF
      ENDIF

      ! Parameter control and print
      ! ---------------------------
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) '              LINEAR-TANGENT HYPOTHESIS TEST-RUN '
         WRITE(numout,*) '              ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ '
         WRITE(numout,*) '       Namelist namhlt'
         WRITE(numout,*) '             run stage                        nstg       = ', nstg
         IF( nstg == 1 ) THEN
            WRITE(numout,*) '             temperature increment switch     ln_hltt    = ', ln_hltt
            WRITE(numout,*) '             salinity increment switch        ln_hlts    = ', ln_hlts
            WRITE(numout,*) '             velocity incr. switch            ln_hltuv   = ', ln_hltuv
            WRITE(numout,*) '             sea surface height incr. switch  ln_hltssh  = ', ln_hltssh
         ENDIF
         WRITE(numout,*)
         WRITE(numout,*)
      ENDIF


      !                                     ! ============================== !
      !                                     !  Model general initialization  !
      !                                     ! ============================== !

      IF(lwp) WRITE(numout,cform_aaa)       ! Flag AAAAAAA

      CALL opa_hlt_flg                          ! Control prints & Benchmark

                                            ! Domain decomposition
      IF( jpni*jpnj == jpnij ) THEN
         CALL mpp_init                          ! standard cutting out
      ELSE
         CALL mpp_init2                         ! eliminate land processors
      ENDIF
      
      CALL phy_cst                          ! Physical constants

      CALL dom_cfg                          ! Domain configuration
      
      CALL dom_init                         ! Domain

      IF( lk_c1d    ) THEN                      ! adaptation for 1D configuration 
         CALL cor_c1d                                ! redefine Coriolis at T-point
         umask(:,:,:) = tmask(:,:,:)                 ! U, V and T-points are the same
         vmask(:,:,:) = tmask(:,:,:)                 ! 
      ENDIF

      IF( ln_ctl    )   CALL prt_ctl_init   ! Print control

      IF( lk_obc    )   CALL obc_init       ! Open boundaries 

      IF( lk_bdy    )   CALL bdy_init       ! Unstructured open boundaries

      CALL istate_init                      ! ocean initial state (Dynamics and tracers)

      !                                     ! Ocean physics

      CALL ldf_dyn_init                         ! Lateral ocean momentum physics

      CALL ldf_tra_init                         ! Lateral ocean tracer physics

      CALL zdf_init                             ! Vertical ocean physics

      CALL trd_mod_init                         ! Mixed-layer/Vorticity/Integral constraints trends


#if defined key_top
      CALL ctl_stop( '       Key_top not available')
#endif

#if defined key_coupled && ! defined key_oasis3 && ! defined key_oasis4
      CALL ctl_stop( '       Key_coupled not available')
#endif

#if defined key_oasis3 || defined key_oasis4
      CALL ctl_stop( '       Key_oasis3 and key_oasis4 not available')
#endif
!      IF( lk_diaobs ) THEN
!         CALL dia_obs_init                  ! Initialize observational data
!         CALL dia_obs( nit000 - 1 )         ! Observation operator for restart
!      ENDIF

!      CALL dia_ptr_init                     ! Poleward TRansports initialization
      IF( lk_asminc ) CALL asm_inc_init     ! Initialize assimilation increments
      CALL tam_trj_ini
      IF(lwp) WRITE(numout,*)'Euler time step switch is ', neuler

!      CALL bias_init                        ! Initialize bias options
      !                                     ! =============== !
      !                                     !  time stepping  !
      !                                     ! =============== !

      IF(lwp) WRITE(numout,cform_aaa)       ! Flag AAAAAAA

   END SUBROUTINE opa_hlt_init


   SUBROUTINE opa_hlt_flg
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa  ***
      !!
      !! ** Purpose :   Initialize logical flags that control the choice of
      !!      some algorithm or control print
      !!
      !! ** Method  :    Read in namilist namflg logical flags
      !!
      !! History :
      !!   9.0  !  03-11  (G. Madec)  Original code
      !!----------------------------------------------------------------------
      !! * Local declarations

      NAMELIST/namflg/ ln_dynhpg_imp, nn_dynhpg_rst
      !!----------------------------------------------------------------------

      ! Parameter control and print
      ! ---------------------------
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'opa_flg: Control prints & Benchmark'
         WRITE(numout,*) '~~~~~~~ '
         WRITE(numout,*) '          Namelist namctl'
         WRITE(numout,*) '             run control (for debugging)     ln_ctl    = ', ln_ctl
         WRITE(numout,*) '             level of print                  nprint    = ', nprint
         WRITE(numout,*) '             Start i indice for SUM control  nictls    = ', nictls
         WRITE(numout,*) '             End i indice for SUM control    nictle    = ', nictle
         WRITE(numout,*) '             Start j indice for SUM control  njctls    = ', njctls
         WRITE(numout,*) '             End j indice for SUM control    njctle    = ', njctle
         WRITE(numout,*) '             number of proc. following i     isplt     = ', isplt
         WRITE(numout,*) '             number of proc. following j     jsplt     = ', jsplt
         WRITE(numout,*) '             benchmark parameter (0/1)       nbench    = ', nbench
         WRITE(numout,*) '             bit comparison mode (0/1)       nbit_cmp  = ', nbit_cmp
      ENDIF

      ! ... Control the sub-domain area indices for the control prints
      IF( ln_ctl )   THEN
         IF( lk_mpp )   THEN
            ! the domain is forced to the real splitted domain in MPI
            isplt = jpni ; jsplt = jpnj ; ijsplt = jpni*jpnj
         ELSE
            IF( isplt == 1 .AND. jsplt == 1  ) THEN
               CALL ctl_warn( '          - isplt & jsplt are equal to 1',   &
                    &         '          - the print control will be done over the whole domain' )
            ENDIF

            ! compute the total number of processors ijsplt
            ijsplt = isplt*jsplt
         ENDIF

         IF(lwp) WRITE(numout,*)'          - The total number of processors over which the'
         IF(lwp) WRITE(numout,*)'            print control will be done is ijsplt : ', ijsplt

         ! Control the indices used for the SUM control
         IF( nictls+nictle+njctls+njctle == 0 )   THEN
            ! the print control is done over the default area
            lsp_area = .FALSE.
         ELSE
            ! the print control is done over a specific  area
            lsp_area = .TRUE.
            IF( nictls < 1 .OR. nictls > jpiglo )   THEN
               CALL ctl_warn( '          - nictls must be 1<=nictls>=jpiglo, it is forced to 1' )
               nictls = 1
            ENDIF

            IF( nictle < 1 .OR. nictle > jpiglo )   THEN
               CALL ctl_warn( '          - nictle must be 1<=nictle>=jpiglo, it is forced to jpiglo' )
               nictle = jpiglo
            ENDIF

            IF( njctls < 1 .OR. njctls > jpjglo )   THEN
               CALL ctl_warn( '          - njctls must be 1<=njctls>=jpjglo, it is forced to 1' )
               njctls = 1
            ENDIF

            IF( njctle < 1 .OR. njctle > jpjglo )   THEN
               CALL ctl_warn( '          - njctle must be 1<=njctle>=jpjglo, it is forced to jpjglo' )
               njctle = jpjglo
            ENDIF

         ENDIF          ! IF( nictls+nictle+njctls+njctle == 0 )
       ENDIF            ! IF(ln_ctl)

      IF( nbench == 1 )   THEN
         SELECT CASE ( cp_cfg )
         CASE ( 'gyre' )
            CALL ctl_warn( '          The Benchmark is activated ' )
         CASE DEFAULT
            CALL ctl_stop( '          The Benchmark is based on the GYRE configuration: key_gyre must &
               &                      be used or set nbench = 0' )
         END SELECT
      ENDIF

      IF( nbit_cmp == 1 )   THEN
         CALL ctl_warn( '          Bit comparison enabled. Single and multiple processor results must bit compare', &
              &         '          WARNING: RESULTS ARE NOT PHYSICAL.' )
      ENDIF


      ! Read Namelist namflg : algorithm FLaG
      ! --------------------
      REWIND ( numnam )
      READ   ( numnam, namflg )

      ! Parameter control and print
      ! ---------------------------
      IF(lwp) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'opa_flg : Hydrostatic pressure gradient algorithm'
         WRITE(numout,*) '~~~~~~~'
         WRITE(numout,*) '          Namelist namflg : set algorithm flags'
         WRITE(numout,*) '             centered (F) or semi-implicit (T)   ln_dynhpg_imp = ', ln_dynhpg_imp
         WRITE(numout,*) '             hydrostatic pressure gradient'
         WRITE(numout,*) '             add dynhpg implicit variable        nn_dynhpg_rst = ', nn_dynhpg_rst
         WRITE(numout,*) '             in restart ot not nn_dynhpg_rst'
      ENDIF
      IF( .NOT. ln_dynhpg_imp )   nn_dynhpg_rst = 0      ! force no adding dynhpg implicit variables in restart

   END SUBROUTINE opa_hlt_flg


   SUBROUTINE opa_hlt_closefile
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE opa_closefile  ***
      !!
      !! ** Purpose :   Close the files
      !!
      !! ** Method  :
      !!
      !! History :
      !!   9.0  !  05-01  (O. Le Galloudec)  Original code
      !!----------------------------------------------------------------------
      !! * Modules used
      !!----------------------------------------------------------------------

      IF ( lk_mpp ) CALL mppsync

      ! 1. Unit close
      ! -------------

      CLOSE( numnam )           ! namelist
      CLOSE( numout )           ! standard model output file

      IF(lwp) CLOSE( numstp )   ! time-step file
      IF(lwp) CLOSE( numsol )

   END SUBROUTINE opa_hlt_closefile

   !!======================================================================
#endif
END MODULE opahlt
